/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqdb.h"
#include "plugins/dir/dbdir_l.h"
#include "plugins/file/dbfile.h"

#include <aqdatabase/aqdb_db_be.h>
#include <aqdatabase/user/aqdb_userdb.h>

#include <gwenhywfar/gwenhywfar.h>
#include <gwenhywfar/pathmanager.h>
#include <gwenhywfar/debug.h>






int AQDB_DbFactory(const char *s, AQDB_DB **pDb)
{
  if (s && *s) {
    GWEN_URL *url;
    const char *p;
    AQDB_DB *db=NULL;

    url=GWEN_Url_fromString(s);
    if (url==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid URL");
      return GWEN_ERROR_INVALID;
    }

    p=GWEN_Url_GetProtocol(url);
    if (!p || *p==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "No protocol in URL");
      GWEN_Url_free(url);
      return GWEN_ERROR_INVALID;
    }

    if (strcasecmp(p, "dir")==0)
      db=AQDB_DbDir_new(s);
    else if (strcasecmp(p, "file")==0)
      db=AQDB_DbFile_new(s);
    else if (strcasecmp(p, "ufile")==0) {
      int rv;

      db=AQDB_DbFile_new(s);
      rv=AQDB_UserDb_Extend(db);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Error extending db module [%s] with userdb", p+1);
        GWEN_Url_free(url);
        return rv;
      }
    }
    else {
      DBG_ERROR(AQDB_LOGDOMAIN, "Module [%s] not found", p);
      GWEN_Url_free(url);
      return GWEN_ERROR_NOT_FOUND;
    }
    *pDb=db;
    GWEN_Url_free(url);
    return 0;
  }
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "Empty URL");
    return GWEN_ERROR_INVALID;
  }
}



int AQDB_Init()
{
  int rv;
  const char *s;

  rv=GWEN_Init();
  if (rv) {
    DBG_ERROR_ERR(AQDB_LOGDOMAIN, rv);
    return rv;
  }
  if (!GWEN_Logger_IsOpen(AQDB_LOGDOMAIN)) {
    GWEN_Logger_Open(AQDB_LOGDOMAIN,
                     "aqdatabase", 0,
                     GWEN_LoggerType_Console,
                     GWEN_LoggerFacility_User);
  }

  s=getenv("AQDB_LOGLEVEL");
  if (s && *s) {
    GWEN_LOGGER_LEVEL ll;

    ll=GWEN_Logger_Name2Level(s);
    GWEN_Logger_SetLevel(AQDB_LOGDOMAIN, ll);
  }
  else
    GWEN_Logger_SetLevel(AQDB_LOGDOMAIN, GWEN_LoggerLevel_Notice);

  DBG_INFO(AQDB_LOGDOMAIN,
           "AqDatabase v%s: initialising",
           AQDB_VERSION_FULL_STRING);

  /* define locale paths */
  GWEN_PathManager_DefinePath(AQDB_PM_LIBNAME, AQDB_PM_LOCALEDIR);
#if defined(OS_WIN32) || defined(ENABLE_LOCAL_INSTALL)
  /* add folder relative to EXE */
  GWEN_PathManager_AddRelPath(AQDB_PM_LIBNAME,
                              AQDB_PM_LIBNAME,
                              AQDB_PM_LOCALEDIR,
                              LOCALEDIR,
                              GWEN_PathManager_RelModeExe);
#else
  /* add absolute folder */
  GWEN_PathManager_AddPath(AQDB_PM_LIBNAME,
                           AQDB_PM_LIBNAME,
                           AQDB_PM_LOCALEDIR,
                           LOCALEDIR);
#endif

  /* define data paths */
  GWEN_PathManager_DefinePath(AQDB_PM_LIBNAME, AQDB_PM_DATADIR);
#if defined(OS_WIN32) || defined(ENABLE_LOCAL_INSTALL)
  /* add folder relative to EXE */
  GWEN_PathManager_AddRelPath(AQDB_PM_LIBNAME,
                              AQDB_PM_LIBNAME,
                              AQDB_PM_DATADIR,
                              AQDATABASE_DATA_DIR,
                              GWEN_PathManager_RelModeExe);
#else
  /* add absolute folder */
  GWEN_PathManager_AddPath(AQDB_PM_LIBNAME,
                           AQDB_PM_LIBNAME,
                           AQDB_PM_DATADIR,
                           AQDATABASE_DATA_DIR);
#endif

  if (1) {
    GWEN_STRINGLIST *sl=GWEN_PathManager_GetPaths(AQDB_PM_LIBNAME,
                                                  AQDB_PM_LOCALEDIR);
    if (sl) {
      const char *localedir=GWEN_StringList_FirstString(sl);

      rv=GWEN_I18N_BindTextDomain_Dir(PACKAGE, localedir);
      if (rv) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Could not bind textdomain (%d)", rv);
      }
      else {
        rv=GWEN_I18N_BindTextDomain_Codeset(PACKAGE, "UTF-8");
        if (rv) {
          DBG_ERROR(AQDB_LOGDOMAIN, "Could not set codeset (%d)", rv);
        }
      }

      GWEN_StringList_free(sl);
    }
  }

  return 0;
}



int AQDB_Fini()
{
  /* undefine our own paths */
  GWEN_PathManager_UndefinePath(AQDB_PM_LIBNAME, AQDB_PM_LOCALEDIR);
  GWEN_PathManager_UndefinePath(AQDB_PM_LIBNAME, AQDB_PM_DATADIR);

  /* remove AqBanking additions to all pathmanagers */
  GWEN_PathManager_RemovePaths(AQDB_PM_LIBNAME);
  GWEN_Logger_Close(AQDB_LOGDOMAIN);
  GWEN_Fini();
  return 0;
}



