/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "aqdb_column_p.h"

#include <gwenhywfar/misc.h>
#include <gwenhywfar/debug.h>



GWEN_LIST_FUNCTIONS(AQDB_COLUMN, AQDB_Column)




AQDB_COLUMN *AQDB_Column_new(AQDB_DATA_TYPE t,
                             const char *name,
                             int maxLen)
{
  AQDB_COLUMN *column;

  assert(name);
  GWEN_NEW_OBJECT(AQDB_COLUMN, column);
  GWEN_LIST_INIT(AQDB_COLUMN, column);
  if (column==NULL)
    return NULL;

  column->type=t;
  column->name=strdup(name);
  column->maxLen=maxLen;
  column->isMutable=1;

  return column;
}



AQDB_COLUMN *AQDB_Column_dup(const AQDB_COLUMN *origColumn)
{
  AQDB_COLUMN *column;

  column=AQDB_Column_new(origColumn->type, origColumn->name, origColumn->maxLen);
  if (column==NULL)
    return NULL;
  column->isMutable=origColumn->isMutable;

  return column;
}



void AQDB_Column_free(AQDB_COLUMN *column)
{
  if (column) {
    GWEN_LIST_FINI(AQDB_COLUMN, column);
    free(column->name);
    GWEN_FREE_OBJECT(column);
  }
}



AQDB_DATA_TYPE AQDB_Column_GetType(const AQDB_COLUMN *column)
{
  assert(column);
  return column->type;
}



const char *AQDB_Column_GetName(const AQDB_COLUMN *column)
{
  assert(column);
  return column->name;
}



int AQDB_Column_GetMaxLen(const AQDB_COLUMN *column)
{
  assert(column);
  return column->maxLen;
}



int AQDB_Column_GetIsMutable(const AQDB_COLUMN *column)
{
  assert(column);
  return column->isMutable;
}



void AQDB_Column_SetIsMutable(AQDB_COLUMN *column, int b)
{
  assert(column);
  column->isMutable=b;
}



AQDB_COLUMN *AQDB_Column_fromDb(GWEN_DB_NODE *db)
{
  AQDB_COLUMN *column;
  AQDB_DATA_TYPE t;
  const char *s;
  const char *name;
  int maxLen;

  s=GWEN_DB_GetCharValue(db, "type", 0, "unknown");
  t=AQDB_DataType_fromString(s);
  name=GWEN_DB_GetCharValue(db, "name", 0, NULL);
  maxLen=GWEN_DB_GetIntValue(db, "maxLen", 0, 0);
  column=AQDB_Column_new(t, name, maxLen);
  if (column==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Could not create column from DB");
    return NULL;
  }
  column->isMutable=GWEN_DB_GetIntValue(db, "isMutable", 0, 1);
  return column;
}



void AQDB_Column_toDb(const AQDB_COLUMN *column, GWEN_DB_NODE *db)
{
  assert(column);
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "type",
                       AQDB_DataType_toString(column->type));
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "name", column->name);
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "maxLen", column->maxLen);
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "isMutable", column->isMutable);
}



AQDB_COLUMN_LIST *AQDB_Column_List_dup(const AQDB_COLUMN_LIST *cl)
{
  AQDB_COLUMN_LIST *l;
  AQDB_COLUMN *col;

  l=AQDB_Column_List_new();
  col=AQDB_Column_List_First(cl);
  while (col) {
    AQDB_COLUMN *ncol;

    ncol=AQDB_Column_dup(col);
    AQDB_Column_List_Add(ncol, l);
    col=AQDB_Column_List_Next(col);
  }

  return l;
}



int AQDB_Column_toObject(const AQDB_COLUMN *col, AQDB_OBJECT *o)
{
  int rv;
  const char *s;

  /* store type */
  s=AQDB_DataType_toString(AQDB_Column_GetType(col));
  rv=AQDB_Object_SetFieldString(o, AQDB_COLUMN_FIELD_TYPE, s);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* store name */
  s=AQDB_Column_GetName(col);
  if (s) {
    rv=AQDB_Object_SetFieldString(o, AQDB_COLUMN_FIELD_NAME, s);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  /* store maxlen */
  rv=AQDB_Object_SetFieldInt64(o, AQDB_COLUMN_FIELD_MAXLEN,
                               AQDB_Column_GetMaxLen(col));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* store mutability */
  rv=AQDB_Object_SetFieldInt64(o, AQDB_COLUMN_FIELD_MUTABLE,
                               AQDB_Column_GetIsMutable(col));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* done */
  return 0;
}



int AQDB_Column_fromObject(const AQDB_OBJECT *o, AQDB_COLUMN **pCol)
{
  AQDB_DATA_TYPE t;
  const char *s;
  const char *name;
  int64_t maxLen;
  int64_t isMutable;
  int rv;
  AQDB_COLUMN *col;

  /* get type */
  rv=AQDB_Object_GetFieldString(o, AQDB_COLUMN_FIELD_TYPE, &s);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  t=AQDB_DataType_fromString(s);

  /* get name */
  rv=AQDB_Object_GetFieldString(o, AQDB_COLUMN_FIELD_NAME, &name);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* get maxlen */
  rv=AQDB_Object_GetFieldInt64(o, AQDB_COLUMN_FIELD_MAXLEN, &maxLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* get maxlen */
  rv=AQDB_Object_GetFieldInt64(o, AQDB_COLUMN_FIELD_MUTABLE, &isMutable);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* create column */
  col=AQDB_Column_new(t, name, maxLen);
  assert(col);

  AQDB_Column_SetIsMutable(col, isMutable);

  /* done */
  *pCol=col;
  return 0;
}



AQDB_COLUMN *AQDB_Column_List_GetByName(const AQDB_COLUMN_LIST *columnList, const char *name)
{
  AQDB_COLUMN *column;

  column=AQDB_Column_List_First(columnList);
  while (column) {
    const char *cname;

    cname=AQDB_Column_GetName(column);
    if (cname && strcasecmp(cname, name)==0)
      return column;
    column=AQDB_Column_List_Next(column);
  }

  return NULL;
}




