/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2019 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqdb_db_p.h"

#include <gwenhywfar/gwenhywfar.h>
#include <gwenhywfar/pathmanager.h>
#include <gwenhywfar/debug.h>



GWEN_INHERIT_FUNCTIONS(AQDB_DB)
GWEN_LIST_FUNCTIONS(AQDB_DB, AQDB_DB)
GWEN_LIST2_FUNCTIONS(AQDB_DB, AQDB_DBI)





AQDB_DATA_TYPE AQDB_DataType_fromString(const char *s)
{
  if (s) {
    if (strcasecmp(s, "none")==0)
      return AQDB_DataType_None;
    else if (strcasecmp(s, "int")==0)
      return AQDB_DataType_Int;
    else if (strcasecmp(s, "string")==0)
      return AQDB_DataType_String;
    else if (strcasecmp(s, "bin")==0)
      return AQDB_DataType_Bin;
    else if (strcasecmp(s, "date")==0)
      return AQDB_DataType_Date;
    else if (strcasecmp(s, "timestamp")==0)
      return AQDB_DataType_Timestamp;
    else if (strcasecmp(s, "amount")==0)
      return AQDB_DataType_Amount;
  }
  return AQDB_DataType_Unknown;
}



const char *AQDB_DataType_toString(AQDB_DATA_TYPE t)
{
  switch (t) {
  case AQDB_DataType_None:
    return "none";
  case AQDB_DataType_Int:
    return "int";
  case AQDB_DataType_String:
    return "string";
  case AQDB_DataType_Bin:
    return "bin";
  case AQDB_DataType_Date:
    return "date";
  case AQDB_DataType_Timestamp:
    return "timestamp";
  case AQDB_DataType_Amount:
    return "amount";
  default:
    return "unknown";
  }
}





AQDB_DB *AQDB_DB_new(const char *url)
{
  AQDB_DB *db;

  GWEN_NEW_OBJECT(AQDB_DB, db);
  assert(db);
  db->refCount=1;
  GWEN_INHERIT_INIT(AQDB_DB, db);
  GWEN_LIST_INIT(AQDB_DB, db);

  if (url && *url) {
    db->url=GWEN_Url_fromString(url);
    if (db->url==NULL) {
      DBG_WARN(AQDB_LOGDOMAIN, "Invalid URL");
    }
  }

  return db;
}



void AQDB_DB_Attach(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  db->refCount++;
}



void AQDB_DB_free(AQDB_DB *db)
{
  if (db) {
    assert(db->refCount);
    if (--(db->refCount)==0) {
      GWEN_LIST_FINI(AQDB_DB, db);
      GWEN_INHERIT_FINI(AQDB_DB, db);
      GWEN_Url_free(db->url);
      AQDB_Cache_free(db->cache);
      GWEN_FREE_OBJECT(db);
    }
  }
}



const GWEN_URL *AQDB_DB_GetUrl(const AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->url;
}



uint32_t AQDB_DB_GetCacheHintMaxSize(const AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->cacheHintMaxSize;
}



uint32_t AQDB_DB_GetCacheHintMaxEntries(const AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->cacheHintMaxEntries;
}



int AQDB_DB_Open(AQDB_DB *db, uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->openFn)
    return db->openFn(db, flags);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_Create(AQDB_DB *db, uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->createFn)
    return db->createFn(db, flags);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_Close(AQDB_DB *db, uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->closeFn)
    return db->closeFn(db, flags);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_Check(AQDB_DB *db, GWEN_BUFFER *resultTextBuf)
{
  assert(db);
  assert(db->refCount);
  if (db->checkFn)
    return db->checkFn(db, resultTextBuf);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_ReadObject(AQDB_DB *db,
                       AQDB_ID tableId,
                       AQDB_ID objectId,
                       AQDB_OBJECT **pObject)
{
  assert(db);
  assert(db->refCount);
  if (db->readObjectFn)
    return db->readObjectFn(db, AQDB_ACCESS_FLAGS_ISWRITER, tableId, objectId, pObject);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_ReadObjectWithAccessFlags(AQDB_DB *db,
                                      uint32_t aflags,
                                      AQDB_ID tableId,
                                      AQDB_ID objectId,
                                      AQDB_OBJECT **pObject)
{
  assert(db);
  assert(db->refCount);
  if (db->readObjectFn)
    return db->readObjectFn(db, aflags, tableId, objectId, pObject);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object)
{
  assert(db);
  assert(db->refCount);
  if (db->writeObjectFn)
    return db->writeObjectFn(db, object);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_DeleteObject(AQDB_DB *db,
                         AQDB_ID tableId,
                         AQDB_ID objectId)
{
  assert(db);
  assert(db->refCount);
  if (db->deleteObjectFn)
    return db->deleteObjectFn(db, tableId, objectId);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_AddObject(AQDB_DB *db,
                      AQDB_ID tableId,
                      AQDB_OBJECT *object)
{
  assert(db);
  assert(db->refCount);
  if (db->addObjectFn)
    return db->addObjectFn(db, tableId, object);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_OpenTable(AQDB_DB *db,
                      const char *tname,
                      uint32_t flags,
                      AQDB_ID *pTableId)
{
  assert(db);
  assert(db->refCount);
  if (db->openTableFn)
    return db->openTableFn(db, AQDB_ACCESS_FLAGS_ISWRITER, tname, flags, pTableId);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_CreateTable(AQDB_DB *db,
                        const char *tname,
                        uint32_t flags,
                        const AQDB_COLUMN_LIST *columns)
{
  assert(db);
  assert(db->refCount);
  if (db->createTableFn)
    return db->createTableFn(db, tname, flags, columns);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_CloseTable(AQDB_DB *db,
                       AQDB_ID tableId,
                       uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->closeTableFn)
    return db->closeTableFn(db, tableId, flags);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList)
{
  assert(db);
  assert(db->refCount);
  if (db->getTableNameListFn)
    return db->getTableNameListFn(db, tableNameList);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_QuerySubmit(AQDB_DB *db,
                        AQDB_ID tableId,
                        AQDB_ID refQueryId,
                        const char *queryExpression,
                        uint32_t flags,
                        AQDB_ID *pQueryId)
{
  assert(db);
  assert(db->refCount);
  if (db->querySubmitFn)
    return db->querySubmitFn(db, AQDB_ACCESS_FLAGS_ISWRITER, tableId,
                             refQueryId, queryExpression, flags,
                             pQueryId);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_QuerySubmitWithAccessFlags(AQDB_DB *db,
                                       uint32_t accessFlags,
                                       AQDB_ID tableId,
                                       AQDB_ID refQueryId,
                                       const char *queryExpression,
                                       uint32_t flags,
                                       AQDB_ID *pQueryId)
{
  assert(db);
  assert(db->refCount);
  if (db->querySubmitFn)
    return db->querySubmitFn(db, accessFlags, tableId,
                             refQueryId, queryExpression, flags,
                             pQueryId);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_QueryGetFirst(AQDB_DB *db,
                          AQDB_ID queryId,
                          AQDB_ID *pIdListPtr,
                          int *pIdListLen)
{
  assert(db);
  assert(db->refCount);
  if (db->queryGetFirstFn)
    return db->queryGetFirstFn(db, queryId,
                               pIdListPtr, pIdListLen);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_QueryGetNext(AQDB_DB *db,
                         AQDB_ID queryId,
                         AQDB_ID *pIdListPtr,
                         int *pIdListLen)
{
  assert(db);
  assert(db->refCount);
  if (db->queryGetNextFn)
    return db->queryGetNextFn(db, queryId,
                              pIdListPtr, pIdListLen);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_QueryClose(AQDB_DB *db, AQDB_ID queryId)
{
  assert(db);
  assert(db->refCount);
  if (db->queryCloseFn)
    return db->queryCloseFn(db, queryId);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_BeginEdit(AQDB_DB *db, uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->beginEditFn)
    return db->beginEditFn(db, flags, "unknown");
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_BeginEditWithUserName(AQDB_DB *db, uint32_t flags, const char *userName)
{
  assert(db);
  assert(db->refCount);
  if (db->beginEditFn)
    return db->beginEditFn(db, flags, userName);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_EndEdit(AQDB_DB *db, uint32_t flags)
{
  assert(db);
  assert(db->refCount);
  if (db->endEditFn)
    return db->endEditFn(db, flags);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}




int AQDB_DB_GetRollbackSessions(AQDB_DB *db, AQDB_SESSION_DESCR_LIST *sl)
{
  assert(db);
  assert(db->refCount);
  if (db->getRollbackSessionsFn)
    return db->getRollbackSessionsFn(db, AQDB_ACCESS_FLAGS_ISWRITER, sl);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



int AQDB_DB_UndoSession(AQDB_DB *db, AQDB_ID sid)
{
  assert(db);
  assert(db->refCount);
  if (db->undoSessionFn)
    return db->undoSessionFn(db, sid);
  else
    return GWEN_ERROR_NOT_IMPLEMENTED;
}



AQDB_COLUMN_LIST *AQDB_DB_GetColumnList(AQDB_DB *db, AQDB_ID tableId)
{
  assert(db);
  assert(db->refCount);
  if (db->getColumnListFn)
    return db->getColumnListFn(db, AQDB_ACCESS_FLAGS_ISWRITER, tableId);
  else
    return NULL;
}






AQDB_DB_OPEN_FN AQDB_DB_SetOpenFn(AQDB_DB *db,
                                  AQDB_DB_OPEN_FN f)
{
  AQDB_DB_OPEN_FN of;

  assert(db);
  assert(db->refCount);
  of=db->openFn;
  db->openFn=f;
  return of;
}



AQDB_DB_CREATE_FN AQDB_DB_SetCreateFn(AQDB_DB *db,
                                      AQDB_DB_CREATE_FN f)
{
  AQDB_DB_CREATE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->createFn;
  db->createFn=f;
  return of;
}



AQDB_DB_CLOSE_FN AQDB_DB_SetCloseFn(AQDB_DB *db,
                                    AQDB_DB_CLOSE_FN f)
{
  AQDB_DB_CLOSE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->closeFn;
  db->closeFn=f;
  return of;
}



AQDB_DB_CHECK_FN AQDB_DB_SetCheckFn(AQDB_DB *db, AQDB_DB_CHECK_FN f)
{
  AQDB_DB_CHECK_FN of;

  assert(db);
  assert(db->refCount);
  of=db->checkFn;
  db->checkFn=f;
  return of;
}



AQDB_DB_READOBJECT_FN AQDB_DB_SetReadObjectFn(AQDB_DB *db,
                                              AQDB_DB_READOBJECT_FN f)
{
  AQDB_DB_READOBJECT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->readObjectFn;
  db->readObjectFn=f;
  return of;
}



AQDB_DB_WRITEOBJECT_FN AQDB_DB_SetWriteObjectFn(AQDB_DB *db,
                                                AQDB_DB_WRITEOBJECT_FN f)
{
  AQDB_DB_WRITEOBJECT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->writeObjectFn;
  db->writeObjectFn=f;
  return of;
}



AQDB_DB_DELETEOBJECT_FN AQDB_DB_SetDeleteObjectFn(AQDB_DB *db,
                                                  AQDB_DB_DELETEOBJECT_FN f)
{
  AQDB_DB_DELETEOBJECT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->deleteObjectFn;
  db->deleteObjectFn=f;
  return of;
}



AQDB_DB_ADDOBJECT_FN AQDB_DB_SetAddObjectFn(AQDB_DB *db,
                                            AQDB_DB_ADDOBJECT_FN f)
{
  AQDB_DB_ADDOBJECT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->addObjectFn;
  db->addObjectFn=f;
  return of;
}



AQDB_DB_OPENTABLE_FN AQDB_DB_SetOpenTableFn(AQDB_DB *db,
                                            AQDB_DB_OPENTABLE_FN f)
{
  AQDB_DB_OPENTABLE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->openTableFn;
  db->openTableFn=f;
  return of;
}



AQDB_DB_CREATETABLE_FN AQDB_DB_SetCreateTableFn(AQDB_DB *db,
                                                AQDB_DB_CREATETABLE_FN f)
{
  AQDB_DB_CREATETABLE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->createTableFn;
  db->createTableFn=f;
  return of;
}



AQDB_DB_CLOSETABLE_FN AQDB_DB_SetCloseTableFn(AQDB_DB *db,
                                              AQDB_DB_CLOSETABLE_FN f)
{
  AQDB_DB_CLOSETABLE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->closeTableFn;
  db->closeTableFn=f;
  return of;
}



AQDB_DB_GETTABLENAMELIST_FN AQDB_DB_SetGetTableNameListFn(AQDB_DB *db, AQDB_DB_GETTABLENAMELIST_FN f)
{
  AQDB_DB_GETTABLENAMELIST_FN of;

  assert(db);
  assert(db->refCount);
  of=db->getTableNameListFn;
  db->getTableNameListFn=f;
  return of;
}



AQDB_DB_BEGINEDIT_FN AQDB_DB_SetBeginEditFn(AQDB_DB *db,
                                            AQDB_DB_BEGINEDIT_FN f)
{
  AQDB_DB_BEGINEDIT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->beginEditFn;
  db->beginEditFn=f;
  return of;
}



AQDB_DB_ENDEDIT_FN AQDB_DB_SetEndEditFn(AQDB_DB *db,
                                        AQDB_DB_ENDEDIT_FN f)
{
  AQDB_DB_ENDEDIT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->endEditFn;
  db->endEditFn=f;
  return of;
}



AQDB_DB_QUERYSUBMIT_FN AQDB_DB_SetQuerySubmitFn(AQDB_DB *db,
                                                AQDB_DB_QUERYSUBMIT_FN f)
{
  AQDB_DB_QUERYSUBMIT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->querySubmitFn;
  db->querySubmitFn=f;
  return of;
}



AQDB_DB_QUERYGETFIRST_FN AQDB_DB_SetQueryGetFirstFn(AQDB_DB *db,
                                                    AQDB_DB_QUERYGETFIRST_FN f)
{
  AQDB_DB_QUERYGETFIRST_FN of;

  assert(db);
  assert(db->refCount);
  of=db->queryGetFirstFn;
  db->queryGetFirstFn=f;
  return of;
}



AQDB_DB_QUERYGETNEXT_FN AQDB_DB_SetQueryGetNextFn(AQDB_DB *db,
                                                  AQDB_DB_QUERYGETNEXT_FN f)
{
  AQDB_DB_QUERYGETNEXT_FN of;

  assert(db);
  assert(db->refCount);
  of=db->queryGetNextFn;
  db->queryGetNextFn=f;
  return of;
}



AQDB_DB_QUERYCLOSE_FN AQDB_DB_SetQueryCloseFn(AQDB_DB *db,
                                              AQDB_DB_QUERYCLOSE_FN f)
{
  AQDB_DB_QUERYCLOSE_FN of;

  assert(db);
  assert(db->refCount);
  of=db->queryCloseFn;
  db->queryCloseFn=f;
  return of;
}



AQDB_DB_GETROLLBACKSESSIONS_FN AQDB_DB_SetGetRollbackSessionsFn(AQDB_DB *db,
                                                                AQDB_DB_GETROLLBACKSESSIONS_FN f)
{
  AQDB_DB_GETROLLBACKSESSIONS_FN of;

  assert(db);
  assert(db->refCount);
  of=db->getRollbackSessionsFn;
  db->getRollbackSessionsFn=f;
  return of;
}



AQDB_DB_UNDOSESSION_FN AQDB_DB_SetUndoSessionFn(AQDB_DB *db,
                                                AQDB_DB_UNDOSESSION_FN f)
{
  AQDB_DB_UNDOSESSION_FN of;

  assert(db);
  assert(db->refCount);
  of=db->undoSessionFn;
  db->undoSessionFn=f;
  return of;
}



AQDB_DB_GETCOLUMNLIST_FN AQDB_DB_SetGetColumnListFn(AQDB_DB *db,
                                                    AQDB_DB_GETCOLUMNLIST_FN f)
{
  AQDB_DB_GETCOLUMNLIST_FN of;

  assert(db);
  assert(db->refCount);
  of=db->getColumnListFn;
  db->getColumnListFn=f;
  return of;
}



AQDB_DB_OPEN_FN AQDB_DB_GetOpenFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->openFn;
}



AQDB_DB_CREATE_FN AQDB_DB_GetCreateFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->createFn;
}



AQDB_DB_CLOSE_FN AQDB_DB_GetCloseFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->closeFn;
}



AQDB_DB_CHECK_FN AQDB_DB_GetCheckFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->checkFn;
}



AQDB_DB_READOBJECT_FN AQDB_DB_GetReadObjectFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->readObjectFn;
}



AQDB_DB_WRITEOBJECT_FN AQDB_DB_GetWriteObjectFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->writeObjectFn;
}



AQDB_DB_DELETEOBJECT_FN AQDB_DB_GetDeleteObjectFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->deleteObjectFn;
}



AQDB_DB_ADDOBJECT_FN AQDB_DB_GetAddObjectFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->addObjectFn;
}



AQDB_DB_OPENTABLE_FN AQDB_DB_GetOpenTableFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->openTableFn;
}



AQDB_DB_CREATETABLE_FN AQDB_DB_GetCreateTableFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->createTableFn;
}



AQDB_DB_CLOSETABLE_FN AQDB_DB_GetCloseTableFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->closeTableFn;
}



AQDB_DB_GETTABLENAMELIST_FN AQDB_DB_GetGetTableNameListFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->getTableNameListFn;
}



AQDB_DB_BEGINEDIT_FN AQDB_DB_GetBeginEditFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->beginEditFn;
}



AQDB_DB_ENDEDIT_FN AQDB_DB_GetEndEditFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->endEditFn;
}



AQDB_DB_QUERYSUBMIT_FN AQDB_DB_GetQuerySubmitFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->querySubmitFn;
}



AQDB_DB_QUERYGETFIRST_FN AQDB_DB_GetQueryGetFirstFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->queryGetFirstFn;
}



AQDB_DB_QUERYGETNEXT_FN AQDB_DB_GetQueryGetNextFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->queryGetNextFn;
}



AQDB_DB_QUERYCLOSE_FN AQDB_DB_GetQueryCloseFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->queryCloseFn;
}



AQDB_DB_GETROLLBACKSESSIONS_FN AQDB_DB_GetGetRollbackSessionsFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->getRollbackSessionsFn;
}



AQDB_DB_UNDOSESSION_FN AQDB_DB_GetUndoSessionFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->undoSessionFn;
}



AQDB_DB_GETCOLUMNLIST_FN AQDB_DB_GetGetColumnListFn(AQDB_DB *db)
{
  assert(db);
  assert(db->refCount);
  return db->getColumnListFn;
}





int AQDB_DB_SetCache(AQDB_DB *db, uint32_t maxEntries, uint32_t maxSize)
{
  assert(db);

  db->cacheHintMaxSize=maxSize;
  db->cacheHintMaxEntries=maxEntries;

  if (db->cache)
    AQDB_Cache_free(db->cache);
  db->cache=AQDB_Cache_new(maxEntries, maxSize);

  return 0;
}



int AQDB_DB_CacheGetObject(AQDB_DB *db,
                           AQDB_ID tableId,
                           AQDB_ID objectId,
                           AQDB_OBJECT **pObject)
{
  if (db->cache) {
    int rv;

    rv=AQDB_Cache_GetObject(db->cache, tableId, objectId, pObject);
    if (rv<0 && rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return rv;
  }

  return GWEN_ERROR_NOT_FOUND;
}



int AQDB_DB_CacheSetObject(AQDB_DB *db, const AQDB_OBJECT *object, uint32_t osize)
{
  if (db->cache) {
    int rv;

    rv=AQDB_Cache_SetObject(db->cache, object, osize);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  return 0;
}



void AQDB_DB_CachePurgeObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId)
{
  if (db->cache)
    AQDB_Cache_PurgeObject(db->cache, tableId, objectId);
}





int AQDB_DB_ForEveryQueryObject(AQDB_DB *db,
                                AQDB_ID tableId,
                                AQDB_ID refQueryId,
                                const char *queryExpression,
                                uint32_t flags,
                                AQDB_DB_FOR_EVERY_QUERY_OBJECT_FN fn,
                                void *data,
                                void **pResult)
{
  AQDB_ID qid=0;
  int rv;
  AQDB_ID oids[16];
  int len;

  assert(db);

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Bad table id %llu",
              (unsigned long long) tableId);
    return GWEN_ERROR_INVALID;
  }

  /* query */
  rv=AQDB_DB_QuerySubmit(db,
                         tableId,
                         refQueryId,
                         queryExpression,
                         flags, /* query flags */
                         &qid);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* browse all objects */
  len=16;
  rv=AQDB_DB_QueryGetFirst(db, qid, oids, &len);
  while (rv==0) {
    int i;

    for (i=0; i<len; i++) {
      AQDB_OBJECT *o=NULL;
      void *result=NULL;

      rv=AQDB_DB_ReadObject(db,
                            tableId,
                            oids[i],
                            &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error reading object %d (%d)", oids[i], rv);
        AQDB_DB_QueryClose(db, qid);
        return rv;
      }

      result=fn(db, tableId, o, data);
      AQDB_Object_free(o);
      if (result!=NULL) {
        AQDB_DB_QueryClose(db, qid);
        if (pResult)
          *pResult=result;
        return 0;
      }
    }

    len=16;
    rv=AQDB_DB_QueryGetNext(db, qid, oids, &len);
  }
  AQDB_DB_QueryClose(db, qid);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}










