/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "aqdb_object_p.h"
#include <aqdatabase/aqdb_value.h>

#include <gwenhywfar/buffer.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/gwentime.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/base64.h>

#include <inttypes.h>



GWEN_LIST_FUNCTIONS(AQDB_OBJECT, AQDB_Object)
GWEN_LIST2_FUNCTIONS(AQDB_OBJECT, AQDB_Object)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static int _setFieldFromStringInt(AQDB_OBJECT *o, uint32_t idx, const char *data);
static int _setFieldFromStringBin(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier);
static int _setFieldFromStringDate(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier);
static int _setFieldFromStringTime(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier);
static int _setFieldFromStringAmount(AQDB_OBJECT *o, uint32_t idx, const char *data);

static int _getFieldAsStringInt(AQDB_OBJECT *o, uint32_t idx, GWEN_BUFFER *destBuf);
static int _getFieldAsString(AQDB_OBJECT *o, uint32_t idx, GWEN_BUFFER *destBuf);
static int _getFieldAsStringBin(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf);
static int _getFieldAsStringDate(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf);
static int _getFieldAsStringTime(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf);
static int _getFieldAsStringAmount(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf);



/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */



AQDB_OBJECT_FIELD *AQDB_ObjectField_new(const uint8_t *p, uint32_t len)
{
  AQDB_OBJECT_FIELD *field;

  field=(AQDB_OBJECT_FIELD *) GWEN_Memory_malloc(sizeof(AQDB_OBJECT_FIELD)+len);
  assert(field);
  memset(field, 0, sizeof(AQDB_OBJECT_FIELD)+len);

  field->refCount=1;
  field->fieldLen=len;
  field->type=AQDB_DataType_None;

  if (len && p)
    memmove(field->dataBegin, p, len);

  return field;
}



void AQDB_ObjectField_free(AQDB_OBJECT_FIELD *field)
{
  if (field) {
    assert(field->refCount);
    if (field->refCount==1) {
      AQDB_ObjectField_ClearTypedPointer(field);
      field->refCount=0;
      GWEN_Memory_dealloc((void *) field);
    }
    else {
      field->refCount--;
    }
  }
}



void AQDB_ObjectField_ClearTypedPointer(AQDB_OBJECT_FIELD *field)
{
  assert(field);
  assert(field->refCount);
  if (field->typedPointer!=NULL) {
    switch (field->type) {
    case AQDB_DataType_Date:
      GWEN_Date_free((GWEN_DATE *)(field->typedPointer));
      break;
    case AQDB_DataType_Timestamp:
      GWEN_Time_free((GWEN_TIME *)(field->typedPointer));
      break;
    case AQDB_DataType_Amount:
      AQDB_Value_free((AQDB_VALUE *)(field->typedPointer));
      break;
    default:
      break;
    }
    field->typedPointer=NULL;
    field->type=AQDB_DataType_None;
  }
}



void AQDB_ObjectField_Attach(AQDB_OBJECT_FIELD *field)
{
  assert(field);
  assert(field->refCount);
  field->refCount++;
}



AQDB_OBJECT_FIELD *AQDB_ObjectField_dup(const AQDB_OBJECT_FIELD *field)
{
  AQDB_OBJECT_FIELD *nf;

  assert(field);
  assert(field->refCount);

  nf=AQDB_ObjectField_new(field->dataBegin, field->fieldLen);
  if (field->typedPointer!=NULL) {
    nf->type=field->type;

    switch (field->type) {
    case AQDB_DataType_Date:
      nf->typedPointer=GWEN_Date_dup((GWEN_DATE *)(field->typedPointer));
      break;
    case AQDB_DataType_Timestamp:
      nf->typedPointer=GWEN_Time_dup((GWEN_TIME *)(field->typedPointer));
      break;
    case AQDB_DataType_Amount:
      nf->typedPointer=AQDB_Value_dup((AQDB_VALUE *)(field->typedPointer));
      break;

    default:
      break;
    }
  }

  return nf;
}



AQDB_OBJECT_SHARED *AQDB_ObjectShared_new(AQDB_ID tableId,
                                          AQDB_ID objectId,
                                          uint32_t nfields)
{
  AQDB_OBJECT_SHARED *so;
  uint32_t flen;

  assert(nfields>0);
  flen=nfields*sizeof(AQDB_OBJECT_FIELD *);

  so=(AQDB_OBJECT_SHARED *) GWEN_Memory_malloc(sizeof(AQDB_OBJECT_SHARED)+flen);
  assert(so);
  memset(so, 0, sizeof(AQDB_OBJECT_SHARED)+flen);

  so->refCount=1;
  so->tableId=tableId;
  so->objectId=objectId;
  so->fieldCount=nfields;

  return so;
}



void AQDB_ObjectShared_Attach(AQDB_OBJECT_SHARED *so)
{
  assert(so);
  assert(so->refCount);
  so->refCount++;
}




void AQDB_ObjectShared_free(AQDB_OBJECT_SHARED *so)
{
  if (so) {
    assert(so->refCount);
    if (so->refCount==1) {
      int i;

      /* clear field translations and fields */
      for (i=0; i<so->fieldCount; i++) {
        if (so->fieldPointers[i])
          AQDB_ObjectField_free(so->fieldPointers[i]);
        so->fieldPointers[i]=NULL;
      }

      GWEN_Memory_dealloc((void *) so);
    }
    else
      so->refCount--;
  }
}



AQDB_OBJECT_SHARED *AQDB_ObjectShared_dup(const AQDB_OBJECT_SHARED *so)
{
  AQDB_OBJECT_SHARED *nso;
  int i;

  assert(so);
  assert(so->refCount);

  nso=AQDB_ObjectShared_new(so->tableId, so->objectId, so->fieldCount);

  /* copy field translations and fields */
  for (i=0; i<so->fieldCount; i++) {
    if (so->fieldPointers[i]) {
      AQDB_ObjectField_Attach(so->fieldPointers[i]);
      nso->fieldPointers[i]=so->fieldPointers[i];
    }
  }

  return nso;
}



AQDB_OBJECT_SHARED *AQDB_ObjectShared_deepCopy(const AQDB_OBJECT_SHARED *so)
{
  AQDB_OBJECT_SHARED *nso;
  int i;

  assert(so);
  assert(so->refCount);

  nso=AQDB_ObjectShared_new(so->tableId, so->objectId, so->fieldCount);

  /* copy field translations and fields */
  for (i=0; i<so->fieldCount; i++) {
    if (so->fieldPointers[i])
      nso->fieldPointers[i]=AQDB_ObjectField_dup(so->fieldPointers[i]);
    else
      nso->fieldPointers[i]=NULL;
  }

  return nso;
}









AQDB_OBJECT *AQDB_Object_new(AQDB_ID tableId,
                             AQDB_ID objectId,
                             uint32_t nfields)
{
  AQDB_OBJECT *o;

  GWEN_NEW_OBJECT(AQDB_OBJECT, o);
  o->refCount=1;
  GWEN_LIST_INIT(AQDB_OBJECT, o);

  o->shared=AQDB_ObjectShared_new(tableId, objectId, nfields);

  return o;
}



AQDB_OBJECT *AQDB_Object_dup(const AQDB_OBJECT *o)
{
  AQDB_OBJECT *no;

  GWEN_NEW_OBJECT(AQDB_OBJECT, no);
  no->refCount=1;
  GWEN_LIST_INIT(AQDB_OBJECT, no);

  AQDB_ObjectShared_Attach(o->shared);
  no->shared=o->shared;

  return no;
}


AQDB_OBJECT *AQDB_Object_deepCopy(const AQDB_OBJECT *o)
{
  AQDB_OBJECT *no;

  GWEN_NEW_OBJECT(AQDB_OBJECT, no);
  no->refCount=1;
  GWEN_LIST_INIT(AQDB_OBJECT, no);

  no->shared=AQDB_ObjectShared_deepCopy(o->shared);

  return no;
}



void AQDB_Object_Attach(AQDB_OBJECT *o)
{
  assert(o);
  assert(o->refCount);
  o->refCount++;
}




void AQDB_Object_free(AQDB_OBJECT *o)
{
  if (o) {
    assert(o->refCount);
    if ((--o->refCount)==0) {
      GWEN_LIST_FINI(AQDB_OBJECT, o);
      AQDB_ObjectShared_free(o->shared);
      GWEN_FREE_OBJECT(o);
    }
  }
}




int AQDB_Object_GetField(const AQDB_OBJECT *o,
                         uint32_t idx,
                         uint8_t **pData,
                         uint32_t *pLen)
{
  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    *pData=NULL;
    *pLen=0;
    return 0;
  }
  else {
    if (o->shared->fieldPointers[idx]) {
      *pData=o->shared->fieldPointers[idx]->dataBegin;
      *pLen=o->shared->fieldPointers[idx]->fieldLen;
    }
    else {
      *pData=NULL;
      *pLen=0;
    }
    return 0;
  }
}



int AQDB_Object_SetField(AQDB_OBJECT *o,
                         uint32_t idx,
                         const uint8_t *data,
                         uint32_t len)
{
  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }
  else {
    /* create a private copy of the shared data */
    if (o->shared->refCount>1) {
      AQDB_OBJECT_SHARED *nso;

      nso=AQDB_ObjectShared_dup(o->shared);
      AQDB_ObjectShared_free(o->shared);
      o->shared=nso;
    }

    if (o->shared->fieldPointers[idx])
      AQDB_ObjectField_free(o->shared->fieldPointers[idx]);
    o->shared->fieldPointers[idx]=AQDB_ObjectField_new(data, len);
    return 0;
  }
}




AQDB_ID AQDB_Object_GetObjectId(const AQDB_OBJECT *o)
{
  assert(o);
  assert(o->refCount);
  return o->shared->objectId;
}



void AQDB_Object_SetObjectId(AQDB_OBJECT *o, AQDB_ID objectId)
{
  assert(o);
  assert(o->refCount);
  o->shared->objectId=objectId;
}



AQDB_ID AQDB_Object_GetTableId(const AQDB_OBJECT *o)
{
  assert(o);
  assert(o->refCount);
  return o->shared->tableId;
}



void AQDB_Object_SetTableId(AQDB_OBJECT *o, AQDB_ID objectId)
{
  assert(o);
  assert(o->refCount);
  o->shared->tableId=objectId;
}



uint32_t AQDB_Object_GetFieldCount(const AQDB_OBJECT *o)
{
  assert(o);
  assert(o->refCount);
  return o->shared->fieldCount;
}



int AQDB_Object_GetFieldInt64(const AQDB_OBJECT *o,
                              uint32_t idx,
                              int64_t *pVal)
{
  uint8_t *d;
  uint32_t l;
  int64_t result=0;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l>sizeof(int64_t)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Too many bytes for type (%d)", l);
    return GWEN_ERROR_BAD_DATA;
  }

  while (l--) {
    result<<=8;
    result+=*(d++);
  }

  *pVal=result;
  return 0;
}



int AQDB_Object_GetFieldInt32(const AQDB_OBJECT *o,
                              uint32_t idx,
                              int32_t *pVal)
{
  uint8_t *d;
  uint32_t l;
  int32_t result=0;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l>sizeof(int32_t)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Too many bytes for type (%d)", l);
    return GWEN_ERROR_BAD_DATA;
  }

  while (l--) {
    result<<=8;
    result+=*(d++);
  }

  *pVal=result;
  return 0;
}



int AQDB_Object_GetFieldDate(AQDB_OBJECT *o,
                             uint32_t idx,
                             const GWEN_DATE **pVal)
{
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Date) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
        return rv;
      }

      if (l) {
        GWEN_DATE *ti;

        ti=GWEN_Date_fromString((const char *) d);
        if (ti) {
          o->shared->fieldPointers[idx]->type=AQDB_DataType_Date;
          o->shared->fieldPointers[idx]->typedPointer=(void *) ti;
        }
      }
    }

    *pVal=(const GWEN_DATE *)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldDate(AQDB_OBJECT *o,
                             uint32_t idx,
                             const GWEN_DATE *val)
{
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  if (val) {
    rv=AQDB_Object_SetFieldString(o, idx, GWEN_Date_GetString(val));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    rv=AQDB_Object_SetField(o, idx, NULL, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int AQDB_Object_GetFieldTimestamp(AQDB_OBJECT *o,
                                  uint32_t idx,
                                  const GWEN_TIME **pVal)
{
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Timestamp) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
        return rv;
      }

      if (l) {
        GWEN_TIME *ti;

        ti=GWEN_Time_fromString((const char *) d, "YYYYMMDD-hhmmss");
        if (ti) {
          o->shared->fieldPointers[idx]->type=AQDB_DataType_Timestamp;
          o->shared->fieldPointers[idx]->typedPointer=(void *) ti;
        }
      }
    }

    *pVal=(const GWEN_TIME *)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldTimestamp(AQDB_OBJECT *o,
                                  uint32_t idx,
                                  const GWEN_TIME *val)
{
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  if (val) {
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 16, 0, 1);
    rv=GWEN_Time_toString(val, "YYYYMMDD-hhmmss", tbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(tbuf);
      return rv;
    }

    rv=AQDB_Object_SetField(o, idx,
                            (const uint8_t *) GWEN_Buffer_GetStart(tbuf),
                            GWEN_Buffer_GetUsedBytes(tbuf)+1);
    GWEN_Buffer_free(tbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    rv=AQDB_Object_SetField(o, idx, NULL, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int AQDB_Object_GetFieldValue(AQDB_OBJECT *o,
                              uint32_t idx,
                              const AQDB_VALUE **pVal)
{
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
             idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Amount) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
        return rv;
      }

      if (l) {
        AQDB_VALUE *v;

        v=AQDB_Value_fromString((const char *) d);
        if (v) {
          o->shared->fieldPointers[idx]->type=AQDB_DataType_Amount;
          o->shared->fieldPointers[idx]->typedPointer=(void *) v;
        }
      }
    }

    *pVal=(const AQDB_VALUE *)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldValue(AQDB_OBJECT *o,
                              uint32_t idx,
                              const AQDB_VALUE *val)
{
  int rv;

  /* TODO: this should be improved to makeuse of the GWEN_BUFFER unnecessary */
  if (val) {
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 64, 0, 1);
    AQDB_Value_toString(val, tbuf);
    rv=AQDB_Object_SetFieldString(o, idx, GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_free(tbuf);
  }
  else
    rv=AQDB_Object_SetFieldString(o, idx, NULL);

  return rv;
}




int AQDB_Object_SetFieldInt64(AQDB_OBJECT *o,
                              uint32_t idx,
                              int64_t val)
{
  uint8_t buf[8];
  uint8_t *p;
  int rv;
  int i;

  assert(o);
  assert(o->refCount);

  buf[7]=val & 0xff;
  val>>=8;
  buf[6]=val & 0xff;
  val>>=8;
  buf[5]=val & 0xff;
  val>>=8;
  buf[4]=val & 0xff;
  val>>=8;
  buf[3]=val & 0xff;
  val>>=8;
  buf[2]=val & 0xff;
  val>>=8;
  buf[1]=val & 0xff;
  val>>=8;
  buf[0]=val & 0xff;

  p=buf;
  for (i=0; i<7; i++) {
    if (*p!=0)
      break;
    p++;
  }

  rv=AQDB_Object_SetField(o, idx, p, 8-i);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_SetFieldInt32(AQDB_OBJECT *o,
                              uint32_t idx,
                              int32_t val)
{
  uint8_t buf[4];
  uint8_t *p;
  int rv;
  int i;

  assert(o);
  assert(o->refCount);

  buf[3]=val & 0xff;
  val>>=8;
  buf[2]=val & 0xff;
  val>>=8;
  buf[1]=val & 0xff;
  val>>=8;
  buf[0]=val & 0xff;

  p=buf;
  for (i=0; i<3; i++) {
    if (*p!=0)
      break;
    p++;
  }

  rv=AQDB_Object_SetField(o, idx, p, 4-i);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_SetFieldString(AQDB_OBJECT *o,
                               uint32_t idx,
                               const char *s)
{
  int rv;

  assert(o);
  assert(o->refCount);

  if (!s)
    s="";
  rv=AQDB_Object_SetField(o, idx, (const uint8_t *)s, strlen(s)+1);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_GetFieldString(const AQDB_OBJECT *o,
                               uint32_t idx,
                               const char **pVal)
{
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l)
    *pVal=(const char *)d;
  else
    *pVal="";
  return 0;
}



void AQDB_Object_Dump(const AQDB_OBJECT *o, int indent)
{
  int i;
  int j;

  for (i=0; i<indent; i++)
    fprintf(stderr, " ");
  fprintf(stderr, "Object:\n");
  for (i=0; i<indent+2; i++)
    fprintf(stderr, " ");
  fprintf(stderr, "Table Id  : %08x\n", o->shared->tableId);
  for (i=0; i<indent+2; i++)
    fprintf(stderr, " ");
  fprintf(stderr, "Object Id : %08x\n", o->shared->objectId);
  for (i=0; i<indent+2; i++)
    fprintf(stderr, " ");
  fprintf(stderr, "Field num : %d\n", o->shared->fieldCount);

  for (i=0; i<indent+2; i++)
    fprintf(stderr, " ");
  fprintf(stderr, "Fields\n");
  for (j=0; j<o->shared->fieldCount; j++) {
    uint8_t *data;
    uint32_t len;
    int rv;

    rv=AQDB_Object_GetField(o, j, &data, &len);
    if (rv<0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d: %d", j, rv);
    }
    else {
      for (i=0; i<indent+4; i++)
        fprintf(stderr, " ");
      fprintf(stderr, "Field %02d:\n", j);
      GWEN_Text_DumpString((const char *) data, len, indent+6);
    }
  }
}



int AQDB_Object_SetFieldFromStringByType(AQDB_OBJECT *o, uint32_t idx, AQDB_DATA_TYPE t, const char *data,
                                         const char *modifier)
{
  switch (t) {
  case AQDB_DataType_Int:
    return _setFieldFromStringInt(o, idx, data);
  case AQDB_DataType_String:
    return AQDB_Object_SetFieldString(o, idx, data);
  case AQDB_DataType_Bin:
    return _setFieldFromStringBin(o, idx, data, modifier);
  case AQDB_DataType_Date:
    return _setFieldFromStringDate(o, idx, data, modifier);
  case AQDB_DataType_Timestamp:
    return _setFieldFromStringTime(o, idx, data, modifier);
  case AQDB_DataType_Amount:
    return _setFieldFromStringAmount(o, idx, data);
  case AQDB_DataType_Unknown:
  case AQDB_DataType_None:
    break;
  }
  DBG_ERROR(AQDB_LOGDOMAIN, "Unhandled data type %d (%s)", t, AQDB_DataType_toString(t));
  return GWEN_ERROR_INVALID;
}



int _setFieldFromStringInt(AQDB_OBJECT *o, uint32_t idx, const char *data)
{
  if (data && *data) {
#ifdef SCNi64
    int64_t i=0;

    if (sscanf(data, "%"SCNi64, &i)==1)
      return AQDB_Object_SetFieldInt64(o, idx, i);
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "Not a number [%s]", data);
      return GWEN_ERROR_BAD_DATA;
    }
#else
    long int i=0;

    if (sscanf(data, "%li", &i)==1)
      return AQDB_Object_SetFieldInt64(o, idx, (int64_t) i);
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "Not a number [%s]", data);
      return GWEN_ERROR_BAD_DATA;
    }
#endif
  }
  else
    return AQDB_Object_SetField(o, idx, NULL, 0);
}



int _setFieldFromStringDate(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier)
{
  if (data && *data) {
    GWEN_DATE *d;
    int rv;

    if (!(modifier && *modifier))
      d=GWEN_Date_fromString(data);
    else
      d=GWEN_Date_fromStringWithTemplate(data, modifier);
    if (d==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid date spec [%s]", data);
      return GWEN_ERROR_BAD_DATA;
    }

    rv=AQDB_Object_SetFieldDate(o, idx, d);
    GWEN_Date_free(d);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return rv;
  }
  else
    return AQDB_Object_SetField(o, idx, NULL, 0);
}



int _setFieldFromStringTime(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier)
{
  if (data && *data) {
    GWEN_TIME *ti;
    int rv;

    if (!(modifier && *modifier))
      modifier="YYYYMMDD-hh:mm:ss";
    ti=GWEN_Time_fromString(data, modifier);
    if (ti==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid time spec [%s]", data);
      return GWEN_ERROR_BAD_DATA;
    }
    rv=AQDB_Object_SetFieldTimestamp(o, idx, ti);
    GWEN_Time_free(ti);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return rv;
  }
  else
    return AQDB_Object_SetField(o, idx, NULL, 0);
}



int _setFieldFromStringBin(AQDB_OBJECT *o, uint32_t idx, const char *data, const char *modifier)
{
  if (data && *data) {
    GWEN_BUFFER *dbuf=NULL;
    int rv;

    dbuf=GWEN_Buffer_new(0, 256, 0, 1);
    if (!(modifier && *modifier))
      modifier="hex";
    if (strcasecmp(modifier, "hex")==0) {
      rv=GWEN_Text_FromHexBuffer(data, dbuf);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "here (%d)", rv);
        GWEN_Buffer_free(dbuf);
        return rv;
      }
    }
    else if (strcasecmp(modifier, "base64")==0) {
      rv=GWEN_Base64_Decode((const uint8_t *) data, 0, dbuf);
      if (rv<0) {
        DBG_ERROR(AQDB_LOGDOMAIN, "here (%d)", rv);
        GWEN_Buffer_free(dbuf);
        return rv;
      }
    }
    else {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid modifier (%s)", modifier);
      GWEN_Buffer_free(dbuf);
      return GWEN_ERROR_INVALID;
    }

    rv=AQDB_Object_SetField(o, idx, (const uint8_t *) GWEN_Buffer_GetStart(dbuf), GWEN_Buffer_GetUsedBytes(dbuf));
    GWEN_Buffer_free(dbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return rv;
  }
  else
    return AQDB_Object_SetField(o, idx, NULL, 0);
}



int _setFieldFromStringAmount(AQDB_OBJECT *o, uint32_t idx, const char *data)
{
  if (data && *data) {
    AQDB_VALUE *v;
    int rv;

    v=AQDB_Value_fromString(data);
    if (v==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid amount spec [%s]", data);
      return GWEN_ERROR_BAD_DATA;
    }

    rv=AQDB_Object_SetFieldValue(o, idx, v);
    AQDB_Value_free(v);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return rv;
  }
  else
    return AQDB_Object_SetField(o, idx, NULL, 0);
}



int AQDB_Object_GetFieldAsStringByType(AQDB_OBJECT *o,
                                       uint32_t idx,
                                       AQDB_DATA_TYPE t,
                                       const char *modifier,
                                       GWEN_BUFFER *destBuf)
{
  switch (t) {
  case AQDB_DataType_Int:
    return _getFieldAsStringInt(o, idx, destBuf);
  case AQDB_DataType_String:
    return _getFieldAsString(o, idx, destBuf);
  case AQDB_DataType_Bin:
    return _getFieldAsStringBin(o, idx, modifier, destBuf);
  case AQDB_DataType_Date:
    return _getFieldAsStringDate(o, idx, modifier, destBuf);
  case AQDB_DataType_Timestamp:
    return _getFieldAsStringTime(o, idx, modifier, destBuf);
  case AQDB_DataType_Amount:
    return _getFieldAsStringAmount(o, idx, modifier, destBuf);
  case AQDB_DataType_Unknown:
  case AQDB_DataType_None:
    break;
  }
  DBG_ERROR(AQDB_LOGDOMAIN, "Unhandled data type %d (%s)", t, AQDB_DataType_toString(t));
  return GWEN_ERROR_INVALID;

}



int _getFieldAsStringInt(AQDB_OBJECT *o, uint32_t idx, GWEN_BUFFER *destBuf)
{
  int rv;
  int64_t val=0;

  rv=AQDB_Object_GetFieldInt64(o, idx, &val);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
#ifdef PRId64
  rv=GWEN_Buffer_AppendArgs(destBuf, "%"PRId64, val);
#else
  rv=GWEN_Buffer_AppendArgs(destBuf, "%ld", (long int) val);
#endif
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _getFieldAsString(AQDB_OBJECT *o, uint32_t idx, GWEN_BUFFER *destBuf)
{
  int rv;
  const char *val=NULL;

  rv=AQDB_Object_GetFieldString(o, idx, &val);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  if (val)
    GWEN_Buffer_AppendString(destBuf, val);
  return 0;
}



int _getFieldAsStringBin(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf)
{
  int rv;
  uint8_t *valuePointer=NULL;
  uint32_t valueLength=0;

  rv=AQDB_Object_GetField(o, idx, &valuePointer, &valueLength);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  if (valuePointer && valueLength) {
    if (!(modifier && *modifier))
      modifier="hex";

    if (strcasecmp(modifier, "hex")==0)
      rv=GWEN_Text_ToHexBuffer((const char *)valuePointer, valueLength, destBuf, 0, 0, 0);
    else if (strcasecmp(modifier, "base64")==0)
      rv=GWEN_Base64_Encode((const unsigned char *)valuePointer, valueLength, destBuf, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int _getFieldAsStringDate(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf)
{
  int rv;
  const GWEN_DATE *val=NULL;

  rv=AQDB_Object_GetFieldDate(o, idx, &val);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  if (val) {
    if (modifier && *modifier)
      rv=GWEN_Date_toStringWithTemplate(val, modifier, destBuf);
    else
      rv=GWEN_Buffer_AppendString(destBuf, GWEN_Date_GetString(val));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int _getFieldAsStringTime(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf)
{
  int rv;
  const GWEN_TIME *val=NULL;

  rv=AQDB_Object_GetFieldTimestamp(o, idx, &val);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  if (val) {
    if (!(modifier && *modifier))
      modifier="YYYYMMDD-hh:mm:ss";
    rv=GWEN_Time_toString(val, modifier, destBuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int _getFieldAsStringAmount(AQDB_OBJECT *o, uint32_t idx, const char *modifier, GWEN_BUFFER *destBuf)
{
  int rv;
  const AQDB_VALUE *val=NULL;
  int precision=2;

  rv=AQDB_Object_GetFieldValue(o, idx, &val);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  if (modifier && *modifier) {
    if (1!=sscanf(modifier, "%d", &precision)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Bad modifer (not a number)");
    }
  }

  if (val)
    AQDB_Value_toHumanReadableString(val, destBuf, 2);

  return 0;
}







