/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2009-2010 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqdb_query_p.h"

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>

#include <ctype.h>


GWEN_INHERIT_FUNCTIONS(AQDB_QUERY)




AQDB_QUERY_EXPR *AQDB_Query_Expr_new(AQDB_QUERY *q)
{
  AQDB_QUERY_EXPR *expr;

  GWEN_NEW_OBJECT(AQDB_QUERY_EXPR, expr);
  expr->query=q;
  expr->fieldIndex=-1;

  return expr;
}



AQDB_QUERY_EXPR *AQDB_Query_Expr_Data_new(AQDB_QUERY *q, const char *ptr, uint32_t len)
{
  AQDB_QUERY_EXPR *expr;

  expr=AQDB_Query_Expr_new(q);
  expr->op=AQDB_Query_Expr_OperatorData;
  if (len) {
    expr->dataPtr=(char *)malloc(len+1);
    if (expr->dataPtr==NULL) {
      AQDB_Query_Expr_free(expr);
      return NULL;
    }
    memmove(expr->dataPtr, ptr, len);
    expr->dataPtr[len]=0;
  }

  return expr;
}



AQDB_QUERY_EXPR *AQDB_Query_Expr_Field_new(AQDB_QUERY *q, int idx, AQDB_DATA_TYPE t)
{
  AQDB_QUERY_EXPR *expr;

  expr=AQDB_Query_Expr_new(q);
  expr->op=AQDB_Query_Expr_OperatorField;
  expr->fieldIndex=idx;
  expr->resultType=t;

  return expr;
}




void AQDB_Query_Expr_free(AQDB_QUERY_EXPR *expr)
{
  if (expr) {
    AQDB_Query_Expr_free(expr->left);
    AQDB_Query_Expr_free(expr->right);

    switch (expr->resultType) {
    case AQDB_DataType_Int:
    case AQDB_DataType_String:
    case AQDB_DataType_Bin:
      break;
    case AQDB_DataType_Date:
      GWEN_Date_free(expr->typedPtr.date);
      expr->typedPtr.date=NULL;
      break;
    case AQDB_DataType_Timestamp:
      GWEN_Time_free(expr->typedPtr.timestamp);
      expr->typedPtr.timestamp=NULL;
      break;
    case AQDB_DataType_Amount:
      AQDB_Value_free(expr->typedPtr.value);
      expr->typedPtr.value=NULL;
      break;
    case AQDB_DataType_Unknown:
    case AQDB_DataType_None:
      break;
    }

    free(expr->dataPtr);
    expr->dataPtr=NULL;
    GWEN_FREE_OBJECT(expr);
  }
}



int AQDB_Query_GetFieldData(AQDB_QUERY *q, AQDB_QUERY_EXPR *expr)
{
  return q->getFieldDataFn(q, expr);
}



AQDB_QUERY_EXPR *AQDB_Query_MkFieldExpr(AQDB_QUERY *q, const char *name)
{
  return q->mkFieldExprFn(q, name);
}




int AQDB_Query_Expr__readOperand(AQDB_QUERY *q,
                                 const char **pp,
                                 AQDB_QUERY_EXPR **pExp)
{
  AQDB_QUERY_EXPR *expr;
  const char *p;
  const char *pStart;
  int len;

  p=*pp;
  while (*p && isspace(*p))
    p++;

  if (!*p)
    return GWEN_ERROR_EOF;

  /* quoted argument */
  if (*p=='"') {
    p++;
    pStart=p;
    while (*p && *p!='"')
      p++;
    if (!*p) {
      DBG_INFO(AQDB_LOGDOMAIN, "Missing quotation mark at the end");
      *pp=p;
      return GWEN_ERROR_BAD_DATA;
    }

    len=p-pStart;
    expr=AQDB_Query_Expr_Data_new(q, pStart, len);
    expr->resultType=AQDB_DataType_String;
    p++; /* skip quotation mark */
    *pp=p;
    *pExp=expr;
    return 0;
  }
  else if (*p=='$') {
    char *fieldName;

    /* field name */
    p++;
    pStart=p;
    while (*p && isalnum(*p))
      p++;

    len=p-pStart;

    fieldName=(char *)malloc(len+1);
    assert(fieldName);
    memmove(fieldName, pStart, len);
    fieldName[len]=0;

    expr=AQDB_Query_MkFieldExpr(q, fieldName);
    if (expr==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "Field \"%s\" not found", fieldName);
      free(fieldName);
      return GWEN_ERROR_INVALID;
    }
    free(fieldName);

    *pp=p;
    *pExp=expr;
    return 0;
  }
  else if (*p=='(') {
    int rv;

    /* skip opening bracket */
    p++;
    /* read full expression */
    rv=AQDB_Query_Expr__readExpression(q, &p, &expr);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      *pp=p;
      return rv;
    }

    /* read closing bracket */
    while (*p && isspace(*p))
      p++;
    if (*p!=')') {
      DBG_INFO(AQDB_LOGDOMAIN, "Missing closing bracket");
      AQDB_Query_Expr_free(expr);
      *pp=p;
      return GWEN_ERROR_BAD_DATA;
    }
    p++;
    *pp=p;
    *pExp=expr;
    return 0;
  }
  else {
    /* unquoted argument */
    pStart=p;
    while (*p && isalnum(*p))
      p++;

    *pp=p;
    len=p-pStart;
    expr=AQDB_Query_Expr_Data_new(q, pStart, len);
    *pExp=expr;
    return 0;
  }
}



int AQDB_Query_Expr__transformData(AQDB_QUERY_EXPR *expr,
                                   AQDB_DATA_TYPE resultType)
{
  switch (resultType) {
  case AQDB_DataType_Int: {
    register int64_t td=0;
    register const uint8_t *p;
    register uint8_t c;

    p=(const uint8_t *)(expr->dataPtr);
    if (*p) {
      if (*p=='0' && p[1]=='x') {
        /* hexa-decimcal */
        p+=2;
        while ((c=*p)) {
          c-='0';
          if (c>9)
            c-=7;
          if (c>15) {
            DBG_INFO(AQDB_LOGDOMAIN, "Not an integer value");
            return GWEN_ERROR_BAD_DATA;
          }
          td<<=4;
          td+=c;
          p++;
        }
      }
      else {
        /* decimal */
        while ((c=*p)) {
          c-='0';
          if (c>9) {
            DBG_INFO(AQDB_LOGDOMAIN, "Not an integer value");
            return GWEN_ERROR_BAD_DATA;
          }
          td*=10;
          td+=c;
          p++;
        }
      }
    }

    expr->intData=td;
    expr->resultType=AQDB_DataType_Int;
    break;
  }

  case AQDB_DataType_String:
    expr->resultType=AQDB_DataType_String;
    break;

  case AQDB_DataType_Bin: {
    GWEN_BUFFER *buf;
    int rv;

    buf=GWEN_Buffer_new(0, 64, 0, 1);
    rv=GWEN_Text_FromHexBuffer(expr->dataPtr, buf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(buf);
      return rv;
    }
    free(expr->dataPtr);
    expr->dataPtr=(char *)malloc(GWEN_Buffer_GetUsedBytes(buf));
    assert(expr->dataPtr);
    memmove(expr->dataPtr,
            GWEN_Buffer_GetStart(buf),
            GWEN_Buffer_GetUsedBytes(buf));
    expr->dataLen=GWEN_Buffer_GetUsedBytes(buf);
    GWEN_Buffer_free(buf);
    expr->resultType=AQDB_DataType_Bin;
    break;
  }

  case AQDB_DataType_Date: {
    GWEN_DATE *ti;

    ti=GWEN_Date_fromString(expr->dataPtr);
    if (expr->dataPtr && ti==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid date string [%s]", expr->dataPtr);
      return GWEN_ERROR_INVALID;
    }
    expr->typedPtr.date=ti;
    expr->resultType=AQDB_DataType_Date;
    break;
  }

  case AQDB_DataType_Timestamp: {
    GWEN_TIME *ti;

    ti=GWEN_Time_fromString((const char *)(expr->dataPtr), "YYYYMMDD-hhmmss");
    if (expr->dataPtr && ti==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid date string [%s]", expr->dataPtr);
      return GWEN_ERROR_INVALID;
    }
    expr->typedPtr.timestamp=ti;
    expr->resultType=AQDB_DataType_Timestamp;
    break;
  }

  case AQDB_DataType_Amount: {
    AQDB_VALUE *v;

    v=AQDB_Value_fromString(expr->dataPtr);
    if (v==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "Bad data (%s)", expr->dataPtr);
      return GWEN_ERROR_BAD_DATA;
    }
    expr->typedPtr.value=v;
    expr->resultType=AQDB_DataType_Amount;
    break;
  }

  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled type %d", resultType);
    return GWEN_ERROR_INVALID;
  }

  return 0;
}



int AQDB_Query_Expr__readExpression(AQDB_QUERY *q,
                                    const char **pp,
                                    AQDB_QUERY_EXPR **pExp)
{
  AQDB_QUERY_EXPR *expr;
  const char *p;
  int rv;

  p=*pp;

  while (*p && isspace(*p))
    p++;

  if (!*p)
    return GWEN_ERROR_EOF;

  expr=AQDB_Query_Expr_new(q);

  /* read left operand */
  rv=AQDB_Query_Expr__readOperand(q, &p, &(expr->left));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Query_Expr_free(expr);
    *pp=p;
    return rv;
  }

  while (*p && isspace(*p))
    p++;

  if (!*p) {
    AQDB_QUERY_EXPR *leftExpr;

    /* expression only consists of the left part */
    leftExpr=expr->left;
    expr->left=NULL;
    AQDB_Query_Expr_free(expr);
    *pExp=leftExpr;
    return 0;
  }

  /* two-byte operators */
  if (strncmp(p, "<=", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorLesserThanOrEqual;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  else if (strncmp(p, ">=", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorGreaterThanOrEqual;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  else if (strncmp(p, "==", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorEqual;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  else if (strncmp(p, "!=", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorNotEqual;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  else if (strncmp(p, "||", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorOr;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  else if (strncmp(p, "&&", 2)==0) {
    expr->op=AQDB_Query_Expr_OperatorAnd;
    expr->resultType=AQDB_DataType_Int;
    p+=2;
  }
  /* one-byte operators */
  else if (*p=='<') {
    expr->op=AQDB_Query_Expr_OperatorLesserThan;
    expr->resultType=AQDB_DataType_Int;
    p++;
  }
  else if (*p=='>') {
    expr->op=AQDB_Query_Expr_OperatorGreaterThan;
    expr->resultType=AQDB_DataType_Int;
    p++;
  }
  else if (*p=='^') {
    expr->op=AQDB_Query_Expr_OperatorXor;
    expr->resultType=AQDB_DataType_Int;
    p++;
  }
  else if (*p=='?') {
    char *functionName;
    const char *pStart;
    int len;

    /* function name */
    p++;
    pStart=p;
    while (*p && isalnum(*p))
      p++;

    len=p-pStart;

    functionName=(char *)malloc(len+1);
    assert(functionName);
    memmove(functionName, pStart, len);
    functionName[len]=0;

    if (strcasecmp(functionName, "contains")==0) {
      expr->op=AQDB_Query_Expr_OperatorContains;
      expr->resultType=AQDB_DataType_Int;
    }
    else if (strcasecmp(functionName, "matches")==0) {
      expr->op=AQDB_Query_Expr_OperatorMatches;
      expr->resultType=AQDB_DataType_Int;
    }
    else if (strcasecmp(functionName, "isset")==0) {
      expr->op=AQDB_Query_Expr_OperatorIsSet;
      expr->resultType=AQDB_DataType_Int;
    }
    else if (strcasecmp(functionName, "areset")==0) {
      expr->op=AQDB_Query_Expr_OperatorAreSet;
      expr->resultType=AQDB_DataType_Int;
    }
    else if (strcasecmp(functionName, "isclear")==0) {
      expr->op=AQDB_Query_Expr_OperatorIsClear;
      expr->resultType=AQDB_DataType_Int;
    }
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "Unexpected function [%s]", functionName);
      AQDB_Query_Expr_free(expr);
      *pp=p;
      return GWEN_ERROR_BAD_DATA;
    }

    free(functionName);
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "Unexpected character");
    AQDB_Query_Expr_free(expr);
    *pp=p;
    return GWEN_ERROR_BAD_DATA;
  }

  /* read right operand */
  while (*p && isspace(*p))
    p++;

  if (!*p)
    return GWEN_ERROR_EOF;

  rv=AQDB_Query_Expr__readOperand(q, &p, &(expr->right));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Query_Expr_free(expr);
    *pp=p;
    return rv;
  }

  *pp=p;

  /* perform some checks */
  if (expr->left && expr->right) {
    /* transform data if necessary */
    if (expr->left->op==AQDB_Query_Expr_OperatorField) {
      if (expr->right->op==AQDB_Query_Expr_OperatorData) {
        rv=AQDB_Query_Expr__transformData(expr->right, expr->left->resultType);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          AQDB_Query_Expr_free(expr);
          return rv;
        }
      }
    }
    else if (expr->right->op==AQDB_Query_Expr_OperatorField) {
      if (expr->left->op==AQDB_Query_Expr_OperatorData) {
        rv=AQDB_Query_Expr__transformData(expr->left, expr->right->resultType);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          AQDB_Query_Expr_free(expr);
          return rv;
        }
      }
    }

    /* check for same result types on both sides */
    if (expr->left->resultType!=expr->right->resultType) {
      DBG_INFO(AQDB_LOGDOMAIN,
               "Left and right have different types (%d!=%d)",
               expr->left->resultType,
               expr->right->resultType);
      AQDB_Query_Expr_Dump(expr->left, stderr, 2);
      AQDB_Query_Expr_Dump(expr->right, stderr, 2);

      AQDB_Query_Expr_free(expr);
      return GWEN_ERROR_INVALID;
    }
  }

  *pExp=expr;
  return 0;
}



int AQDB_Query_Expr_fromString(AQDB_QUERY *q, const char *s, AQDB_QUERY_EXPR **pExpr)
{
  AQDB_QUERY_EXPR *expr;
  int rv;
  const char *p;

  p=s;
  rv=AQDB_Query_Expr__readExpression(q, &p, &expr);
  if (rv) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  *pExpr=expr;

  return 0;
}



int AQDB_Query_Expr_CalculateInt(AQDB_QUERY_EXPR *expr)
{
  int64_t v1, v2;
  int result=0;

  v1=expr->left->intData;
  v2=expr->right->intData;
  switch (expr->op) {
  case AQDB_Query_Expr_OperatorEqual:
    result=(v1==v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    result=(v1==v2)?0:1;
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    result=(v1>v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    result=(v1>=v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    result=(v1<v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    result=(v1<=v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorContains:
  case AQDB_Query_Expr_OperatorMatches:
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid operator %d for this result type", expr->op);
    return GWEN_ERROR_INVALID;
  case AQDB_Query_Expr_OperatorIsSet:
    result=((v1 & v2)==0)?0:1;
    break;
  case AQDB_Query_Expr_OperatorAreSet:
    result=((v1 & v2)==v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorIsClear:
    result=((v1 & v2)==0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorOr:
    result=(v1|v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorAnd:
    result=(v1&v2)?1:0;
    break;
  case AQDB_Query_Expr_OperatorXor:
    result=(v1^v2)?1:0;
    break;
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  expr->intData=result;

  return 0;
}



const char *AQDB_Query_Expr_StrCaseStr(const char *haystack, const char *needle)
{
  while (*haystack) {
    while (*haystack && tolower(*haystack)!=tolower(*needle))
      haystack++;
    if (*haystack) {
      const char *t;
      const char *s;

      /* now haystack points to an area which begins with *needle */
      t=haystack;
      s=needle;
      t++;
      s++;
      while (*t && *s && (tolower(*t)==tolower(*s))) {
        t++;
        s++;
      }
      if (*s==0)
        return haystack;
    }
    else
      /* not found */
      break;
    /* no match here, advance haystack */
    haystack++;
  }

  /* not found */
  return NULL;
}



int AQDB_Query_Expr_CalculateString(AQDB_QUERY_EXPR *expr)
{
  const char *v1, *v2;
  int result=0;

  v1=expr->left->dataPtr;
  if (v1==NULL)
    v1="";
  v2=expr->right->dataPtr;
  if (v2==NULL)
    v2="";
  switch (expr->op) {
  case AQDB_Query_Expr_OperatorEqual:
    result=(GWEN_Text_ComparePattern(v1, v2, 0)==-1)?0:1;
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    result=(GWEN_Text_ComparePattern(v1, v2, 0)==-1)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    result=(strcasecmp(v1, v2)>0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    result=(strcasecmp(v1, v2)>=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    result=(strcasecmp(v1, v2)<0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    result=(strcasecmp(v1, v2)<=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorContains:
    result=(AQDB_Query_Expr_StrCaseStr(v1, v2)!=NULL);
    break;
  case AQDB_Query_Expr_OperatorMatches:
    result=(GWEN_Text_ComparePattern(v1, v2, 0)!=-1);
    break;
  case AQDB_Query_Expr_OperatorIsSet:
  case AQDB_Query_Expr_OperatorAreSet:
  case AQDB_Query_Expr_OperatorIsClear:
  case AQDB_Query_Expr_OperatorOr:
  case AQDB_Query_Expr_OperatorAnd:
  case AQDB_Query_Expr_OperatorXor:
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  expr->intData=result;

  return 0;
}



int AQDB_Query_Expr_CalculateBin(AQDB_QUERY_EXPR *expr)
{
  DBG_INFO(AQDB_LOGDOMAIN,
           "Binary operations not yet supported");
  return GWEN_ERROR_NOT_SUPPORTED;
}



int AQDB_Query_Expr_CalculateValue(AQDB_QUERY_EXPR *expr)
{
  const AQDB_VALUE *v1, *v2;
  int result=0;

  v1=expr->left->typedPtr.value;
  v2=expr->right->typedPtr.value;
  switch (expr->op) {
  case AQDB_Query_Expr_OperatorEqual:
    result=(AQDB_Value_Compare(v1, v2)==0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    result=(AQDB_Value_Compare(v1, v2)==0)?0:1;
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    result=(AQDB_Value_Compare(v1, v2)>0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    result=(AQDB_Value_Compare(v1, v2)>=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    result=(AQDB_Value_Compare(v1, v2)<0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    result=(AQDB_Value_Compare(v1, v2)<=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorOr:
  case AQDB_Query_Expr_OperatorAnd:
  case AQDB_Query_Expr_OperatorXor:
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  expr->intData=result;

  return 0;
}



int AQDB_Query_Expr_CalculateDate(AQDB_QUERY_EXPR *expr)
{
  const GWEN_DATE *t1, *t2;
  int result=0;

  t1=expr->left->typedPtr.date;
  t2=expr->right->typedPtr.date;
  switch (expr->op) {
  case AQDB_Query_Expr_OperatorEqual:
    result=(GWEN_Date_Compare(t1, t2)==0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    result=(GWEN_Date_Compare(t1, t2)==0)?0:1;
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    result=(GWEN_Date_Compare(t1, t2)>0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    result=(GWEN_Date_Compare(t1, t2)>=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    result=(GWEN_Date_Compare(t1, t2)<0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    result=(GWEN_Date_Compare(t1, t2)<=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorOr:
  case AQDB_Query_Expr_OperatorAnd:
  case AQDB_Query_Expr_OperatorXor:
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  expr->intData=result;

  return 0;
}



int AQDB_Query_Expr_CalculateTimestamp(AQDB_QUERY_EXPR *expr)
{
  const GWEN_TIME *t1, *t2;
  int result=0;

  t1=expr->left->typedPtr.timestamp;
  t2=expr->right->typedPtr.timestamp;
  switch (expr->op) {
  case AQDB_Query_Expr_OperatorEqual:
    result=(GWEN_Time_Compare(t1, t2)==0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    result=(GWEN_Time_Compare(t1, t2)==0)?0:1;
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    result=(GWEN_Time_Compare(t1, t2)>0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    result=(GWEN_Time_Compare(t1, t2)>=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    result=(GWEN_Time_Compare(t1, t2)<0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    result=(GWEN_Time_Compare(t1, t2)<=0)?1:0;
    break;
  case AQDB_Query_Expr_OperatorOr:
  case AQDB_Query_Expr_OperatorAnd:
  case AQDB_Query_Expr_OperatorXor:
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  expr->intData=result;

  return 0;
}



int AQDB_Query_Expr_CalculateField(AQDB_QUERY_EXPR *expr)
{
  return AQDB_Query_GetFieldData(expr->query, expr);
}



int AQDB_Query_Expr__Calculate(AQDB_QUERY_EXPR *expr)
{
  int rv;

  /* calculate left */
  if (expr->left) {
    rv=AQDB_Query_Expr__Calculate(expr->left);
    if (rv) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  /* calculate right */
  if (expr->right) {
    rv=AQDB_Query_Expr__Calculate(expr->right);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  if (expr->op==AQDB_Query_Expr_OperatorData) {
    /* nothing to do */
  }
  else if (expr->op==AQDB_Query_Expr_OperatorField) {
    rv=AQDB_Query_Expr_CalculateField(expr);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else if (expr->op>=AQDB_Query_Expr_OperatorEqual) {
    switch (expr->left->resultType) {
    case AQDB_DataType_Int:
      rv=AQDB_Query_Expr_CalculateInt(expr);
      break;
    case AQDB_DataType_String:
      rv=AQDB_Query_Expr_CalculateString(expr);
      break;
    case AQDB_DataType_Date:
      rv=AQDB_Query_Expr_CalculateDate(expr);
      break;
    case AQDB_DataType_Timestamp:
      rv=AQDB_Query_Expr_CalculateTimestamp(expr);
      break;
    case AQDB_DataType_Bin:
      rv=AQDB_Query_Expr_CalculateBin(expr);
      break;
    case AQDB_DataType_Amount:
      rv=AQDB_Query_Expr_CalculateValue(expr);
      break;
    default:
      DBG_INFO(AQDB_LOGDOMAIN,
               "Unhandled type %d",
               expr->left->resultType);
      return GWEN_ERROR_INVALID;
    }

    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
    return GWEN_ERROR_INVALID;
  }

  return 0;
}



int AQDB_Query_Expr_Calculate(AQDB_QUERY_EXPR *expr)
{
  int rv;

  rv=AQDB_Query_Expr__Calculate(expr);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  if (expr->resultType!=AQDB_DataType_Int) {
    DBG_INFO(AQDB_LOGDOMAIN,
             "Unexpected result type for expression (%d)",
             expr->resultType);
    return GWEN_ERROR_INVALID;
  }

  return (int)(expr->intData);
}



void AQDB_Query_Expr_Dump(const AQDB_QUERY_EXPR *expr, FILE *f, int indent)
{
  int i;

  for (i=0; i<indent; i++)
    fprintf(f, " ");

  switch (expr->op) {
  case AQDB_Query_Expr_OperatorData:
    fprintf(f, "DATA ");
    if (expr->resultType==AQDB_DataType_Int)
      fprintf(f, "%lld\n", (long long int) expr->intData);
    else if (expr->resultType==AQDB_DataType_String ||
             expr->resultType==AQDB_DataType_Date ||
             expr->resultType==AQDB_DataType_Amount)
      fprintf(f, "\"%s\"\n", expr->dataPtr);
    else
      fprintf(f, "Type %d\n", expr->resultType);
    break;
  case AQDB_Query_Expr_OperatorField:
    fprintf(f, "Field %d (Type %d)\n", expr->fieldIndex, expr->resultType);
    break;
  case AQDB_Query_Expr_OperatorEqual:
    fprintf(f, "==\n");
    break;
  case AQDB_Query_Expr_OperatorNotEqual:
    fprintf(f, "!=\n");
    break;
  case AQDB_Query_Expr_OperatorGreaterThan:
    fprintf(f, ">\n");
    break;
  case AQDB_Query_Expr_OperatorGreaterThanOrEqual:
    fprintf(f, ">=\n");
    break;
  case AQDB_Query_Expr_OperatorLesserThan:
    fprintf(f, "<\n");
    break;
  case AQDB_Query_Expr_OperatorLesserThanOrEqual:
    fprintf(f, "<=\n");
    break;
  case AQDB_Query_Expr_OperatorOr:
    fprintf(f, "||\n");
    break;
  case AQDB_Query_Expr_OperatorAnd:
    fprintf(f, "&&\n");
    break;
  case AQDB_Query_Expr_OperatorXor:
    fprintf(f, "^\n");
    break;
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unhandled operator %d", expr->op);
  }

  if (expr->left)
    AQDB_Query_Expr_Dump(expr->left, f, indent+2);
  if (expr->right)
    AQDB_Query_Expr_Dump(expr->right, f, indent+2);
}







AQDB_QUERY *AQDB_Query_new()
{
  AQDB_QUERY *q;

  GWEN_NEW_OBJECT(AQDB_QUERY, q);
  GWEN_INHERIT_INIT(AQDB_QUERY, q);

  return q;
}



void AQDB_Query_free(AQDB_QUERY *q)
{
  if (q) {
    GWEN_INHERIT_FINI(AQDB_QUERY, q);
    AQDB_Query_Expr_free(q->expression);
    GWEN_FREE_OBJECT(q);
  }
}



int AQDB_Query_SetString(AQDB_QUERY *q, const char *s)
{
  int rv;

  rv=AQDB_Query_Expr_fromString(q, s, &(q->expression));
  if (rv<0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error parsing query string [%s]: (%d)", s, rv);
    return rv;
  }
  assert(q->expression);

  return 0;
}



int AQDB_Query_Calculate(AQDB_QUERY *q)
{
  if (q->expression) {
    return AQDB_Query_Expr_Calculate(q->expression);
  }
  else
    return GWEN_ERROR_INVALID;
}



void AQDB_Query_SetMkFieldExprFn(AQDB_QUERY *q, AQDB_QUERY_MKFIELDEXPR_FN f)
{
  q->mkFieldExprFn=f;
}



void AQDB_Query_SetGetFieldDataFn(AQDB_QUERY *q, AQDB_QUERY_GETFIELDDATA_FN f)
{
  q->getFieldDataFn=f;
}






