/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqdb_value_p.h"

#include <gwenhywfar/misc.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/buffer.h>

#include <assert.h>
#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif


#define AQDB_VALUE_STRSIZE 256



AQDB_VALUE *AQDB_Value_new()
{
  AQDB_VALUE *v;

  GWEN_NEW_OBJECT(AQDB_VALUE, v);
  mpq_init(v->value);
  return v;
}



void AQDB_Value_free(AQDB_VALUE *v)
{
  if (v) {
    mpq_clear(v->value);
    GWEN_FREE_OBJECT(v);
  }
}



AQDB_VALUE *AQDB_Value_dup(const AQDB_VALUE *ov)
{
  AQDB_VALUE *v;

  assert(ov);
  v=AQDB_Value_new();
  mpq_set(v->value, ov->value);

  return v;
}



AQDB_VALUE *AQDB_Value_fromDouble(double i)
{
  GWEN_BUFFER *nbuf;
  AQDB_VALUE *v;
  int rv;

  nbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=GWEN_Text_DoubleToBuffer(i, nbuf);
  assert(rv==0);
  v=AQDB_Value_fromString(GWEN_Buffer_GetStart(nbuf));
  GWEN_Buffer_free(nbuf);
  return v;
}



AQDB_VALUE *AQDB_Value_fromString(const char *s)
{
  AQDB_VALUE *v;
  int isNeg=0;
  char *t;
  int conversion_succeeded=1;
  char *tmpString=NULL;

  while (*s && *s<33)
    s++;
  if (*s=='-') {
    isNeg=1;
    s++;
  }
  else if (*s=='+') {
    s++;
  }

  tmpString=strdup(s);

  v=AQDB_Value_new();
  t=strchr(tmpString, ',');
  if (t)
    *t='.';
  t=strchr(tmpString, '.');
  if (t) {
    // remove comma and calculate denominator
    unsigned long denominator = 1;
    char *next;
    do {
      next=t+1;
      *t=*next;
      if (*next != 0)
        denominator *= 10;
      t++;
    }
    while (*next);

    // set denominator to the calculated value
    mpz_set_ui(mpq_denref(v->value), denominator);

    // set numerator to the resulting integer string without comma
    if (mpz_set_str(mpq_numref(v->value), tmpString, 10)==-1) {
      conversion_succeeded = 0;
    }
  }
  else {
    conversion_succeeded=(mpq_set_str(v->value, tmpString, 10)==0);
  }
  free(tmpString);

  if (!conversion_succeeded) {
    DBG_ERROR(AQDB_LOGDOMAIN, "[%s] is not a valid value", s);
    AQDB_Value_free(v);
    return NULL;
  }

  /* canonicalize */
  mpq_canonicalize(v->value);

  if (isNeg)
    mpq_neg(v->value, v->value);

  return v;
}



void AQDB_Value_toString(const AQDB_VALUE *v, GWEN_BUFFER *buf)
{
  int rv;
  uint32_t size;
  char *p;

  assert(v);
  GWEN_Buffer_AllocRoom(buf, AQDB_VALUE_STRSIZE);
  p=GWEN_Buffer_GetPosPointer(buf);
  size=GWEN_Buffer_GetMaxUnsegmentedWrite(buf);
  rv=gmp_snprintf(p, size, "%Qi", v->value);
  assert(rv<size);
  GWEN_Buffer_IncrementPos(buf, rv+1);
  GWEN_Buffer_AdjustUsedBytes(buf);
}



void AQDB_Value_toHumanReadableString(const AQDB_VALUE *v,
                                      GWEN_BUFFER *buf,
                                      int prec)
{
  char numbuf[128];
  double num;
  int rv;
#ifdef HAVE_SETLOCALE
  const char *orig_locale = setlocale(LC_NUMERIC, NULL);
  char *currentLocale = strdup(orig_locale ? orig_locale : "C");
  setlocale(LC_NUMERIC, "C");
#endif

  num=AQDB_Value_GetValueAsDouble(v);
  rv=snprintf(numbuf, sizeof(numbuf), "%.*f",
              prec?prec:2, num);

#ifdef HAVE_SETLOCALE
  setlocale(LC_NUMERIC, currentLocale);
  free(currentLocale);
#endif

  if (rv<1 || rv>=sizeof(numbuf)) {
    assert(0);
  }
  GWEN_Buffer_AppendString(buf, numbuf);
}



void AQDB_Value_toLocalHumanReadableString(const AQDB_VALUE *v,
                                           GWEN_BUFFER *buf,
                                           int prec)
{
  char numbuf[128];
  double num;
  int rv;

  num=AQDB_Value_GetValueAsDouble(v);
  rv=snprintf(numbuf, sizeof(numbuf), "%.*f",
              prec?prec:2, num);

  if (rv<1 || rv>=sizeof(numbuf)) {
    assert(0);
  }
  GWEN_Buffer_AppendString(buf, numbuf);
}



double AQDB_Value_GetValueAsDouble(const AQDB_VALUE *v)
{
  assert(v);
  if (mpz_fits_slong_p(mpq_numref(v->value)) && mpz_fits_slong_p(mpq_denref(v->value))) {
    return (double)(mpz_get_d(mpq_numref(v->value)) / mpz_get_d(mpq_denref(v->value)));
  }
  else {
    return mpq_get_d(v->value);
  }
}



void AQDB_Value_SetValueFromDouble(AQDB_VALUE *v, double i)
{
  assert(v);
  mpq_set_d(v->value, i);
}



void AQDB_Value_SetZero(AQDB_VALUE *v)
{
  assert(v);
  mpq_clear(v->value);
  mpq_init(v->value);
}



int AQDB_Value_IsZero(const AQDB_VALUE *v)
{
  assert(v);
  return (mpq_sgn(v->value)==0);
}



int AQDB_Value_IsNegative(const AQDB_VALUE *v)
{
  assert(v);
  return (mpq_sgn(v->value)<0);
}



int AQDB_Value_IsPositive(const AQDB_VALUE *v)
{
  assert(v);
  return (mpq_sgn(v->value)>=0);
}



int AQDB_Value_Compare(const AQDB_VALUE *v1, const AQDB_VALUE *v2)
{
  assert(v1);
  assert(v2);

  return mpq_cmp(v1->value, v2->value);
}



int AQDB_Value_AddValue(AQDB_VALUE *v1, const AQDB_VALUE *v2)
{
  assert(v1);
  assert(v2);

  mpq_add(v1->value, v1->value, v2->value);
  return 0;
}



int AQDB_Value_SubValue(AQDB_VALUE *v1, const AQDB_VALUE *v2)
{
  assert(v1);
  assert(v2);
  mpq_sub(v1->value, v1->value, v2->value);
  return 0;
}



int AQDB_Value_MultValue(AQDB_VALUE *v1, const AQDB_VALUE *v2)
{
  assert(v1);
  assert(v2);

  mpq_mul(v1->value, v1->value, v2->value);
  return 0;
}



int AQDB_Value_DivValue(AQDB_VALUE *v1, const AQDB_VALUE *v2)
{
  assert(v1);
  assert(v2);

  mpq_div(v1->value, v1->value, v2->value);
  return 0;
}



int AQDB_Value_Negate(AQDB_VALUE *v)
{
  assert(v);
  mpq_neg(v->value, v->value);
  return 0;
}



void AQDB_Value_Dump(const AQDB_VALUE *v, FILE *f, unsigned int indent)
{
  unsigned int i;

  for (i=0; i<indent; i++)
    fprintf(f, " ");
  fprintf(f, "Value: ");
  if (v) {
    GWEN_BUFFER *nbuf;

    nbuf=GWEN_Buffer_new(0, 128, 0, 1);
    AQDB_Value_toHumanReadableString(v, nbuf, 2);
    gmp_fprintf(f, "%Qi (%s)\n", v->value, GWEN_Buffer_GetStart(nbuf));
    GWEN_Buffer_free(nbuf);
  }
  else
    fprintf(f, "[none]\n");
}











