/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2011 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "aqdb_cache_p.h"

#include <aqdatabase/aqdb_object_be.h>

#include <gwenhywfar/buffer.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>



GWEN_LIST_FUNCTIONS(AQDB_CACHE_ENTRY, AQDB_CacheEntry)
GWEN_IDMAP_FUNCTIONS(AQDB_CACHE_ENTRY, AQDB_CacheEntry)

GWEN_LIST_FUNCTIONS(AQDB_CACHE_TABLE, AQDB_CacheTable)
GWEN_IDMAP_FUNCTIONS(AQDB_CACHE_TABLE, AQDB_CacheTable)



GWEN_LIST_FUNCTIONS(AQDB_CACHE, AQDB_Cache)
GWEN_LIST2_FUNCTIONS(AQDB_CACHE, AQDB_Cache)






AQDB_CACHE_ENTRY *AQDB_CacheEntry_new()
{
  AQDB_CACHE_ENTRY *ce;

  GWEN_NEW_OBJECT(AQDB_CACHE_ENTRY, ce)
  ce->refCount=1;
  GWEN_LIST_INIT(AQDB_CACHE_ENTRY, ce)

  return ce;
}



void AQDB_CacheEntry_free(AQDB_CACHE_ENTRY *ce)
{
  if (ce) {

    assert(ce->refCount);
    if (ce->refCount==1) {
      /*DBG_DEBUG(AQDB_LOGDOMAIN, "Removing cache entry after %lu hits", (unsigned long int) (ce->accessCounter));*/
      GWEN_LIST_FINI(AQDB_CACHE_ENTRY, ce)
      if (ce->objectPointer)
        AQDB_Object_free(ce->objectPointer);
      ce->objectPointer=NULL;

      ce->refCount=0;
      GWEN_FREE_OBJECT(ce);
    }
    else
      ce->refCount--;
  }
}






AQDB_CACHE_TABLE *AQDB_CacheTable_new()
{
  AQDB_CACHE_TABLE *ct;

  GWEN_NEW_OBJECT(AQDB_CACHE_TABLE, ct)
  ct->refCount=1;
  GWEN_LIST_INIT(AQDB_CACHE_TABLE, ct)

  ct->entryMap=AQDB_CacheEntry_IdMap_new(GWEN_IdMapAlgo_Hex4);

  return ct;
}



void AQDB_CacheTable_free(AQDB_CACHE_TABLE *ct)
{
  if (ct) {
    assert(ct->refCount);
    if (ct->refCount==1) {
      GWEN_LIST_FINI(AQDB_CACHE_TABLE, ct)
      AQDB_CacheEntry_IdMap_free(ct->entryMap);

      ct->refCount=0;
      GWEN_FREE_OBJECT(ct);
    }
    else
      ct->refCount--;
  }
}













AQDB_CACHE *AQDB_Cache_new(uint32_t maxEntries, uint32_t maxSize)
{
  AQDB_CACHE *cc;

  GWEN_NEW_OBJECT(AQDB_CACHE, cc)
  cc->refCount=1;
  GWEN_INHERIT_INIT(AQDB_CACHE, cc)
  GWEN_LIST_INIT(AQDB_CACHE, cc)

  cc->maxEntryCount=maxEntries;
  cc->maxTotalSize=maxSize;

  cc->tableMap=AQDB_CacheTable_IdMap_new(GWEN_IdMapAlgo_Hex4);
  cc->tableList=AQDB_CacheTable_List_new();
  cc->entryList=AQDB_CacheEntry_List_new();

  return cc;
}



void AQDB_Cache_Attach(AQDB_CACHE *cc)
{
  assert(cc);
  assert(cc->refCount);
  cc->refCount++;
}



void AQDB_Cache_free(AQDB_CACHE *cc)
{
  if (cc) {
    assert(cc->refCount);
    if (cc->refCount==1) {

      DBG_NOTICE(AQDB_LOGDOMAIN, "Cache usage: %lu MB, %lu Entries; Hits: %lu, Misses: %lu, Evicted: %lu",
                 (unsigned long int)((cc->statMaxUsedTotalSize)/(1024*1024)),
                 (unsigned long int)(cc->statMaxUsedEntries),
                 (unsigned long int)(cc->statHits),
                 (unsigned long int)(cc->statMisses),
                 (unsigned long int)(cc->statEvicted));

      GWEN_LIST_FINI(AQDB_CACHE, cc)
      GWEN_INHERIT_FINI(AQDB_CACHE, cc)

      AQDB_CacheTable_IdMap_free(cc->tableMap);
      AQDB_CacheTable_List_free(cc->tableList);

      AQDB_CacheEntry_List_free(cc->entryList);
      cc->refCount=0;
      GWEN_FREE_OBJECT(cc);
    }
    else
      cc->refCount--;
  }
}



int AQDB_Cache_GetObject(AQDB_CACHE *cc,
                         AQDB_ID tableId,
                         AQDB_ID objectId,
                         AQDB_OBJECT **pObject)
{

  AQDB_CACHE_TABLE *ct;

  assert(cc);
  assert(cc->refCount);

  ct=AQDB_CacheTable_IdMap_Find(cc->tableMap, tableId);
  if (ct) {
    AQDB_CACHE_ENTRY *ce;

    ce=AQDB_CacheEntry_IdMap_Find(ct->entryMap, objectId);
    if (ce) {
      /* entry found, move to front of list */
      AQDB_CacheEntry_List_Del(ce);
      AQDB_CacheEntry_List_Insert(ce, cc->entryList);
      ce->accessCounter++;

      cc->statHits++;

      *pObject=AQDB_Object_dup(ce->objectPointer);
      return 0;
    }
  }

  cc->statMisses++;
  return GWEN_ERROR_NOT_FOUND;
}



void AQDB_Cache_DelEntry(AQDB_CACHE *cc, AQDB_CACHE_ENTRY *ce)
{
  AQDB_CACHE_TABLE *ct;

  ct=AQDB_CacheTable_IdMap_Find(cc->tableMap, ce->tableId);
  assert(ct);

  AQDB_CacheEntry_List_Del(ce);
  AQDB_CacheEntry_IdMap_Remove(ct->entryMap, ce->objectId);
  cc->usedTotalSize-=ce->objectSize;
  AQDB_CacheEntry_free(ce);
}



int AQDB_Cache_AddEntry(AQDB_CACHE *cc, AQDB_CACHE_ENTRY *ce)
{
  AQDB_CACHE_TABLE *ct;
  uint32_t cnt;

  if ((AQDB_CacheEntry_List_GetCount(cc->entryList)+1)>=cc->maxEntryCount) {
    AQDB_CACHE_ENTRY *dce;

    /* need to free one entry */
    dce=AQDB_CacheEntry_List_Last(cc->entryList);
    assert(dce);
    /*DBG_ERROR(AQDB_LOGDOMAIN, "Removing entry %p with %d bytes", dce, dce->objectSize); */
    cc->statEvicted++;
    AQDB_Cache_DelEntry(cc, dce);
  }

  if ((cc->usedTotalSize+ce->objectSize)>=cc->maxTotalSize) {
    /* need to free at least some bytes worth of entries */
    while ((cc->usedTotalSize+ce->objectSize)>=cc->maxTotalSize) {
      if (AQDB_CacheEntry_List_GetCount(cc->entryList)<1) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Unable to free enough entries to accomodate new one");
        return GWEN_ERROR_INTERNAL;
      }
      else {
        AQDB_CACHE_ENTRY *dce;

        /* need to free one entry */
        dce=AQDB_CacheEntry_List_Last(cc->entryList);
        assert(dce);
        DBG_ERROR(AQDB_LOGDOMAIN, "Removing entry %p with %d bytes", dce, dce->objectSize);
        AQDB_Cache_DelEntry(cc, dce);
      }
    }
  }

  ct=AQDB_CacheTable_IdMap_Find(cc->tableMap, ce->tableId);
  assert(ct);

  AQDB_CacheEntry_IdMap_Insert(ct->entryMap, ce->objectId, ce);
  AQDB_CacheEntry_List_Insert(ce, cc->entryList);
  cc->usedTotalSize+=ce->objectSize;
  if (cc->usedTotalSize>cc->statMaxUsedTotalSize)
    cc->statMaxUsedTotalSize=cc->usedTotalSize;

  cnt=AQDB_CacheEntry_List_GetCount(cc->entryList);
  if (cnt>cc->statMaxUsedEntries)
    cc->statMaxUsedEntries=cnt;

  return 0;
}



int AQDB_Cache_SetObject(AQDB_CACHE *cc, const AQDB_OBJECT *object, uint32_t osize)
{
  AQDB_ID tid;
  AQDB_ID oid;
  AQDB_CACHE_TABLE *ct;
  AQDB_CACHE_ENTRY *ce;
  AQDB_OBJECT *o;
  int rv;

  assert(cc);
  assert(cc->refCount);

  o=AQDB_Object_dup(object);
  assert(o);

  tid=AQDB_Object_GetTableId(object);
  oid=AQDB_Object_GetObjectId(object);
  ct=AQDB_CacheTable_IdMap_Find(cc->tableMap, tid);
  if (ct==NULL) {
    ct=AQDB_CacheTable_new();
    ct->tableId=tid;
    AQDB_CacheTable_IdMap_Insert(cc->tableMap, tid, ct);
    AQDB_CacheTable_List_Add(ct, cc->tableList);
  }
  assert(ct);

  /* remove possibly existing entry */
  ce=AQDB_CacheEntry_IdMap_Find(ct->entryMap, oid);
  if (ce)
    AQDB_Cache_DelEntry(cc, ce);

  /* create and add new entry */
  ce=AQDB_CacheEntry_new();
  ce->tableId=tid;
  ce->objectId=oid;

  ce->objectPointer=o;
  ce->objectSize=osize;

  rv=AQDB_Cache_AddEntry(cc, ce);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_CacheEntry_free(ce);
    return rv;
  }

  return 0;
}



void AQDB_Cache_PurgeObject(AQDB_CACHE *cc, AQDB_ID tableId, AQDB_ID objectId)
{
  AQDB_CACHE_TABLE *ct;

  assert(cc);
  assert(cc->refCount);

  ct=AQDB_CacheTable_IdMap_Find(cc->tableMap, tableId);
  if (ct) {
    AQDB_CACHE_ENTRY *ce;

    ce=AQDB_CacheEntry_IdMap_Find(ct->entryMap, objectId);
    if (ce)
      AQDB_Cache_DelEntry(cc, ce);
  }
}









