/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqdb_objectquery_p.h"

#include <aqdatabase/aqdb_query_be.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>


GWEN_INHERIT(AQDB_QUERY, AQDB_OBJECT_QUERY)




static AQDB_QUERY_EXPR GWENHYWFAR_CB *AQDB_ObjectQuery_MkFieldExpr(AQDB_QUERY *q, const char *name)
{
  AQDB_OBJECT_QUERY *xq;
  AQDB_COLUMN *col;
  int idx=0;

  assert(q);
  xq=GWEN_INHERIT_GETDATA(AQDB_QUERY, AQDB_OBJECT_QUERY, q);
  assert(xq);

  col=AQDB_Column_List_First(xq->columnList);
  while (col) {
    const char *s;

    s=AQDB_Column_GetName(col);
    if (s && strcasecmp(s, name)==0)
      break;
    idx++;
    col=AQDB_Column_List_Next(col);
  }
  if (col==NULL)
    return NULL;
  return AQDB_Query_Expr_Field_new(q, idx, AQDB_Column_GetType(col));
}




static int GWENHYWFAR_CB AQDB_ObjectQuery_GetFieldData(AQDB_QUERY *q, AQDB_QUERY_EXPR *expr)
{
  AQDB_OBJECT_QUERY *xq;
  AQDB_OBJECT *o;
  int rv;

  assert(q);
  xq=GWEN_INHERIT_GETDATA(AQDB_QUERY, AQDB_OBJECT_QUERY, q);
  assert(xq);

  if (expr->dataPtr) {
    free(expr->dataPtr);
    expr->dataPtr=NULL;
    expr->dataLen=0;
  }

  o=(AQDB_OBJECT *) q->object;

  switch (expr->resultType) {
  case AQDB_DataType_Int: {
    int64_t val;

    rv=AQDB_Object_GetFieldInt64(o, expr->fieldIndex, &val);
    if (rv<0)
      return rv;
    expr->intData=val;
    break;
  }

  case AQDB_DataType_String: {
    const char *val;

    rv=AQDB_Object_GetFieldString(o, expr->fieldIndex, &val);
    if (rv<0)
      return rv;
    expr->dataPtr=strdup(val);
    expr->dataLen=strlen(val);
    break;
  }

  case AQDB_DataType_Bin: {
    uint8_t *p;
    uint32_t len;

    rv=AQDB_Object_GetField(o, expr->fieldIndex, &p, &len);
    if (rv<0)
      return rv;
    if (p && len) {
      expr->dataPtr=(char *) malloc(len);
      memmove(expr->dataPtr, p, len);
      expr->dataLen=len;
    }
    break;
  }

  case AQDB_DataType_Date: {
    const GWEN_DATE *val;

    rv=AQDB_Object_GetFieldDate(o, expr->fieldIndex, &val);
    if (rv<0)
      return rv;
    if (expr->typedPtr.date)
      GWEN_Date_free(expr->typedPtr.date);
    if (val)
      expr->typedPtr.date=GWEN_Date_dup(val);
    else
      expr->typedPtr.date=NULL;
    break;
  }

  case AQDB_DataType_Timestamp: {
    const GWEN_TIME *val;

    rv=AQDB_Object_GetFieldTimestamp(o, expr->fieldIndex, &val);
    if (rv<0)
      return rv;
    if (expr->typedPtr.timestamp)
      GWEN_Time_free(expr->typedPtr.timestamp);
    if (val)
      expr->typedPtr.timestamp=GWEN_Time_dup(val);
    else
      expr->typedPtr.timestamp=NULL;
    break;
  }

  case AQDB_DataType_Amount: {
    const AQDB_VALUE *val;

    rv=AQDB_Object_GetFieldValue(o, expr->fieldIndex, &val);
    if (rv<0)
      return rv;
    if (expr->typedPtr.value)
      AQDB_Value_free(expr->typedPtr.value);
    if (val)
      expr->typedPtr.value=AQDB_Value_dup(val);
    else
      expr->typedPtr.value=NULL;
    break;
  }

  default:
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field %d", q->expression->fieldIndex);
    assert(0);
    return GWEN_ERROR_INVALID;
  }

  return 0;
}



void GWENHYWFAR_CB AQDB_ObjectQuery_FreeData(void *bp, void *p)
{
  AQDB_OBJECT_QUERY *xq;

  xq=(AQDB_OBJECT_QUERY *) p;
  assert(xq);

  AQDB_Column_List_free(xq->columnList);
  GWEN_FREE_OBJECT(xq);
}



AQDB_QUERY *AQDB_ObjectQuery_new(const char *s, const AQDB_COLUMN_LIST *cl)
{
  AQDB_QUERY *q;
  AQDB_OBJECT_QUERY *xq;
  int rv;

  q=AQDB_Query_new();
  GWEN_NEW_OBJECT(AQDB_OBJECT_QUERY, xq);

  GWEN_INHERIT_SETDATA(AQDB_QUERY, AQDB_OBJECT_QUERY, q, xq,
                       AQDB_ObjectQuery_FreeData);
  q->object=NULL;
  if (cl)
    xq->columnList=AQDB_Column_List_dup(cl);

  AQDB_Query_SetMkFieldExprFn(q, AQDB_ObjectQuery_MkFieldExpr);
  AQDB_Query_SetGetFieldDataFn(q, AQDB_ObjectQuery_GetFieldData);
  rv=AQDB_Query_SetString(q, s);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    AQDB_Query_free(q);
    return NULL;
  }

  return q;
}



int AQDB_ObjectQuery_Check(AQDB_QUERY *q, AQDB_OBJECT *object)
{
  AQDB_OBJECT_QUERY *xq;
  int rv;

  assert(q);
  xq=GWEN_INHERIT_GETDATA(AQDB_QUERY, AQDB_OBJECT_QUERY, q);
  assert(xq);

  q->object=object;
  rv=AQDB_Query_Calculate(q);
  q->object=NULL;

  return rv;
}








