/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqdb_query_result_p.h"

#include <aqdatabase/aqdb_db_be.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/misc.h>

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>



GWEN_LIST_FUNCTIONS(AQDB_QUERY_RESULT, AQDB_QueryResult)




AQDB_QUERY_RESULT *AQDB_QueryResult_new(AQDB_ID qid)
{
  AQDB_QUERY_RESULT *qr;

  GWEN_NEW_OBJECT(AQDB_QUERY_RESULT, qr);
  qr->refCount=1;
  GWEN_LIST_INIT(AQDB_QUERY_RESULT, qr);

  qr->queryId=qid;
  qr->idList=GWEN_IdList64_newWithSteps(AQDB_QUERY_RESULT_IDLIST_ENTRIES);
  qr->it=GWEN_IdList64_Iterator_new(qr->idList);

  return qr;
}



void AQDB_QueryResult_free(AQDB_QUERY_RESULT *qr)
{
  if (qr) {
    assert(qr->refCount);
    if (--(qr->refCount)==0) {
      GWEN_LIST_FINI(AQDB_QUERY_RESULT, qr);
      GWEN_IdList64_Iterator_free(qr->it);
      GWEN_IdList64_free(qr->idList);

      GWEN_FREE_OBJECT(qr);
    }
  }
}



GWEN_IDLIST64 *AQDB_QueryResult_GetIdList(const AQDB_QUERY_RESULT *qr)
{
  assert(qr);
  assert(qr->refCount);

  return qr->idList;
}



void AQDB_QueryResult_SetIdList(AQDB_QUERY_RESULT *qr, GWEN_IDLIST64 *idl)
{
  assert(qr);
  assert(qr->refCount);

  if (qr->idList && qr->idList!=idl) {
    GWEN_IdList64_Iterator_free(qr->it);
    GWEN_IdList64_free(qr->idList);
    qr->idList=idl;
    qr->it=GWEN_IdList64_Iterator_new(qr->idList);
  }
}



uint32_t AQDB_QueryResult_GetQueryId(const AQDB_QUERY_RESULT *qr)
{
  assert(qr);
  assert(qr->refCount);
  return qr->queryId;
}



void AQDB_QueryResult_AddId(AQDB_QUERY_RESULT *qr, AQDB_ID id)
{
  assert(qr);
  assert(qr->refCount);
  GWEN_IdList64_AddId(qr->idList, id);
}



int AQDB_QueryResult_GetFirstIds(AQDB_QUERY_RESULT *qr,
                                 AQDB_ID *idListPtr,
                                 int *idListLen)
{
  int i;

  if (*idListLen<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid buffer size (%d)", *idListLen);
    return GWEN_ERROR_INVALID;
  }

  for (i=0; i<*idListLen; i++) {
    uint64_t id;

    if (i==0)
      id=GWEN_IdList64_Iterator_GetFirstId(qr->it);
    else
      id=GWEN_IdList64_Iterator_GetNextId(qr->it);
    if (id==0)
      break;
    idListPtr[i]=id;
  }

  *idListLen=i;

  if (i==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "No first id");
    return GWEN_ERROR_NOT_FOUND;
  }

  return 0;
}



int AQDB_QueryResult_GetNextIds(AQDB_QUERY_RESULT *qr,
                                AQDB_ID *idListPtr,
                                int *idListLen)
{
  int i;

  for (i=0; i<*idListLen; i++) {
    uint64_t id;

    id=GWEN_IdList64_Iterator_GetNextId(qr->it);
    if (id==0)
      break;
    idListPtr[i]=id;
  }

  *idListLen=i;

  if (i==0) {
    DBG_DEBUG(AQDB_LOGDOMAIN, "No next id");
    return GWEN_ERROR_NOT_FOUND;
  }

  return 0;
}



AQDB_QUERY_RESULT *AQDB_QueryResult_List_GetById(AQDB_QUERY_RESULT_LIST *qrl, AQDB_ID id)
{
  AQDB_QUERY_RESULT *qr;

  qr=AQDB_QueryResult_List_First(qrl);
  while (qr) {
    if (id==AQDB_QueryResult_GetQueryId(qr))
      return qr;
    qr=AQDB_QueryResult_List_Next(qr);
  }

  return NULL;
}








