/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

/* included from "layer.c" */


int AQDB_IoLayerCached_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object)
{
  AQDB_IOLAYER_CACHED *xio;
  GWEN_MULTICACHE_TYPE *typeCache;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->addObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=xio->addObjectFn(ioLayer, dbt, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_IoLayerCached_LockCacheMutex(ioLayer);           /* begin critical section */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  typeCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
  if (typeCache)
    _storeCopyInCache(ioLayer, typeCache, dbt, object);
  AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);            /* end critical section */

  return 0;
}



int AQDB_IoLayerCached_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId)
{
  AQDB_IOLAYER_CACHED *xio;
  GWEN_MULTICACHE_TYPE *typeCache;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->deleteObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=AQDB_IoLayerCached_LockCacheMutex(ioLayer);           /* begin critical section */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  typeCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
  if (typeCache)
    GWEN_MultiCache_Type_PurgeData(typeCache, objectId);

  rv=xio->deleteObjectFn(ioLayer, dbt, objectId);
  AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);            /* end critical section */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerCached_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                          const AQDB_OBJECT *object)
{
  AQDB_IOLAYER_CACHED *xio;
  GWEN_MULTICACHE_TYPE *typeCache;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->writeObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=xio->writeObjectFn(ioLayer, dbt, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_IoLayerCached_LockCacheMutex(ioLayer);           /* begin critical section */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  typeCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
  if (typeCache)
    _storeCopyInCache(ioLayer, typeCache, dbt, object);
  AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);            /* end critical section */

  return 0;
}



int AQDB_IoLayerCached_ReadObject(AQDB_IOLAYER *ioLayer, uint32_t aflags, AQDB_TABLE *dbt,
                                         AQDB_ID objectId,
                                         AQDB_OBJECT **pObject)
{
  AQDB_IOLAYER_CACHED *xio;
  int isWriter;
  GWEN_MULTICACHE_TYPE *typeCache;
  AQDB_OBJECT *object=NULL;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->readObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  isWriter=(aflags & AQDB_ACCESS_FLAGS_ISWRITER);

  rv=AQDB_IoLayerCached_LockCacheMutex(ioLayer);             /* begin critical section */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  if (isWriter)
    typeCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
  else
    typeCache=AQDB_IoLayerCached_Table_GetReaderCache(dbt);
  if (typeCache) {
    object=(AQDB_OBJECT *) GWEN_MultiCache_Type_GetData(typeCache, objectId);
    if (object) {
      AQDB_OBJECT *copyObject;

#ifdef AQDB_IOLAYER_CACHED_DEEP_COPY
      copyObject=AQDB_Object_deepCopy(object);
#else
      copyObject=AQDB_Object_dup(object);
#endif

      AQDB_Object_free(object);
      *pObject=copyObject;
      AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);
      return 0;
    }
  }
  AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);              /* end critical section */

  /* read from file */
  rv=xio->readObjectFn(ioLayer, aflags, dbt, objectId, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  else {
    rv=AQDB_IoLayerCached_LockCacheMutex(ioLayer);           /* begin critical section */
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    if (isWriter)
      typeCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
    else
      typeCache=AQDB_IoLayerCached_Table_GetReaderCache(dbt);
    if (typeCache)
      _storeCopyInCache(ioLayer, typeCache, dbt, object);
    AQDB_IoLayerCached_UnlockCacheMutex(ioLayer);            /* end critical section */
  }

  *pObject=object;
  return 0;
}




void _storeCopyInCache(AQDB_IOLAYER *ioLayer, GWEN_MULTICACHE_TYPE *typeCache, AQDB_TABLE *dbt,
                       const AQDB_OBJECT *object)
{
  AQDB_OBJECT *copyObject;
  AQDB_ID objectId;
  uint32_t objectLen;

  objectId=AQDB_Object_GetObjectId(object);
  objectLen=AQDB_Table_GetMaxObjectSize(dbt);

#if 0
  DBG_ERROR(AQDB_LOGDOMAIN, "Storing object in cache (table id %lu, object id %lu, size: %lu)",
            (long unsigned) AQDB_Table_GetTableId(dbt), (long unsigned) objectId,
            (long unsigned) objectLen);
#endif

#ifdef AQDB_IOLAYER_CACHED_DEEP_COPY
  copyObject=AQDB_Object_deepCopy(object);
#else
  copyObject=AQDB_Object_dup(object);
#endif

  GWEN_MultiCache_Type_SetData(typeCache, objectId, copyObject, objectLen);
  AQDB_Object_free(copyObject);
}



