/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2021 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

/* included from "layer.c" */


AQDB_TABLE *AQDB_IoLayerCached_CreateTable(AQDB_IOLAYER *ioLayer,
                                           const char *name,
                                           uint32_t flags,
                                           const AQDB_COLUMN_LIST *columns)
{
  AQDB_IOLAYER_CACHED *xio;
  AQDB_TABLE *dbt;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->createTableFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return NULL;
  }

  dbt=xio->createTableFn(ioLayer, name, flags, columns);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  if (AQDB_IoLayerCached_Table_IsOfType(dbt)==0) {
    DBG_DEBUG(AQDB_LOGDOMAIN, "Extending table %s for cached layer", AQDB_Table_GetName(dbt));
    rv=AQDB_IoLayerCached_Table_Extend(dbt);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return NULL;
    }
    AQDB_IoLayerCached_CreateTableCaches(ioLayer, dbt);
  }

  /* table has been created, so we're in editMode, let table reflect that */
  rv=AQDB_IoLayerCached_BeginEditTable(ioLayer, dbt);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    /* dont free table here, it has already been added to the list */
    return NULL;
  }

  return dbt;
}




AQDB_TABLE *AQDB_IoLayerCached_OpenTable(AQDB_IOLAYER *ioLayer,
                                         uint32_t aflags,
                                         const char *tname,
                                         uint32_t flags,
                                         AQDB_ID *pTableId)
{
  AQDB_IOLAYER_CACHED *xio;
  AQDB_TABLE *dbt;
  AQDB_ID tableId=0;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->openTableFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return NULL;
  }

  dbt=xio->openTableFn(ioLayer, aflags, tname, flags, &tableId);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  if (AQDB_IoLayerCached_Table_IsOfType(dbt)==0) {
    int rv;

    DBG_DEBUG(AQDB_LOGDOMAIN, "Extending table %s for cached layer", AQDB_Table_GetName(dbt));
    rv=AQDB_IoLayerCached_Table_Extend(dbt);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return NULL;
    }
    AQDB_IoLayerCached_CreateTableCaches(ioLayer, dbt);
  }

  *pTableId=tableId;
  return dbt;
}



int AQDB_IoLayerCached_SetObjectPos(AQDB_IOLAYER *ioLayer,
                                    AQDB_TABLE *dbt,
                                    uint64_t index,
                                    uint64_t objectPos)
{
  AQDB_IOLAYER_CACHED *xio;
  GWEN_MULTICACHE_TYPE *writerCache;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->setObjectPosFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  rv=xio->setObjectPosFn(ioLayer, dbt, index, objectPos);

  writerCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);
  if (writerCache) {
    GWEN_MultiCache_Type_PurgeData(writerCache, index+1);
  }

  return rv;
}




int AQDB_IoLayerCached_BeginEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt)
{
  /* nothing to do for now */
  return 0;
}



int AQDB_IoLayerCached_EndEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, int abandon)
{
  GWEN_MULTICACHE_TYPE *readerCache;
  GWEN_MULTICACHE_TYPE *writerCache;

  readerCache=AQDB_IoLayerCached_Table_GetReaderCache(dbt);
  writerCache=AQDB_IoLayerCached_Table_GetWriterCache(dbt);

  if (abandon==0) {
    if (writerCache && readerCache) {
      GWEN_IDLIST64 *idList;

      idList=GWEN_MultiCache_Type_GetIdsInCache(writerCache);
      if (idList) {
        GWEN_IDLIST64_ITERATOR *iter;
        int cnt=0;

        /* only remove objects from readerCache which have changed and so landed also in writerCache */
        iter=GWEN_IdList64_Iterator_new(idList);
        if (iter) {
          AQDB_ID id;

          id=(AQDB_ID) GWEN_IdList64_Iterator_GetFirstId(iter);
          while (id) {
            cnt++;
            GWEN_MultiCache_Type_PurgeData(readerCache, id);
            id=(AQDB_ID) GWEN_IdList64_Iterator_GetNextId(iter);
          }
          GWEN_IdList64_Iterator_free(iter);
        }
        /*DBG_ERROR(AQDB_LOGDOMAIN, "Purged %d entries", cnt);*/
        GWEN_IdList64_free(idList);
      }

      /*DBG_ERROR(AQDB_LOGDOMAIN, "Purging writer cache");*/
      GWEN_MultiCache_Type_PurgeAll(writerCache);
    }
  }
  else {
    /* abandon, purge all caches */
    if (readerCache)
      GWEN_MultiCache_Type_PurgeAll(readerCache);
    if (writerCache)
      GWEN_MultiCache_Type_PurgeAll(writerCache);
  }

  return 0;
}



void AQDB_IoLayerCached_CreateTableCaches(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt)
{
  GWEN_MULTICACHE_TYPE *mct;

  mct=AQDB_IoLayerCached_CreateCacheType(ioLayer);
  if (mct)
    AQDB_IoLayerCached_Table_SetReaderCache(dbt, mct);

  mct=AQDB_IoLayerCached_CreateCacheType(ioLayer);
  if (mct)
    AQDB_IoLayerCached_Table_SetWriterCache(dbt, mct);
}



