/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG
//#define AQDB_IOLAYER_CACHED_DEEP_COPY


#include "aqdatabase/backendsupport/iocached/layer_p.h"

#include "aqdatabase/backendsupport/iocached/table.h"

#include <aqdatabase/aqdb_object_be.h>

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_IOLAYER, AQDB_IOLAYER_CACHED)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB AQDB_IoLayerCached_FreeData(void *bp, void *p);
static void _unattachFromTables(AQDB_IOLAYER *ioLayer);

static int AQDB_IoLayerCached_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user);
static int AQDB_IoLayerCached_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags);

static AQDB_TABLE *AQDB_IoLayerCached_CreateTable(AQDB_IOLAYER *ioLayer,
                                                  const char *name,
                                                  uint32_t flags,
                                                  const AQDB_COLUMN_LIST *columns);
static AQDB_TABLE *AQDB_IoLayerCached_OpenTable(AQDB_IOLAYER *ioLayer,
                                                uint32_t aflags,
                                                const char *tname,
                                                uint32_t flags,
                                                AQDB_ID *pTableId);
static int AQDB_IoLayerCached_SetObjectPos(AQDB_IOLAYER *ioLayer,
                                           AQDB_TABLE *dbt,
                                           uint64_t index,
                                           uint64_t objectPos);

static int AQDB_IoLayerCached_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object);
static int AQDB_IoLayerCached_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId);
static int AQDB_IoLayerCached_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                          const AQDB_OBJECT *object);
static int AQDB_IoLayerCached_ReadObject(AQDB_IOLAYER *ioLayer, uint32_t aflags, AQDB_TABLE *dbt,
                                         AQDB_ID objectId,
                                         AQDB_OBJECT **pObject);


static int AQDB_IoLayerCached_BeginEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt);
static int AQDB_IoLayerCached_EndEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, int abandon);

static void AQDB_IoLayerCached_CreateTableCaches(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt);

static int _attachCacheObject(const GWEN_MULTICACHE_TYPE *ct, void *p);
static int _freeCacheObject(GWEN_UNUSED const GWEN_MULTICACHE_TYPE *ct, void *p);

static void _storeCopyInCache(AQDB_IOLAYER *ioLayer, GWEN_MULTICACHE_TYPE *typeCache, AQDB_TABLE *dbt,
                              const AQDB_OBJECT *object);


/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerCached_Extend(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_CACHED *xio;
  AQDB_DB *database;
  uint32_t cacheHintMaxSize;

  AQDB_IoLayer_SetName(ioLayer, "cached");

  GWEN_NEW_OBJECT(AQDB_IOLAYER_CACHED, xio);
  if (xio==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer, xio, AQDB_IoLayerCached_FreeData);

  database=AQDB_IoLayer_GetDatabase(ioLayer);
  cacheHintMaxSize=AQDB_DB_GetCacheHintMaxSize(database);
  if (cacheHintMaxSize) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Enabling database cache (%lu bytes).", (unsigned long) cacheHintMaxSize);
    xio->multiCache=GWEN_MultiCache_new((uint64_t) cacheHintMaxSize);
  }

  xio->cacheMutex=GWEN_Mutex_new();

  xio->beginEditFn=AQDB_IoLayer_SetBeginEditFn(ioLayer, AQDB_IoLayerCached_BeginEdit);
  xio->endEditFn=AQDB_IoLayer_SetEndEditFn(ioLayer, AQDB_IoLayerCached_EndEdit);

  xio->createTableFn=AQDB_IoLayer_SetCreateTableFn(ioLayer, AQDB_IoLayerCached_CreateTable);
  xio->openTableFn=AQDB_IoLayer_SetOpenTableFn(ioLayer, AQDB_IoLayerCached_OpenTable);
  xio->closeTableFn=AQDB_IoLayer_GetCloseTableFn(ioLayer);

  xio->setObjectPosFn=AQDB_IoLayer_SetSetObjectPosFn(ioLayer, AQDB_IoLayerCached_SetObjectPos);

  xio->addObjectFn=AQDB_IoLayer_SetAddObjectFn(ioLayer, AQDB_IoLayerCached_AddObject);
  xio->deleteObjectFn=AQDB_IoLayer_SetDeleteObjectFn(ioLayer, AQDB_IoLayerCached_DeleteObject);
  xio->writeObjectFn=AQDB_IoLayer_SetWriteObjectFn(ioLayer, AQDB_IoLayerCached_WriteObject);
  xio->readObjectFn=AQDB_IoLayer_SetReadObjectFn(ioLayer, AQDB_IoLayerCached_ReadObject);

  return 0;
}



void AQDB_IoLayerCached_FreeData(void *bp, void *p)
{
  AQDB_IOLAYER *ioLayer;
  AQDB_IOLAYER_CACHED *xio;

  ioLayer=(AQDB_IOLAYER*) bp;
  xio=(AQDB_IOLAYER_CACHED *) p;

  _unattachFromTables(ioLayer);

  if (xio->multiCache) {
    char usageBuffer[512];

    if (GWEN_MultiCache_GetUsageString(xio->multiCache, usageBuffer, sizeof(usageBuffer)-1)>=0) {
      DBG_NOTICE(AQDB_LOGDOMAIN, "%s", usageBuffer);
    }

    GWEN_MultiCache_free(xio->multiCache);
    xio->multiCache=NULL;
  }

  GWEN_Mutex_free(xio->cacheMutex);

  GWEN_FREE_OBJECT(xio);
}



void _unattachFromTables(AQDB_IOLAYER *ioLayer)
{
  AQDB_TABLE_LIST *tl;

  tl=AQDB_IoLayer_GetDbTypeTables(ioLayer);
  if (tl) {
    AQDB_TABLE *dbt;

    dbt=AQDB_Table_List_First(tl);
    while(dbt) {
      AQDB_IoLayerCached_Table_Unextend(dbt);
      dbt=AQDB_Table_List_Next(dbt);
    }
  }
}



GWEN_MULTICACHE_TYPE *AQDB_IoLayerCached_CreateCacheType(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_CACHED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->multiCache) {
    GWEN_MULTICACHE_TYPE *cacheType;

    cacheType=GWEN_MultiCache_Type_new(xio->multiCache);
    GWEN_MultiCache_Type_SetAttachObjectFn(cacheType, _attachCacheObject);
    GWEN_MultiCache_Type_SetFreeObjectFn(cacheType,  _freeCacheObject);
    return cacheType;
  }

  return NULL;
}



int AQDB_IoLayerCached_LockCacheMutex(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_CACHED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->cacheMutex) {
    int rv;

    rv=GWEN_Mutex_Lock(xio->cacheMutex);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No mutex, not locking");
  }

  return 0;
}



int AQDB_IoLayerCached_UnlockCacheMutex(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_CACHED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_CACHED, ioLayer);
  assert(xio);

  if (xio->cacheMutex) {
    int rv;

    rv=GWEN_Mutex_Unlock(xio->cacheMutex);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No mutex, not unlocking");
  }

  return 0;
}



int _attachCacheObject(GWEN_UNUSED const GWEN_MULTICACHE_TYPE *ct, void *p)
{
  AQDB_OBJECT *o;

  o=(AQDB_OBJECT *)p;

  AQDB_Object_Attach(o);
  return 0;
}



int _freeCacheObject(GWEN_UNUSED const GWEN_MULTICACHE_TYPE *ct, void *p)
{
  AQDB_OBJECT *o;

  o=(AQDB_OBJECT *)p;

  AQDB_Object_free(o);
  return 0;
}



#include "io_table.c"
#include "io_object.c"
#include "io_session.c"


