/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqdatabase/backendsupport/iocached/table_p.h"

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void AQDB_IoLayerCached_Table_FreeData(void *bp, void *p);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */


int AQDB_IoLayerCached_Table_Extend(AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  GWEN_NEW_OBJECT(AQDB_IOLAYER_CACHED_TABLE, xt);
  if (xt==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt, xt,
                       AQDB_IoLayerCached_Table_FreeData);

  return 0;
}



void AQDB_IoLayerCached_Table_Unextend(AQDB_TABLE *dbt)
{
  if (AQDB_IoLayerCached_Table_IsOfType(dbt)) {
    GWEN_INHERIT_UNLINK(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  }
}



void AQDB_IoLayerCached_Table_FreeData(void *bp, void *p)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  xt=(AQDB_IOLAYER_CACHED_TABLE *) p;
  if (xt->writerCache)
    GWEN_MultiCache_Type_free(xt->writerCache);
  if (xt->readerCache)
    GWEN_MultiCache_Type_free(xt->readerCache);
  GWEN_FREE_OBJECT(xt);
}



int AQDB_IoLayerCached_Table_IsOfType(const AQDB_TABLE *dbt)
{
  return GWEN_INHERIT_ISOFTYPE(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
}



void AQDB_IoLayerCached_Table_PurgeCaches(AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  assert(xt);

  if (xt->readerCache) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Purging reader cache");
    GWEN_MultiCache_Type_PurgeAll(xt->readerCache);
  }
  if (xt->writerCache) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Purging writer cache");
    GWEN_MultiCache_Type_PurgeAll(xt->writerCache);
  }
}


GWEN_MULTICACHE_TYPE *AQDB_IoLayerCached_Table_GetReaderCache(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  assert(xt);

  return xt->readerCache;
}



void AQDB_IoLayerCached_Table_SetReaderCache(AQDB_TABLE *dbt, GWEN_MULTICACHE_TYPE *mct)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  assert(xt);

  if (xt->readerCache)
    GWEN_MultiCache_Type_free(xt->readerCache);
  xt->readerCache=mct;
}


GWEN_MULTICACHE_TYPE *AQDB_IoLayerCached_Table_GetWriterCache(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  assert(xt);

  return xt->writerCache;
}



void AQDB_IoLayerCached_Table_SetWriterCache(AQDB_TABLE *dbt, GWEN_MULTICACHE_TYPE *mct)
{
  AQDB_IOLAYER_CACHED_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_CACHED_TABLE, dbt);
  assert(xt);

  if (xt->writerCache)
    GWEN_MultiCache_Type_free(xt->writerCache);
  xt->writerCache=mct;
}





