/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqdatabase/backendsupport/ioquery/io_query.h"

#include "aqdatabase/backendsupport/ioquery/layer.h"

#include <aqdatabase/aqdb_db.h>
#include <aqdatabase/aqdb_object_be.h>
#include <aqdatabase/aqdb_query.h>
#include <aqdatabase/backendsupport/aqdb_objectquery.h>
#include <aqdatabase/backendsupport/aqdb_query_result.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>




/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */


static uint64_t _getNewQueryIdLocked(AQDB_IOLAYER *ioLayer);
static GWEN_IDLIST64 *_getListOfInputIdsLocked(AQDB_IOLAYER *ioLayer,
                                               uint32_t accessFlags,
                                               AQDB_TABLE *dbt,
                                               AQDB_ID refQueryId,
                                               uint32_t queryFlags);
static int _addQueryResultLocked(AQDB_IOLAYER *ioLayer, AQDB_QUERY_RESULT *qr);

static GWEN_IDLIST64 *_getRefQueryResults(AQDB_IOLAYER *ioLayer, AQDB_ID refQueryId);
static AQDB_QUERY_RESULT *_queryWithExpression(AQDB_IOLAYER *ioLayer,
                                               AQDB_TABLE *dbt,
                                               GWEN_IDLIST64 *idl,
                                               const char *queryExpression,
                                               uint32_t accessFlags,
                                               uint32_t queryFlags,
                                               uint64_t newQueryId);


/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */





int AQDB_IoLayerQuery_QuerySubmit(AQDB_IOLAYER *ioLayer,
                                  uint32_t accessFlags,
                                  AQDB_TABLE *dbt,
                                  AQDB_ID refQueryId,
                                  const char *queryExpression,
                                  uint32_t queryFlags,
                                  AQDB_ID *pQueryId)
{
  GWEN_IDLIST64 *idl;
  AQDB_QUERY_RESULT *qr;
  uint64_t newQueryId;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Submit query for table %s: [%s]",
            AQDB_Table_GetName(dbt),
            queryExpression?queryExpression:"-empty-");

  newQueryId=_getNewQueryIdLocked(ioLayer);
  if (newQueryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Could not acquire new query id");
    return GWEN_ERROR_GENERIC;
  }

  idl=_getListOfInputIdsLocked(ioLayer, accessFlags, dbt, refQueryId, queryFlags);
  if (idl==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No objects to search in [%s]", AQDB_Table_GetName(dbt));
    return GWEN_ERROR_NOT_FOUND;
  }

  /* create a query from expression (if any) */
  if (queryExpression)
    qr=_queryWithExpression(ioLayer, dbt, idl, queryExpression, accessFlags, queryFlags, newQueryId);
  else { /* just add all entries from the current id list to the QueryResult object */
    qr=AQDB_QueryResult_new(newQueryId);
    assert(qr);

    if (queryFlags & AQDB_QUERY_FLAGS_STOPATFIRST) {
      GWEN_IDLIST64 *idlResults;

      idlResults=AQDB_QueryResult_GetIdList(qr);
      if (idlResults!=NULL && GWEN_IdList64_GetEntryCount(idlResults)>0) {
        /* just add the first id to the list */
        AQDB_QueryResult_AddId(qr, GWEN_IdList64_GetIdAt(idl, 0));
      }
    }
    else
      /* add all ids to the list */
      AQDB_QueryResult_SetIdList(qr, idl);
  }

  /* check for results */
  if (qr) {
    GWEN_IDLIST64 *idlResults;

    idlResults=AQDB_QueryResult_GetIdList(qr);
    if (idlResults==NULL || GWEN_IdList64_GetEntryCount(idlResults)==0) {
      /* empty, discard result list */
      DBG_INFO(AQDB_LOGDOMAIN, "No results while searching in [%s]", AQDB_Table_GetName(dbt));
      AQDB_QueryResult_free(qr);
      return GWEN_ERROR_NOT_FOUND;
    }

    /* at least one result, add to query result list */
    rv=_addQueryResultLocked(ioLayer, qr);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_QueryResult_free(qr);
      return rv;
    }

    if (pQueryId)
      *pQueryId=newQueryId;

    DBG_DEBUG(AQDB_LOGDOMAIN, "%d objects match expression in table %s",
              (int) GWEN_IdList64_GetEntryCount(idlResults),
              AQDB_Table_GetName(dbt));
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No query generated for table [%s]", AQDB_Table_GetName(dbt));
    return GWEN_ERROR_NOT_FOUND;
  }
}



uint64_t _getNewQueryIdLocked(AQDB_IOLAYER *ioLayer)
{
  uint64_t newQueryId;
  int rv;

  rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return 0;
  }
  newQueryId=AQDB_IoLayerQuery_GetLastQueryId(ioLayer)+1;
  AQDB_IoLayerQuery_SetLastQueryId(ioLayer, newQueryId);
  AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);

  return newQueryId;
}



GWEN_IDLIST64 *_getListOfInputIdsLocked(AQDB_IOLAYER *ioLayer,
                                        uint32_t accessFlags,
                                        AQDB_TABLE *dbt,
                                        AQDB_ID refQueryId,
                                        uint32_t queryFlags)
{
  GWEN_IDLIST64 *idl=NULL;

  /* fill id list (either list of available ids or result of a previous query) */
  if (refQueryId) {/* get results from previous query */
    int rv;

    rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return NULL;
    }
    idl=_getRefQueryResults(ioLayer, refQueryId);
    AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);
  }
  else /* get list of active ids */
    idl=AQDB_IoLayerQuery_GetActiveIds(ioLayer, accessFlags, dbt, queryFlags);

  if (idl==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No objects to search in [%s]", AQDB_Table_GetName(dbt));
    return NULL;
  }
  else {
    DBG_DEBUG(AQDB_LOGDOMAIN, "Got %lu objects to check for table %s",
              (unsigned long int) GWEN_IdList64_GetEntryCount(idl),
              AQDB_Table_GetName(dbt));
  }

  return idl;
}



int _addQueryResultLocked(AQDB_IOLAYER *ioLayer, AQDB_QUERY_RESULT *qr)
{
  int rv;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Adding query result %lu",
            (unsigned long int) AQDB_QueryResult_GetQueryId(qr));
  rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  queryResultList=AQDB_IoLayerQuery_GetQueryResultList(ioLayer);
  AQDB_QueryResult_List_Add(qr, queryResultList);
  AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);

  return 0;
}



GWEN_IDLIST64 *_getRefQueryResults(AQDB_IOLAYER *ioLayer, AQDB_ID refQueryId)
{
  AQDB_QUERY_RESULT *qrRef;
  GWEN_IDLIST64 *idl;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  queryResultList=AQDB_IoLayerQuery_GetQueryResultList(ioLayer);
  qrRef=AQDB_QueryResult_List_GetById(queryResultList, refQueryId);
  if (qrRef==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No active query with id %d", (int) refQueryId);
    return NULL;
  }
  idl=AQDB_QueryResult_GetIdList(qrRef);
  assert(idl);
  idl=GWEN_IdList64_dup(idl);  /* copy here for thread-safety (otherwise GWEN_IdList64_Attach(idl) would suffice) */
  return idl;
}



AQDB_QUERY_RESULT *_queryWithExpression(AQDB_IOLAYER *ioLayer,
                                        AQDB_TABLE *dbt,
                                        GWEN_IDLIST64 *idl,
                                        const char *queryExpression,
                                        uint32_t accessFlags,
                                        uint32_t queryFlags,
                                        uint64_t newQueryId)
{
  AQDB_QUERY_RESULT *qr;
  AQDB_COLUMN_LIST *cl;
  AQDB_QUERY *dbq;
  GWEN_IDLIST64_ITERATOR *it;
  uint64_t id;

  cl=AQDB_Table_GetColumnList(dbt);
  if (cl==NULL || AQDB_Column_List_GetCount(cl)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table has no columns");
    return NULL;;
  }

  /* parse expression */
  dbq=AQDB_ObjectQuery_new(queryExpression, cl);
  if (dbq==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No query generated");
    return NULL;
  }

  qr=AQDB_QueryResult_new(newQueryId);
  assert(qr);

  it=GWEN_IdList64_Iterator_new(idl);
  id=GWEN_IdList64_Iterator_GetFirstId(it);
  while (id) {
    AQDB_OBJECT *o=NULL;
    int rv;

    rv=AQDB_IoLayerQuery_ReadObject(ioLayer, accessFlags, dbt, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    }
    else {
      /* an object has been read, check whether it matches */
      rv=AQDB_ObjectQuery_Check(dbq, o);
      AQDB_Object_free(o);
      if (rv) {
        /* it actually matches, add it to the idlist */
        AQDB_QueryResult_AddId(qr, id);
        if (queryFlags & AQDB_QUERY_FLAGS_STOPATFIRST)
          break;
      }
    }

    /* try next id */
    id=GWEN_IdList64_Iterator_GetNextId(it);
  } /* while */
  AQDB_Query_free(dbq);
  GWEN_IdList64_Iterator_free(it);
  GWEN_IdList64_free(idl);
  return qr;
}






int AQDB_IoLayerQuery_QueryGetFirst(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                    int *pIdListLen)
{
  int rv;
  AQDB_QUERY_RESULT *qr;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  queryResultList=AQDB_IoLayerQuery_GetQueryResultList(ioLayer);
  qr=AQDB_QueryResult_List_GetById(queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_QueryResult_GetFirstIds(qr, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerQuery_QueryGetNext(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                   int *pIdListLen)
{
  int rv;
  AQDB_QUERY_RESULT *qr;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  queryResultList=AQDB_IoLayerQuery_GetQueryResultList(ioLayer);

  if (queryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id %d", queryId);
    return GWEN_ERROR_INVALID;
  }

  qr=AQDB_QueryResult_List_GetById(queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }


  rv=AQDB_QueryResult_GetNextIds(qr, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_DEBUG(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerQuery_QueryClose(AQDB_IOLAYER *ioLayer, AQDB_ID queryId)
{
  AQDB_QUERY_RESULT *qr;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  queryResultList=AQDB_IoLayerQuery_GetQueryResultList(ioLayer);

  if (queryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id %d", queryId);
    return GWEN_ERROR_INVALID;
  }

  qr=AQDB_QueryResult_List_GetById(queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }

  AQDB_QueryResult_List_Del(qr);
  AQDB_QueryResult_free(qr);

  return 0;
}




