/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "aqdatabase/backendsupport/ioquery/layer_p.h"

#include "aqdatabase/backendsupport/ioquery/io_query.h"

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_IOLAYER, AQDB_IOLAYER_QUERY)




/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB _freeData(void *bp, void *p);

static int AQDB_IoLayerQuery_Open(AQDB_IOLAYER *ioLayer, uint32_t flags);
static int AQDB_IoLayerQuery_Create(AQDB_IOLAYER *ioLayer, uint32_t flags);
static int AQDB_IoLayerQuery_Close(AQDB_IOLAYER *ioLayer, uint32_t flags);

static int _queryGetFirstLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen);
static int _queryGetNextLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen);
static int _queryCloseLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerQuery_Extend(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  AQDB_IoLayer_SetName(ioLayer, "query");

  GWEN_NEW_OBJECT(AQDB_IOLAYER_QUERY, xio);
  if (xio==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer, xio, _freeData);

  xio->queryResultList=AQDB_QueryResult_List_new();
  xio->mutexForQueries=GWEN_Mutex_new();

  xio->createFn=AQDB_IoLayer_SetCreateFn(ioLayer, AQDB_IoLayerQuery_Create);
  xio->openFn=AQDB_IoLayer_SetOpenFn(ioLayer, AQDB_IoLayerQuery_Open);
  xio->closeFn=AQDB_IoLayer_SetCloseFn(ioLayer, AQDB_IoLayerQuery_Close);

  xio->querySubmitFn=AQDB_IoLayer_SetQuerySubmitFn(ioLayer, AQDB_IoLayerQuery_QuerySubmit);
  xio->queryGetFirstFn=AQDB_IoLayer_SetQueryGetFirstFn(ioLayer, _queryGetFirstLocked);
  xio->queryGetNextFn=AQDB_IoLayer_SetQueryGetNextFn(ioLayer, _queryGetNextLocked);
  xio->queryCloseFn=AQDB_IoLayer_SetQueryCloseFn(ioLayer, _queryCloseLocked);

  xio->getActiveIdsForTableFn=AQDB_IoLayer_GetGetActiveIdsForTableFn(ioLayer);
  xio->readObjectFn=AQDB_IoLayer_GetReadObjectFn(ioLayer);


  return 0;
}



void _freeData(void *bp, void *p)
{
  AQDB_IOLAYER_QUERY *xio;

  xio=(AQDB_IOLAYER_QUERY *) p;
  assert(xio);

  GWEN_Mutex_free(xio->mutexForQueries);

  AQDB_QueryResult_List_free(xio->queryResultList);

  GWEN_FREE_OBJECT(xio);
}



AQDB_QUERY_RESULT_LIST *AQDB_IoLayerQuery_GetQueryResultList(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  return xio->queryResultList;
}



void AQDB_IoLayerQuery_ClearQueryResultList(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->queryResultList)
    AQDB_QueryResult_List_Clear(xio->queryResultList);
}



uint64_t AQDB_IoLayerQuery_GetLastQueryId(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  return xio->lastQueryId;
}



void AQDB_IoLayerQuery_SetLastQueryId(AQDB_IOLAYER *ioLayer, uint64_t id)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  xio->lastQueryId=id;
}



int AQDB_IoLayerQuery_LockQueryMutex(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->mutexForQueries) {
    int rv;

    rv=GWEN_Mutex_Lock(xio->mutexForQueries);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No mutex, not locking");
  }

  return 0;
}



int AQDB_IoLayerQuery_UnlockQueryMutex(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->mutexForQueries) {
    int rv;

    rv=GWEN_Mutex_Unlock(xio->mutexForQueries);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No mutex, not unlocking");
  }

  return 0;
}



int AQDB_IoLayerQuery_Open(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->openFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Virtual function not implemented in base class");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=xio->openFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  AQDB_IoLayerQuery_ClearQueryResultList(ioLayer);
  return 0;
}



int AQDB_IoLayerQuery_Create(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->createFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Virtual function not implemented in base class");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=xio->createFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  AQDB_IoLayerQuery_ClearQueryResultList(ioLayer);
  return 0;
}



int AQDB_IoLayerQuery_Close(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->closeFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Virtual function not implemented in base class");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  rv=xio->closeFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  AQDB_IoLayerQuery_ClearQueryResultList(ioLayer);
  return 0;
}



GWEN_IDLIST64 *AQDB_IoLayerQuery_GetActiveIds(AQDB_IOLAYER *ioLayer, uint32_t accessFlags, AQDB_TABLE *dbt, uint32_t queryFlags)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->getActiveIdsForTableFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Virtual function not implemented in base class");
    return NULL;
  }

  return xio->getActiveIdsForTableFn(ioLayer, accessFlags, dbt, queryFlags);
}



int AQDB_IoLayerQuery_ReadObject(AQDB_IOLAYER *ioLayer,
                                 uint32_t accessFlags,
                                 AQDB_TABLE *dbt,
                                 AQDB_ID objectId,
                                 AQDB_OBJECT **pObject)
{
  AQDB_IOLAYER_QUERY *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  if (xio->readObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Virtual function not implemented in base class");
    return GWEN_ERROR_NOT_IMPLEMENTED;
  }

  return xio->readObjectFn(ioLayer, accessFlags, dbt, objectId, pObject);
}



int _queryGetFirstLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  rv=AQDB_IoLayerQuery_QueryGetFirst(ioLayer, queryId, pIdListPtr, pIdListLen);
  AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);
  return rv;
}



int _queryGetNextLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  rv=AQDB_IoLayerQuery_QueryGetNext(ioLayer, queryId, pIdListPtr, pIdListLen);
  AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);
  return rv;
}



int _queryCloseLocked(AQDB_IOLAYER *ioLayer, AQDB_ID queryId)
{
  AQDB_IOLAYER_QUERY *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_QUERY, ioLayer);
  assert(xio);

  rv=AQDB_IoLayerQuery_LockQueryMutex(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  rv=AQDB_IoLayerQuery_QueryClose(ioLayer, queryId);
  AQDB_IoLayerQuery_UnlockQueryMutex(ioLayer);
  return rv;
}



