/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2021 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "aqdatabase/backendsupport/iotimed/layer_p.h"
#include "aqdatabase/version.h"

#include <gwenhywfar/debug.h>

#include <time.h>



GWEN_INHERIT(AQDB_IOLAYER, AQDB_IOLAYER_TIMED)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB AQDB_IoLayerTimed_FreeData(void *bp, void *p);

static int AQDB_IoLayerTimed_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user);
static int AQDB_IoLayerTimed_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags);

static int AQDB_IoLayerTimed_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object);
static int AQDB_IoLayerTimed_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId);
static int AQDB_IoLayerTimed_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                         const AQDB_OBJECT *object);
static int AQDB_IoLayerTimed_ReadObject(AQDB_IOLAYER *ioLayer, uint32_t aflags, AQDB_TABLE *dbt,
                                        AQDB_ID objectId,
                                        AQDB_OBJECT **pObject);
static int AQDB_IoLayerTimed_QuerySubmit(AQDB_IOLAYER *ioLayer,
                                         uint32_t accessFlags,
                                         AQDB_TABLE *dbt,
                                         AQDB_ID refQueryId,
                                         const char *queryExpression,
                                         uint32_t queryFlags,
                                         AQDB_ID *pQueryId);
static int AQDB_IoLayerTimed_QueryGetFirst(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                           int *pIdListLen);
static int AQDB_IoLayerTimed_QueryGetNext(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                          int *pIdListLen);
static int AQDB_IoLayerTimed_QueryClose(AQDB_IOLAYER *ioLayer, AQDB_ID queryId);

static struct timespec _startTimer();
static long int _endTimer(struct timespec start_time);

static timerArray *timerArray_new(void);
static void timerArray_free(timerArray *ta);
static void timerArray_AddMeasure(timerArray *ta, long int meas);
static void timerArray_WriteStatsToFile(const timerArray *ta, FILE *f, const char *title);

static char *numToCommaString(unsigned long long n, int maxLength);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerTimed_Extend(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_TIMED *xio;

  AQDB_IoLayer_SetName(ioLayer, "timed");

  GWEN_NEW_OBJECT(AQDB_IOLAYER_TIMED, xio);
  if (xio==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer, xio, AQDB_IoLayerTimed_FreeData);

  xio->beginEditFn=AQDB_IoLayer_SetBeginEditFn(ioLayer, AQDB_IoLayerTimed_BeginEdit);
  xio->endEditFn=AQDB_IoLayer_SetEndEditFn(ioLayer, AQDB_IoLayerTimed_EndEdit);

  xio->addObjectFn=AQDB_IoLayer_SetAddObjectFn(ioLayer, AQDB_IoLayerTimed_AddObject);
  xio->deleteObjectFn=AQDB_IoLayer_SetDeleteObjectFn(ioLayer, AQDB_IoLayerTimed_DeleteObject);
  xio->writeObjectFn=AQDB_IoLayer_SetWriteObjectFn(ioLayer, AQDB_IoLayerTimed_WriteObject);
  xio->readObjectFn=AQDB_IoLayer_SetReadObjectFn(ioLayer, AQDB_IoLayerTimed_ReadObject);

  xio->querySubmitFn=AQDB_IoLayer_SetQuerySubmitFn(ioLayer, AQDB_IoLayerTimed_QuerySubmit);
  xio->queryGetFirstFn=AQDB_IoLayer_SetQueryGetFirstFn(ioLayer, AQDB_IoLayerTimed_QueryGetFirst);
  xio->queryGetNextFn=AQDB_IoLayer_SetQueryGetNextFn(ioLayer, AQDB_IoLayerTimed_QueryGetNext);
  xio->queryCloseFn=AQDB_IoLayer_SetQueryCloseFn(ioLayer, AQDB_IoLayerTimed_QueryClose);

  xio->beginEditTimers=timerArray_new();
  xio->endEditTimers=timerArray_new();

  xio->querySubmitTimers=timerArray_new();
  xio->queryGetFirstTimers=timerArray_new();
  xio->queryGetNextTimers=timerArray_new();
  xio->queryCloseTimers=timerArray_new();

  xio->addObjectTimers=timerArray_new();
  xio->deleteObjectTimers=timerArray_new();
  xio->writeObjectTimers=timerArray_new();
  xio->readObjectTimers=timerArray_new();

  return 0;
}



void AQDB_IoLayerTimed_FreeData(void *bp, void *p)
{
  AQDB_IOLAYER_TIMED *xio;

  xio=(AQDB_IOLAYER_TIMED *) p;
  assert(xio);

  if (1) {
    FILE *f;
  
    f=fopen("LOG_database_stats.log", "a");
    if (f==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Error creating file \"LOG_database_stats.log\", not writing stats");
    }
    else {
      GWEN_TIME *ti;
      GWEN_BUFFER *buf;

      ti=GWEN_CurrentTime();
      buf=GWEN_Buffer_new(0, 64, 0, 1);
      GWEN_Time_toString(ti, "YYYY/MM/DD hh:mm:ss", buf);

      fprintf(f, "\nAqDatabase Statistics (v %s) on %s\n",
	      AQDATABASE_VERSION_FULL_STRING,
	      GWEN_Buffer_GetStart(buf));
      fprintf(f, "----------------------------------------------------------------------\n");
      GWEN_Buffer_free(buf);
      GWEN_Time_free(ti);

      timerArray_WriteStatsToFile(xio->beginEditTimers, f, "BeginEdit");
      timerArray_WriteStatsToFile(xio->endEditTimers, f, "EndEdit");
      timerArray_WriteStatsToFile(xio->querySubmitTimers, f, "QuerySubmit");
      timerArray_WriteStatsToFile(xio->queryGetFirstTimers, f, "QueryGetFirst");
      timerArray_WriteStatsToFile(xio->queryGetNextTimers, f, "QueryGetNext");
      timerArray_WriteStatsToFile(xio->queryCloseTimers, f, "QueryClose");
      timerArray_WriteStatsToFile(xio->addObjectTimers, f, "AddObject");
      timerArray_WriteStatsToFile(xio->deleteObjectTimers, f, "DeleteObject");
      timerArray_WriteStatsToFile(xio->writeObjectTimers, f, "WriteObject");
      timerArray_WriteStatsToFile(xio->readObjectTimers, f, "ReadObject");

      fprintf(f, "\n");

      fclose(f);
    }
  }

  timerArray_free(xio->beginEditTimers);
  timerArray_free(xio->endEditTimers);
  timerArray_free(xio->querySubmitTimers);
  timerArray_free(xio->queryGetFirstTimers);
  timerArray_free(xio->queryGetNextTimers);
  timerArray_free(xio->queryCloseTimers);
  timerArray_free(xio->addObjectTimers);
  timerArray_free(xio->deleteObjectTimers);
  timerArray_free(xio->writeObjectTimers);
  timerArray_free(xio->readObjectTimers);

  GWEN_FREE_OBJECT(xio);
}



int AQDB_IoLayerTimed_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->beginEditFn(ioLayer, editFlags, user);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->beginEditTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->endEditFn(ioLayer, flags);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->endEditTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->addObjectFn(ioLayer, dbt, object);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->addObjectTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->deleteObjectFn(ioLayer, dbt, objectId);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->deleteObjectTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                         const AQDB_OBJECT *object)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->writeObjectFn(ioLayer, dbt, object);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->writeObjectTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_ReadObject(AQDB_IOLAYER *ioLayer, uint32_t aflags, AQDB_TABLE *dbt,
                                 AQDB_ID objectId,
                                 AQDB_OBJECT **pObject)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->readObjectFn(ioLayer, aflags, dbt, objectId, pObject);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->readObjectTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_QuerySubmit(AQDB_IOLAYER *ioLayer,
                                  uint32_t accessFlags,
                                  AQDB_TABLE *dbt,
                                  AQDB_ID refQueryId,
                                  const char *queryExpression,
                                  uint32_t queryFlags,
                                  AQDB_ID *pQueryId)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->querySubmitFn(ioLayer, accessFlags, dbt, refQueryId, queryExpression, queryFlags, pQueryId);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->querySubmitTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_QueryGetFirst(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                    int *pIdListLen)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->queryGetFirstFn(ioLayer, queryId, pIdListPtr, pIdListLen);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->queryGetFirstTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_QueryGetNext(AQDB_IOLAYER *ioLayer, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                     int *pIdListLen)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->queryGetNextFn(ioLayer, queryId, pIdListPtr, pIdListLen);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->queryGetNextTimers, nsecs);
  return rv;
}



int AQDB_IoLayerTimed_QueryClose(AQDB_IOLAYER *ioLayer, AQDB_ID queryId)
{
  AQDB_IOLAYER_TIMED *xio;
  struct timespec ts;
  long int nsecs;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_TIMED, ioLayer);
  assert(xio);

  ts=_startTimer();
  rv=xio->queryCloseFn(ioLayer, queryId);
  nsecs=_endTimer(ts);
  timerArray_AddMeasure(xio->queryCloseTimers, nsecs);
  return rv;
}


/* call this function to start a nanosecond-resolution timer */
struct timespec _startTimer()
{
  struct timespec start_time;

  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start_time);
  return start_time;
}



/* call this function to end a timer, returning nanoseconds elapsed as a long */
long int _endTimer(struct timespec start_time)
{
  struct timespec end_time;
  long int diffInNanos;

  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end_time);
  diffInNanos=(end_time.tv_sec-start_time.tv_sec)*(long)1e9+(end_time.tv_nsec-start_time.tv_nsec);
  return diffInNanos;
}



timerArray *timerArray_new()
{
  timerArray *ta;

  GWEN_NEW_OBJECT(timerArray, ta);
  assert(ta);
  return ta;
}


void timerArray_free(timerArray *ta)
{
  free(ta);
}



void timerArray_AddMeasure(timerArray *ta, long int meas)
{

  if (ta->numMeasurements==0) {
    ta->minMeasurement=meas;
    ta->maxMeasurement=meas;
  }
  else {
    if (meas<ta->minMeasurement)
      ta->minMeasurement=meas;

    if (meas>ta->maxMeasurement)
      ta->maxMeasurement=meas;
  }

  ta->numMeasurements++;
  ta->sumMeasurements+=meas;
}



void timerArray_WriteStatsToFile(const timerArray *ta, FILE *f, const char *title)
{
  if (ta->numMeasurements>0) {
    unsigned long long int medium;
    GWEN_BUFFER *tbuf;

    medium=(ta->sumMeasurements)/(ta->numMeasurements);
    tbuf=GWEN_Buffer_new(0, 256, 0, 1);

    GWEN_Buffer_AppendString(tbuf, " COUNT=");
    GWEN_Buffer_AppendString(tbuf, numToCommaString(ta->numMeasurements, 12));

    GWEN_Buffer_AppendString(tbuf, "   AVG=");
    GWEN_Buffer_AppendString(tbuf, numToCommaString(medium, 14));
    GWEN_Buffer_AppendString(tbuf, " nsecs");

    GWEN_Buffer_AppendString(tbuf, "   MIN=");
    GWEN_Buffer_AppendString(tbuf, numToCommaString(ta->minMeasurement, 14));
    GWEN_Buffer_AppendString(tbuf, " nsecs");

    GWEN_Buffer_AppendString(tbuf, "   MAX=");
    GWEN_Buffer_AppendString(tbuf, numToCommaString(ta->maxMeasurement, 14));
    GWEN_Buffer_AppendString(tbuf, " nsecs");

    GWEN_Buffer_AppendString(tbuf, "   TOTAL=");
    GWEN_Buffer_AppendString(tbuf, numToCommaString(ta->sumMeasurements, 20));
    GWEN_Buffer_AppendString(tbuf, " nsecs");


    fprintf(f, "%16s: %s\n", title, GWEN_Buffer_GetStart(tbuf));

    GWEN_Buffer_free(tbuf);
  }
  else {
    fprintf(f, "%16s: no measurements\n", title);
  }
}



char *numToCommaString(unsigned long long n, int maxLength)
{
  static int comma=' ';
  static char retbuf[30];
  char *p;
  int i=0;
  int len=0;

  p=&retbuf[sizeof(retbuf)-1];

  *p = '\0';

  do {
    int currentDigit;
    int nextN;

    currentDigit=n % 10;
    nextN=n/10;

    if(i%3 == 0 && i != 0) {
      *--p = comma;
      len++;
    }
    if (!(currentDigit==0 && nextN==0)) {
      *--p = '0' + currentDigit;
      len++;
    }
    n=nextN;
    i++;
  } while(n != 0 && len<(sizeof(retbuf)-1));

  if (maxLength && len<maxLength) {
    while(len<maxLength && len<(sizeof(retbuf)-1)) {
      *--p='_';
      len++;
    }
  }

  return p;
}


