/****************************************************************************
 * This file is part of the project AqFinance.
 * AqDataBase (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQDB_TABLE_H
#define AQDB_TABLE_H

#include <aqdatabase/aqdb_db_be.h>
#include <aqdatabase/aqdb_column.h>

#include <gwenhywfar/list1.h>
#include <gwenhywfar/inherit.h>


/** @file aqdatabase/backendsupport/table.h
 *
 * This file containts a AQDB_TABLE. It contains references to the column descriptions and
 * the data objects of a database type table.
 *
 * Most functions here are very low level, the functions normally needed by higher levels are:
 * - @ref AQDB_Table_GetColumnList
 * - @ref AQDB_Table_GetName
 * and of course the constructor, destructor, reading and writing stuff.
 *
 * The main job of this class is to organize the AQDB_COLUMN_LIST (see @ref AQDB_Table_GetColumnList),
 * which contains a descriptions of the columns this database type contains, and a AQFS_TABLE_LIST which
 * references the list of objects of this database type stored in the AQFS file (see @ref AQDB_Table_GetTableListObjects).
 */

typedef struct AQDB_TABLE AQDB_TABLE;
GWEN_INHERIT_FUNCTION_DEFS(AQDB_TABLE)
GWEN_LIST_FUNCTION_DEFS(AQDB_TABLE, AQDB_Table)


/** @name Construction, destruction, reading, writing
 *
 */
/*@{*/

/**
 *  Constructor.
 */
AQDB_TABLE *AQDB_Table_new();

/**
 * Destructor.
 * @param dbt pointer to this AQDB_TABLE object
 */
void AQDB_Table_free(AQDB_TABLE *dbt);

/*@}*/



/** @name Information stored on disk.
 *
 */
/*@{*/

/**
 * Get the name of the type stored in this object.
 * @param dbt pointer to this AQDB_TABLE object
 */
const char *AQDB_Table_GetName(const AQDB_TABLE *dbt);

/**
 * Set the name of the type stored in this object.
 * @param dbt pointer to this AQDB_TABLE object
 * @param s new value
 */
void AQDB_Table_SetName(AQDB_TABLE *dbt, const char *s);

/*@}*/



/** @name Runtime data, not stored on disk.
 *
 */
/*@{*/

/**
 * Get the current id of this object temporarily assigned by the database.
 *
 * This is is used when writing back the content of this object.
 * @param dbt pointer to this AQDB_TABLE object
 */
AQDB_ID AQDB_Table_GetTableId(const AQDB_TABLE *dbt);

/**
 * Set the current id of this object temporarily assigned by the database.
 * @param dbt pointer to this AQDB_TABLE object
 * @param i new value
 */
void AQDB_Table_SetTableId(AQDB_TABLE *dbt, AQDB_ID i);


/**
 * Get open flags (see @ref AQDB_ACTION_FLAGS_READ et al).
 * @param dbt pointer to this AQDB_TABLE object
 */
uint32_t AQDB_Table_GetOpenFlags(const AQDB_TABLE *dbt);

/**
 * Set open flags (see @ref AQDB_ACTION_FLAGS_READ et al).
 * @param dbt pointer to this AQDB_TABLE object
 * @param i new value
 */
void AQDB_Table_SetOpenFlags(AQDB_TABLE *dbt, uint32_t i);

/**
 * Add open flags (see @ref AQDB_ACTION_FLAGS_READ et al).
 * @param dbt pointer to this AQDB_TABLE object
 * @param i flags to add
 */
void AQDB_Table_AddOpenFlags(AQDB_TABLE *dbt, uint32_t i);

/**
 * Subtract open flags (see @ref AQDB_ACTION_FLAGS_READ et al).
 * @param dbt pointer to this AQDB_TABLE object
 * @param i flags to subtract
 */
void AQDB_Table_SubOpenFlags(AQDB_TABLE *dbt, uint32_t i);


/**
 * Get flags.
 * @param dbt pointer to this AQDB_TABLE object
 */
uint32_t AQDB_Table_GetFlags(const AQDB_TABLE *dbt);

/**
 * Set flags.
 * @param dbt pointer to this AQDB_TABLE object
 * @param i new value
 */
void AQDB_Table_SetFlags(AQDB_TABLE *dbt, uint32_t i);

/**
 * Add flags.
 * @param dbt pointer to this AQDB_TABLE object
 * @param i flags to add
 */
void AQDB_Table_AddFlags(AQDB_TABLE *dbt, uint32_t i);

/**
 * Subtract flags.
 * @param dbt pointer to this AQDB_TABLE object
 * @param i flags to subtract
 */
void AQDB_Table_SubFlags(AQDB_TABLE *dbt, uint32_t i);


/**
 * Get list of column descriptions.
 *
 * This list is generated by browing through the entries in the AQFS_TABLE for columns.
 * @param dbt pointer to this AQDB_TABLE object
 */
AQDB_COLUMN_LIST *AQDB_Table_GetColumnList(const AQDB_TABLE *dbt);

/**
 * Set list of column descriptions.
 * @param dbt pointer to this AQDB_TABLE object
 * @param cl new value
 */
void AQDB_Table_SetColumnList(AQDB_TABLE *dbt, AQDB_COLUMN_LIST *cl);

/**
 * Get number of column descriptions for the database type described by this object.
 * @param dbt pointer to this AQDB_TABLE object
 */
int AQDB_Table_GetColumnCount(const AQDB_TABLE *dbt);


uint32_t AQDB_Table_GetMaxObjectSize(const AQDB_TABLE *dbt);
void AQDB_Table_SetMaxObjectSize(AQDB_TABLE *dbt, uint32_t i);


/*@}*/




#endif
