/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "./fromxml.h"

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static int _readTables(AQDB_DB *database, GWEN_XMLNODE *xmlSource);
static int _readTable(AQDB_DB *database, const char *sTableName, GWEN_XMLNODE *xmlSource);
static AQDB_COLUMN_LIST *_readColumns(GWEN_XMLNODE *xmlSource);
static AQDB_COLUMN *_readColumn(GWEN_XMLNODE *xmlSource);
static int _readObjects(AQDB_DB *database, AQDB_ID tableId, const AQDB_COLUMN_LIST *columnList,
                        GWEN_XMLNODE *xmlSource);
static int _readObject(AQDB_DB *database, AQDB_ID tableId, const AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlSource);
static int _readField(AQDB_OBJECT *object, AQDB_DATA_TYPE t, int fieldIdx, GWEN_XMLNODE *xmlNode);




/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_Db_FromXml(AQDB_DB *database, GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;
  int rv;

  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "aqdatabase", NULL, NULL);
  if (xmlNode==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No aqdatabase element in XML tree");
    return GWEN_ERROR_BAD_DATA;
  }

  rv=_readTables(database, xmlNode);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _readTables(AQDB_DB *database, GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;

  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "table", NULL, NULL);
  while (xmlNode) {
    int rv;
    const char *sTableName;

    sTableName=GWEN_XMLNode_GetProperty(xmlNode, "name", NULL);
    if (!(sTableName && *sTableName)) {
      DBG_INFO(AQDB_LOGDOMAIN, "No table name");
      return GWEN_ERROR_BAD_DATA;
    }
    rv=_readTable(database, sTableName, xmlNode);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    xmlNode=GWEN_XMLNode_FindNextTag(xmlNode, "table", NULL, NULL);
  }

  return 0;
}



int _readTable(AQDB_DB *database, const char *sTableName, GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;
  AQDB_COLUMN_LIST *columnList=NULL;
  AQDB_ID tableId=0;
  int rv;

  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "columns", NULL, NULL);
  if (xmlNode==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No columns node in XML");
    return GWEN_ERROR_BAD_DATA;
  }

  /* read columns */
  columnList=_readColumns(xmlNode);
  if (columnList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No columns read from XML");
    return GWEN_ERROR_BAD_DATA;
  }

  /* create table */
  rv=AQDB_DB_CreateTable(database, sTableName, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, columnList);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Column_List_free(columnList);
    return rv;
  }

  /* open table */
  rv=AQDB_DB_OpenTable(database, sTableName, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Column_List_free(columnList);
    return rv;
  }

  /* read objects */
  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "objects", NULL, NULL);
  if (xmlNode) {
    rv=_readObjects(database, tableId, columnList, xmlNode);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Column_List_free(columnList);
      return rv;
    }
  }

  AQDB_Column_List_free(columnList);

  return 0;
}



AQDB_COLUMN_LIST *_readColumns(GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;
  AQDB_COLUMN_LIST *columnList;

  columnList=AQDB_Column_List_new();
  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "column", NULL, NULL);
  while (xmlNode) {
    AQDB_COLUMN *column;

    column=_readColumn(xmlNode);
    if (column==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "No column read");
      AQDB_Column_List_free(columnList);
      return NULL;
    }
    AQDB_Column_List_Add(column, columnList);
    xmlNode=GWEN_XMLNode_FindNextTag(xmlNode, "column", NULL, NULL);
  }

  if (AQDB_Column_List_GetCount(columnList)<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Empty column list");
    AQDB_Column_List_free(columnList);
    return NULL;
  }

  return columnList;
}



AQDB_COLUMN *_readColumn(GWEN_XMLNODE *xmlSource)
{
  AQDB_COLUMN *column;
  const char *sName;
  const char *sType;
  AQDB_DATA_TYPE t;
  int maxLen;
  int i;

  sName=GWEN_XMLNode_GetCharValue(xmlSource, "name", NULL);
  if (!(sName && *sName)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No name in column");
    return NULL;
  }

  sType=GWEN_XMLNode_GetCharValue(xmlSource, "type", NULL);
  if (!(sType && *sType)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No type in column");
    return NULL;
  }
  t=AQDB_DataType_fromString(sType);
  if (!(t>AQDB_DataType_None)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Bad type spec (%s) in column \"%s\"", sType, sName);
    return NULL;
  }

  maxLen=GWEN_XMLNode_GetIntValue(xmlSource, "maxLen", 0);

  column=AQDB_Column_new(t, sName, maxLen);

  i=GWEN_XMLNode_GetIntValue(xmlSource, "mutable", 1);
  AQDB_Column_SetIsMutable(column, i);

  return column;
}



int _readObjects(AQDB_DB *database, AQDB_ID tableId, const AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;

  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "object", NULL, NULL);
  while (xmlNode) {
    int rv;

    rv=_readObject(database, tableId, columnList, xmlNode);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    xmlNode=GWEN_XMLNode_FindNextTag(xmlNode, "object", NULL, NULL);
  }

  return 0;
}



int _readObject(AQDB_DB *database, AQDB_ID tableId, const AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlSource)
{
  GWEN_XMLNODE *xmlNode;
  const AQDB_COLUMN *column;
  int fieldIdx=0;
  AQDB_OBJECT *object;
  int rv;

  object=AQDB_Object_new(tableId, 0, AQDB_Column_List_GetCount(columnList));
  column=AQDB_Column_List_First(columnList);
  xmlNode=GWEN_XMLNode_FindFirstTag(xmlSource, "field", NULL, NULL);

  while (xmlNode && column) {
    rv=_readField(object, AQDB_Column_GetType(column), fieldIdx, xmlNode);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(object);
      return rv;
    }

    fieldIdx++;
    column=AQDB_Column_List_Next(column);
    xmlNode=GWEN_XMLNode_FindNextTag(xmlNode, "field", NULL, NULL);
  }
  if (xmlNode) {
    DBG_WARN(AQDB_LOGDOMAIN, "There are more fields in the object than columns in the type description.");
  }
  if (column) {
    DBG_WARN(AQDB_LOGDOMAIN, "There are more columns in the type description than there are fields in the objects.");
  }

  rv=AQDB_DB_AddObject(database, tableId, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  AQDB_Object_free(object);
  return 0;
}



int _readField(AQDB_OBJECT *object, AQDB_DATA_TYPE t, int fieldIdx, GWEN_XMLNODE *xmlNode)
{
  const char *s;

  s=GWEN_XMLNode_GetCharValue(xmlNode, NULL, NULL);
  if (s) {
    const char *modifier;
    int rv;
    GWEN_BUFFER *dbuf;

    switch (t) {
    case AQDB_DataType_Timestamp:
      modifier="YYYYMMDD-hh:mm:ss";
      break;
    case AQDB_DataType_Bin:
      modifier="hex";
      break;
    case AQDB_DataType_Amount:
      modifier="0";
      break;
    default:
      modifier=NULL;
      break;
    }

    dbuf=GWEN_Buffer_new(0, 256, 0, 1);
    rv=GWEN_Text_UnescapeXmlToBuffer(s, dbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(dbuf);
      return rv;
    }

    rv=AQDB_Object_SetFieldFromStringByType(object, fieldIdx, t, GWEN_Buffer_GetStart(dbuf), modifier);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(dbuf);
      return rv;
    }

    GWEN_Buffer_free(dbuf);

  }

  return 0;
}





