/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "./toxml.h"

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */


static int _writeDatabaseToXml(AQDB_DB *database, GWEN_XMLNODE *xmlDest);
static int _dumpTable(AQDB_DB *database, const char *sTableName, GWEN_XMLNODE *xmlDest);
static int _dumpColumns(AQDB_DB *database, AQDB_ID tableId, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest);
static void _dumpColumn(const AQDB_COLUMN *column, GWEN_XMLNODE *xmlDest);
static int _dumpObjects(AQDB_DB *database, AQDB_ID tableId, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest);
static int _dumpObject(AQDB_OBJECT *object, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest);
static int _dumpField(AQDB_OBJECT *object, int fieldIdx, AQDB_DATA_TYPE t, GWEN_XMLNODE *xmlDest);
static GWEN_XMLNODE *_generateXmlDocument(void);



/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */




GWEN_XMLNODE *AQDB_Db_ToXml(AQDB_DB *database)
{
  GWEN_XMLNODE *xmlRoot;
  GWEN_XMLNODE *xmlNode;
  int rv;

  xmlRoot=_generateXmlDocument();
  if (xmlRoot==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Cout not create XML document");
    return NULL;
  }

  xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "aqdatabase");
  GWEN_XMLNode_AddChild(xmlRoot, xmlNode);

  rv=_writeDatabaseToXml(database, xmlNode);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_XMLNode_free(xmlNode);
    return NULL;
  }

  return xmlRoot;
}



int _writeDatabaseToXml(AQDB_DB *database, GWEN_XMLNODE *xmlDest)
{
  GWEN_STRINGLIST *tableNameList;
  int rv;

  tableNameList=GWEN_StringList_new();
  rv=AQDB_DB_GetTableNameList(database, tableNameList);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Error getting table name list (%d)", rv);
    GWEN_StringList_free(tableNameList);
    return rv;
  }

  if (GWEN_StringList_Count(tableNameList)) {
    GWEN_STRINGLISTENTRY *se;

    se=GWEN_StringList_FirstEntry(tableNameList);
    while (se) {
      const char *s;

      s=GWEN_StringListEntry_Data(se);
      if (s && *s && *s!='+') {
        rv=_dumpTable(database, s, xmlDest);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "Error dumping table [%s] (%d)", s, rv);
          GWEN_StringList_free(tableNameList);
          return rv;
        }
      }
      se=GWEN_StringListEntry_Next(se);
    }
  }
  GWEN_StringList_free(tableNameList);
  return 0;
}



int _dumpTable(AQDB_DB *database, const char *sTableName, GWEN_XMLNODE *xmlDest)
{
  AQDB_ID tableId=0;
  AQDB_COLUMN_LIST *columnList;
  GWEN_XMLNODE *xmlNodeTable;
  GWEN_XMLNODE *xmlNodeColumns;
  GWEN_XMLNODE *xmlNodeObjects;
  int rv;

  rv=AQDB_DB_OpenTable(database, sTableName, AQDB_ACTION_FLAGS_READ, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  columnList=AQDB_DB_GetColumnList(database, tableId);
  if (columnList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Error getting column list for table");
    return GWEN_ERROR_GENERIC;
  }

  xmlNodeTable=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "table");
  GWEN_XMLNode_AddChild(xmlDest, xmlNodeTable);
  GWEN_XMLNode_SetProperty(xmlNodeTable, "name", sTableName);

  xmlNodeColumns=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "columns");
  GWEN_XMLNode_AddChild(xmlNodeTable, xmlNodeColumns);

  rv=_dumpColumns(database, tableId, columnList, xmlNodeColumns);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Column_List_free(columnList);
    return rv;
  }

  xmlNodeObjects=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "objects");
  GWEN_XMLNode_AddChild(xmlNodeTable, xmlNodeObjects);

  rv=_dumpObjects(database, tableId, columnList, xmlNodeObjects);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Column_List_free(columnList);
    return rv;
  }
  AQDB_Column_List_free(columnList);

  return 0;
}



int _dumpColumns(AQDB_DB *database, AQDB_ID tableId, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest)
{
  AQDB_COLUMN *column;

  column=AQDB_Column_List_First(columnList);
  while (column) {
    const char *s;

    s=AQDB_Column_GetName(column);
    if (s && *s) {
      GWEN_XMLNODE *xmlNode;

      xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "column");
      GWEN_XMLNode_AddChild(xmlDest, xmlNode);
      _dumpColumn(column, xmlNode);
    }
    else {
      DBG_WARN(AQDB_LOGDOMAIN, "Column with no name, ignoring");
    }

    column=AQDB_Column_List_Next(column);
  }

  return 0;
}



void _dumpColumn(const AQDB_COLUMN *column, GWEN_XMLNODE *xmlDest)
{
  const char *s;

  s=AQDB_Column_GetName(column);
  if (s && *s)
    GWEN_XMLNode_SetCharValue(xmlDest, "name", s);
  s=AQDB_DataType_toString(AQDB_Column_GetType(column));
  if (s && *s)
    GWEN_XMLNode_SetCharValue(xmlDest, "type", s);
  GWEN_XMLNode_SetIntValue(xmlDest, "maxLen", AQDB_Column_GetMaxLen(column));
  GWEN_XMLNode_SetIntValue(xmlDest, "isMutable", AQDB_Column_GetIsMutable(column));
}



int _dumpObjects(AQDB_DB *database, AQDB_ID tableId, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest)
{
  AQDB_ID qid=0;
  int rv;
  AQDB_ID oids[16];
  int len;

  /* query */
  rv=AQDB_DB_QuerySubmit(database, tableId, 0, NULL, 0, &qid);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }

  /* browse all objects */
  len=16;
  rv=AQDB_DB_QueryGetFirst(database, qid, oids, &len);
  while (rv==0) {
    int i;

    for (i=0; i<len; i++) {
      AQDB_OBJECT *o=NULL;
      GWEN_XMLNODE *xmlNode;

      rv=AQDB_DB_ReadObject(database, tableId, oids[i], &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error reading object %d (%d)", oids[i], rv);
        AQDB_DB_QueryClose(database, qid);
        return rv;
      }

      xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "object");
      GWEN_XMLNode_AddChild(xmlDest, xmlNode);
      rv=_dumpObject(o, columnList, xmlNode);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error writing object %d (%d)", oids[i], rv);
        AQDB_DB_QueryClose(database, qid);
        return rv;
      }
      AQDB_Object_free(o);
    }

    len=16;
    rv=AQDB_DB_QueryGetNext(database, qid, oids, &len);
  }
  AQDB_DB_QueryClose(database, qid);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;

}



int _dumpObject(AQDB_OBJECT *object, AQDB_COLUMN_LIST *columnList, GWEN_XMLNODE *xmlDest)
{
  AQDB_COLUMN *column;
  int fieldIdx=0;

  column=AQDB_Column_List_First(columnList);
  while (column) {
    GWEN_XMLNODE *xmlNode;
    int rv;

    xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "field");
    GWEN_XMLNode_AddChild(xmlDest, xmlNode);
    rv=_dumpField(object, fieldIdx, AQDB_Column_GetType(column), xmlNode);
    if (rv<0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    fieldIdx++;
    column=AQDB_Column_List_Next(column);
  }

  return 0;
}



int _dumpField(AQDB_OBJECT *object, int fieldIdx, AQDB_DATA_TYPE t, GWEN_XMLNODE *xmlDest)
{
  const char *modifier;
  int rv;
  GWEN_BUFFER *dbuf;
  GWEN_BUFFER *xbuf;
  GWEN_XMLNODE *xmlNode;

  switch (t) {
  case AQDB_DataType_Timestamp:
    modifier="YYYYMMDD-hh:mm:ss";
    break;
  case AQDB_DataType_Bin:
    modifier="hex";
    break;
  case AQDB_DataType_Amount:
    modifier="0";
    break;
  default:
    modifier=NULL;
    break;
  }

  dbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AQDB_Object_GetFieldAsStringByType(object, fieldIdx, t, modifier, dbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(dbuf);
    return rv;
  }

  xbuf=GWEN_Buffer_new(0, GWEN_Buffer_GetUsedBytes(dbuf)+10, 0, 1);
  rv=GWEN_Text_EscapeXmlToBuffer(GWEN_Buffer_GetStart(dbuf), xbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(xbuf);
    GWEN_Buffer_free(dbuf);
    return rv;
  }

  xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeData, GWEN_Buffer_GetStart(xbuf));
  GWEN_XMLNode_AddChild(xmlDest, xmlNode);

  GWEN_Buffer_free(xbuf);
  GWEN_Buffer_free(dbuf);

  return 0;
}








GWEN_XMLNODE *_generateXmlDocument(void)
{
  GWEN_XMLNODE *xmlRoot;
  GWEN_XMLNODE *xmlNode;

  xmlRoot=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "root");
  xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "?xml");
  GWEN_XMLNode_AddHeader(xmlRoot, xmlNode);
  GWEN_XMLNode_SetProperty(xmlNode, "version", "1.0");
  GWEN_XMLNode_SetProperty(xmlNode, "encoding", "UTF-8");

  return xmlRoot;
}

