/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqdb_session_p.h"

#include <aqdatabase/user/aqdb_userdb.h>
#include <aqdatabase/aqdb_object_be.h>
#include <aqdatabase/aqdb_db_be.h>

#include <gwenhywfar/debug.h>



GWEN_INHERIT_FUNCTIONS(AQDB_USER_SESSION)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static AQDB_USER_TABLE *_getTableByName(AQDB_USER_SESSION *userSession, const char *name);
int _openAllTables(AQDB_USER_SESSION *userSession);
static int _openTable(AQDB_USER_SESSION *userSession,
                      const char *name,
                      const AQDB_ACL_LIST *userAclList,
                      const AQDB_ACL_LIST *groupAclList);
static uint32_t _getPermissions(AQDB_USER_SESSION *userSession,
                                const char *tableName,
                                const AQDB_ACL_LIST *userAclList,
                                const AQDB_ACL_LIST *groupAclList);
static uint32_t _getPermissionsFromAclList(AQDB_ID subjectId, const char *tableName, const AQDB_ACL_LIST *aclList);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



AQDB_USER_SESSION *AQDB_UserSession_new(AQDB_DB *database, const char *userName)
{
  AQDB_USER_SESSION *userSession;

  GWEN_NEW_OBJECT(AQDB_USER_SESSION, userSession);
  assert(userSession);
  GWEN_INHERIT_INIT(AQDB_USER_SESSION, userSession);

  userSession->database=database;

  return userSession;
}



void AQDB_UserSession_free(AQDB_USER_SESSION *userSession)
{
  if (userSession) {
    int i;

    GWEN_INHERIT_FINI(AQDB_USER_SESSION, userSession);
    for (i=0; i<AQDB_USER_SESSION_MAX_TABLES; i++) {
      AQDB_UserTable_free(userSession->openTables[i]);
      userSession->openTables[i]=NULL;
    }

    free(userSession->userName);
    free(userSession->sessionName);
    GWEN_FREE_OBJECT(userSession);
  }
}



const char *AQDB_UserSession_GetSessionName(const AQDB_USER_SESSION *userSession)
{
  assert(userSession);
  return userSession->userName;
}



void AQDB_UserSession_SetSessionName(AQDB_USER_SESSION *userSession, const char *s)
{
  assert(userSession);
  if (userSession->sessionName)
    free(userSession->sessionName);
  if (s)
    userSession->sessionName=strdup(s);
  else
    userSession->sessionName=NULL;
}



const char *AQDB_UserSession_GetUserName(const AQDB_USER_SESSION *userSession)
{
  assert(userSession);
  return userSession->userName;
}



void AQDB_UserSession_SetUserName(AQDB_USER_SESSION *userSession, const char *s)
{
  assert(userSession);
  if (userSession->userName)
    free(userSession->userName);
  if (s)
    userSession->userName=strdup(s);
  else
    userSession->userName=NULL;
}



AQDB_ID AQDB_UserSession_GetUserId(const AQDB_USER_SESSION *userSession)
{
  assert(userSession);
  return userSession->userId;
}



void AQDB_UserSession_SetUserId(AQDB_USER_SESSION *userSession, AQDB_ID id)
{
  assert(userSession);
  userSession->userId=id;
}



int AQDB_UserSession_GetMaxNumGroupIds(GWEN_UNUSED const AQDB_USER_SESSION *userSession)
{
  return AQDB_USER_SESSION_MAX_GROUPS;
}



AQDB_ID AQDB_UserSession_GetGroupIdAt(const AQDB_USER_SESSION *userSession, int idx)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_GROUPS) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return GWEN_ERROR_INVALID;
  }
  return userSession->groupIds[idx];
}



int AQDB_UserSession_SetGroupIdAt(AQDB_USER_SESSION *userSession, int idx, AQDB_ID id)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_GROUPS) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return GWEN_ERROR_INVALID;
  }
  userSession->groupIds[idx]=id;
  return 0;
}



int AQDB_UserSession_AddGroupId(AQDB_USER_SESSION *userSession, AQDB_ID id)
{
  int i;

  assert(userSession);
  for (i=0; i<AQDB_USER_SESSION_MAX_GROUPS; i++) {
    if (userSession->groupIds[i]==0) {
      userSession->groupIds[i]=id;
      return i;
    }
  }

  return GWEN_ERROR_MEMORY_FULL;
}



int AQDB_UserSession_GetMaxNumOpenTables(GWEN_UNUSED const AQDB_USER_SESSION *userSession)
{
  return AQDB_USER_SESSION_MAX_TABLES;
}



AQDB_USER_TABLE *AQDB_UserSession_GetOpenTableAt(const AQDB_USER_SESSION *userSession, int idx)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return NULL;
  }
  return userSession->openTables[idx];
}



int AQDB_UserSession_SetOpenTableAt(AQDB_USER_SESSION *userSession, int idx, AQDB_USER_TABLE *userTable)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return GWEN_ERROR_INVALID;
  }
  if (userSession->openTables[idx])
    AQDB_UserTable_free(userSession->openTables[idx]);
  userSession->openTables[idx]=userTable;
  return 0;
}



int AQDB_UserSession_AddOpenTable(AQDB_USER_SESSION *userSession, AQDB_USER_TABLE *userTable)
{
  int i;

  assert(userSession);
  for (i=0; i<AQDB_USER_SESSION_MAX_TABLES; i++) {
    if (userSession->openTables[i]==0) {
      userSession->openTables[i]=userTable;
      return i;
    }
  }

  return GWEN_ERROR_MEMORY_FULL;
}




AQDB_ID AQDB_UserSession_GetMaxNumOpenQueryIds(GWEN_UNUSED const AQDB_USER_SESSION *userSession)
{
  return AQDB_USER_SESSION_MAX_QUERIES;
}



AQDB_ID AQDB_UserSession_GetOpenQueryIdAt(const AQDB_USER_SESSION *userSession, int idx)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_QUERIES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return GWEN_ERROR_INVALID;
  }
  return userSession->openQueryIds[idx];
}



int AQDB_UserSession_SetOpenQueryIdAt(AQDB_USER_SESSION *userSession, int idx, AQDB_ID id)
{
  assert(userSession);
  if (idx>=AQDB_USER_SESSION_MAX_QUERIES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Index %d out of array", idx);
    return GWEN_ERROR_INVALID;
  }
  userSession->openQueryIds[idx]=id;
  return 0;
}



int AQDB_UserSession_AddOpenQueryId(AQDB_USER_SESSION *userSession, AQDB_ID id)
{
  int i;

  assert(userSession);
  for (i=0; i<AQDB_USER_SESSION_MAX_QUERIES; i++) {
    if (userSession->openQueryIds[i]==0) {
      userSession->openQueryIds[i]=id;
      return i;
    }
  }

  return GWEN_ERROR_MEMORY_FULL;
}




/* ------------------------------------------- database functions ----------------------------------- */



int AQDB_UserSession_OpenTable(AQDB_USER_SESSION *userSession, const char *name, uint32_t flags, AQDB_ID *pTableId)
{
  if (flags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE)) {
    AQDB_USER_TABLE *userTable;

    DBG_INFO(AQDB_LOGDOMAIN, "Opening table \"%s\"", name);
    userTable=_getTableByName(userSession, name);
    if (userTable) {
      uint32_t permissions;

      permissions=AQDB_UserTable_GetPermissions(userTable);
      /* only open tables for which we have the permissions to open */
      if (permissions & AQDB_PERMISSIONS_OPEN) {
        if (!(permissions & AQDB_PERMISSIONS_READ))
          flags&=~AQDB_ACTION_FLAGS_READ;
        if (!(permissions & (AQDB_PERMISSIONS_MODIFY | AQDB_PERMISSIONS_ADD)))
          flags&=~AQDB_ACTION_FLAGS_WRITE;
        if (!(flags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE))) {
          DBG_ERROR(AQDB_LOGDOMAIN,
                    "No remaining open flags after applying permissions for table \"%s\" (permissions=%08x), rejecting",
                    name, permissions);
          return GWEN_ERROR_PERMISSIONS;
        }
        AQDB_UserTable_AddOpenFlags(userTable, flags);

        DBG_INFO(AQDB_LOGDOMAIN, "Table \"%s\" open (permissions=%08x)", name, permissions);
        return AQDB_UserTable_GetTableId(userTable);
      }
      else {
        DBG_ERROR(AQDB_LOGDOMAIN, "No permissions to open table \"%s\"", name);
        return GWEN_ERROR_NOT_FOUND;
      }
    }
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "Table \"%s\" not found", name);
      return GWEN_ERROR_NOT_FOUND;
    }
  }
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not opening table \"%s\" because of empty open-flags", name);
    return GWEN_ERROR_INVALID;
  }
}



int AQDB_UserSession_CloseTable(AQDB_USER_SESSION *userSession, AQDB_ID tableId, uint32_t flags)
{
  AQDB_USER_TABLE *userTable;
  uint32_t openFlags;
  const char *name;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }

  name=AQDB_UserTable_GetTableName(userTable);
  openFlags=AQDB_UserTable_GetOpenFlags(userTable);
  if ((openFlags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE))==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", name);
    return GWEN_ERROR_NOT_OPEN;
  }
  AQDB_UserTable_SubOpenFlags(userTable, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);

  /* dont do anything else here, especially do not close the table */

  return 0;
}



int AQDB_UserSession_GetTableNameList(AQDB_USER_SESSION *userSession, GWEN_STRINGLIST *tableNameList)
{
  int i;
  int count=0;

  for (i=0; i<AQDB_USER_SESSION_MAX_TABLES; i++) {
    AQDB_USER_TABLE *userTable;

    userTable=userSession->openTables[i];
    if (userTable) {
      uint32_t permissions;
      const char *name;

      name=AQDB_UserTable_GetTableName(userTable);
      permissions=AQDB_UserTable_GetPermissions(userTable);
      if (name && *name && (permissions & AQDB_PERMISSIONS_OPEN)) {
        GWEN_StringList_AppendString(tableNameList, name, 0, 1);
        count++;
      }
    }
  } /* for */

  if (count==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No tables available");
    return GWEN_ERROR_NOT_FOUND;
  }

  return 0;
}




int AQDB_UserSession_AddObject(AQDB_USER_SESSION *userSession, AQDB_ID tableId, AQDB_OBJECT *object)
{
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  int rv;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_ADD)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (ADD)", tableName);
    return GWEN_ERROR_PERMISSIONS;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_WRITE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return GWEN_ERROR_NOT_OPEN;
  }

  rv=AQDB_DB_AddObject(userSession->database, AQDB_UserTable_GetDbTableId(userTable), object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  return 0;
}



int AQDB_UserSession_DeleteObject(AQDB_USER_SESSION *userSession, AQDB_ID tableId, AQDB_ID objectId)
{
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  int rv;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_DELETE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (DELETE)", tableName);
    return GWEN_ERROR_PERMISSIONS;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_WRITE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return GWEN_ERROR_NOT_OPEN;
  }

  rv=AQDB_DB_DeleteObject(userSession->database, AQDB_UserTable_GetDbTableId(userTable), objectId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  return 0;
}



int AQDB_UserSession_WriteObject(AQDB_USER_SESSION *userSession, const AQDB_OBJECT *object)
{
  AQDB_ID tableId;
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  int rv;

  tableId=AQDB_Object_GetTableId(object);

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_MODIFY)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (MODIFY)", tableName);
    return GWEN_ERROR_PERMISSIONS;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_WRITE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return GWEN_ERROR_NOT_OPEN;
  }

  if (1) {
    AQDB_OBJECT *copyOfObject;

    copyOfObject=AQDB_Object_dup(object);
    AQDB_Object_SetTableId(copyOfObject, AQDB_UserTable_GetDbTableId(userTable));

    rv=AQDB_DB_WriteObject(userSession->database, copyOfObject);
    AQDB_Object_free(copyOfObject);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int AQDB_UserSession_ReadObject(AQDB_USER_SESSION *userSession, AQDB_ID tableId, AQDB_ID objectId,
                                AQDB_OBJECT **pObject)
{
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  int rv;
  AQDB_OBJECT *localObject=NULL;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (READ)", tableName);
    return GWEN_ERROR_PERMISSIONS;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return GWEN_ERROR_NOT_OPEN;
  }

  rv=AQDB_DB_ReadObjectWithAccessFlags(userSession->database,
                                       (userSession->isInEditMode)?AQDB_ACCESS_FLAGS_ISWRITER:0,
                                       AQDB_UserTable_GetDbTableId(userTable), objectId, &localObject);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  if (1) {
    AQDB_OBJECT *copyOfObject;

    copyOfObject=AQDB_Object_dup(localObject);
    AQDB_Object_SetTableId(copyOfObject, AQDB_UserTable_GetTableId(userTable));
    *pObject=copyOfObject;
  }
  AQDB_Object_free(localObject);

  return 0;
}



int AQDB_UserSession_BeginEdit(AQDB_USER_SESSION *userSession, uint32_t editFlags)
{
  int rv;

  if (userSession->isInEditMode) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Already in edit mode");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DB_BeginEditWithUserName(userSession->database, editFlags, userSession->userName);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  userSession->isInEditMode=1;
  return 0;
}



int AQDB_UserSession_EndEdit(AQDB_USER_SESSION *userSession, uint32_t flags)
{
  int rv;

  if (userSession->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not in edit mode");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DB_EndEdit(userSession->database, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  userSession->isInEditMode=0;
  return 0;
}



AQDB_COLUMN_LIST *AQDB_UserSession_GetColumnList(AQDB_USER_SESSION *userSession, AQDB_ID tableId)
{
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  AQDB_COLUMN_LIST *columnList;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return NULL;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return NULL;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (READ)", tableName);
    return NULL;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return NULL;
  }

  columnList=AQDB_DB_GetColumnList(userSession->database, AQDB_UserTable_GetDbTableId(userTable));
  if (columnList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  return columnList;
}



int AQDB_UserSession_QuerySubmit(AQDB_USER_SESSION *userSession,
                                 AQDB_ID tableId,
                                 AQDB_ID refQueryId,
                                 const char *queryExpression,
                                 uint32_t flags,
                                 AQDB_ID *pQueryId)
{
  AQDB_USER_TABLE *userTable;
  const char *tableName;
  int rv;
  AQDB_ID localQueryId=0;
  AQDB_ID dbRefQueryId;

  if (tableId<1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }
  userTable=AQDB_UserSession_GetOpenTableAt(userSession, tableId-1);
  if (userTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d", tableId-1);
    return GWEN_ERROR_INVALID;
  }
  tableName=AQDB_UserTable_GetTableName(userTable);

  if ((AQDB_UserTable_GetPermissions(userTable) & AQDB_PERMISSIONS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for table \"%s\" (READ)", tableName);
    return GWEN_ERROR_PERMISSIONS;
  }

  if ((AQDB_UserTable_GetOpenFlags(userTable) & AQDB_ACTION_FLAGS_READ)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table \"%s\" not open", tableName);
    return GWEN_ERROR_NOT_OPEN;
  }

  if (refQueryId>0)
    dbRefQueryId=AQDB_UserSession_GetOpenQueryIdAt(userSession, refQueryId-1);
  else
    dbRefQueryId=0;

  rv=AQDB_DB_QuerySubmitWithAccessFlags(userSession->database,
                                        (userSession->isInEditMode)?AQDB_ACCESS_FLAGS_ISWRITER:0,
                                        AQDB_UserTable_GetDbTableId(userTable),
                                        dbRefQueryId, queryExpression, flags, &localQueryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_UserSession_AddOpenQueryId(userSession, localQueryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  *pQueryId=(rv+1);
  return 0;
}



int AQDB_UserSession_QueryGetFirst(AQDB_USER_SESSION *userSession, AQDB_ID queryId, AQDB_ID *pIdListPtr,
                                   int *pIdListLen)
{
  int rv;
  AQDB_ID localQueryId;

  if (queryId<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id 0");
    return GWEN_ERROR_INVALID;
  }
  localQueryId=AQDB_UserSession_GetOpenQueryIdAt(userSession, queryId-1);
  if (localQueryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id (%lu)", (unsigned long int) queryId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DB_QueryGetFirst(userSession->database, localQueryId, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserSession_QueryGetNext(AQDB_USER_SESSION *userSession, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen)
{
  int rv;
  AQDB_ID localQueryId;

  if (queryId<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id 0");
    return GWEN_ERROR_INVALID;
  }
  localQueryId=AQDB_UserSession_GetOpenQueryIdAt(userSession, queryId-1);
  if (localQueryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id (%lu)", (unsigned long int) queryId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DB_QueryGetNext(userSession->database, localQueryId, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserSession_QueryClose(AQDB_USER_SESSION *userSession, AQDB_ID queryId)
{
  int rv;
  AQDB_ID localQueryId;

  if (queryId<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id 0");
    return GWEN_ERROR_INVALID;
  }
  localQueryId=AQDB_UserSession_GetOpenQueryIdAt(userSession, queryId-1);
  if (localQueryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id (%lu)", (unsigned long int) queryId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DB_QueryClose(userSession->database, localQueryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* remove query id */
  AQDB_UserSession_SetOpenQueryIdAt(userSession, queryId-1, 0);

  return 0;
}



int AQDB_UserSession_GetRollbackSessions(AQDB_USER_SESSION *userSession, AQDB_SESSION_DESCR_LIST *sl)
{
  int rv;

  if (!(userSession->userPermissions & AQDB_PERMISSIONS_UNDO)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for rollback handling");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_DB_GetRollbackSessions(userSession->database, sl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserSession_UndoSession(AQDB_USER_SESSION *userSession, AQDB_ID sessionId)
{
  int rv;

  if (!(userSession->userPermissions & AQDB_PERMISSIONS_UNDO)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Permission denied for rollback handling");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_DB_UndoSession(userSession->database, sessionId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



AQDB_USER_TABLE *_getTableByName(AQDB_USER_SESSION *userSession, const char *name)
{
  int i;

  for (i=0; i<AQDB_USER_SESSION_MAX_TABLES; i++) {
    AQDB_USER_TABLE *userTable;

    userTable=userSession->openTables[i];
    if (userTable) {
      const char *s;

      s=AQDB_UserTable_GetTableName(userTable);
      if (s && *s && strcasecmp(s, name)==0)
        return userTable;
    }
  }

  return NULL;
}


int _openAllTables(AQDB_USER_SESSION *userSession)
{
  GWEN_STRINGLIST *tableNameList;
  AQDB_ACL_LIST *userAclList;
  AQDB_ACL_LIST *groupAclList;
  int rv;

  userAclList=AQDB_UserDb_ReadAllUserAcl(userSession->database);
  groupAclList=AQDB_UserDb_ReadAllGroupAcl(userSession->database);

  tableNameList=GWEN_StringList_new();
  rv=AQDB_DB_GetTableNameList(userSession->database, tableNameList);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_StringList_free(tableNameList);
    AQDB_Acl_List_free(groupAclList);
    AQDB_Acl_List_free(userAclList);
    return rv;
  }
  else {
    GWEN_STRINGLISTENTRY *se;

    se=GWEN_StringList_FirstEntry(tableNameList);
    while (se) {
      const char *s;

      s=GWEN_StringListEntry_Data(se);
      if (s && *s) {
        DBG_INFO(AQDB_LOGDOMAIN, "Opening table \"%s\"", s);
        rv=_openTable(userSession, s, userAclList, groupAclList);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          GWEN_StringList_free(tableNameList);
          AQDB_Acl_List_free(groupAclList);
          AQDB_Acl_List_free(userAclList);
          return rv;
        }
      }
      se=GWEN_StringListEntry_Next(se);
    }
  }
  GWEN_StringList_free(tableNameList);
  AQDB_Acl_List_free(groupAclList);
  AQDB_Acl_List_free(userAclList);

  return 0;
}



int _openTable(AQDB_USER_SESSION *userSession, const char *name, const AQDB_ACL_LIST *userAclList,
               const AQDB_ACL_LIST *groupAclList)
{
  AQDB_USER_TABLE *userTable;
  int tableIdx;
  AQDB_ID dbTableId=0;
  int rv;
  uint32_t permissions=0;

  userTable=AQDB_UserTable_new();
  AQDB_UserTable_SetTableName(userTable, name);
  rv=AQDB_UserSession_AddOpenTable(userSession, userTable);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_UserTable_free(userTable);
    return rv;
  }
  tableIdx=rv;
  AQDB_UserTable_SetTableId(userTable, tableIdx+1);
  rv=AQDB_DB_OpenTable(userSession->database, name, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &dbTableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_UserSession_SetOpenTableAt(userSession, tableIdx, NULL);
    return rv;
  }
  AQDB_UserTable_SetDbTableId(userTable, dbTableId);

  permissions=_getPermissions(userSession, name, userAclList, groupAclList);
  AQDB_UserTable_SetPermissions(userTable, permissions);

  return 0;
}



uint32_t _getPermissions(AQDB_USER_SESSION *userSession,
                         const char *tableName,
                         const AQDB_ACL_LIST *userAclList,
                         const AQDB_ACL_LIST *groupAclList)
{
  uint32_t permissions=0;

  if (userAclList && userSession->userId!=0)
    permissions|=_getPermissionsFromAclList(userSession->userId, tableName, userAclList);
  if (groupAclList) {
    if (userSession->primaryGroupId>0)
      permissions|=_getPermissionsFromAclList(userSession->primaryGroupId, tableName, groupAclList);
    if (userSession->secondaryGroupId>0)
      permissions|=_getPermissionsFromAclList(userSession->secondaryGroupId, tableName, groupAclList);
  }

  return permissions;
}



uint32_t _getPermissionsFromAclList(AQDB_ID wantedSubjectId, const char *wantedTableName, const AQDB_ACL_LIST *aclList)
{
  const AQDB_ACL *acl;
  uint32_t permissions=0;

  acl=AQDB_Acl_List_First(aclList);
  while (acl) {
    AQDB_ID aclSubjectId;

    aclSubjectId=AQDB_Acl_GetSubjectId(acl);
    if (aclSubjectId>0 && aclSubjectId==wantedSubjectId) {
      const char *aclTableName;

      aclTableName=AQDB_Acl_GetTableName(acl);
      if (aclTableName && *aclTableName && strcasecmp(aclTableName, wantedTableName)==0)
        permissions|=AQDB_Acl_GetPermissions(acl);
    }

    acl=AQDB_Acl_List_Next(acl);
  }

  return permissions;
}




