/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqdb_userdb_p.h"
#include "aqdb_acl.h"
#include "aqdb_iddef.h"
#include "aqdb_group.h"
#include "aqdb_user.h"

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_DB, AQDB_USERDB)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB _freeData(void *bp, void *p);
static int GWENHYWFAR_CB AQDB_DbUser_Create(AQDB_DB *db, uint32_t flags);
static int GWENHYWFAR_CB AQDB_DbUser_Open(AQDB_DB *db, uint32_t flags);
static int GWENHYWFAR_CB AQDB_DbUser_Close(AQDB_DB *db, uint32_t flags);
static int GWENHYWFAR_CB AQDB_DbUser_CreateTable(AQDB_DB *db, const char *name, uint32_t flags,
                                                 const AQDB_COLUMN_LIST *columns);
static int GWENHYWFAR_CB AQDB_DbUser_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList);

static int _createAclTypes(AQDB_DB *db);
static int _createAclTables(AQDB_DB *db);
static int _openAclTables(AQDB_DB *db);

static AQDB_ID _getFirstDbIdForQuery(AQDB_DB *db, AQDB_ID tableId, const char *qexpr, uint32_t aflags);
static int _getIncrementedCounter(AQDB_DB *db, const char *counterName);
static int _readPermissionsFromAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName);
static AQDB_ACL_LIST *_readAclList(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName);
static AQDB_ACL_LIST *_readAclListWithExpr(AQDB_DB *db, AQDB_ID idForAclTable, const char *qexpr);
static int _addAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName, uint32_t permissions);
static int _deleteAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName);
static AQDB_OBJECT_LIST *_getListOfMatchingObjects(AQDB_DB *db,
                                                   AQDB_ID tableId,
                                                   uint32_t accessFlags,
                                                   const char *qexpr,
                                                   uint32_t queryFlags);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */


int AQDB_UserDb_Extend(AQDB_DB *db)
{
  AQDB_USERDB *xdb;

  GWEN_NEW_OBJECT(AQDB_USERDB, xdb)
  GWEN_INHERIT_SETDATA(AQDB_DB, AQDB_USERDB, db, xdb, _freeData)

  xdb->createFn=AQDB_DB_SetCreateFn(db, AQDB_DbUser_Create);
  xdb->openFn=AQDB_DB_SetOpenFn(db, AQDB_DbUser_Open);
  xdb->closeFn=AQDB_DB_SetCloseFn(db, AQDB_DbUser_Close);

  xdb->createTableFn=AQDB_DB_SetCreateTableFn(db, AQDB_DbUser_CreateTable);
  xdb->getTableNameListFn=AQDB_DB_SetGetTableNameListFn(db, AQDB_DbUser_GetTableNameList);

  xdb->readObjectFn=AQDB_DB_GetReadObjectFn(db);
  xdb->writeObjectFn=AQDB_DB_GetWriteObjectFn(db);
  xdb->addObjectFn=AQDB_DB_GetAddObjectFn(db);
  xdb->deleteObjectFn=AQDB_DB_GetDeleteObjectFn(db);

  xdb->querySubmitFn=AQDB_DB_GetQuerySubmitFn(db);
  xdb->queryGetFirstFn=AQDB_DB_GetQueryGetFirstFn(db);
  xdb->queryGetNextFn=AQDB_DB_GetQueryGetNextFn(db);
  xdb->queryCloseFn=AQDB_DB_GetQueryCloseFn(db);

  return 0;
}



void AQDB_UserDb_Unextend(AQDB_DB *db)
{
  AQDB_USERDB *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  AQDB_DB_SetCreateFn(db, xdb->createFn);
  AQDB_DB_SetOpenFn(db, xdb->openFn);
  AQDB_DB_SetCloseFn(db, xdb->closeFn);

  AQDB_DB_SetCreateTableFn(db, xdb->createTableFn);
  AQDB_DB_SetGetTableNameListFn(db, xdb->getTableNameListFn);
}



void GWENHYWFAR_CB _freeData(void *bp, void *p)
{
  AQDB_USERDB *xdb;

  xdb=(AQDB_USERDB *) p;
  assert(xdb);

  GWEN_FREE_OBJECT(xdb);
}



int AQDB_DbUser_Create(AQDB_DB *db, uint32_t flags)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->createFn(db, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=_createAclTypes(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbUser_Open(AQDB_DB *db, uint32_t flags)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->openFn(db, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=_openAclTables(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbUser_Close(AQDB_DB *db, uint32_t flags)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->closeFn(db, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* TODO: close user tables */
  return 0;
}



int AQDB_DbUser_CreateTable(AQDB_DB *db, const char *name, uint32_t flags, const AQDB_COLUMN_LIST *columns)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  if (name) {
    if (strcasecmp(name, AQDB_USERDB_TABLENAME_USERS)==0 ||
        strcasecmp(name, AQDB_USERDB_TABLENAME_GROUPS)==0 ||
        strcasecmp(name, AQDB_USERDB_TABLENAME_USERACL)==0 ||
        strcasecmp(name, AQDB_USERDB_TABLENAME_GROUPACL)==0||
        strcasecmp(name, AQDB_USERDB_TABLENAME_IDS)==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Direct creating of access management tables not allowed.");
      return GWEN_ERROR_INVALID;
    }
  }

  rv=xdb->createTableFn(db, name, flags, columns);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbUser_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList)
{
  AQDB_USERDB *xdb;
  GWEN_STRINGLIST *localTableNameList;
  GWEN_STRINGLISTENTRY *se;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  localTableNameList=GWEN_StringList_new();
  rv=xdb->getTableNameListFn(db, localTableNameList);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_StringList_free(localTableNameList);
    return rv;
  }

  /* filter out names of access management tables */
  se=GWEN_StringList_FirstEntry(localTableNameList);
  while (se) {
    const char *name;

    name=GWEN_StringListEntry_Data(se);
    if (name && *name) {
      if (strcasecmp(name, AQDB_USERDB_TABLENAME_USERS)==0 ||
          strcasecmp(name, AQDB_USERDB_TABLENAME_GROUPS)==0 ||
          strcasecmp(name, AQDB_USERDB_TABLENAME_USERACL)==0 ||
          strcasecmp(name, AQDB_USERDB_TABLENAME_GROUPACL)==0 ||
          strcasecmp(name, AQDB_USERDB_TABLENAME_IDS)==0) {
        DBG_DEBUG(AQDB_LOGDOMAIN, "Not adding access management table \"%s\".", name);
      }
      else
        GWEN_StringList_AppendString(tableNameList, name, 0, 0);
    }
    se=GWEN_StringListEntry_Next(se);
  } /* while */
  GWEN_StringList_free(localTableNameList);

  return 0;
}



int _createAclTypes(AQDB_DB *db)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->openFn(db, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xdb->closeFn(db, 0);
    return rv;
  }

  rv=_createAclTables(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_EndEdit(db, AQDB_ACTION_FLAGS_ABORT);
    xdb->closeFn(db, 0);
    return rv;
  }

  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_EndEdit(db, AQDB_ACTION_FLAGS_ABORT);
    xdb->closeFn(db, 0);
    return rv;
  }

  rv=xdb->closeFn(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _createAclTables(AQDB_DB *db)
{
  AQDB_USERDB *xdb;
  AQDB_COLUMN_LIST *cl;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  cl=AQDB_IdDef_CreateColumnList();
  rv=xdb->createTableFn(db, AQDB_USERDB_TABLENAME_IDS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return rv;
  }

  cl=AQDB_User_CreateColumnList();
  rv=xdb->createTableFn(db, AQDB_USERDB_TABLENAME_USERS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  cl=AQDB_Group_CreateColumnList();
  rv=xdb->createTableFn(db, AQDB_USERDB_TABLENAME_GROUPS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  cl=AQDB_Acl_CreateColumnList();
  rv=xdb->createTableFn(db, AQDB_USERDB_TABLENAME_USERACL, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  cl=AQDB_Acl_CreateColumnList();
  rv=xdb->createTableFn(db, AQDB_USERDB_TABLENAME_GROUPACL, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}


int _openAclTables(AQDB_DB *db)
{
  AQDB_USERDB *xdb;
  AQDB_ID tableId;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=AQDB_DB_OpenTable(db, AQDB_USERDB_TABLENAME_IDS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return rv;
  }
  xdb->tableIdIdDef=tableId;

  rv=AQDB_DB_OpenTable(db, AQDB_USERDB_TABLENAME_USERS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->tableIdUsers=tableId;

  rv=AQDB_DB_OpenTable(db, AQDB_USERDB_TABLENAME_GROUPS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->tableIdGroups=tableId;

  rv=AQDB_DB_OpenTable(db, AQDB_USERDB_TABLENAME_USERACL, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->tableIdUserAcl=tableId;

  rv=AQDB_DB_OpenTable(db, AQDB_USERDB_TABLENAME_GROUPACL, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->tableIdGroupAcl=tableId;

  return 0;
}



AQDB_USER *AQDB_UserDb_ReadUserByName(AQDB_DB *db, const char *userName)
{
  AQDB_USERDB *xdb;
  AQDB_ID dbIdForObject;
  AQDB_OBJECT *object=NULL;
  AQDB_USER *user=NULL;
  int rv;
  char qexpr[256];

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "$userName==%s", userName);
  qexpr[sizeof(qexpr)-1]=0;
  dbIdForObject=_getFirstDbIdForQuery(db, xdb->tableIdUsers, qexpr, 0);
  if (dbIdForObject==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  rv=xdb->readObjectFn(db, 0, xdb->tableIdUsers, dbIdForObject, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  rv=AQDB_User_fromObject(object, &user);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return NULL;
  }
  AQDB_Object_free(object);
  AQDB_User_SetDbId(user, dbIdForObject);

  return user;
}



AQDB_USER *AQDB_UserDb_ReadUserById(AQDB_DB *db, AQDB_ID userId)
{
  AQDB_USERDB *xdb;
  AQDB_ID dbIdForObject;
  AQDB_OBJECT *object=NULL;
  AQDB_USER *user=NULL;
  int rv;
  char qexpr[256];

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "$userName==%lu", (unsigned long int) userId);
  qexpr[sizeof(qexpr)-1]=0;
  dbIdForObject=_getFirstDbIdForQuery(db, xdb->tableIdUsers, qexpr, 0);
  if (dbIdForObject==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  rv=xdb->readObjectFn(db, 0, xdb->tableIdUsers, dbIdForObject, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  rv=AQDB_User_fromObject(object, &user);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return NULL;
  }
  AQDB_Object_free(object);
  AQDB_User_SetDbId(user, dbIdForObject);

  return user;
}



int AQDB_UserDb_UpdateUser(AQDB_DB *db, const AQDB_USER *user)
{
  AQDB_USERDB *xdb;
  AQDB_ID objectId;
  AQDB_OBJECT *object;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  objectId=AQDB_User_GetDbId(user);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }
  rv=AQDB_User_CreateObject(db, xdb->tableIdUsers, objectId, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_User_WriteObject(user, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  rv=xdb->writeObjectFn(db, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_Object_free(object);

  return 0;
}



int AQDB_UserDb_AddUser(AQDB_DB *db, AQDB_USER *user)
{
  AQDB_USERDB *xdb;
  AQDB_ID userId;
  AQDB_OBJECT *object;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_getIncrementedCounter(db, "user");
  if (rv<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  userId=(AQDB_ID) rv;

  AQDB_User_SetUserId(user, userId);

  rv=AQDB_User_CreateObject(db, xdb->tableIdUsers, 0, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_User_WriteObject(user, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  rv=xdb->addObjectFn(db, xdb->tableIdUsers, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_User_SetDbId(user, AQDB_Object_GetObjectId(object));
  AQDB_Object_free(object);

  return 0;
}



AQDB_GROUP *AQDB_UserDb_ReadGroupByName(AQDB_DB *db, const char *groupName)
{
  AQDB_USERDB *xdb;
  AQDB_ID dbIdForObject;
  AQDB_OBJECT *object=NULL;
  AQDB_GROUP *group=NULL;
  int rv;
  char qexpr[256];

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "$groupName==%s", groupName);
  qexpr[sizeof(qexpr)-1]=0;
  dbIdForObject=_getFirstDbIdForQuery(db, xdb->tableIdGroups, qexpr, 0);
  if (dbIdForObject==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  rv=xdb->readObjectFn(db, 0, xdb->tableIdGroups, dbIdForObject, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  rv=AQDB_Group_fromObject(object, &group);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return NULL;
  }
  AQDB_Object_free(object);
  AQDB_Group_SetDbId(group, dbIdForObject);

  return group;
}



AQDB_GROUP *AQDB_UserDb_ReadGroupById(AQDB_DB *db, AQDB_ID groupId)
{
  AQDB_USERDB *xdb;
  AQDB_ID dbIdForObject;
  AQDB_OBJECT *object=NULL;
  AQDB_GROUP *group=NULL;
  int rv;
  char qexpr[256];

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "$groupId==%lu", (unsigned long int) groupId);
  qexpr[sizeof(qexpr)-1]=0;
  dbIdForObject=_getFirstDbIdForQuery(db, xdb->tableIdGroups, qexpr, 0);
  if (dbIdForObject==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  rv=xdb->readObjectFn(db, 0, xdb->tableIdGroups, dbIdForObject, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  rv=AQDB_Group_fromObject(object, &group);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return NULL;
  }
  AQDB_Object_free(object);
  AQDB_Group_SetDbId(group, dbIdForObject);

  return group;
}



int AQDB_UserDb_UpdateGroup(AQDB_DB *db, const AQDB_GROUP *group)
{
  AQDB_USERDB *xdb;
  AQDB_ID objectId;
  AQDB_OBJECT *object;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  objectId=AQDB_Group_GetDbId(group);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }
  rv=AQDB_Group_CreateObject(db, xdb->tableIdGroups, objectId, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_Group_WriteObject(group, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  rv=xdb->writeObjectFn(db, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_Object_free(object);

  return 0;
}



int AQDB_UserDb_AddGroup(AQDB_DB *db, AQDB_GROUP *group)
{
  AQDB_USERDB *xdb;
  AQDB_ID groupId;
  AQDB_OBJECT *object;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_getIncrementedCounter(db, "group");
  if (rv<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  groupId=(AQDB_ID) rv;

  AQDB_Group_SetGroupId(group, groupId);

  rv=AQDB_Group_CreateObject(db, xdb->tableIdGroups, 0, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_Group_WriteObject(group, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  rv=xdb->addObjectFn(db, xdb->tableIdGroups, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_Group_SetDbId(group, AQDB_Object_GetObjectId(object));
  AQDB_Object_free(object);

  return 0;
}



AQDB_ID _getFirstDbIdForQuery(AQDB_DB *db, AQDB_ID tableId, const char *qexpr, uint32_t aflags)
{
  AQDB_USERDB *xdb;
  AQDB_ID qid=0;
  AQDB_ID result=0;
  int len;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->querySubmitFn(db, aflags, tableId, 0, qexpr, AQDB_QUERY_FLAGS_STOPATFIRST, &qid);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return 0;
  }

  /* get single first id */
  len=1;
  rv=xdb->queryGetFirstFn(db, qid, &result, &len);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xdb->queryCloseFn(db, qid);
    return rv;
  }
  xdb->queryCloseFn(db, qid);

  if (len!=1) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Query did not return an object id");
    return GWEN_ERROR_GENERIC;
  }

  return result;
}



int AQDB_UserDb_ReadUserPermissions(AQDB_DB *db, AQDB_ID userId, const char *tableName)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_readPermissionsFromAcl(db, xdb->tableIdUserAcl, userId, tableName);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return rv;
}



int AQDB_UserDb_ReadGroupPermissions(AQDB_DB *db, AQDB_ID groupId, const char *tableName)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_readPermissionsFromAcl(db, xdb->tableIdGroupAcl, groupId, tableName);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return rv;
}



int AQDB_UserDb_AddUserAcl(AQDB_DB *db, AQDB_ID userId, const char *tableName, uint32_t permissions)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_addAcl(db, xdb->tableIdUserAcl, userId, tableName, permissions);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserDb_AddGroupAcl(AQDB_DB *db, AQDB_ID groupId, const char *tableName, uint32_t permissions)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_addAcl(db, xdb->tableIdGroupAcl, groupId, tableName, permissions);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserDb_DeleteUserAcl(AQDB_DB *db, AQDB_ID userId, const char *tableName)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_deleteAcl(db, xdb->tableIdUserAcl, userId, tableName);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_UserDb_DeleteGroupAcl(AQDB_DB *db, AQDB_ID groupId, const char *tableName)
{
  AQDB_USERDB *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_deleteAcl(db, xdb->tableIdGroupAcl, groupId, tableName);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



AQDB_ACL_LIST *AQDB_UserDb_ReadAllUserAcl(AQDB_DB *db)
{
  AQDB_USERDB *xdb;
  AQDB_ACL_LIST *aclList;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  aclList=_readAclListWithExpr(db, xdb->tableIdUserAcl, NULL);
  if (aclList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  return aclList;
}



AQDB_ACL_LIST *AQDB_UserDb_ReadAllGroupAcl(AQDB_DB *db)
{
  AQDB_USERDB *xdb;
  AQDB_ACL_LIST *aclList;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  aclList=_readAclListWithExpr(db, xdb->tableIdGroupAcl, NULL);
  if (aclList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  return aclList;
}



int _deleteAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName)
{
  AQDB_USERDB *xdb;
  AQDB_ACL_LIST *aclList;
  AQDB_ACL *acl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  aclList=_readAclList(db, idForAclTable, subjectId, tableName);
  if (aclList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  acl=AQDB_Acl_List_First(aclList);
  while (acl) {
    int rv;

    rv=xdb->deleteObjectFn(db, idForAclTable, AQDB_Acl_GetDbId(acl));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Acl_List_free(aclList);
      return rv;
    }

    acl=AQDB_Acl_List_Next(acl);
  }
  AQDB_Acl_List_free(aclList);
  return 0;
}



int _addAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName, uint32_t permissions)
{
  AQDB_USERDB *xdb;
  AQDB_ACL *acl;
  AQDB_OBJECT *object;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=_getIncrementedCounter(db, "useracl");
  if (rv<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  acl=AQDB_Acl_new();
  AQDB_Acl_SetSubjectId(acl, subjectId);
  AQDB_Acl_SetTableName(acl, tableName);
  AQDB_Acl_SetPermissions(acl, permissions);

  rv=AQDB_Acl_CreateObject(db, idForAclTable, 0, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Acl_free(acl);
    return rv;
  }

  rv=AQDB_Acl_WriteObject(acl, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    AQDB_Acl_free(acl);
    return rv;
  }

  rv=xdb->addObjectFn(db, idForAclTable, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_Object_free(object);
  AQDB_Acl_free(acl);

  return 0;
}



int _readPermissionsFromAcl(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName)
{
  AQDB_ACL_LIST *aclList;
  AQDB_ACL *acl;
  int permissions=0;

  aclList=_readAclList(db, subjectId, subjectId, tableName);
  if (aclList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  acl=AQDB_Acl_List_First(aclList);
  while (acl) {
    permissions|=AQDB_Acl_GetPermissions(acl);
    acl=AQDB_Acl_List_Next(acl);
  }
  AQDB_Acl_List_free(aclList);

  return permissions;
}



AQDB_ACL_LIST *_readAclList(AQDB_DB *db, AQDB_ID idForAclTable, AQDB_ID subjectId, const char *tableName)
{
  AQDB_USERDB *xdb;
  AQDB_ACL_LIST *aclList;
  char qexpr[256];

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "($subjectId==%lu)&&($tableName==%s)", (unsigned long int) subjectId, tableName);
  qexpr[sizeof(qexpr)-1]=0;

  aclList=_readAclListWithExpr(db, idForAclTable, qexpr);
  if (aclList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }
  return aclList;
}



AQDB_ACL_LIST *_readAclListWithExpr(AQDB_DB *db, AQDB_ID idForAclTable, const char *qexpr)
{
  AQDB_USERDB *xdb;
  AQDB_OBJECT_LIST *ol;
  AQDB_OBJECT *object;
  AQDB_ACL_LIST *aclList;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  ol=_getListOfMatchingObjects(db, idForAclTable, 0, qexpr, 0);
  if (ol==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }

  aclList=AQDB_Acl_List_new();
  object=AQDB_Object_List_First(ol);
  while (object) {
    AQDB_ACL *acl;
    int rv;

    acl=AQDB_Acl_new();
    rv=AQDB_Acl_ReadObject(acl, object);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Acl_List_free(aclList);
      AQDB_Object_List_free(ol);
      return NULL;
    }
    AQDB_Acl_List_Add(acl, aclList);
    object=AQDB_Object_List_Next(object);
  }
  AQDB_Object_List_free(ol);

  return aclList;
}




AQDB_OBJECT_LIST *_getListOfMatchingObjects(AQDB_DB *db,
                                            AQDB_ID tableId,
                                            uint32_t accessFlags,
                                            const char *qexpr,
                                            uint32_t queryFlags)
{
  AQDB_USERDB *xdb;
  AQDB_OBJECT_LIST *ol;
  AQDB_ID queryId;
  AQDB_ID resultList[64];
  int idListLen;
  int i;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  rv=xdb->querySubmitFn(db, accessFlags, tableId, 0, qexpr, queryFlags, &queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  idListLen=64;
  rv=xdb->queryGetFirstFn(db, queryId, resultList, &idListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xdb->queryCloseFn(db, queryId);
    return NULL;
  }

  ol=AQDB_Object_List_new();

  for (;;) {
    for (i=0; i<idListLen; i++) {
      AQDB_OBJECT *o=NULL;

      rv=xdb->readObjectFn(db, accessFlags, tableId, resultList[i], &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        xdb->queryCloseFn(db, queryId);
        AQDB_Object_List_free(ol);
        return NULL;
      }
      assert(o);
      AQDB_Object_List_Add(o, ol);
    }

    idListLen=64;
    rv=xdb->queryGetNextFn(db, queryId, resultList, &idListLen);
    if (rv<0) {
      if (rv!=GWEN_ERROR_NOT_FOUND) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        xdb->queryCloseFn(db, queryId);
        AQDB_Object_List_free(ol);
        return NULL;
      }
      else
        break;
    }
  }

  rv=xdb->queryCloseFn(db, queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_List_free(ol);
    return NULL;
  }

  return ol;
}



int _getIncrementedCounter(AQDB_DB *db, const char *counterName)
{
  AQDB_USERDB *xdb;
  AQDB_ID dbIdForObject;
  AQDB_OBJECT *object=NULL;
  AQDB_IDDEF *idDef=NULL;
  int rv;
  char qexpr[256];
  int counter;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_USERDB, db);
  assert(xdb);

  snprintf(qexpr, sizeof(qexpr)-1, "$idName==%s", counterName);
  qexpr[sizeof(qexpr)-1]=0;
  dbIdForObject=_getFirstDbIdForQuery(db, xdb->tableIdIdDef, qexpr, 0);
  if (dbIdForObject) {
    rv=xdb->readObjectFn(db, 0, xdb->tableIdIdDef, dbIdForObject, &object);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    rv=AQDB_IdDef_fromObject(object, &idDef);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(object);
      return rv;
    }
    AQDB_Object_free(object);
    object=NULL;
    AQDB_IdDef_SetDbId(idDef, dbIdForObject);

    counter=AQDB_IdDef_GetIdValue(idDef);
  }
  else {
    counter=0;
    idDef=AQDB_IdDef_new();
    AQDB_IdDef_SetIdName(idDef, counterName);
    AQDB_IdDef_SetIdValue(idDef, counter);
  }

  counter++;
  AQDB_IdDef_SetIdValue(idDef, counter);

  rv=AQDB_IdDef_CreateObject(db, xdb->tableIdIdDef, dbIdForObject, &object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_IdDef_WriteObject(idDef, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }

  if (dbIdForObject)
    rv=xdb->writeObjectFn(db, object);
  else
    rv=xdb->addObjectFn(db, xdb->tableIdIdDef, object);

  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(object);
    return rv;
  }
  AQDB_Object_free(object);

  return counter;
}




