/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2008 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "dbdir_p.h"
#include "table_l.h"
#include "rollback_l.h"
#include <aqdatabase/aqdb_db_be.h>
#include <aqdatabase/aqdb_object_be.h>
#include <aqdatabase/backendsupport/aqdb_sessionlog.h>
#include <aqdatabase/aqdb_sessiondescr.h>

#include <gwenhywfar/misc.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/stringlist.h>

#include <errno.h>
#include <string.h>


GWEN_INHERIT(AQDB_DB, AQDB_DB_DIR)

#define AQDB_DBDIR_LOCK_TIMEOUT 30000



AQDB_DB *AQDB_DbDir_new(const char *urlString)
{
  AQDB_DB *db;
  AQDB_DB_DIR *xdb;
  const GWEN_URL *url;

  db=AQDB_DB_new(urlString);
  if (db==NULL)
    return NULL;

  GWEN_NEW_OBJECT(AQDB_DB_DIR, xdb);
  if (xdb==NULL) {
    AQDB_DB_free(db);
    return NULL;
  }
  GWEN_INHERIT_SETDATA(AQDB_DB, AQDB_DB_DIR, db, xdb,
                       AQDB_DbDir_FreeData);

  url=AQDB_DB_GetUrl(db);
  if (url) {
    GWEN_BUFFER *pbuf;
    const char *s;

    pbuf=GWEN_Buffer_new(0, 256, 0, 1);
    s=GWEN_Url_GetServer(url);
    if (s)
      GWEN_Buffer_AppendString(pbuf, s);
    s=GWEN_Url_GetPath(url);
    if (s) {
      if (GWEN_Buffer_GetUsedBytes(pbuf))
        GWEN_Buffer_AppendString(pbuf, GWEN_DIR_SEPARATOR_S);
      GWEN_Buffer_AppendString(pbuf, s);
    }
    xdb->dataFolder=strdup(GWEN_Buffer_GetStart(pbuf));
    GWEN_Buffer_free(pbuf);
  }

  AQDB_DB_SetOpenFn(db, AQDB_DbDir_Open);
  AQDB_DB_SetCreateFn(db, AQDB_DbDir_Create);
  AQDB_DB_SetCloseFn(db, AQDB_DbDir_Close);

  AQDB_DB_SetOpenTableFn(db, AQDB_DbDir_OpenTable);
  AQDB_DB_SetCreateTableFn(db, AQDB_DbDir_CreateTable);
  AQDB_DB_SetCloseTableFn(db, AQDB_DbDir_CloseTable);
  AQDB_DB_SetGetTableNameListFn(db, AQDB_DbDir_GetTableNameList);

  AQDB_DB_SetReadObjectFn(db, AQDB_DbDir_ReadObject);
  AQDB_DB_SetWriteObjectFn(db, AQDB_DbDir_WriteObject);
  AQDB_DB_SetDeleteObjectFn(db, AQDB_DbDir_DeleteObject);
  AQDB_DB_SetAddObjectFn(db, AQDB_DbDir_AddObject);

  AQDB_DB_SetQuerySubmitFn(db, AQDB_DbDir_QuerySubmit);
  AQDB_DB_SetQueryGetFirstFn(db, AQDB_DbDir_QueryGetFirst);
  AQDB_DB_SetQueryGetNextFn(db, AQDB_DbDir_QueryGetNext);
  AQDB_DB_SetQueryCloseFn(db, AQDB_DbDir_QueryClose);

  AQDB_DB_SetBeginEditFn(db, AQDB_DbDir_BeginEdit);
  AQDB_DB_SetEndEditFn(db, AQDB_DbDir_EndEdit);

  AQDB_DB_SetGetColumnListFn(db, AQDB_DbDir_GetColumnList);

  AQDB_DB_SetGetRollbackSessionsFn(db, AQDB_DbDir_GetRollbackSessions);
  AQDB_DB_SetUndoSessionFn(db, AQDB_DbDir_UndoSession);

  return db;
}



void GWENHYWFAR_CB AQDB_DbDir_FreeData(void *bp, void *p)
{
  AQDB_DB_DIR *xdb;

  xdb=(AQDB_DB_DIR *)p;
  assert(xdb);

  if (xdb->dirLock) {
    DBG_WARN(AQDB_LOGDOMAIN, "Database still locked, unlocking");
    GWEN_FSLock_Unlock(xdb->dirLock);
    GWEN_FSLock_free(xdb->dirLock);
  }

  free(xdb->dataFolder);
  free(xdb->currentUserName);

  GWEN_FREE_OBJECT(xdb);
}



const char *AQDB_DbDir_GetCurrentUserName(const AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  return xdb->currentUserName;
}



void AQDB_DbDir_SetCurrentUserName(AQDB_DB *db, const char *s)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  free(xdb->currentUserName);
  if (s)
    xdb->currentUserName=strdup(s);
  else
    xdb->currentUserName=NULL;
}



const char *AQDB_DbDir_GetDataFolder(const AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  return xdb->dataFolder;
}



int GWENHYWFAR_CB AQDB_DbDir_Open(AQDB_DB *db, uint32_t flags)
{
  if (!(flags & AQDB_ACTION_FLAGS_READ) &&
      !(flags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN,
              "Bad combination of flags given");
    return GWEN_ERROR_INVALID;
  }
  else {
    AQDB_DB_DIR *xdb;
    int rv;

    assert(db);
    xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
    assert(xdb);

    if (xdb->dataFolder==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN,
                "No data folder in URL");
      return GWEN_ERROR_INVALID;
    }

    rv=GWEN_Directory_GetPath(xdb->dataFolder,
                              GWEN_PATH_FLAGS_PATHMUSTEXIST |
                              GWEN_PATH_FLAGS_NAMEMUSTEXIST |
                              GWEN_PATH_FLAGS_CHECKROOT);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return GWEN_ERROR_NOT_FOUND;
    }

    xdb->openFlags=flags;

    return 0;
  }
}



int GWENHYWFAR_CB AQDB_DbDir_Create(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_DIR *xdb;
  int rv;
  AQDB_COLUMN_LIST *cl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->dataFolder==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN,
              "No data folder in URL");
    return GWEN_ERROR_INVALID;
  }

  if (1) {
    GWEN_STRINGLIST *sl;

    DBG_INFO(AQDB_LOGDOMAIN,
             "Checking whether database at [%s] already exists",
             xdb->dataFolder);
    sl=GWEN_StringList_new();
    rv=GWEN_Directory_GetFileEntries(xdb->dataFolder, sl, "*.table");
    if (rv==0 && GWEN_StringList_Count(sl)) {
      GWEN_StringList_free(sl);
      DBG_ERROR(AQDB_LOGDOMAIN, "Database already exists");
      return GWEN_ERROR_FOUND;
    }
    GWEN_StringList_free(sl);
  }

  rv=GWEN_Directory_GetPath(xdb->dataFolder,
                            /*GWEN_PATH_FLAGS_NAMEMUSTNOTEXIST | */
                            GWEN_PATH_FLAGS_CHECKROOT);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return GWEN_ERROR_FOUND;
  }

  /* open or create session table */
  xdb->sessionTable=AQDB_DbDir_Table_new(db, AQDB_TABLE_NAME_SESSIONTABLE, AQDB_DB_DIR_TID_SESSIONLOG);
  /* create table */
  cl=AQDB_SessionLog_CreateColumns();
  rv=AQDB_DbDir_TableCreate(xdb->sessionTable,
                            AQDB_ACTION_FLAGS_READ |
                            AQDB_ACTION_FLAGS_WRITE |
                            AQDB_DBDIR_TABLE_FLAGS_VARIABLE_LENGTH,
                            cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbDir_Table_free(xdb->sessionTable);
    xdb->sessionTable=NULL;
    return rv;
  }

  /* open or create rollback table */
  xdb->rollbackTable=AQDB_DbDir_Table_new(db, AQDB_TABLE_NAME_ROLLBACKTABLE, AQDB_DB_DIR_TID_ROLLBACKLOG);
  /* create table */
  cl=AQDB_SessionLog_CreateColumns();
  rv=AQDB_DbDir_TableCreate(xdb->rollbackTable,
                            AQDB_ACTION_FLAGS_READ |
                            AQDB_ACTION_FLAGS_WRITE |
                            AQDB_DBDIR_TABLE_FLAGS_VARIABLE_LENGTH,
                            cl);
  AQDB_Column_List_free(cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbDir_Table_free(xdb->sessionTable);
    xdb->sessionTable=NULL;
    AQDB_DbDir_Table_free(xdb->rollbackTable);
    xdb->rollbackTable=NULL;
    return rv;
  }

  /* done, reset the data pointers */
  AQDB_DbDir_Table_free(xdb->sessionTable);
  xdb->sessionTable=NULL;
  AQDB_DbDir_Table_free(xdb->rollbackTable);
  xdb->rollbackTable=NULL;

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_Close(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->openFlags) {
    int i;

    for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
      AQDB_DBDIR_TABLE *table=xdb->tableList[i];

      if (table) {
        AQDB_DbDir_Table_free(table);
        xdb->tableList[i]=NULL;
      }
    }

    AQDB_DbDir_Table_free(xdb->sessionTable);
    xdb->sessionTable=NULL;
    AQDB_DbDir_Table_free(xdb->rollbackTable);
    xdb->rollbackTable=NULL;

    return 0;
  }
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "DB not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}



int AQDB_DbDir_FindTableByName(AQDB_DB *db,
                               const char *tname,
                               AQDB_DBDIR_TABLE **pTable)
{
  AQDB_DB_DIR *xdb;
  int i;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
    AQDB_DBDIR_TABLE *table=xdb->tableList[i];

    if (table && strcasecmp(AQDB_DbDir_Table_GetName(table), tname)==0) {
      *pTable=table;
      return i;
    }
  }

  return GWEN_ERROR_NOT_FOUND;
}



int GWENHYWFAR_CB AQDB_DbDir_OpenTable(AQDB_DB *db,
                                       uint32_t aflags,
                                       const char *tname,
                                       uint32_t flags,
                                       AQDB_ID *pTableId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;
  AQDB_ID tid=0;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tname && tname[0]=='_') {
    if (strcasecmp(tname, AQDB_TABLE_NAME_SESSIONTABLE)==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be opened by applications");
      return GWEN_ERROR_INVALID;
    }
    else if (strcasecmp(tname, AQDB_TABLE_NAME_ROLLBACKTABLE)==0) {
      rv=AQDB_DbDir_EnsureRollbackTable(db);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
      *pTableId=AQDB_DB_DIR_TID_ROLLBACKLOG;
      return 0;
    }
  }

  rv=AQDB_DbDir_FindTableByName(db, tname, &table);
  if (rv==GWEN_ERROR_NOT_FOUND) {
    int i;

    for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
      if (xdb->tableList[i]==NULL)
        break;
    }
    if (i>=AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN,
                "Too many tables in use");
      return GWEN_ERROR_GENERIC;
    }

    tid=i+1;
    table=AQDB_DbDir_Table_new(db, tname, tid);
    xdb->tableList[i]=table;
  }
  else
    tid=rv+1;

  /* return id in any case */
  *pTableId=tid;
  rv=AQDB_DbDir_TableOpen(table, (flags & xdb->openFlags));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



AQDB_COLUMN_LIST *GWENHYWFAR_CB AQDB_DbDir_GetColumnList(AQDB_DB *db,
                                                         uint32_t aflags,
                                                         AQDB_ID tableId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  const AQDB_COLUMN_LIST *clOrig;
  AQDB_COLUMN_LIST *clNew;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG ||
      tableId==AQDB_DB_DIR_TID_ROLLBACKLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Unable to close internal table %d", (int) tableId);
    return NULL;
  }

  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return NULL;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d (unused)", tableId);
    return NULL;
  }

  clOrig=AQDB_DbDir_TableGetColumnList(table);
  if (clOrig==NULL || AQDB_Column_List_GetCount(clOrig)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table id %d has no columns", tableId);
    return NULL;
  }

  /* copy the column list and return that copy */
  clNew=AQDB_Column_List_dup(clOrig);
  return clNew;
}



int AQDB_DbDir_GetColumnCount(AQDB_DB *db, AQDB_ID tableId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  const AQDB_COLUMN_LIST *clOrig;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG)
    table=xdb->sessionTable;
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
      return GWEN_ERROR_INVALID;
    }
    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d (unused)", tableId);
    return GWEN_ERROR_INVALID;
  }

  clOrig=AQDB_DbDir_TableGetColumnList(table);
  if (clOrig==NULL || AQDB_Column_List_GetCount(clOrig)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table id %d has no columns", tableId);
    return 0;
  }

  return AQDB_Column_List_GetCount(clOrig);
}



int GWENHYWFAR_CB AQDB_DbDir_CreateTable(AQDB_DB *db,
                                         const char *tname,
                                         uint32_t flags,
                                         const AQDB_COLUMN_LIST *columns)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tname && tname[0]=='_') {
    if (strcasecmp(tname, AQDB_TABLE_NAME_SESSIONTABLE)==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Unable to create internal table [%s]", tname);
      return GWEN_ERROR_INVALID;
    }
    else if (strcasecmp(tname, AQDB_TABLE_NAME_ROLLBACKTABLE)==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Unable to create internal table [%s]", tname);
      return GWEN_ERROR_INVALID;
    }
  }

  rv=AQDB_DbDir_FindTableByName(db, tname, &table);
  if (rv==GWEN_ERROR_NOT_FOUND) {
    int i;

    for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
      if (xdb->tableList[i]==NULL)
        break;
    }
    if (i>=AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN,
                "Too many tables in use");
      return GWEN_ERROR_GENERIC;
    }

    table=AQDB_DbDir_Table_new(db, tname, i+1);
    xdb->tableList[i]=table;
  }

  rv=AQDB_DbDir_TableCreate(table,
                            (flags & xdb->openFlags),
                            columns);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_CloseTable(AQDB_DB *db,
                                        AQDB_ID tableId,
                                        uint32_t flags)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG ||
      tableId==AQDB_DB_DIR_TID_ROLLBACKLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Unable to close internal table %d", (int) tableId);
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return GWEN_ERROR_INVALID;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d (unused)", tableId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableClose(table, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbDir_ReallyReadObject(AQDB_DB *db,
                                AQDB_ID tableId,
                                AQDB_ID objectId,
                                AQDB_OBJECT **pObject)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG)
    table=xdb->sessionTable;
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableReadObject(table, objectId, pObject);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_ReadObject(AQDB_DB *db,
                                        uint32_t aflags,
                                        AQDB_ID tableId,
                                        AQDB_ID objectId,
                                        AQDB_OBJECT **pObject)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be read");
    return GWEN_ERROR_INVALID;
  }
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->dirLock!=NULL) {
    /* in edit mode, try to read the object from the session log first */
    rv=AQDB_DbDir_ReadObjectFromSessionLog(db, tableId, objectId, pObject);
    if (rv==0) {
      DBG_DEBUG(AQDB_LOGDOMAIN, "Returning modified object");
      return rv;
    }
    else if (rv!=GWEN_ERROR_NO_DATA) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  /* only arrives here if not in edit mode or object not found in session log */
  rv=AQDB_DbDir_ReallyReadObject(db, tableId, objectId, pObject);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  return 0;
}



int AQDB_DbDir_ReallyWriteObject(AQDB_DB *db, const AQDB_OBJECT *object)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  AQDB_ID tableId;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (%lu out of range)",
              (unsigned long int) tableId);
    return GWEN_ERROR_INVALID;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableWriteObject(table, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  AQDB_ID tableId;
  AQDB_ID objectId;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return GWEN_ERROR_INVALID;
  }

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id (0)");
    abort();
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->dirLock==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_WriteSessionLog(db, AQDB_SessionLogType_ChgObject,
                                AQDB_Object_GetTableId(object),
                                AQDB_Object_GetObjectId(object),
                                object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  xdb->sessionModifications++;

  return 0;
}



int AQDB_DbDir_ReallyDeleteObject(AQDB_DB *db,
                                  AQDB_ID tableId,
                                  AQDB_ID objectId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG) {
    rv=AQDB_DbDir_EnsureRollbackTable(db);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    table=xdb->rollbackTable;
  }
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (%lu out of range)",
                (unsigned long int) tableId);
      abort();
      return GWEN_ERROR_INVALID;
    }
    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableDeleteObject(table, objectId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_DeleteObject(AQDB_DB *db,
                                          AQDB_ID tableId,
                                          AQDB_ID objectId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return GWEN_ERROR_INVALID;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->dirLock==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_WriteSessionLog(db, AQDB_SessionLogType_DelObject, tableId, objectId, NULL);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  xdb->sessionDeletions++;

  /* mark as "deleted" */
  AQDB_DbDir_Table_EditObjectDeleted(table, objectId);

  return 0;
}



int AQDB_DbDir_ReallyAddObject(AQDB_DB *db,
                               AQDB_ID tableId,
                               AQDB_OBJECT *object)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return GWEN_ERROR_INVALID;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableAddObject(table, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_AddObject(AQDB_DB *db,
                                       AQDB_ID tableId,
                                       AQDB_OBJECT *object)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
    return GWEN_ERROR_INVALID;
  }

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->dirLock==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked");
    return GWEN_ERROR_INVALID;
  }

  /* only reserve object, so that it gets a new id */
  DBG_INFO(AQDB_LOGDOMAIN, "Reserving entry in table [%s]", AQDB_DbDir_Table_GetName(table));
  rv=AQDB_DbDir_TableReserveObject(table, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_DbDir_WriteSessionLog(db, AQDB_SessionLogType_AddObject,
                                AQDB_Object_GetTableId(object),
                                AQDB_Object_GetObjectId(object),
                                object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  xdb->sessionAdditions++;

  /* mark as "added" */
  AQDB_DbDir_Table_EditObjectAdded(table, AQDB_Object_GetObjectId(object));

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_QuerySubmit(AQDB_DB *db,
                                         uint32_t aflags,
                                         AQDB_ID tableId,
                                         AQDB_ID refQueryId,
                                         const char *queryExpression,
                                         uint32_t flags,
                                         AQDB_ID *pQueryId)
{
  AQDB_DB_DIR *xdb;
  AQDB_ID refTableId;
  AQDB_DBDIR_TABLE *table;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  refTableId=(refQueryId>>(AQDB_ID_BITS/2));
  if (refTableId && refTableId!=tableId) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid reference query id (not the same table)");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be read");
    return GWEN_ERROR_INVALID;
  }
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG) {
    DBG_ERROR(0, "Using rollback log");
    rv=AQDB_DbDir_EnsureRollbackTable(db);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    table=xdb->rollbackTable;
  }
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %lu (unused)", (unsigned long int) tableId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableQuerySubmit(table,
                                 (refQueryId & AQDB_ID_BOTTOMHALFMASK),
                                 queryExpression,
                                 flags,
                                 pQueryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  *pQueryId|=(tableId << (AQDB_ID_BITS/2));

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_QueryGetFirst(AQDB_DB *db,
                                           AQDB_ID queryId,
                                           AQDB_ID *pIdListPtr,
                                           int *pIdListLen)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;
  AQDB_ID tableId;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tableId=(queryId>>(AQDB_ID_BITS/2));

  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be read");
    return GWEN_ERROR_INVALID;
  }
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d (out of range)", tableId);
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableQueryGetFirst(table,
                                   (queryId & AQDB_ID_BOTTOMHALFMASK),
                                   pIdListPtr,
                                   pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_QueryGetNext(AQDB_DB *db,
                                          AQDB_ID queryId,
                                          AQDB_ID *pIdListPtr,
                                          int *pIdListLen)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;
  AQDB_ID tableId;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tableId=(queryId>>(AQDB_ID_BITS/2));
  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be read");
    return GWEN_ERROR_INVALID;
  }
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id %d (out of range)", tableId);
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableQueryGetNext(table,
                                  (queryId & AQDB_ID_BOTTOMHALFMASK),
                                  pIdListPtr,
                                  pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_QueryClose(AQDB_DB *db, AQDB_ID queryId)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  int rv;
  AQDB_ID tableId;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tableId=(queryId>>(AQDB_ID_BITS/2));
  if (tableId==AQDB_DB_DIR_TID_SESSIONLOG) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Internal session table can't be read");
    return GWEN_ERROR_INVALID;
  }
  else if (tableId==AQDB_DB_DIR_TID_ROLLBACKLOG)
    table=xdb->rollbackTable;
  else {
    if (tableId==0 || tableId>AQDB_DB_DIR_MAX_OPEN_TABLES) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (out of range)");
      return GWEN_ERROR_INVALID;
    }

    table=xdb->tableList[tableId-1];
  }

  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_DbDir_TableQueryClose(table, (queryId & AQDB_ID_BOTTOMHALFMASK));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbDir_GetNextSessionId(AQDB_DB *db, AQDB_ID *pId)
{
  AQDB_DB_DIR *xdb;
  GWEN_BUFFER *tbuf;
  FILE *f;
  AQDB_ID id=0;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  GWEN_Buffer_AppendString(tbuf, xdb->dataFolder);
  GWEN_Buffer_AppendString(tbuf, GWEN_DIR_SEPARATOR_S);
  GWEN_Buffer_AppendString(tbuf, "_lastsession");

  f=fopen(GWEN_Buffer_GetStart(tbuf), "r+");
  if (f) {
    unsigned long int v;

    if (1!=fscanf(f, "%lu", &v))
      v=0;
    id=(AQDB_ID) v;
  }
  else {
    f=fopen(GWEN_Buffer_GetStart(tbuf), "w+");
    if (f==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Unable to create file [%s]: %s",
                GWEN_Buffer_GetStart(tbuf),
                strerror(errno));
      GWEN_Buffer_free(tbuf);
      return GWEN_ERROR_IO;
    }
  }

  rewind(f);
  id++;

  fprintf(f, "%lu\n", (unsigned long int) id);
  if (fclose(f)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Unable to close file [%s]: %s",
              GWEN_Buffer_GetStart(tbuf),
              strerror(errno));
    GWEN_Buffer_free(tbuf);
    return GWEN_ERROR_IO;
  }
  GWEN_Buffer_free(tbuf);

  *pId=id;
  return 0;
}



int AQDB_DbDir_WriteSessionDescr(AQDB_DB *db, int t, AQDB_ID *pId)
{
  AQDB_DB_DIR *xdb;
  AQDB_SESSION_DESCR *sd;
  AQDB_SESSION_LOG *sl;
  AQDB_OBJECT *o;
  GWEN_TIME *ti;
  GWEN_BUFFER *buf;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* create session descriptor */
  ti=GWEN_CurrentTime();
  sd=AQDB_SessionDescr_new();
  AQDB_SessionDescr_SetSessionId(sd, xdb->currentSessionId);
  AQDB_SessionDescr_SetUserName(sd, xdb->currentUserName);
  AQDB_SessionDescr_SetDateTime(sd, ti);
  GWEN_Time_free(ti);
  if (t==AQDB_SessionLogType_EndEdit) {
    AQDB_SessionDescr_SetAdditions(sd, xdb->sessionAdditions);
    AQDB_SessionDescr_SetDeletions(sd, xdb->sessionDeletions);
    AQDB_SessionDescr_SetModifications(sd, xdb->sessionModifications);
  }

  rv=AQDB_SessionDescr_CreateObject(db, 0, 0, &o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_SessionDescr_free(sd);
    return rv;
  }
  rv=AQDB_SessionDescr_toObject(sd, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    AQDB_SessionDescr_free(sd);
    return rv;
  }
  AQDB_SessionDescr_free(sd);

  buf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AQDB_DbDir_Object_toBuffer(o, buf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(buf);
    AQDB_Object_free(o);
    AQDB_SessionDescr_free(sd);
    return rv;
  }

  /* create session log */
  sl=AQDB_SessionLog_new();
  AQDB_SessionLog_SetType(sl, t);
  AQDB_SessionLog_SetSessionId(sl, xdb->currentSessionId);
  AQDB_SessionLog_SetData(sl, GWEN_Buffer_GetStart(buf), GWEN_Buffer_GetUsedBytes(buf));
  GWEN_Buffer_free(buf);
  AQDB_Object_free(o);
  o=NULL;

  /* write session log to object */
  rv=AQDB_SessionLog_CreateObject(db, AQDB_DB_DIR_TID_SESSIONLOG, 0, &o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_SessionLog_free(sl);
    return rv;
  }
  rv=AQDB_SessionLog_toObject(sl, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    AQDB_SessionLog_free(sl);
    return rv;
  }
  AQDB_SessionLog_free(sl);

  /* directly append object to table */
  rv=AQDB_DbDir_TableAddObject(xdb->sessionTable, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return rv;
  }
  if (pId)
    *pId=AQDB_Object_GetObjectId(o);
  xdb->lastSessionObjectId=AQDB_Object_GetObjectId(o);
  AQDB_Object_free(o);

  return 0;
}



int AQDB_DbDir_WriteSessionLog(AQDB_DB *db, int t,
                               AQDB_ID tableId,
                               AQDB_ID objectId,
                               const AQDB_OBJECT *object)
{
  AQDB_DB_DIR *xdb;
  AQDB_SESSION_LOG *sl;
  AQDB_OBJECT *o;
  AQDB_DBDIR_TABLE *table;
  AQDB_DBDIR_OBJDESCR *od;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* create session log */
  sl=AQDB_SessionLog_new();
  AQDB_SessionLog_SetType(sl, t);
  AQDB_SessionLog_SetSessionId(sl, xdb->currentSessionId);
  AQDB_SessionLog_SetTableName(sl, AQDB_DbDir_Table_GetName(xdb->tableList[tableId-1]));
  AQDB_SessionLog_SetTableId(sl, tableId);
  AQDB_SessionLog_SetObjectId(sl, objectId);

  if (object) {
    GWEN_BUFFER *buf;

    buf=GWEN_Buffer_new(0, 256, 0, 1);
    rv=AQDB_DbDir_Object_toBuffer(object, buf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(buf);
      AQDB_SessionLog_free(sl);
      return rv;
    }

    AQDB_SessionLog_SetData(sl, GWEN_Buffer_GetStart(buf), GWEN_Buffer_GetUsedBytes(buf));
    GWEN_Buffer_free(buf);
  }

  /* write session log to object */
  rv=AQDB_SessionLog_CreateObject(db, AQDB_DB_DIR_TID_SESSIONLOG, 0, &o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_SessionLog_free(sl);
    return rv;
  }
  rv=AQDB_SessionLog_toObject(sl, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    AQDB_SessionLog_free(sl);
    return rv;
  }
  AQDB_SessionLog_free(sl);

  /* directly append object to table */
  rv=AQDB_DbDir_TableAddObject(xdb->sessionTable, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return rv;
  }

  /* speedup: cache ids of objects in session log */
  table=xdb->tableList[tableId-1];
  assert(table);
  od=AQDB_DbDir_ObjectDescr_new();
  AQDB_DbDir_ObjectDescr_SetSessionLogId(od, AQDB_Object_GetObjectId(o));
  AQDB_DbDir_ObjectDescr_SetObjectTableId(od, tableId);
  AQDB_DbDir_ObjectDescr_SetObjectId(od, objectId);
  AQDB_DbDir_ObjectDescr_SetEntryType(od, t);
  AQDB_DbDir_Table_AddSessionObjectDescr(table, od);

  xdb->lastSessionObjectId=AQDB_Object_GetObjectId(o);
  AQDB_Object_free(o);

  return 0;
}



int GWENHYWFAR_CB AQDB_DbDir_BeginEdit(AQDB_DB *db, uint32_t flags, const char *userName)
{
  AQDB_DB_DIR *xdb;
  GWEN_FSLOCK *lck;
  GWEN_FSLOCK_RESULT res;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->dirLock!=NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Already locked");
    return GWEN_ERROR_INVALID;
  }

  xdb->firstSessionObjectId=0;
  xdb->lastSessionObjectId=0;
  xdb->sessionAdditions=0;
  xdb->sessionDeletions=0;
  xdb->sessionModifications=0;

  lck=GWEN_FSLock_new(xdb->dataFolder, GWEN_FSLock_TypeDir);
  res=GWEN_FSLock_Lock(lck, AQDB_DBDIR_LOCK_TIMEOUT, 0);
  switch (res) {
  case GWEN_FSLock_ResultOk:
    xdb->dirLock=lck;
    break;
  case GWEN_FSLock_ResultError:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock error");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_IO;
  case GWEN_FSLock_ResultTimeout:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_TIMEOUT;
  case GWEN_FSLock_ResultBusy:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout (busy)");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_TIMEOUT;
  case GWEN_FSLock_ResultNoLock:
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid lock");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_IO;
  case GWEN_FSLock_ResultUserAbort:
    DBG_INFO(AQDB_LOGDOMAIN, "User aborted");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_USER_ABORTED;
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unknown error");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_GENERIC;
  }

  xdb->sessionFlags=flags;

  /* set username for session logs */
  AQDB_DbDir_SetCurrentUserName(db, userName);

  if (!(xdb->sessionFlags & AQDB_DB_DIR_SESSION_FLAGS_ROLLBACK)) {
    /* make sure there are session tables */
    rv=AQDB_DbDir_EnsureSessionTables(db);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_FSLock_Unlock(xdb->dirLock);
      GWEN_FSLock_free(xdb->dirLock);
      xdb->dirLock=NULL;
      return rv;
    }

    /* create session id */
    rv=AQDB_DbDir_GetNextSessionId(db, &(xdb->currentSessionId));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_FSLock_Unlock(xdb->dirLock);
      GWEN_FSLock_free(xdb->dirLock);
      xdb->dirLock=NULL;
      return rv;
    }

    /* add entry to session log */
    rv=AQDB_DbDir_WriteSessionDescr(db, AQDB_SessionLogType_BeginEdit, &(xdb->firstSessionObjectId));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_FSLock_Unlock(xdb->dirLock);
      GWEN_FSLock_free(xdb->dirLock);
      xdb->dirLock=NULL;
      return rv;
    }
  }

  return 0;
}



int AQDB_DbDir_Flush(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int i;
  int errors=0;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
    AQDB_DBDIR_TABLE *table=xdb->tableList[i];

    if (table) {
      rv=AQDB_DbDir_Table_Flush(table);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        errors++;
      }
    }
  }

  if (xdb->sessionTable) {
    rv=AQDB_DbDir_Table_Flush(xdb->sessionTable);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      errors++;
    }
  }

  if (xdb->rollbackTable) {
    rv=AQDB_DbDir_Table_Flush(xdb->rollbackTable);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      errors++;
    }
  }

  if (errors) {
    DBG_INFO(AQDB_LOGDOMAIN, "Some errors occurred (%d errors)", errors);
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}



void AQDB_DbDir_AllTablesEditEnd(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int i;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  for (i=0; i<AQDB_DB_DIR_MAX_OPEN_TABLES; i++) {
    AQDB_DBDIR_TABLE *table=xdb->tableList[i];

    if (table)
      AQDB_DbDir_Table_EditEnd(table);
  }
}



int GWENHYWFAR_CB AQDB_DbDir_EndEdit(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_DIR *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->dirLock==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked");
    return GWEN_ERROR_INVALID;
  }

  /* add entry to session log */
  if (!(flags & AQDB_ACTION_FLAGS_ABORT)) {
    if (!(xdb->sessionFlags & AQDB_DB_DIR_SESSION_FLAGS_ROLLBACK)) {
      rv=AQDB_DbDir_WriteSessionDescr(db, AQDB_SessionLogType_EndEdit, NULL);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }

      /* create rollback log from session log */
      rv=AQDB_DbDir_Session2Rollback(db);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }

      /* now commit the changes from the session log */
      rv=AQDB_DbDir_CommitSession(db);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
    }

    AQDB_DbDir_AllTablesEditEnd(db);

    /* flush all tables */
    rv=AQDB_DbDir_Flush(db);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    /* abort */
    AQDB_DbDir_AllTablesEditEnd(db);
    /* flush but ignore return value */
    AQDB_DbDir_Flush(db);
  }

  GWEN_FSLock_Unlock(xdb->dirLock);
  GWEN_FSLock_free(xdb->dirLock);
  xdb->dirLock=NULL;

  return 0;
}



int AQDB_DbDir_WriteRollbackLog(AQDB_DB *db, const AQDB_SESSION_LOG *sl)
{
  AQDB_DB_DIR *xdb;
  AQDB_OBJECT *o;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* write session log to object */
  rv=AQDB_SessionLog_CreateObject(db, AQDB_DB_DIR_TID_ROLLBACKLOG, 0, &o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  rv=AQDB_SessionLog_toObject(sl, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return rv;
  }

  /* directly append object to table */
  rv=AQDB_DbDir_TableAddObject(xdb->rollbackTable, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return rv;
  }
  AQDB_Object_free(o);

  return 0;
}



int AQDB_DbDir_Session2Rollback(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int rv;
  AQDB_ID id;
  int done=0;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* create a rollback log for every session log */
  for (id=xdb->firstSessionObjectId; done==0; id++) {
    AQDB_OBJECT *o=NULL;
    AQDB_SESSION_LOG *sl=NULL;
    int doWrite=1;

    /* read session log */
    rv=AQDB_DbDir_TableReadObject(xdb->sessionTable, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    rv=AQDB_SessionLog_fromObject(o, &sl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      return rv;
    }
    AQDB_Object_free(o);
    o=NULL;

    /* modify session log for rollback */
    switch (AQDB_SessionLog_GetType(sl)) {
    case AQDB_SessionLogType_BeginEdit:
#if 0
      DBG_DEBUG(AQDB_LOGDOMAIN, "BeginEdit");
#endif
      /* no modification necessary */
      break;

    case AQDB_SessionLogType_EndEdit:
#if 0
      DBG_DEBUG(AQDB_LOGDOMAIN, "EndEdit");
#endif
      /* no modification necessary, but last object of this session */
      done=1;
      break;

    case AQDB_SessionLogType_AddObject:
      /* an object was added, so we don't need an object for rollback (because
       * it didn't exist before this operation) */
#if 0
      DBG_DEBUG(AQDB_LOGDOMAIN, "AddObject [%s]", AQDB_SessionLog_GetTableName(sl));
#endif
      AQDB_SessionLog_SetType(sl, AQDB_SessionLogType_DelObject);
      AQDB_SessionLog_SetData(sl, NULL, 0);
      break;

    case AQDB_SessionLogType_DelObject: {
      GWEN_BUFFER *buf;

#if 0
      DBG_DEBUG(AQDB_LOGDOMAIN, "DelObject [%s]", AQDB_SessionLog_GetTableName(sl));
#endif
      /* object was deleted, so we store the object */
      rv=AQDB_DbDir_ReallyReadObject(db, AQDB_SessionLog_GetTableId(sl), AQDB_SessionLog_GetObjectId(sl), &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_SessionLog_free(sl);
        return rv;
      }

      buf=GWEN_Buffer_new(0, 256, 0, 1);
      rv=AQDB_DbDir_Object_toBuffer(o, buf);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        GWEN_Buffer_free(buf);
        AQDB_SessionLog_free(sl);
        return rv;
      }

      AQDB_SessionLog_SetData(sl, GWEN_Buffer_GetStart(buf), GWEN_Buffer_GetUsedBytes(buf));
      GWEN_Buffer_free(buf);

      /* reverse operation of DelObject is AddObject */
      AQDB_SessionLog_SetType(sl, AQDB_SessionLogType_AddObject);

      AQDB_Object_free(o);
      o=NULL;
      break;
    }

    case AQDB_SessionLogType_ChgObject: {
      GWEN_BUFFER *buf;
      AQDB_ID tableId;
      AQDB_ID objectId;
      AQDB_DBDIR_TABLE *table;

#if 0
      DBG_DEBUG(AQDB_LOGDOMAIN, "ChgObject [%s]", AQDB_SessionLog_GetTableName(sl));
#endif
      tableId=AQDB_SessionLog_GetTableId(sl);
      objectId=AQDB_SessionLog_GetObjectId(sl);
      assert(objectId);
      table=xdb->tableList[tableId-1];
      if (AQDB_DbDir_Table_EditIsObjectAdded(table, objectId)==0) {
        /* object was modified, so we store the previous object */
        rv=AQDB_DbDir_ReallyReadObject(db, tableId, objectId, &o);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          AQDB_SessionLog_free(sl);
          return rv;
        }

        buf=GWEN_Buffer_new(0, 256, 0, 1);
        rv=AQDB_DbDir_Object_toBuffer(o, buf);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          GWEN_Buffer_free(buf);
          AQDB_SessionLog_free(sl);
          return rv;
        }

        AQDB_SessionLog_SetData(sl, GWEN_Buffer_GetStart(buf), GWEN_Buffer_GetUsedBytes(buf));
        GWEN_Buffer_free(buf);

        AQDB_Object_free(o);
        o=NULL;
      }
      else {
#if 0
        DBG_DEBUG(AQDB_LOGDOMAIN,
                  "Object just added, not writing previous object (because there was none)");
#endif
        doWrite=0;
      }
      break;
    }

    case AQDB_SessionLogType_Unknown:
      DBG_ERROR(AQDB_LOGDOMAIN, "Unknown session log entry type %d", AQDB_SessionLog_GetType(sl));
      AQDB_SessionLog_free(sl);
      return GWEN_ERROR_BAD_DATA;
    }

    /* now add object to rollback table */
    if (doWrite) {
      rv=AQDB_DbDir_WriteRollbackLog(db, sl);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_SessionLog_free(sl);
        return rv;
      }
    }
    AQDB_SessionLog_free(sl);
  } /* for */

  return 0;
}



int AQDB_DbDir_CommitSession(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int rv;
  AQDB_ID id;
  int done=0;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* create a rollback log for every session log */
  for (id=xdb->firstSessionObjectId; done==0; id++) {
    AQDB_OBJECT *o=NULL;
    AQDB_SESSION_LOG *sl=NULL;

    /* read session log */
    rv=AQDB_DbDir_TableReadObject(xdb->sessionTable, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    assert(o);

    rv=AQDB_SessionLog_fromObject(o, &sl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      return rv;
    }
    AQDB_Object_free(o);
    o=NULL;

    /* commit session log */
    switch (AQDB_SessionLog_GetType(sl)) {
    case AQDB_SessionLogType_BeginEdit:
      /* nothing to do */
      break;

    case AQDB_SessionLogType_EndEdit:
      /* nothing to do, but last object of this session */
      done=1;
      break;

    case AQDB_SessionLogType_AddObject: {
      int cnt;

      /* now really add object */
      cnt=AQDB_DbDir_GetColumnCount(db, AQDB_SessionLog_GetTableId(sl));
      if (cnt<1) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", cnt);
        AQDB_SessionLog_free(sl);
        return cnt;
      }
      o=AQDB_DbDir_Object_fromBuffer(AQDB_SessionLog_GetTableId(sl),
                                     AQDB_SessionLog_GetObjectId(sl),
                                     cnt,
                                     AQDB_SessionLog_GetDataPtr(sl),
                                     AQDB_SessionLog_GetDataLen(sl));
      assert(o);
      /* only write object here, because it has already been added to the table with
       * a status!=ACTIVE, so it already is there; we just have to overwrite it now */
      rv=AQDB_DbDir_ReallyWriteObject(db, o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_Object_free(o);
        AQDB_SessionLog_free(sl);
        return rv;
      }
      AQDB_Object_free(o);
      o=NULL;
      break;
    }

    case AQDB_SessionLogType_DelObject: {
      /* really delete object now */
      rv=AQDB_DbDir_ReallyDeleteObject(db, AQDB_SessionLog_GetTableId(sl), AQDB_SessionLog_GetObjectId(sl));
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_SessionLog_free(sl);
        return rv;
      }
      break;
    }

    case AQDB_SessionLogType_ChgObject: {
      int cnt;

      /* now really write object */
      cnt=AQDB_DbDir_GetColumnCount(db, AQDB_SessionLog_GetTableId(sl));
      if (cnt<1) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", cnt);
        AQDB_SessionLog_free(sl);
        return cnt;
      }
      o=AQDB_DbDir_Object_fromBuffer(AQDB_SessionLog_GetTableId(sl),
                                     AQDB_SessionLog_GetObjectId(sl),
                                     cnt,
                                     AQDB_SessionLog_GetDataPtr(sl),
                                     AQDB_SessionLog_GetDataLen(sl));
      assert(o);
      rv=AQDB_DbDir_ReallyWriteObject(db, o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_Object_free(o);
        AQDB_SessionLog_free(sl);
        return rv;
      }
      AQDB_Object_free(o);
      o=NULL;
      break;
    }

    case AQDB_SessionLogType_Unknown:
      DBG_ERROR(AQDB_LOGDOMAIN, "Unknown session log entry type %d", AQDB_SessionLog_GetType(sl));
      AQDB_SessionLog_free(sl);
      return GWEN_ERROR_BAD_DATA;
    }
    AQDB_SessionLog_free(sl);
  } /* for */

  return 0;
}



int AQDB_DbDir_ReadObjectFromSessionLog(AQDB_DB *db,
                                        AQDB_ID tableId,
                                        AQDB_ID objectId,
                                        AQDB_OBJECT **pObject)
{
  AQDB_DB_DIR *xdb;
  int rv;
  AQDB_OBJECT *objectFound=NULL;
  AQDB_ID id;
  AQDB_DBDIR_TABLE *table;
  int colCount;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  /* speedup: directly get session object id instead of searching through all */
  table=xdb->tableList[tableId-1];
  assert(table);

  colCount=AQDB_DbDir_TableGetColumnCount(table);
  assert(colCount);

  id=AQDB_DbDir_Table_GetSessionObjectId(table, objectId);
  if (id) {
    AQDB_OBJECT *o=NULL;
    AQDB_SESSION_LOG *sl=NULL;

    /* read session log */
    rv=AQDB_DbDir_TableReadObject(xdb->sessionTable, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    assert(o);

    rv=AQDB_SessionLog_fromObject(o, &sl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      return rv;
    }
    AQDB_Object_free(o);
    o=NULL;

    /* commit session log */
    switch (AQDB_SessionLog_GetType(sl)) {
    case AQDB_SessionLogType_BeginEdit:
    case AQDB_SessionLogType_EndEdit:
      /* should not happen */
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid session object type, aborting");
      abort();
      break;

    case AQDB_SessionLogType_AddObject:
    case AQDB_SessionLogType_ChgObject:
      if (AQDB_SessionLog_GetTableId(sl)==tableId &&
          AQDB_SessionLog_GetObjectId(sl)==objectId) {
        /* read object from session log */
        o=AQDB_DbDir_Object_fromBuffer(tableId, objectId,
                                       colCount,
                                       AQDB_SessionLog_GetDataPtr(sl),
                                       AQDB_SessionLog_GetDataLen(sl));
        assert(o);
        if (objectFound)
          AQDB_Object_free(objectFound);
        objectFound=o;
      }
      break;

    case AQDB_SessionLogType_DelObject:
      /* object deleted */
      return GWEN_ERROR_NOT_FOUND;

    case AQDB_SessionLogType_Unknown:
      DBG_ERROR(AQDB_LOGDOMAIN, "Unknown session log entry type %d", AQDB_SessionLog_GetType(sl));
      AQDB_SessionLog_free(sl);
      return GWEN_ERROR_BAD_DATA;
    }
    AQDB_SessionLog_free(sl);

    *pObject=objectFound;
    return 0;
  }
  else
    return GWEN_ERROR_NO_DATA;
}



int AQDB_DbDir_EnsureSessionTable(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->sessionTable==NULL) {
    /* open or create session table */
    xdb->sessionTable=AQDB_DbDir_Table_new(db, AQDB_TABLE_NAME_SESSIONTABLE, AQDB_DB_DIR_TID_SESSIONLOG);
    rv=AQDB_DbDir_TableOpen(xdb->sessionTable, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
    if (rv==GWEN_ERROR_NOT_FOUND) {
      AQDB_COLUMN_LIST *cl;

      /* create table */
      DBG_INFO(AQDB_LOGDOMAIN, "Creating session table");
      cl=AQDB_SessionLog_CreateColumns();
      rv=AQDB_DbDir_TableCreate(xdb->sessionTable,
                                AQDB_ACTION_FLAGS_READ |
                                AQDB_ACTION_FLAGS_WRITE |
                                AQDB_DBDIR_TABLE_FLAGS_VARIABLE_LENGTH,
                                cl);
      AQDB_Column_List_free(cl);
      if (rv>=0)
        rv=AQDB_DbDir_TableOpen(xdb->sessionTable, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
    }
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbDir_Table_free(xdb->sessionTable);
      xdb->sessionTable=NULL;
      return rv;
    }
  }

  /* reset session table */
  rv=AQDB_DbDir_TableReset(xdb->sessionTable);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbDir_Table_free(xdb->sessionTable);
    xdb->sessionTable=NULL;
    return rv;
  }

  return 0;
}



int AQDB_DbDir_EnsureRollbackTable(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->rollbackTable==NULL) {
    /* open or create rollback table */
    xdb->rollbackTable=AQDB_DbDir_Table_new(db, AQDB_TABLE_NAME_ROLLBACKTABLE, AQDB_DB_DIR_TID_ROLLBACKLOG);
    rv=AQDB_DbDir_TableOpen(xdb->rollbackTable, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
    if (rv==GWEN_ERROR_NOT_FOUND) {
      AQDB_COLUMN_LIST *cl;

      /* create table */
      DBG_INFO(AQDB_LOGDOMAIN, "Creating missing table");
      cl=AQDB_SessionLog_CreateColumns();
      rv=AQDB_DbDir_TableCreate(xdb->rollbackTable,
                                AQDB_ACTION_FLAGS_READ |
                                AQDB_ACTION_FLAGS_WRITE |
                                AQDB_DBDIR_TABLE_FLAGS_VARIABLE_LENGTH,
                                cl);
      AQDB_Column_List_free(cl);
      if (rv>=0)
        rv=AQDB_DbDir_TableOpen(xdb->rollbackTable, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
    }
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbDir_Table_free(xdb->rollbackTable);
      xdb->rollbackTable=NULL;
      return rv;
    }
  }

  return 0;
}



int AQDB_DbDir_EnsureSessionTables(AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  rv=AQDB_DbDir_EnsureSessionTable(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_DbDir_EnsureRollbackTable(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbDir_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList)
{
  AQDB_DB_DIR *xdb;
  int rv;
  const GWEN_URL *url;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  url=AQDB_DB_GetUrl(db);
  if (url) {
    GWEN_BUFFER *pbuf;
    const char *s;
    GWEN_STRINGLIST *fileNameList;

    pbuf=GWEN_Buffer_new(0, 256, 0, 1);
    s=GWEN_Url_GetServer(url);
    if (s)
      GWEN_Buffer_AppendString(pbuf, s);
    s=GWEN_Url_GetPath(url);
    if (s) {
      if (GWEN_Buffer_GetUsedBytes(pbuf))
        GWEN_Buffer_AppendString(pbuf, GWEN_DIR_SEPARATOR_S);
      GWEN_Buffer_AppendString(pbuf, s);
    }

    fileNameList=GWEN_StringList_new();
    rv=GWEN_Directory_GetFileEntries(GWEN_Buffer_GetStart(pbuf), fileNameList, "*.table");
    if (rv<0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_StringList_free(fileNameList);
      GWEN_Buffer_free(pbuf);
      return rv;
    }
    GWEN_Buffer_free(pbuf);

    if (GWEN_StringList_Count(fileNameList)) {
      GWEN_STRINGLISTENTRY *se;

      se=GWEN_StringList_FirstEntry(fileNameList);
      while (se) {
        const char *s;

        s=GWEN_StringListEntry_Data(se);
        if (s && *s && *s!='_') {
          GWEN_BUFFER *tbuf;
          char *t;

          tbuf=GWEN_Buffer_new(0, 256, 0, 1);
          GWEN_Text_UnescapeToBufferTolerant(s, tbuf);
          t=strstr(GWEN_Buffer_GetStart(tbuf), ".table");
          if (t)
            *t=0;
          GWEN_StringList_AppendString(tableNameList, GWEN_Buffer_GetStart(tbuf), 0, 1);
          GWEN_Buffer_free(tbuf);
        }
        se=GWEN_StringListEntry_Next(se);
      }
    }
    GWEN_StringList_free(fileNameList);
  }

  return 0;
}



int AQDB_DbDir_AdjustIdList(AQDB_DB *db, AQDB_ID tableId, GWEN_IDLIST64 *idl)
{
  AQDB_DB_DIR *xdb;
  AQDB_DBDIR_TABLE *table;
  AQDB_DBDIR_OBJDESCR_LIST *odl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  if (xdb->sessionFlags & AQDB_DB_DIR_SESSION_FLAGS_ROLLBACK)
    /* don't adjust anything in rollback mode */
    return 0;

  table=xdb->tableList[tableId-1];
  if (table==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id (unused)");
    return GWEN_ERROR_INVALID;
  }

  odl=AQDB_DbDir_Table_GetSessionObjectList(table);
  if (odl) {
    AQDB_DBDIR_OBJDESCR *od;

    od=AQDB_DbDir_ObjectDescr_List_First(odl);
    while (od) {
      AQDB_ID objectId;

      objectId=AQDB_DbDir_ObjectDescr_GetObjectId(od);

      switch (AQDB_DbDir_ObjectDescr_GetEntryType(od)) {
      case AQDB_SessionLogType_BeginEdit:
      case AQDB_SessionLogType_EndEdit:
        break;

      case AQDB_SessionLogType_AddObject:
      case AQDB_SessionLogType_ChgObject:
        if (GWEN_IdList64_HasId(idl, objectId)==0)
          GWEN_IdList64_AddId(idl, objectId);
        break;

      case AQDB_SessionLogType_DelObject:
        /* object deleted */
        GWEN_IdList64_DelId(idl, objectId);
        break;

      case AQDB_SessionLogType_Unknown:
        DBG_ERROR(AQDB_LOGDOMAIN, "Unknown session log entry type %d",
                  AQDB_DbDir_ObjectDescr_GetEntryType(od));
        return GWEN_ERROR_BAD_DATA;
      }
      od=AQDB_DbDir_ObjectDescr_List_Next(od);
    }
  }

  return 0;
}



int AQDB_DbDir_IsInEditMode(const AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  return (xdb->dirLock==NULL)?0:1;
}



uint32_t AQDB_DbDir_GetSessionFlags(const AQDB_DB *db)
{
  AQDB_DB_DIR *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_DIR, db);
  assert(xdb);

  return xdb->sessionFlags;
}



AQDB_OBJECT *AQDB_DbDir_Object_fromBuffer(AQDB_ID tableId,
                                          AQDB_ID objectId,
                                          uint32_t nfields,
                                          const uint8_t *pStart,
                                          uint32_t bsize)
{
  AQDB_OBJECT *o;
  uint32_t j;
  const uint8_t *p;
  const uint8_t *pEnd;

  assert(nfields);

  o=AQDB_Object_new(tableId, objectId, nfields);
  assert(o);

  /* scan buffer and set fields accordingly */
  p=pStart;
  pEnd=p+bsize;
  if (pEnd<p) {
    DBG_INFO(AQDB_LOGDOMAIN, "Empty buffer");
    AQDB_Object_free(o);
    return NULL;
  }

  j=0;
  while (p<pEnd) {
    uint32_t len;
    uint8_t c;

    c=*(p++);
    switch (c) {
    case 0xf1:
      len=*(p++);
      break;

    case 0xf2:
      len=(*(p++))<<8;
      len+=*(p++);
      break;

    case 0xf3:
      len=(*(p++))<<16;
      len+=(*(p++))<<8;
      len+=*(p++);
      break;

    case 0xf4:
      len=(*(p++))<<24;
      len+=(*(p++))<<16;
      len+=(*(p++))<<8;
      len+=*(p++);
      break;

    case 0xf5:
    case 0xf6:
    case 0xf7:
    case 0xf8:
    case 0xf9:
    case 0xfa:
    case 0xfb:
    case 0xfc:
    case 0xfd:
    case 0xfe:
    case 0xff:
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size specifier for field (%d)", j);
      AQDB_Object_free(o);
      return NULL;

    default:
      len=c<<8;
      len+=*(p++);
      break;
    }

    AQDB_Object_SetField(o, j, p, len);
    p+=len;
    j++;
  }
  if (p!=pEnd) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Incomplete data in buffer");
    AQDB_Object_free(o);
    return NULL;
  }

  return o;
}



int AQDB_DbDir_Object_toBuffer(const AQDB_OBJECT *o, GWEN_BUFFER *buf)
{
  uint32_t fieldCount;
  uint32_t j;

  fieldCount=AQDB_Object_GetFieldCount(o);
  for (j=0; j<fieldCount; j++) {
    uint8_t *pData;
    uint32_t len;
    int rv;
    uint8_t lh[5];

    rv=AQDB_Object_GetField(o, j, &pData, &len);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    if (len>0xffffff) {
      lh[0]=0xf4;
      lh[1]=(len>>24) & 0xff;
      lh[2]=(len>>16) & 0xff;
      lh[3]=(len>>8) & 0xff;
      lh[4]=len & 0xff;
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 5);
    }
    else if (len>0xffff) {
      lh[0]=0xf3;
      lh[1]=(len>>16) & 0xff;
      lh[2]=(len>>8) & 0xff;
      lh[3]=len & 0xff;
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 4);
    }
    else if (len>0xf000) {
      lh[0]=0xf2;
      lh[1]=(len>>8) & 0xff;
      lh[2]=len & 0xff;
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 3);
    }
    else {
      lh[0]=(len>>8) & 0xff;
      lh[1]=len & 0xff;
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 2);
    }

    if (len>0)
      GWEN_Buffer_AppendBytes(buf, (const char *) pData, len);
  }

  return 0;
}




