/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


/*
 * This tool scans a given AQFS file and checks whether it contains valid objects.
 */


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include <aqfilestorage/aqfs_file.h>
#include <aqfilestorage/aqfs_item_header.h>


#include <gwenhywfar/gwenhywfar.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>



#define MAX_ITEM_LEN (1024*1024*10)

//#define DEBUG_OUTPUT


static uint64_t _getRealFileSize(const char *fileName);
static int checkDataItemAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level);
static int checkTableItemOnlyAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level);
static int checkTableItemDeepAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level);
static int64_t checkItemAt(AQFS_FILE *fsf, uint64_t offset, int withRecursion, int level);


static uint64_t g_num_of_tables=0;
static uint64_t g_num_of_items=0;



uint64_t _getRealFileSize(const char *fileName)
{
  struct stat sb;

  if (lstat(fileName, &sb)==-1) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error on lstat: %d", errno);
    return 0;
  }

  return (uint64_t) sb.st_size;
}



int AQFS_CheckFileSize(const char *fileName)
{
  AQFS_FILE *fsf;
  int rv;
  uint64_t fileSize;
  uint64_t realFileSize;

  realFileSize=_getRealFileSize(fileName);
  if (realFileSize==0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error getting real file size");
    return 2;
  }

  fsf=AQFS_File_new(fileName);

  rv=AQFS_File_Open(fsf, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Could not open file (%d)\n", rv);
    return 2;
  }

  fileSize=AQFS_File_GetFileSize(fsf);
  if (fileSize==0) {
    fprintf(stderr, "ERROR: Empty file.\n");
    AQFS_File_Close(fsf);
    AQFS_File_free(fsf);
    return 2;
  }

  AQFS_File_Close(fsf);
  AQFS_File_free(fsf);

  if (fileSize>realFileSize) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Noted filesize is greater than real file size (%llu > %llu)",
              (unsigned long long int) fileSize,
              (unsigned long long int) realFileSize);
    return GWEN_ERROR_BAD_DATA;
  }
  else if (fileSize<realFileSize) {
    DBG_WARN(AQFS_LOGDOMAIN, "Noted filesize is lesser than real file size (%llu < %llu)",
             (unsigned long long int) fileSize,
             (unsigned long long int) realFileSize);
  }

  return 0;
}



int AQFS_CheckFileSequentially(const char *fileName, uint64_t *pNumTables, uint64_t *pNumItems)
{
  AQFS_FILE *fsf;
  int rv;
  uint64_t offset;
  uint64_t realFileSize;
  uint64_t fileSize;

  g_num_of_tables=0;
  g_num_of_items=0;


  realFileSize=_getRealFileSize(fileName);
  if (realFileSize==0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error getting real file size");
    return 2;
  }

  fsf=AQFS_File_new(fileName);

  rv=AQFS_File_Open(fsf, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Could not open file (%d)\n", rv);
    return 2;
  }

  fileSize=AQFS_File_GetFileSize(fsf);
  if (fileSize==0) {
    fprintf(stderr, "ERROR: Empty file.\n");
    AQFS_File_Close(fsf);
    AQFS_File_free(fsf);
    return 2;
  }

  offset=AQFS_File_GetFirstItemOffset(fsf);
  if (offset==0) {
    fprintf(stderr, "ERROR: First offset is zero, invalid or empty file\n");
    AQFS_File_Close(fsf);
    AQFS_File_free(fsf);
    return 2;
  }

  while (offset<realFileSize) { /* fileSize */
    int64_t checkResult;

    checkResult=checkItemAt(fsf, offset, 0, 1); /* no recursion */
    if (checkResult<0) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "ERROR: Could not check item at offset %llu, file size is %llu (%d)",
                (unsigned long long int) offset,
                (unsigned long long int) fileSize,
                (int) checkResult);
      return 2;
    }
    offset+=checkResult;
  }

  AQFS_File_Close(fsf);

  AQFS_File_free(fsf);

  if (pNumTables)
    *pNumTables=g_num_of_tables;
  if (pNumItems)
    *pNumItems=g_num_of_items;

  return 0;
}



int AQFS_CheckFileStructure(const char *fileName, uint64_t *pNumTables, uint64_t *pNumItems)
{
  AQFS_FILE *fsf;
  int rv;
  uint64_t offset;
  uint64_t fileSize;
  uint64_t realFileSize;

  g_num_of_tables=0;
  g_num_of_items=0;

  realFileSize=_getRealFileSize(fileName);
  if (realFileSize==0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error getting real file size");
    return 2;
  }

  fsf=AQFS_File_new(fileName);

  rv=AQFS_File_Open(fsf, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Could not open file (%d)\n", rv);
    return 2;
  }

  fileSize=AQFS_File_GetFileSize(fsf);
  if (fileSize==0) {
    fprintf(stderr, "ERROR: Empty file.\n");
    AQFS_File_Close(fsf);
    AQFS_File_free(fsf);
    return 2;
  }

  offset=AQFS_File_GetFirstItemOffset(fsf);
  if (offset==0) {
    fprintf(stderr, "ERROR: First offset is zero, invalid or empty file\n");
    AQFS_File_Close(fsf);
    AQFS_File_free(fsf);
    return 2;
  }

  while (offset<fileSize) {
    AQFS_ITEM_HEADER *ih;

    ih=AQFS_ItemHeader_new();
    rv=AQFS_File_ReadItemHeaderAt(fsf, offset, ih);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "ERROR: Could not read file at offset %llu (%d)",
                (unsigned long long int) offset, rv);
      return rv;
    }

    if (AQFS_ItemHeader_GetDataType(ih)==AQFS_ITEMHEADER_TYPE_GENERIC_TABLE) {
      rv=checkTableItemDeepAt(fsf, offset, ih, 1);
      if (rv<0) {
        DBG_ERROR(AQFS_LOGDOMAIN,
                  "ERROR: Could not check table item at offset %llu (%d)",
                  (unsigned long long int) offset, rv);
        AQFS_ItemHeader_free(ih);
        return rv;
      }
    }
    offset+=AQFS_ItemHeader_GetSizeAllocated(ih)+AQFS_ITEM_HEADER_SIZE;
    AQFS_ItemHeader_free(ih);
  } /* while */

  AQFS_File_Close(fsf);

  AQFS_File_free(fsf);

  if (pNumTables)
    *pNumTables=g_num_of_tables;
  if (pNumItems)
    *pNumItems=g_num_of_items;

  return 0;
}



int64_t checkItemAt(AQFS_FILE *fsf, uint64_t offset, int withRecursion, int level)
{
  AQFS_ITEM_HEADER *ih;
  int rv;
  int64_t sizeOfItem;

  ih=AQFS_ItemHeader_new();
  rv=AQFS_File_ReadItemHeaderAt(fsf, offset, ih);
  if (rv<0) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "ERROR[%d]: Could not read file at offset %llu (%d)",
              level, (unsigned long long int) offset, rv);
    return rv;
  }

  switch (AQFS_ItemHeader_GetDataType(ih)) {
  case AQFS_ITEMHEADER_TYPE_DELETED:
#ifdef DEBUG_OUTPUT
    fprintf(stderr, "%08llu[%03d]: Deleted item or table\n", (unsigned long long int) offset, level);
#endif
    /* deleted object, don't handle further */
    break;

  case AQFS_ITEMHEADER_TYPE_GENERIC_DATA:
#ifdef DEBUG_OUTPUT
    fprintf(stderr, "%08llu[%03d]: Item\n", (unsigned long long int) offset, level);
#endif
    rv=checkDataItemAt(fsf, offset, ih, level);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "ERROR[%d]: Could not check data item at offset %llu (%d)",
                level, (unsigned long long int) offset, rv);
      AQFS_ItemHeader_free(ih);
      return rv;
    }
    break;

  case AQFS_ITEMHEADER_TYPE_GENERIC_TABLE:
#ifdef DEBUG_OUTPUT
    fprintf(stderr, "%08llu[%03d]: Table\n", (unsigned long long int) offset, level);
#endif
    if (withRecursion)
      rv=checkTableItemDeepAt(fsf, offset, ih, level);
    else
      rv=checkTableItemOnlyAt(fsf, offset, ih, level);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "ERROR[%d]: Could not check table item at offset %llu (%d)",
                level, (unsigned long long int) offset, rv);
      AQFS_ItemHeader_free(ih);
      return rv;
    }
    break;

  default:
    DBG_ERROR(AQFS_LOGDOMAIN,
              "ERROR[%d]: Invalid data type at offset %llu (%d)",
              level,
              (unsigned long long int) offset,
              (int) AQFS_ItemHeader_GetDataType(ih));
    AQFS_ItemHeader_free(ih);
    return GWEN_ERROR_BAD_DATA;
  }

  sizeOfItem=AQFS_ItemHeader_GetSizeAllocated(ih)+AQFS_ITEM_HEADER_SIZE;

  AQFS_ItemHeader_free(ih);
  return sizeOfItem;
}



int checkDataItemAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level)
{
  int64_t sizeRead;
  uint8_t *ptr=NULL;

  g_num_of_items++;

  sizeRead=AQFS_File_ReadDataAt(fsf, offset, &ptr, MAX_ITEM_LEN);
  if (sizeRead<0) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "ERROR[%d]: Could not read data item at offset %llu (%d)",
              level,
              (unsigned long long int) offset,
              (int) sizeRead);
    return GWEN_ERROR_BAD_DATA;
  }
  free(ptr);

  return 0;
}



int checkTableItemOnlyAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level)
{
  AQFS_TABLE *ft;
  int rv;

  g_num_of_tables++;

  ft=AQFS_Table_new();
  rv=AQFS_File_ReadTableAt(fsf, offset, ft);
  if (rv<0) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "ERROR[%d]: Could not read file at offset %llu (%d)",
              level, (unsigned long long int) offset, rv);
    AQFS_Table_free(ft);
    return rv;
  }

  AQFS_Table_free(ft);
  return 0;
}



int checkTableItemDeepAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih, int level)
{
  int rv;
  uint64_t nextOffset;
  AQFS_TABLE *ft;
  uint64_t maxTableEntries;

  g_num_of_tables++;

  ft=AQFS_Table_new();
  rv=AQFS_File_ReadTableAt(fsf, offset, ft);
  if (rv<0) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "ERROR[%d]: Could not read file at offset %llu (%d)",
              level, (unsigned long long int) offset, rv);
    AQFS_Table_free(ft);
    return rv;
  }

  maxTableEntries=AQFS_Table_GetMaxEntries(ft);
  if (maxTableEntries) {
    uint64_t *tableEntries;

    tableEntries=AQFS_Table_GetPtrEntries(ft);
    if (tableEntries) {
      uint64_t idx;

      for (idx=0; idx<maxTableEntries; idx++) {
        uint64_t tableEntryOffset;

        tableEntryOffset=tableEntries[idx];
        if ((tableEntryOffset!=0) &&
            (tableEntryOffset!=0xffffffffffffffffL)) {
          int64_t checkResult;

          checkResult=checkItemAt(fsf, tableEntryOffset, 1, level+1); /* recursion */
          if (checkResult<0) {
            DBG_ERROR(AQFS_LOGDOMAIN,
                      "ERROR[%d]: Could not check item at offset %llu, index %llu (%d)",
                      level,
                      (unsigned long long int) tableEntryOffset,
                      (unsigned long long int) idx,
                      (int) checkResult);
            AQFS_Table_free(ft);
            return (int) checkResult;
          }
        }
      } /* for */
    }
  }

  /* check pointer to next table */
  nextOffset=AQFS_Table_GetOffsetNextTable(ft);
  AQFS_Table_free(ft);

  if (nextOffset) {
    rv=checkTableItemOnlyAt(fsf, nextOffset, ih, level);
    g_num_of_tables--; /* try to avoid double counting */
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "ERROR[%d]: Next table pointer invalid at offset %llu (%d)",
                level,
                (unsigned long long int) offset,
                (int) rv);
      return rv;
    }
  }

  return 0;
}




