/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFinance (c) by 2015 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqfs_file_p.h"


#include "aqfs_file_io.c"


#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>



#define AQFS_FILE_MAX_TABLE_SIZE (64*1024)


/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static uint64_t _getRealFileSize(const char *fileName);
static int _setFileSize(AQFS_FILE *fsf, uint64_t newFileSize);

static int _readFileHeader(AQFS_FILE *fsf, AQFS_FILE_HEADER *fh);
static int _writeFileHeader(AQFS_FILE *fsf, const AQFS_FILE_HEADER *fh);

static int _writeForcedAt(AQFS_FILE *fsf, uint64_t offset, const uint8_t *ptr, uint64_t len);




/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */



AQFS_FILE *AQFS_File_new(const char *fname)
{
  AQFS_FILE *fsf;

  GWEN_NEW_OBJECT(AQFS_FILE, fsf)
  fsf->refCount=1;

  if (fname && *fname)
    fsf->dataFileName=strdup(fname);

  return fsf;
}



void AQFS_File_free(AQFS_FILE *fsf)
{
  if (fsf) {
    assert(fsf->refCount);
    if (fsf->refCount==1) {
      /* close if not already done */
      if (fsf->isOpen)
        AQFS_File_close(fsf);

      /* release resources */
      AQFS_FileHeader_free(fsf->fileHeader);
      fsf->fileHeader=NULL;
      free(fsf->dataFileName);
      fsf->dataFileName=NULL;
      fsf->refCount=0;
      GWEN_FREE_OBJECT(fsf);
    }
    else
      fsf->refCount--;
  }
}



uint64_t AQFS_File_GetFirstItemOffset(AQFS_FILE *fsf)
{
  assert(fsf);
  assert(fsf->refCount);

  return AQFS_FILE_HEADER_SIZE;
}



uint64_t AQFS_File_GetFileSize(const AQFS_FILE *fsf)
{
  assert(fsf);
  assert(fsf->refCount);

  if (!(fsf->isOpen)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "File not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  assert(fsf->fileHeader);
  return AQFS_FileHeader_GetFileSize(fsf->fileHeader);
}




int AQFS_File_Open(AQFS_FILE *fsf, int rw)
{
  int rv;

  assert(fsf);
  assert(fsf->refCount);

  if (fsf->isOpen) {
    DBG_ERROR(AQFS_LOGDOMAIN, "File already open");
    return GWEN_ERROR_OPEN;
  }
  else {
    uint64_t realFileSize;

    realFileSize=_getRealFileSize(fsf->dataFileName);

    /* open file */
    rv=AQFS_File_open(fsf, rw);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    fsf->rw=rw;

    /* read file header */
    if (fsf->fileHeader)
      AQFS_FileHeader_free(fsf->fileHeader);
    fsf->fileHeader=AQFS_FileHeader_new();
    rv=_readFileHeader(fsf, fsf->fileHeader);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    if (realFileSize) {
      uint64_t notedFileSize;

      notedFileSize=AQFS_FileHeader_GetFileSize(fsf->fileHeader);
      if (notedFileSize!=realFileSize) {
        DBG_WARN(AQFS_LOGDOMAIN, "Noted filesize does not equal real file size (noted = %llu, real = %llu), fixing",
                 (unsigned long long int) notedFileSize,
                 (unsigned long long int) realFileSize);
        if (rw) {
          DBG_WARN(AQFS_LOGDOMAIN, "Fixing noted filesize");
          rv=_setFileSize(fsf, realFileSize);
          if (rv<0) {
            DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
            AQFS_File_close(fsf);
            if (fsf->fileHeader)
              AQFS_FileHeader_free(fsf->fileHeader);
            fsf->fileHeader=NULL;
            return rv;
          }
        }
      }
    }

    return 0;
  }
}



int AQFS_File_Create(AQFS_FILE *fsf)
{
  int rv;

  assert(fsf);
  assert(fsf->refCount);

  if (fsf->isOpen) {
    DBG_ERROR(AQFS_LOGDOMAIN, "File already open");
    return GWEN_ERROR_OPEN;
  }
  else {
    /* open file */
    rv=AQFS_File_creat(fsf);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    fsf->rw=1;

    /* create and write file header */
    if (fsf->fileHeader)
      AQFS_FileHeader_free(fsf->fileHeader);
    fsf->fileHeader=AQFS_FileHeader_new();
    AQFS_FileHeader_SetVersion(fsf->fileHeader, AQFS_FILE_VERSION);
    AQFS_FileHeader_SetFileSize(fsf->fileHeader, AQFS_FILE_HEADER_SIZE);

    rv=_writeFileHeader(fsf, fsf->fileHeader);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      AQFS_File_close(fsf);
      return rv;
    }

    return 0;
  }
}



int AQFS_File_Close(AQFS_FILE *fsf)
{
  int rv;

  assert(fsf);
  assert(fsf->refCount);

  if (fsf->isOpen) {
    if (fsf->rw) {
      /* write file header */
      assert(fsf->fileHeader);
      rv=_writeFileHeader(fsf, fsf->fileHeader);
      if (rv<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
        AQFS_File_close(fsf);
        return rv;
      }
    }

    /* close file */
    rv=AQFS_File_close(fsf);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    return 0;
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}



int64_t AQFS_File_ReadAt(AQFS_FILE *fsf, uint64_t offset, uint8_t *ptr, uint64_t len)
{
  if (fsf->isOpen) {
    int rv;
    size_t rd;

    /* seek to position */
    rv=AQFS_File_seek(fsf, offset);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "here (%d)", rv);
      return (int64_t) rv;
    }

    /* read data from that position */
    rd=AQFS_File_read(fsf, ptr, 1, len);
    if (rd<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) rd);
    }

    return (int64_t) rd;
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}




int AQFS_File_ReadForcedAt(AQFS_FILE *fsf, uint64_t offset, uint8_t *ptr, uint64_t len)
{
  if (fsf->isOpen) {
    int rv;
    size_t rd;

    /* seek to position */
    rv=AQFS_File_seek(fsf, offset);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "here (%d)", rv);
      return (int64_t) rv;
    }

    /* read data from that position */
    rd=AQFS_File_read(fsf, ptr, len, 1);
    if (rd<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    else if (rd!=1) {
      DBG_ERROR(AQFS_LOGDOMAIN, "EOF met");
      return GWEN_ERROR_EOF;
    }

    return 0;
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}



int AQFS_File_WriteForcedAt(AQFS_FILE *fsf, uint64_t offset, const uint8_t *ptr, uint64_t len)
{
  int rv;

  if (offset<AQFS_FILE_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Writing data into first %d bytes would overwrite file header, rejecting!! (%llu)",
              AQFS_FILE_HEADER_SIZE,
              (long long unsigned int) offset);
    return GWEN_ERROR_INVALID;
  }

  rv=_writeForcedAt(fsf, offset, ptr, len);
  if (rv<0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _writeForcedAt(AQFS_FILE *fsf, uint64_t offset, const uint8_t *ptr, uint64_t len)
{
  if (fsf->isOpen) {
    if (fsf->rw) {
      int rv;
      size_t rd;

      /* seek to position */
      rv=AQFS_File_seek(fsf, offset);
      if (rv<0) {
        DBG_ERROR(AQFS_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }

      /* write data to that position */
      rd=AQFS_File_write(fsf, ptr, len, 1);
      if (rd<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) rd);
        return (int) rd;
      }
      else if (rd!=1) {
        DBG_ERROR(AQFS_LOGDOMAIN, "Disk full");
        return GWEN_ERROR_IO;
      }

      return 0;
    }
    else {
      DBG_ERROR(AQFS_LOGDOMAIN, "File not open for R/W");
      return GWEN_ERROR_NOT_OPEN;
    }
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}



int AQFS_File_Sync(AQFS_FILE *fsf)
{
  if (fsf->isOpen) {
    int rv;

    /* sync */
    rv=AQFS_File_sync(fsf);
    if (rv<0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
    return 0;
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }
}




int AQFS_File_ReadItemHeaderAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih)
{
  int rv;
  uint64_t fileSize;
  uint64_t l;
  uint32_t itemDataType;
  uint8_t hbuf[AQFS_ITEM_HEADER_SIZE];

  fileSize=AQFS_File_GetFileSize(fsf);

  /* read data bytes */
  rv=AQFS_File_ReadForcedAt(fsf, offset, hbuf, AQFS_ITEM_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* parse data bytes */
  rv=AQFS_ItemHeader_ReadFromBuffer(ih, hbuf, AQFS_ITEM_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  AQFS_ItemHeader_SetOffsetHeaderStart(ih, offset);
  AQFS_ItemHeader_SetOffsetDataStart(ih, offset+AQFS_ITEM_HEADER_SIZE);

  /* check item type (there are only two!) */
  itemDataType=AQFS_ItemHeader_GetDataType(ih);
  if (itemDataType!=AQFS_ITEMHEADER_TYPE_GENERIC_DATA &&
      itemDataType!=AQFS_ITEMHEADER_TYPE_GENERIC_TABLE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Invalid item data type (%llu)", (unsigned long long int) itemDataType);
    return GWEN_ERROR_BAD_DATA;
  }

  /* check sizes */
  l=AQFS_ItemHeader_GetSizeAllocated(ih);
  if ((offset+l+AQFS_ITEM_HEADER_SIZE)>fileSize) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Invalid allocated object size (%llu)", (unsigned long long int) l);
    return GWEN_ERROR_BAD_DATA;
  }

  l=AQFS_ItemHeader_GetSizeUsed(ih);
  if ((offset+l+AQFS_ITEM_HEADER_SIZE)>fileSize) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Invalid used object size (%llu)", (unsigned long long int) l);
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}



int AQFS_File_WriteItemHeaderAt(AQFS_FILE *fsf, uint64_t offset, const AQFS_ITEM_HEADER *ih)
{
  int rv;
  uint8_t hbuf[AQFS_ITEM_HEADER_SIZE];

  /* store header data in temporary buffer */
  rv=AQFS_ItemHeader_toBuffer(ih, hbuf, AQFS_ITEM_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* write temporary buffer to file */
  rv=AQFS_File_WriteForcedAt(fsf, offset, hbuf, AQFS_ITEM_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _readFileHeader(AQFS_FILE *fsf, AQFS_FILE_HEADER *fh)
{
  int rv;
  uint8_t hbuf[AQFS_FILE_HEADER_SIZE];

  /* read data bytes */
  rv=AQFS_File_ReadForcedAt(fsf, 0, hbuf, AQFS_FILE_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* parse data bytes */
  rv=AQFS_FileHeader_ReadFromBuffer(fh, hbuf, AQFS_FILE_HEADER_SIZE);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _writeFileHeader(AQFS_FILE *fsf, const AQFS_FILE_HEADER *fh)
{
  if (fsf->isOpen) {
    if (fsf->rw) {
      int rv;
      uint8_t hbuf[AQFS_FILE_HEADER_SIZE];

      /* store header data in temporary buffer */
      rv=AQFS_FileHeader_toBuffer(fh, hbuf, AQFS_FILE_HEADER_SIZE);
      if (rv<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }

      /* write temporary buffer to file */
      rv=_writeForcedAt(fsf, 0, hbuf, AQFS_FILE_HEADER_SIZE);
      if (rv<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
    }
    else {
      DBG_ERROR(AQFS_LOGDOMAIN, "File not open for R/W");
      return GWEN_ERROR_NOT_OPEN;
    }
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "File is not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  return 0;
}



int AQFS_File_WriteDataAt(AQFS_FILE *fsf, uint64_t offset, uint32_t t, const uint8_t *ptr, uint64_t len)
{
  AQFS_ITEM_HEADER *ih;
  int rv;

  /* read stored item header from file */
  ih=AQFS_ItemHeader_new();
  rv=AQFS_File_ReadItemHeaderAt(fsf, offset, ih);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    AQFS_ItemHeader_free(ih);
    return rv;
  }

  /* check whether the item on file is a generic data item */
  if (AQFS_ItemHeader_GetDataType(ih)!=t) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Item in file has a different data type (is %llu, expected %llu)",
              (long long unsigned int) AQFS_ItemHeader_GetDataType(ih),
              (long long unsigned int) t);
    AQFS_ItemHeader_free(ih);
    return GWEN_ERROR_INTERNAL;
  }

  /* check whether new data fits into the item in file */
  if (AQFS_ItemHeader_GetSizeAllocated(ih)<len) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Size of item in file is smaller than current data");
    AQFS_ItemHeader_free(ih);
    return GWEN_ERROR_MEMORY_FULL;
  }

  /* write data if there is any */
  if (len>0) {
    rv=AQFS_File_WriteForcedAt(fsf, AQFS_ItemHeader_GetOffsetDataStart(ih), ptr, len);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      AQFS_ItemHeader_free(ih);
      return rv;
    }
  }

  /* update item header */
  AQFS_ItemHeader_SetSizeUsed(ih, len);
  AQFS_ItemHeader_SetDataType(ih, t);

  /* store updated item header */
  rv=AQFS_File_WriteItemHeaderAt(fsf, offset, ih);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    AQFS_ItemHeader_free(ih);
    return rv;
  }
  AQFS_ItemHeader_free(ih);

  return 0;
}



int64_t AQFS_File_AppendData(AQFS_FILE *fsf, uint32_t t, const uint8_t *ptr, uint64_t len)
{
  AQFS_ITEM_HEADER *ih;
  uint64_t offset;
  int rv;

  assert(fsf);
  assert(fsf->refCount);
  assert(fsf->fileHeader);

  offset=AQFS_FileHeader_GetFileSize(fsf->fileHeader);

  /* create item header */
  ih=AQFS_ItemHeader_new();
  AQFS_ItemHeader_SetSizeAllocated(ih, len);
  AQFS_ItemHeader_SetSizeUsed(ih, len);
  AQFS_ItemHeader_SetDataType(ih, t);
  AQFS_ItemHeader_SetOffsetHeaderStart(ih, offset);
  AQFS_ItemHeader_SetOffsetDataStart(ih, offset+AQFS_ITEM_HEADER_SIZE);

  /* store new item header */
  rv=AQFS_File_WriteItemHeaderAt(fsf, offset, ih);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    AQFS_ItemHeader_free(ih);
    return rv;
  }

  /* write data if there is any */
  if (len>0) {
    rv=AQFS_File_WriteForcedAt(fsf, AQFS_ItemHeader_GetOffsetDataStart(ih), ptr, len);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      AQFS_ItemHeader_free(ih);
      return rv;
    }
  }
  AQFS_ItemHeader_free(ih);

  /* update file header */
  DBG_VERBOUS(AQFS_LOGDOMAIN, "Appended data to file at pos %llu (%llu bytes, new file size: %llu)",
              (unsigned long long int) offset,
              (unsigned long long int) len,
              (unsigned long long int)(offset+AQFS_ITEM_HEADER_SIZE+len));
  AQFS_FileHeader_SetFileSize(fsf->fileHeader, offset+AQFS_ITEM_HEADER_SIZE+len);

  /* write updated file header */
  rv=_writeFileHeader(fsf, fsf->fileHeader);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return offset;
}



int64_t AQFS_File_ReadDataAt(AQFS_FILE *fsf, uint64_t offset, uint8_t **pptr, uint64_t maxLen)
{
  AQFS_ITEM_HEADER *ih;
  uint64_t len;
  int rv;

  ih=AQFS_ItemHeader_new();
  rv=AQFS_File_ReadItemHeaderAt(fsf, offset, ih);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    AQFS_ItemHeader_free(ih);
    return rv;
  }

  /* read payload */
  len=AQFS_ItemHeader_GetSizeUsed(ih);
  if (len) {
    if (maxLen && len>maxLen) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Item too big (%d bytes, maximum is %d bytes)",
                (int) len, (int) maxLen);
      AQFS_ItemHeader_free(ih);
      return GWEN_ERROR_BUFFER_OVERFLOW;
    }
    else {
      uint8_t *ptr;

      ptr=(uint8_t *) malloc(len);
      assert(ptr);

      rv=AQFS_File_ReadForcedAt(fsf, AQFS_ItemHeader_GetOffsetDataStart(ih), ptr, len);
      if (rv<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
        AQFS_ItemHeader_free(ih);
        free(ptr);
        return rv;
      }

      if (pptr)
        *pptr=ptr;
      else
        free(ptr);
    }
  }
  AQFS_ItemHeader_free(ih);

  return len;
}



int AQFS_File_ReadTableAt(AQFS_FILE *fsf, uint64_t offset, AQFS_TABLE *ft)
{
  int rv;
  uint8_t *ptr=NULL;
  int64_t rd;

  /* read data into buffer, which will be allocated by the called function */
  rd=AQFS_File_ReadDataAt(fsf, offset, &ptr, AQFS_FILE_MAX_TABLE_SIZE);
  if (rd<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  else if (rd==0) {
    DBG_INFO(AQFS_LOGDOMAIN, "Empty data");
    return GWEN_ERROR_BAD_DATA;
  }

  /* now ptr contains a pointer to the table data */
  rv=AQFS_Table_ReadFromBuffer(ft, ptr, rd);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    free(ptr);
    return rv;
  }

  /* store table position inside table for later write-back */
  AQFS_Table_SetOffsetTableHeaderStart(ft, offset);

  /* release buffer */
  free(ptr);
  return 0;
}



int AQFS_File_WriteTableAt(AQFS_FILE *fsf, uint64_t offset, const AQFS_TABLE *ft)
{
  int rv;
  uint8_t *ptr=NULL;
  uint64_t len;

  if (offset<AQFS_FILE_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Invalid offset to write a table at (%d)", (int) offset);
    assert(0);
    return GWEN_ERROR_INVALID;
  }

  len=AQFS_Table_Calculate_Size(ft);
  ptr=(uint8_t *) malloc(len);
  len=AQFS_Table_toBuffer(ft, ptr, len);
  if (len==0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Should not happen.");
    free(ptr);
    return GWEN_ERROR_INTERNAL;
  }

  rv=AQFS_File_WriteDataAt(fsf, offset, AQFS_ITEMHEADER_TYPE_GENERIC_TABLE, ptr, len);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    free(ptr);
    return rv;
  }

  /* release buffer */
  free(ptr);
  return 0;
}



int64_t AQFS_File_AppendTable(AQFS_FILE *fsf, AQFS_TABLE *ft)
{
  uint8_t *ptr=NULL;
  uint64_t len;
  int64_t offset;

  len=AQFS_Table_Calculate_Size(ft);
  ptr=(uint8_t *) malloc(len);
  len=AQFS_Table_toBuffer(ft, ptr, len);
  if (len==0) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Should not happen.");
    free(ptr);
    return GWEN_ERROR_INTERNAL;
  }

  offset=AQFS_File_AppendData(fsf, AQFS_ITEMHEADER_TYPE_GENERIC_TABLE, ptr, len);
  if (offset<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) offset);
    free(ptr);
    return (int) offset;
  }
  free(ptr);

  /* store offset for later use */
  AQFS_Table_SetOffsetTableHeaderStart(ft, offset);

  return offset;
}



int AQFS_File_ReadTableList(AQFS_FILE *fsf, uint64_t firstOffset, AQFS_TABLE_LIST *ftl)
{
  uint64_t offset;

  offset=firstOffset;
  while (offset) {
    AQFS_TABLE *ft;
    int rv;

    /* read current table */
    ft=AQFS_Table_new();
    rv=AQFS_File_ReadTableAt(fsf, offset, ft);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    /* next offset is that of the next table, use that */
    offset=AQFS_Table_GetOffsetNextTable(ft);
    AQFS_Table_List_Add(ft, ftl);
  }

  return 0;
}



int AQFS_File_ReadTablePtrList(AQFS_FILE *fsf, uint64_t firstOffset, AQFS_TABLE_PTRLIST *ftl)
{
  uint64_t offset;

  offset=firstOffset;
  while (offset) {
    AQFS_TABLE *ft;
    int rv;
    int64_t tableIdx;

    /* read current table */
    ft=AQFS_Table_new();
    rv=AQFS_File_ReadTableAt(fsf, offset, ft);
    if (rv<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    /* next offset is that of the next table, use that */
    offset=AQFS_Table_GetOffsetNextTable(ft);
    tableIdx=AQFS_Table_PtrList_AddTable(ftl, ft);
    if (tableIdx<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) tableIdx);
      return (int) tableIdx;
    }
    AQFS_Table_free(ft);
  }

  return 0;
}



int64_t AQFS_File_AddTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t entry)
{
  AQFS_TABLE *ft;
  AQFS_TABLE *ftPredecessor=NULL;
  int64_t index=0;
  int i;
  uint64_t *ptr;
  uint64_t maxEntries=AQFS_TABLE_MAX_ENTRIES;

  ftPredecessor=NULL;
  ft=AQFS_Table_List_First(ftl);
  while (ft) {
    if (AQFS_Table_GetFreeEntries(ft))
      break;
    index+=AQFS_Table_GetMaxEntries(ft);
    ftPredecessor=ft;
    ft=AQFS_Table_List_Next(ft);
  }

  if (ft==NULL) {
    int64_t offset;

    if (ftPredecessor)
      maxEntries=AQFS_Table_GetMaxEntries(ftPredecessor);

    /* create new table, add it to the list */
    ft=AQFS_Table_Create(maxEntries);
    assert(ft);

    /* write empty table to file */
    offset=AQFS_File_AppendTable(fsf, ft);
    if (offset<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) offset);
      AQFS_Table_free(ft);
      return (int) offset;
    }

    /* store offset of this table as next table in predecessor, if any */
    if (ftPredecessor) {
      /* set offset of next table */
      AQFS_Table_SetOffsetNextTable(ftPredecessor, offset);
      AQFS_Table_AddRuntimeFlags(ftPredecessor, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
    }

    /* append to list */
    AQFS_Table_List_Add(ft, ftl);
    AQFS_Table_AddRuntimeFlags(ft, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
  }


  /* find free entry */
  ptr=AQFS_Table_GetPtrEntries(ft);
  for (i=0; i<AQFS_Table_GetMaxEntries(ft); i++) {
    if (ptr[i]==0)
      break;
  }

  if (i>=AQFS_Table_GetMaxEntries(ft)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "No free entry found, should not happen");
    assert(0);
    return GWEN_ERROR_INTERNAL;
  }

  /* store new entry, get index */
  ptr[i]=entry;
  index+=i;
  AQFS_Table_DecFreeEntries(ft);
  AQFS_Table_CheckAndSetHighestEntry(ft, i);
  AQFS_Table_AddRuntimeFlags(ft, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);

  return index;
}



int64_t AQFS_File_GetTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t index)
{
  AQFS_TABLE *ft;
  uint64_t origIndex;

  origIndex=index;
  ft=AQFS_Table_List_First(ftl);
  while (ft) {
    if (index<AQFS_Table_GetMaxEntries(ft))
      break;
    index-=AQFS_Table_GetMaxEntries(ft);
    ft=AQFS_Table_List_Next(ft);
  }

  if (ft) {
    uint64_t *ptr;

    ptr=AQFS_Table_GetPtrEntries(ft);
    return ptr[index];
  }

  DBG_ERROR(AQFS_LOGDOMAIN, "Index %ld not found", (long int) origIndex);
  return GWEN_ERROR_NOT_FOUND;
}



int64_t AQFS_File_SetTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t index, uint64_t entry)
{
  AQFS_TABLE *ft;
  uint64_t origIndex;

  origIndex=index;
  ft=AQFS_Table_List_First(ftl);
  while (ft) {
    if (index<AQFS_Table_GetMaxEntries(ft))
      break;
    index-=AQFS_Table_GetMaxEntries(ft);
    ft=AQFS_Table_List_Next(ft);
  }

  if (ft) {
    uint64_t *ptr;
    uint64_t oldVal;

    ptr=AQFS_Table_GetPtrEntries(ft);
    oldVal=ptr[index];
    ptr[index]=entry;
    AQFS_Table_AddRuntimeFlags(ft, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
    return oldVal;
  }

  DBG_ERROR(AQFS_LOGDOMAIN, "Index %ld not found", (long int) origIndex);
  return GWEN_ERROR_NOT_FOUND;
}



int AQFS_File_WriteDirtyTables(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl)
{
  int rv;
  AQFS_TABLE *ft;

  ft=AQFS_Table_List_First(ftl);
  while (ft) {
    if (AQFS_Table_GetRuntimeFlags(ft) & AQFS_TABLE_RUNTIME_FLAGS_DIRTY) {
      rv=AQFS_File_WriteTableAt(fsf, AQFS_Table_GetOffsetTableHeaderStart(ft), ft);
      if (rv<0) {
        DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
      AQFS_Table_SubRuntimeFlags(ft, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
    }
    ft=AQFS_Table_List_Next(ft);
  }

  return 0;
}



int AQFS_File_WritDirtyTablesFromPtrList(AQFS_FILE *fsf, AQFS_TABLE_PTRLIST *ftl)
{
  int64_t idxTableLast;

  idxTableLast=AQFS_Table_PtrList_GetLastTablePos(ftl);
  if (idxTableLast>=0) {
    uint64_t idxTableCurrent;

    for (idxTableCurrent=0; idxTableCurrent<=idxTableLast; idxTableCurrent++) {
      AQFS_TABLE *pTableCurrent;

      pTableCurrent=AQFS_Table_PtrList_GetTableAt(ftl, idxTableCurrent);
      if (pTableCurrent) {

        if (AQFS_Table_GetRuntimeFlags(pTableCurrent) & AQFS_TABLE_RUNTIME_FLAGS_DIRTY) {
          uint64_t tableOffset;
          int rv;

          DBG_DEBUG(AQDB_LOGDOMAIN, "Table is dirty, writing to file.");
          tableOffset=AQFS_Table_GetOffsetTableHeaderStart(pTableCurrent);
          if (tableOffset==0) {
            DBG_ERROR(AQDB_LOGDOMAIN, "No offset in table, SNH!");
            return GWEN_ERROR_INTERNAL;
          }

          rv=AQFS_File_WriteTableAt(fsf, tableOffset, pTableCurrent);
          if (rv<0) {
            DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
            return rv;
          }

          AQFS_Table_SubRuntimeFlags(pTableCurrent, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
        } /* if dirty */
      } /* if (pTableCurrent) */
    } /* for idxTableCurrent */
  } /* if (idxTableLast>=0) */

  return 0;
}



int AQFS_File_AppendNewTablesFromPtrListToFile(AQFS_FILE *fsf, AQFS_TABLE_PTRLIST *ftl)
{
  int64_t idxTableLast;

  idxTableLast=AQFS_Table_PtrList_GetLastTablePos(ftl);
  if (idxTableLast>=0) {
    uint64_t idxTableCurrent;

    for (idxTableCurrent=0; idxTableCurrent<=idxTableLast; idxTableCurrent++) {
      AQFS_TABLE *pTableCurrent;

      pTableCurrent=AQFS_Table_PtrList_GetTableAt(ftl, idxTableCurrent);
      if (pTableCurrent) {
        uint64_t tableOffset;

        tableOffset=AQFS_Table_GetOffsetTableHeaderStart(pTableCurrent);
        if (tableOffset==0) {
          int64_t offset;

          DBG_DEBUG(AQDB_LOGDOMAIN, "Table is new, appending.");
          offset=AQFS_File_AppendTable(fsf, pTableCurrent);
          if (offset<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) offset);
            AQFS_Table_free(pTableCurrent);
            return offset;
          }

          if (idxTableCurrent>0) { /* set nextPointer in previous table */
            AQFS_TABLE *pTablePrevious;

            pTablePrevious=AQFS_Table_PtrList_GetTableAt(ftl, idxTableCurrent-1);
            if (pTablePrevious) {
              DBG_DEBUG(AQDB_LOGDOMAIN, "Setting pointer in preceeding table, marking that as dirty.");
              AQFS_Table_SetOffsetNextTable(pTablePrevious, offset);
              AQFS_Table_AddRuntimeFlags(pTablePrevious, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
            } /* if (pTablePrevious) */
          } /* if (idxTableCurrent>0) */
        } /* if (tableOffset==0) */
      } /* if (pTableCurrent) */
    } /* for idxTableCurrent */

  } /* if (idxTableLast>=0) */

  return 0;
}



uint64_t _getRealFileSize(const char *fileName)
{
  struct stat sb;

  if (lstat(fileName, &sb)==-1) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error on lstat: %d", errno);
    return 0;
  }

  return (uint64_t) sb.st_size;
}



int _setFileSize(AQFS_FILE *fsf, uint64_t newFileSize)
{
  int rv;

  assert(fsf);
  assert(fsf->refCount);

  if (!(fsf->isOpen)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "File not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  if (!(fsf->rw)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "File not open for writing");
    return GWEN_ERROR_NOT_OPEN;
  }

  assert(fsf->fileHeader);
  AQFS_FileHeader_SetFileSize(fsf->fileHeader, newFileSize);

  /* write updated file header */
  rv=_writeFileHeader(fsf, fsf->fileHeader);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}




