/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQFS_FILE_H
#define AQFS_FILE_H


/* the only place where the AQDB headers are used */
#include <aqdatabase/aqdb.h>


#define AQFS_LOGDOMAIN AQDB_LOGDOMAIN /* re-use the AQDB logdomain */


#define AQFS_FILE_VERSION 0x01010100

typedef struct AQFS_FILE AQFS_FILE;


#include <aqfilestorage/aqfs_item_header.h>
#include <aqfilestorage/aqfs_table.h>
#include <aqfilestorage/aqfs_table_ptrlist.h>


/** @page AQFS_File AQFS_File Documentation Page
 *
 * AqFileStore is a rather simple library. It just stores a @ref AQFS_FILE_HEADER followed by a list of data objects.
 *
 * Reading and writing of objects is endian-aware.
 * This library is more or less platform independent, it is known to work on unix-like systems (especially Linux) and Windows.
 *
 * AqFS_FILE only knows two types: Items and tables.
 *
 *
 * Items
 *
 * Items are the basic storage units within a AQFS_FILE. They consist of a @ref AQFS_ITEM_HEADER object followed by some
 * arbitrary data the user wants to store.
 *
 *
 * Tables
 *
 * Tables (type AQFS_TABLE) are basically just lists of offsets, used to store the positions of other objects or tables in a AQFS file.
 *
 * A table is stored within the AQFS file as an item (see above).
 * So within a AQFS_FILE there are only the file header and items, AQFS_FILE doesn't know about any other type of data within a
 * file.
 *
 * Tables can be used to organize objects within a AQFS_FILE. For that purpose they contain a number of "entries" (i.e. just some
 * numbers, which represent offsets within a AQFS file). A table also knows about the number of entries it can contain at most, the
 * number of used entries within a table and a pointer to the next table.
 *
 * The API doesn't let you add entries (offsets) to a single AQFS_TABLE but rather to a AQFS_TABLE_LIST. If a table is full
 * when trying to add an entry, the system just creates another table, adds it to the list and notes it in the "next" pointer
 * inside the AQFS_TABLE. The entry is then added to the newly created table (see @ref AQFS_File_AddTableEntry).
 *
 * Tables are not automatically written to the file. Whenever a table is changed (e.g. when adding a newly created table to
 * a AQFS_TABLE_LIST while calling @ref AQFS_File_AddTableEntry or when modifying an entry of a table) a dirty flag
 * is set (@ref AQFS_TABLE_RUNTIME_FLAGS_DIRTY). Projects using AQFS_FILE need to call @ref AQFS_File_WriteDirtyTables
 * to make AQFS_FILE actually commit the table data to disk.
 *
 */


/** @name Constructor, Destructor, Open, Close and Create
 *
 */
/*@{*/

/**
 * Constructor.
 */
AQFS_FILE *AQFS_File_new(const char *fname);


/**
 * Destructor.
 */
void AQFS_File_free(AQFS_FILE *fsf);


/**
 * Opens an existing file for reading or for reading and writing.
 * This function reads the file header.
 *
 * @return 0 if ok, error code otherwise
 * @param fsf file
 * @param rw if 0, then the file will be opened read-only, otherwise read-write
 */
int AQFS_File_Open(AQFS_FILE *fsf, int rw);


/**
 * Creates a new data file. If a file of the same name already exists, an error will be returned.
 * On successfull return the file will be open in read-write mode.
 * This function also creates and writes the file header.
 *
 * @return 0 if ok, error code otherwise
 * @param fsf file
 */
int AQFS_File_Create(AQFS_FILE *fsf);


/**
 * Closes the data file.
 * If in read-write mode the file header is updated and written.
 *
 * @return 0 if ok, error code otherwise
 * @param fsf file
 */
int AQFS_File_Close(AQFS_FILE *fsf);

/*@}*/





/** @name Low-Level functions
 *
 */
/*@{*/

/**
 * Directly read from the data file at the given position.
 *
 * @return number of bytes read (or error code). Cave: Might be less than requested!
 * @param fsf file
 * @param offset position within the data file
 * @param ptr pointer to a buffer receiving the data read
 * @param len number of bytes to read (maximum)
 */
int64_t AQFS_File_ReadAt(AQFS_FILE *fsf, uint64_t offset, uint8_t *ptr, uint64_t len);


/**
 * Directly read exactly the given number of bytes from the data file at the given position.
 *
 * @return 0 if the number of bytes have been read, error code otherwise
 * @param fsf file
 * @param offset position within the data file
 * @param ptr pointer to a buffer receiving the data read
 * @param len exact number of bytes to read
 */
int AQFS_File_ReadForcedAt(AQFS_FILE *fsf, uint64_t offset, uint8_t *ptr, uint64_t len);


/**
 * Directly write exactly the given number of bytes to the data file at the given position.
 *
 * This function can not write into the file header (currently the first 128 bytes of the db file).
 *
 * @return 0 if the number of bytes have been written, error code on error
 * @param fsf file
 * @param offset position within the data file
 * @param ptr pointer to a buffer holding the data to write
 * @param len number of bytes to write (maximum)
 */
int AQFS_File_WriteForcedAt(AQFS_FILE *fsf, uint64_t offset, const uint8_t *ptr, uint64_t len);



/**
 * Flush all pending data for the given file.
 *
 * @return 0 if okay, !=0 otherwise
 * @param fsf pointer to AQFS_FILE object
 */
int AQFS_File_Sync(AQFS_FILE *fsf);

/*@}*/





/** @name Data Item Reading, Writing and Appending
 *
 */
/*@{*/

/**
 * Returns the offset of the first item. This is the first item after the file header.
 *
 * @param fsf file
 * @return offset to the first data item.
 */
uint64_t AQFS_File_GetFirstItemOffset(AQFS_FILE *fsf);


/**
 * Return the number of bytes currently in use by the database.
 * In most cases this equals the real file size.
 * @param fsf file
 * @return number of bytes in use (i.e. file size).
 */
uint64_t AQFS_File_GetFileSize(const AQFS_FILE *fsf);


/**
 * Reads an item header from the given position and parses it into the given AQFS_ITEM_HEADER object.
 *
 * @return 0 if ok, error code otherwise
 * @param fsf file
 * @param offset position within the data file
 * @param ih item header object to parse the data from file into
 */
int AQFS_File_ReadItemHeaderAt(AQFS_FILE *fsf, uint64_t offset, AQFS_ITEM_HEADER *ih);

/**
 * Writes an item header to the given position using data from the given AQFS_ITEM_HEADER object.
 *
 * @return 0 if ok, error code otherwise
 * @param fsf file
 * @param offset position within the data file
 * @param ih item header object to take data from
 */
int AQFS_File_WriteItemHeaderAt(AQFS_FILE *fsf, uint64_t offset, const AQFS_ITEM_HEADER *ih);


/**
 * Create a data item and add it to the file.
 * Space for the item header and the data will be allocated at the end of the file. An item header
 * will be created and written before writing the data. The file header will be updated to reflect the
 * new file size.
 *
 * @return offset of the newly written item within the file, error code on error
 * @param fsf file
 * @param t type of the item to write (0 means deleted item, all other values free to use)
 * @param ptr pointer to the data of the item to write
 * @param len size of the data to write
 */
int64_t AQFS_File_AppendData(AQFS_FILE *fsf, uint32_t t, const uint8_t *ptr, uint64_t len);

/**
 * Reads an item of maximum maxlen bytes in size and creates a buffer for that item. The pointer
 * to the buffer is returned in pptr. The caller is responsible for freeing the data returned, if any.
 *
 * This function reads the item header from the given position and creates a buffer according to the info
 * from the header. If the data size does not exceed maxLen bytes a buffer will be allocated and the
 * data read into it.
 *
 * @return size of the item data returned (if more than 0 bytes then pptr will point to that data upon return)
 * @param fsf file
 * @param offset position of the item header in the file
 * @param pptr pointer to a pointer receiving the newly allocated buffer upon return
 * @param maxLen if not zero it contains the maximum size of the item to be read (i.e. items which are bigger
 *               will not be read)
 */
int64_t AQFS_File_ReadDataAt(AQFS_FILE *fsf, uint64_t offset, uint8_t **pptr, uint64_t maxLen);


/**
 * Write data to the given position in the file.
 * This function will read the current item header, check for the length of allocated data, update
 * the used size field in the item header and writes the new header back after writing the data to the file.
 *
 * @return 0 if ok, error code on error
 * @param fsf file
 * @param offset position of the item inside the data file
 * @param t type of the item to write (0 means deleted item, all other values free to use)
 * @param ptr pointer to the data of the item to write
 * @param len size of the data to write
 */
int AQFS_File_WriteDataAt(AQFS_FILE *fsf, uint64_t offset, uint32_t t, const uint8_t *ptr, uint64_t len);

/*@}*/





/** @name Table Handling
 *
 */
/*@{*/

/**
 * Directly reads a single table from the given position (including item header).
 *
 * @return 0 if ok, error code on error
 * @param fsf file
 * @param offset position of the item inside the data file
 * @param ft table to read data into
 */
int AQFS_File_ReadTableAt(AQFS_FILE *fsf, uint64_t offset, AQFS_TABLE *ft);


/**
 * Directly writes a single table to the given position (including item header).
 *
 * This function reads the AQFS_ITEM_HEADER at the given offset and checks whether the new data
 * fits into the allocated data portion of the item. If it does the new data is written and the item header
 * updated on disk.
 * @return 0 if ok, error code on error
 * @param fsf file
 * @param offset position of the item inside the data file
 * @param ft table to write data from
 */
int AQFS_File_WriteTableAt(AQFS_FILE *fsf, uint64_t offset, const AQFS_TABLE *ft);


/**
 * Create a data item from the given table and add it to the file.
 * Space for the item header and the data will be allocated at the end of the file. An item header
 * will be created and written before writing the data. The file header will be updated to reflect the
 * new file size.
 *
 * @return offset of the newly written item within the file, error code on error
 * @param fsf file
 * @param ft pointer to the table to add to the file. Upon return the table's member "OffsetTableHeaderStart"
 *           will be set to the location of the appended table
 */

int64_t AQFS_File_AppendTable(AQFS_FILE *fsf, AQFS_TABLE *ft);


/**
 * Read the chained list of tables starting at "firstOffset" and store the tables read in the
 * given list.
 *
 * @return 0 if ok, error code on error
 * @param fsf file
 * @param firstOffset offset of the first table inside the file
 * @param ftl table list to read tables to
 */
int AQFS_File_ReadTableList(AQFS_FILE *fsf, uint64_t firstOffset, AQFS_TABLE_LIST *ftl);


/**
 * Read the chained list of tables starting at "firstOffset" and store the tables read in the
 * given pointer list.
 *
 * @return 0 if ok, error code on error
 * @param fsf file
 * @param firstOffset offset of the first table inside the file
 * @param ftl pointer list to add tables to
 */
int AQFS_File_ReadTablePtrList(AQFS_FILE *fsf, uint64_t firstOffset, AQFS_TABLE_PTRLIST *ftl);


/**
 * Add an entry to the given list of tables.
 * If needed a new table is created and appended to the given table list.
 * All affected tables get the "DIRTY" flag set.
 *
 * @return index at which the new entry has been assigned starting with 0, error code on error
 * @param fsf file
 * @param ftl table list to store entry to
 * @param entry data to store in the created entry
 */
int64_t AQFS_File_AddTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t entry);


/**
 * Get the stored entry from the given index.
 * The entry is looked up by traversing the given table list until the entry is reached. The content is then
 * returned.
 *
 * @return index at which the entry is to be read starting with 0, error code on error
 * @param fsf file
 * @param ftl table list to read entry from
 * @param index index of the entry to read (starting with 0)
 */
int64_t AQFS_File_GetTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t index);


/**
 * Change the stored entry from the given index.
 * The entry is looked up by traversing the given table list until the entry is reached. The content is then
 * returned.
 * All affected tables get the "DIRTY" flag set.
 *
 * @return index at which the entry is to be changed starting with 0, error code on error
 * @param fsf file
 * @param ftl table list to change entry in
 * @param index index of the entry to modify (starting with 0)
 * @param entry new entry (starting with 0)
 */
int64_t AQFS_File_SetTableEntry(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl, uint64_t index, uint64_t entry);


/**
 * Writes all tables from the given list which have the runtime flag "dirty" set. This flag is then reset
 * on those tables.
 *
 * @param fsf file
 * @param ftl list of tables to write
 */
int AQFS_File_WriteDirtyTables(AQFS_FILE *fsf, AQFS_TABLE_LIST *ftl);


/**
 * Writes all tables from the given pointer list which have the runtime flag "dirty" set. This flag is then reset
 * on those tables.
 *
 * @param fsf file
 * @param ftl pointer list of tables to write
 */
int AQFS_File_WritDirtyTablesFromPtrList(AQFS_FILE *fsf, AQFS_TABLE_PTRLIST *ftl);


/**
 * Append every table in the ptr list which has no offset (i.e. is new and not yet in the file) to the file.
 *
 * Set the "NEXT" pointer in predeceeding tables and set their DIRTY flag.
 * You should call this function before @ref AQFS_File_WritDirtyTablesFromPtrList().
 *
 * @param fsf file
 * @param ftl pointer list of tables to write
 */
int AQFS_File_AppendNewTablesFromPtrListToFile(AQFS_FILE *fsf, AQFS_TABLE_PTRLIST *ftl);



/*@}*/






#endif
