/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqfs_file_header_p.h"
#include "aqfs_file.h"

#include <gwenhywfar/memory.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>





AQFS_FILE_HEADER *AQFS_FileHeader_new()
{
  AQFS_FILE_HEADER *fh;

  GWEN_NEW_OBJECT(AQFS_FILE_HEADER, fh);
  fh->refCount=1;
  return fh;
}



void AQFS_FileHeader_free(AQFS_FILE_HEADER *fh)
{
  if (fh) {
    assert(fh->refCount);
    if (fh->refCount==1) {
      fh->refCount=0;
      GWEN_FREE_OBJECT(fh);
    }
    else
      fh->refCount--;
  }
}



int AQFS_FileHeader_ReadFromBuffer(AQFS_FILE_HEADER *fh, const uint8_t *buffer, uint64_t len)
{
  if (len<AQFS_FILE_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for valid header data (%d < %d)",
              (int) len, (int) AQFS_FILE_HEADER_SIZE);
    return GWEN_ERROR_BAD_DATA;
  }

  fh->version=GWEN_ENDIAN_LE32TOH(*((uint32_t *)(buffer+AQFS_FILE_HEADER_OFFS_VERSION)));
  fh->fileSize=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_FILE_HEADER_OFFS_FILE_SIZE)));

  return 0;
}



AQFS_FILE_HEADER *AQFS_FileHeader_fromBuffer(const uint8_t *buffer, uint64_t len)
{
  AQFS_FILE_HEADER *fh;
  int rv;

  fh=AQFS_FileHeader_new();
  rv=AQFS_FileHeader_ReadFromBuffer(fh, buffer, len);
  if (rv<0) {
    AQFS_FileHeader_free(fh);
    return NULL;
  }

  return fh;
}



uint64_t AQFS_FileHeader_toBuffer(const AQFS_FILE_HEADER *fh, uint8_t *buffer, uint64_t len)
{
  if (len<AQFS_FILE_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small to store header data (%d < %d)",
              (int) len, (int) AQFS_FILE_HEADER_SIZE);
    return 0;
  }

  memset(buffer, 0, AQFS_FILE_HEADER_SIZE);

  buffer[AQFS_FILE_HEADER_OFFS_MAGIC+0]='A';
  buffer[AQFS_FILE_HEADER_OFFS_MAGIC+1]='Q';
  buffer[AQFS_FILE_HEADER_OFFS_MAGIC+2]='F';
  buffer[AQFS_FILE_HEADER_OFFS_MAGIC+3]='S';

  *((uint32_t *)(buffer+AQFS_FILE_HEADER_OFFS_VERSION))=GWEN_ENDIAN_HTOLE32(fh->version);
  *((uint64_t *)(buffer+AQFS_FILE_HEADER_OFFS_FILE_SIZE))=GWEN_ENDIAN_HTOLE64(fh->fileSize);

  return AQFS_FILE_HEADER_SIZE;
}



uint32_t AQFS_FileHeader_GetVersion(const AQFS_FILE_HEADER *fh)
{
  assert(fh);
  assert(fh->refCount);
  return fh->version;
}



void AQFS_FileHeader_SetVersion(AQFS_FILE_HEADER *fh, uint32_t i)
{
  assert(fh);
  assert(fh->refCount);
  fh->version=i;
}



uint64_t AQFS_FileHeader_GetFileSize(const AQFS_FILE_HEADER *fh)
{
  assert(fh);
  assert(fh->refCount);
  return fh->fileSize;
}



void AQFS_FileHeader_SetFileSize(AQFS_FILE_HEADER *fh, uint64_t i)
{
  assert(fh);
  assert(fh->refCount);
  if (i<fh->fileSize) {
    DBG_ERROR(AQFS_LOGDOMAIN, "New file size is smaller than the current, refusing to set newvalue (%llu < %llu)",
              (unsigned long long int) i,
              (unsigned long long int) fh->fileSize);
  }
  else
    fh->fileSize=i;
}



