/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQFS_FILE_HEADER_H
#define AQFS_FILE_HEADER_H


#include <stdint.h>

/**
 * This structure is the header of every AQFS file. It contains information about the
 * library version and the actually used size of the file.
 * Its size is currently 128 bytes with enough space to add extra data fields later if
 * necessary. All unused bytes are set to "0".
 */
typedef struct AQFS_FILE_HEADER AQFS_FILE_HEADER;


#define AQFS_FILE_HEADER_SIZE                128


/** @name Construction, destruction, reading, writing.
 */
/*@{*/

/**
 * Constructor.
 */
AQFS_FILE_HEADER *AQFS_FileHeader_new();

/**
 * Destructor.
 */
void AQFS_FileHeader_free(AQFS_FILE_HEADER *fh);


/**
 * Read a AQFS_FILE_HEADER object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param fh AQFS_FILE_HEADER object to read data into
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
int AQFS_FileHeader_ReadFromBuffer(AQFS_FILE_HEADER *fh, const uint8_t *buffer, uint64_t len);


/**
 * Create a AQFS_FILE_HEADER object from a buffer.
 * @return AQFS_FILE_HEADER object created, NULL on error
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
AQFS_FILE_HEADER *AQFS_FileHeader_fromBuffer(const uint8_t *buffer, uint64_t len);

/**
 * Write content of a AQFS_FILE_HEADER object to a buffer.
 * @return number of bytes written, <1 on error
 * @param fh pointer to AQFS_FILE_HEADER object
 * @param buffer pointer to a buffer to write to
 * @param len size of the buffer pointed to by buffer
 */
uint64_t AQFS_FileHeader_toBuffer(const AQFS_FILE_HEADER *fh, uint8_t *buffer, uint64_t len);

/*@}*/



/** @name Information stored in the file.
 */
/*@{*/

/**
 * Return version number store in the AQFS_FILE_HEADER object.
 * This is an uint32_t element, each byte contains a part of the version number:
 * Most significant byte: major version
 * Next lower byte: minor version
 * Next lower byte: patchlevel
 * Least significant byte: currently not used (maybe later for GIT version etc). Always 0.
 * Example: 0x01010100 means "1.1.1.0".
 * @param fh pointer to AQFS_FILE_HEADER object
  */
uint32_t AQFS_FileHeader_GetVersion(const AQFS_FILE_HEADER *fh);

/**
 * Set version field in the header.
 * @param fh pointer to AQFS_FILE_HEADER object
 * @param i new value
 */
void AQFS_FileHeader_SetVersion(AQFS_FILE_HEADER *fh, uint32_t i);

/**
 * Return the number of bytes currently in use by the database.
 * In most cases this equals the real file size.
 * @param fh pointer to AQFS_FILE_HEADER object
 */
uint64_t AQFS_FileHeader_GetFileSize(const AQFS_FILE_HEADER *fh);

/**
 * Set the file size field in the header.
 * @param fh pointer to AQFS_FILE_HEADER object
 * @param i new value
 */
void AQFS_FileHeader_SetFileSize(AQFS_FILE_HEADER *fh, uint64_t i);

/*@}*/





#endif
