/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqfs_file_p.h"


#include <gwenhywfar/error.h>
#include <gwenhywfar/debug.h>



#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <assert.h>




#ifdef OS_WIN32


/* ------------------------------------- Windows ------------------------------------ */

int AQFS_File_open(AQFS_FILE *fsf, int rw)
{
  DWORD flags=0;
  HANDLE fh;

  assert(fsf);

  if (rw)
    flags=GENERIC_READ | GENERIC_WRITE;
  else
    flags=GENERIC_READ;

  fh=CreateFile(fsf->dataFileName,
                flags,                              /* dwDesiredAccess */
                FILE_SHARE_READ | FILE_SHARE_WRITE, /* dwShareMode */
                NULL,                               /* pSecurityAttributes */
                OPEN_EXISTING,                      /* dwCreationDistribution */
                FILE_FLAG_RANDOM_ACCESS,            /* dwFlagsAndAttrs */
                NULL);                              /* hTemplateFile */
  if (fh==INVALID_HANDLE_VALUE) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Error opening file[%s]: %d",
              fsf->dataFileName,
              (int)GetLastError());
    return GWEN_ERROR_IO;
  }

  fsf->fh=fh;
  fsf->isOpen=1;
  return 0;
}



int AQFS_File_creat(AQFS_FILE *fsf)
{
  HANDLE fh;

  assert(fsf);

  fh=CreateFile(fsf->dataFileName,
                GENERIC_READ | GENERIC_WRITE,       /* dwDesiredAccess */
                FILE_SHARE_READ | FILE_SHARE_WRITE, /* dwShareMode */
                NULL,                               /* pSecurityAttributes */
                CREATE_NEW,                         /* dwCreationDistribution */
                FILE_FLAG_RANDOM_ACCESS |           /* dwFlagsAndAttrs */
                FILE_ATTRIBUTE_NORMAL,
                NULL);                              /* hTemplateFile */
  if (fh==INVALID_HANDLE_VALUE) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Error creating file[%s]: %d",
              fsf->dataFileName,
              (int)GetLastError());
    return GWEN_ERROR_IO;
  }

  fsf->fh=fh;
  fsf->isOpen=1;
  return 0;
}



int AQFS_File_close(AQFS_FILE *fsf)
{
  assert(fsf);

  if (!CloseHandle(fsf->fh)) {
    fsf->fh=NULL;
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Error closing file[%s]: %d",
              fsf->dataFileName,
              (int)GetLastError());
    return GWEN_ERROR_IO;
  }
  fsf->fh=NULL;
  fsf->isOpen=0;
  return 0;
}



int AQFS_File_seek(AQFS_FILE *fsf, uint64_t offs)
{
  LARGE_INTEGER li;

  assert(fsf);

  li.QuadPart = offs;
  li.LowPart=SetFilePointer(fsf->fh,
                            li.LowPart,
                            &li.HighPart,
                            FILE_BEGIN);
  if (li.LowPart==INVALID_SET_FILE_POINTER) {
    DWORD rv;

    rv=GetLastError();
    if (rv!=NO_ERROR) {
      DBG_ERROR(AQFS_LOGDOMAIN,
                "Error seeking in file[%s] at %llu: %d",
                fsf->dataFileName,
                (unsigned long long) offs,
                (int)rv);
      return GWEN_ERROR_IO;
    }
  }
  return 0;
}



size_t AQFS_File_read(AQFS_FILE *fsf, uint8_t *ptr, size_t size, size_t nmember)
{
  size_t r=size*nmember;

  assert(fsf);

  while (r) {
    DWORD bytesRead=0;

    if (!ReadFile(fsf->fh, ptr, r, &bytesRead, NULL)) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Error on ReadFile(%s): %d",
                fsf->dataFileName,
                (int)GetLastError());
      return -1;
    }

    if (bytesRead==0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "EOF met on read(%s)",
                fsf->dataFileName);
      return 0;
    }
    else {
      ptr+=bytesRead;
      r-=bytesRead;
    }
  }

  return nmember;
}



size_t AQFS_File_write(AQFS_FILE *fsf, const uint8_t *ptr, size_t size, size_t nmember)
{
  size_t r=size*nmember;

  assert(fsf);

  while (r) {
    DWORD bytesWritten=0;

    if (!WriteFile(fsf->fh, ptr, r, &bytesWritten, NULL)) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Error on WriteFile(%s): %d",
                fsf->dataFileName,
                (int)GetLastError());
      return -1;
    }

    if (bytesWritten==0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Broken pipe on WriteFile(%s)",
                fsf->dataFileName);
      return 0;
    }
    else {
      ptr+=bytesWritten;
      r-=bytesWritten;
    }
  }

  return nmember;
}


int AQFS_File_sync(AQFS_FILE *fsf)
{
  BOOL rv;

  rv=FlushFileBuffers(fsf->fh);
  if (!rv) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Error on FlushFileBuffers(%s): %d",
              fsf->dataFileName,
              (int)GetLastError());
    return -1;
  }
  return 0;
}


/* ------------------------------------- Windows ------------------------------------ */

#else

/* -------------------------------------- Linux ------------------------------------- */


int AQFS_File_seek(AQFS_FILE *fsf, uint64_t offs)
{
  assert(fsf);

  if (lseek(fsf->fd, offs, SEEK_SET)==-1)
    return GWEN_ERROR_IO;
  return 0;
}



int AQFS_File_open(AQFS_FILE *fsf, int rw)
{
  int flags=0;
  int fd;

  assert(fsf);

  if (rw)
    flags=O_RDWR;
  else
    flags=O_RDONLY;
#ifdef O_BINARY
  flags|=O_BINARY;
#endif

  fd=open(fsf->dataFileName, flags);
  if (fd==-1) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Error opening file[%s]: %s",
              fsf->dataFileName,
              strerror(errno));
    return GWEN_ERROR_IO;
  }

  fsf->fd=fd;
  fsf->isOpen=1;
  return 0;
}



int AQFS_File_creat(AQFS_FILE *fsf)
{
  int fd;

  assert(fsf);

  fd=open(fsf->dataFileName,
#ifdef O_BINARY
          O_BINARY |
#endif
          O_RDWR | O_CREAT | O_EXCL,
          S_IRUSR | S_IWUSR
#ifdef S_IRGRP
          | S_IRGRP
#endif
#ifdef S_IWGRP
          | S_IWGRP
#endif
         );
  if (fd==-1) {
    DBG_ERROR(AQFS_LOGDOMAIN,
              "Error creating file[%s]: %s",
              fsf->dataFileName,
              strerror(errno));
    return GWEN_ERROR_IO;
  }

  fsf->fd=fd;
  fsf->isOpen=1;
  return 0;
}



int AQFS_File_close(AQFS_FILE *fsf)
{
  int rv;

  assert(fsf);

  rv=close(fsf->fd);
  fsf->fd=-1;
  fsf->isOpen=0;
  return rv;
}



size_t AQFS_File_read(AQFS_FILE *fsf, uint8_t *ptr, size_t size, size_t nmember)
{
  size_t total=size*nmember;
  size_t r=total;

  assert(fsf);

  while (r) {
    ssize_t t;

    t=read(fsf->fd, ptr, r);
    if (t<0) {
      if (errno!=EINTR) {
        DBG_ERROR(AQFS_LOGDOMAIN, "Error on read(%s): %s",
                  fsf->dataFileName,
                  strerror(errno));
        return GWEN_ERROR_IO;
      }
    }
    else if (t==0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "EOF met on read(%s)",
                fsf->dataFileName);
      return 0;
    }
    else {
      ptr+=t;
      r-=t;
    }
  }

  return nmember;
}



size_t AQFS_File_write(AQFS_FILE *fsf, const uint8_t *ptr, size_t size, size_t nmember)
{
  size_t total=size*nmember;
  size_t r=total;

  assert(fsf);

  while (r) {
    ssize_t t;

    t=write(fsf->fd, ptr, r);
    if (t<0) {
      if (errno!=EINTR) {
        DBG_ERROR(AQFS_LOGDOMAIN, "Error on write(%s): %s",
                  fsf->dataFileName,
                  strerror(errno));
        return -1;
      }
    }
    else if (t==0) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Pipe broken on write(%s)",
                fsf->dataFileName);
      return 0;
    }
    else {
      ptr+=t;
      r-=t;
    }
  }

  return nmember;
}



int AQFS_File_sync(AQFS_FILE *fsf)
{
  int rv;

  assert(fsf);
  rv=fsync(fsf->fd);
  if (rv) {
    if (errno!=EINTR) {
      DBG_ERROR(AQFS_LOGDOMAIN, "Error on fsync(%s): %s",
                fsf->dataFileName,
                strerror(errno));
      return -1;
    }
  }

  return 0;
}

#endif
