/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQFS_FILE_P_H
#define AQFS_FILE_P_H


#include "aqfs_file.h"

#include <aqfilestorage/aqfs_file_header.h>


#include <stdio.h>
#include <stdint.h>


#ifdef OS_WIN32
# include <windows.h>
#endif


/** @file aqfs_file_p.h
 *
 * Private functions and definitons for AQFS_FILE.
 */


/**
 * Definitions of a AQFS_FILE object.
 */
struct AQFS_FILE {
  char *dataFileName;
  int isOpen;
  int rw;

  /* file handle */
#ifdef OS_WIN32
  HANDLE fh;
#else
  int fd;
#endif

  AQFS_FILE_HEADER *fileHeader;

  uint32_t refCount;
};



/** @name Operating System Abstraction
 *
 * AQFS contains implementations of these functions for unix-like systems and for Windows in aqfs_file_io.c.
 */
/*@{*/

/**
 * Physically open the AQFS file.
 * @return 0 on success, error code otherwise
 * @param fsf pointer to AQFS_FILE object
 */
static int AQFS_File_open(AQFS_FILE *fsf, int rw);

/**
 * Physically create AQFS file on disk.
 * @return 0 on success, error code otherwise
 * @param fsf pointer to AQFS_FILE object
 */
static int AQFS_File_creat(AQFS_FILE *fsf);

/**
 * Physically close AQFS file.
 * @return 0 on success, error code otherwise
 * @param fsf pointer to AQFS_FILE object
 */
static int AQFS_File_close(AQFS_FILE *fsf);

/**
 * Seek a position inside a AQFS file for further reading and writing.
 * @return 0 on success, error code otherwise
 * @param fsf pointer to AQFS_FILE object
 * @param offs position to seek
 */
static int AQFS_File_seek(AQFS_FILE *fsf, uint64_t offs);

/**
 * Read bytes from an open AQFS file.
 * Bytes a read in multiples of "members". A member is described by its "size".
 * So if you want to read up to 10 objects each with 100 bytes size you call this functions as
 *  AQFS_File_read(fsf, ptr, 100, 10);
 * Of course you can also read bytewise, in this case set the size of a member to 1.
 * @return number of members read
 * @param fsf pointer to AQFS_FILE object
 * @param ptr pointer to a buffer to receive data
 * @param size size of a member
 * @param number of members the buffer pointer to by ptr can receive
 */
static size_t AQFS_File_read(AQFS_FILE *fsf, uint8_t *ptr, size_t size, size_t nmember);

/**
 * Write bytes to an open AQFS file.
 * Bytes a written in multiples of "members". A member is described by its "size".
 * So if you want to write 10 objects each with 100 bytes size you call this functions as
 *  AQFS_File_write(fsf, ptr, 100, 10);
 * Of course you can also write bytewise, in this case set the size of a member to 1.
 * @return number of members read
 * @param fsf pointer to AQFS_FILE object
 * @param ptr pointer to a buffer to write
 * @param size size of a member
 * @param nmember of members to write
 */
static size_t AQFS_File_write(AQFS_FILE *fsf, const uint8_t *ptr, size_t size, size_t nmember);


/**
 * Flush all pending data for the given file (using fsync on POSIX and FlushFileBuffers on WIN32).
 * @return 0 if okay, !=0 otherwise
 * @param fsf pointer to AQFS_FILE object
 */
static int AQFS_File_sync(AQFS_FILE *fsf);

/*@}*/



#endif
