/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "aqfs_item_header_p.h"
#include "aqfs_file.h"

#include <gwenhywfar/memory.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>







AQFS_ITEM_HEADER *AQFS_ItemHeader_new()
{
  AQFS_ITEM_HEADER *ih;

  GWEN_NEW_OBJECT(AQFS_ITEM_HEADER, ih);
  ih->refCount=1;
  return ih;
}



void AQFS_ItemHeader_free(AQFS_ITEM_HEADER *ih)
{
  if (ih) {
    assert(ih->refCount);
    if (ih->refCount==1) {
      ih->refCount=0;
      GWEN_FREE_OBJECT(ih);
    }
    else
      ih->refCount--;
  }
}



int AQFS_ItemHeader_ReadFromBuffer(AQFS_ITEM_HEADER *ih, const uint8_t *buffer, uint64_t len)
{
  if (len<AQFS_ITEM_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for valid header data (%d < %d)",
              (int) len, (int) AQFS_ITEM_HEADER_SIZE);
    return GWEN_ERROR_BAD_DATA;
  }

  ih->sizeAllocated=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_ITEM_HEADER_OFFS_SIZE_ALLOCATED)));
  ih->sizeUsed=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_ITEM_HEADER_OFFS_SIZE_USED)));
  ih->dataType=GWEN_ENDIAN_LE32TOH(*((uint32_t *)(buffer+AQFS_ITEM_HEADER_OFFS_DATA_TYPE)));

  return 0;
}



AQFS_ITEM_HEADER *AQFS_ItemHeader_fromBuffer(const uint8_t *buffer, uint64_t len)
{
  AQFS_ITEM_HEADER *ih;
  int rv;

  ih=AQFS_ItemHeader_new();
  rv=AQFS_ItemHeader_ReadFromBuffer(ih, buffer, len);
  if (rv<0) {
    AQFS_ItemHeader_free(ih);
    return NULL;
  }

  return ih;
}



uint64_t AQFS_ItemHeader_toBuffer(const AQFS_ITEM_HEADER *ih, uint8_t *buffer, uint64_t len)
{
  if (len<AQFS_ITEM_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small to store header data (%d < %d)",
              (int) len, (int) AQFS_ITEM_HEADER_SIZE);
    return 0;
  }

  *((uint64_t *)(buffer+AQFS_ITEM_HEADER_OFFS_SIZE_ALLOCATED))=GWEN_ENDIAN_HTOLE64(ih->sizeAllocated);
  *((uint64_t *)(buffer+AQFS_ITEM_HEADER_OFFS_SIZE_USED))=GWEN_ENDIAN_HTOLE64(ih->sizeUsed);
  *((uint32_t *)(buffer+AQFS_ITEM_HEADER_OFFS_DATA_TYPE))=GWEN_ENDIAN_HTOLE32(ih->dataType);

  return AQFS_ITEM_HEADER_SIZE;
}



uint64_t AQFS_ItemHeader_GetSizeAllocated(const AQFS_ITEM_HEADER *ih)
{
  assert(ih);
  assert(ih->refCount);
  return ih->sizeAllocated;
}



void AQFS_ItemHeader_SetSizeAllocated(AQFS_ITEM_HEADER *ih, uint64_t i)
{
  assert(ih);
  assert(ih->refCount);
  ih->sizeAllocated=i;
}



uint64_t AQFS_ItemHeader_GetSizeUsed(const AQFS_ITEM_HEADER *ih)
{
  assert(ih);
  assert(ih->refCount);
  return ih->sizeUsed;
}



void AQFS_ItemHeader_SetSizeUsed(AQFS_ITEM_HEADER *ih, uint64_t i)
{
  assert(ih);
  assert(ih->refCount);
  ih->sizeUsed=i;
}



uint32_t AQFS_ItemHeader_GetDataType(const AQFS_ITEM_HEADER *ih)
{
  assert(ih);
  assert(ih->refCount);
  return ih->dataType;
}



void AQFS_ItemHeader_SetDataType(AQFS_ITEM_HEADER *ih, uint32_t i)
{
  assert(ih);
  assert(ih->refCount);
  ih->dataType=i;
}



uint64_t AQFS_ItemHeader_GetOffsetHeaderStart(const AQFS_ITEM_HEADER *ih)
{
  assert(ih);
  assert(ih->refCount);
  return ih->offsetHeaderStart;
}



void AQFS_ItemHeader_SetOffsetHeaderStart(AQFS_ITEM_HEADER *ih, uint64_t i)
{
  assert(ih);
  assert(ih->refCount);
  ih->offsetHeaderStart=i;
}



uint64_t AQFS_ItemHeader_GetOffsetDataStart(const AQFS_ITEM_HEADER *ih)
{
  assert(ih);
  assert(ih->refCount);
  return ih->offsetDataStart;
}



void AQFS_ItemHeader_SetOffsetDataStart(AQFS_ITEM_HEADER *ih, uint64_t i)
{
  assert(ih);
  assert(ih->refCount);
  ih->offsetDataStart=i;
}









