/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQFS_ITEM_HEADER_H
#define AQFS_ITEM_HEADER_H


#include <stdint.h>



/**
 * This structure is the header of every data item in a AQFS file. It contains information about the
 * data type, size allocated, size used and offset to the data part.
 */
typedef struct AQFS_ITEM_HEADER AQFS_ITEM_HEADER;


#define AQFS_ITEM_HEADER_SIZE                20


/**
 * Currently defined item types.
 */
enum {
  /** deleted item */
  AQFS_ITEMHEADER_TYPE_DELETED=0,
  /** generic data item */
  AQFS_ITEMHEADER_TYPE_GENERIC_DATA = (1 | ('D'<<8) | ('A'<<16) | ('T'<<24)),
  /** @ref AQFS_TABLE item */
  AQFS_ITEMHEADER_TYPE_GENERIC_TABLE= (2 | ('T'<<8) | ('A'<<16) | ('B'<<24))
};



/** @name Construction, destruction, reading, writing.
 */
/*@{*/

/**
 * Constructor.
 */
AQFS_ITEM_HEADER *AQFS_ItemHeader_new();

/**
 * Destructor.
 * @param ih pointer to AQFS_ITEM_HEADER object (may be NULL)
 */
void AQFS_ItemHeader_free(AQFS_ITEM_HEADER *ih);


/**
 * Read a AQFS_ITEM_HEADER object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param ih AQFS_ITEM_HEADER object to read data into
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
int AQFS_ItemHeader_ReadFromBuffer(AQFS_ITEM_HEADER *ih, const uint8_t *buffer, uint64_t len);

/**
 * Create a AQFS_ITEM_HEADER object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
AQFS_ITEM_HEADER *AQFS_ItemHeader_fromBuffer(const uint8_t *buffer, uint64_t len);

/**
 * Write content of a AQFS_ITEM_HEADER object to a buffer.
 * @return number of bytes written, <1 on error
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param buffer pointer to a buffer to write to
 * @param len size of the buffer pointed to by buffer
 */
uint64_t AQFS_ItemHeader_toBuffer(const AQFS_ITEM_HEADER *ih, uint8_t *buffer, uint64_t len);

/*@}*/



/** @name Information stored in the file.
 */
/*@{*/

/**
 * Return the number of bytes allocated for this item.
 * @param ih pointer to AQFS_ITEM_HEADER object
 */
uint64_t AQFS_ItemHeader_GetSizeAllocated(const AQFS_ITEM_HEADER *ih);

/**
 * Set the number of bytes allocated for this item.
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param i new value
 */
void AQFS_ItemHeader_SetSizeAllocated(AQFS_ITEM_HEADER *ih, uint64_t i);


/**
 * Return the number of bytes actually used for this item.
 * @param ih pointer to AQFS_ITEM_HEADER object
 */
uint64_t AQFS_ItemHeader_GetSizeUsed(const AQFS_ITEM_HEADER *ih);

/**
 * Set the number of bytes actually used for this item.
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param i new value
 */
void AQFS_ItemHeader_SetSizeUsed(AQFS_ITEM_HEADER *ih, uint64_t i);


/**
 * Return data type (see @ref AQFS_ITEMHEADER_TYPE_GENERIC_DATA et al).
 * @param ih pointer to AQFS_ITEM_HEADER object
 */
uint32_t AQFS_ItemHeader_GetDataType(const AQFS_ITEM_HEADER *ih);

/**
 * Set data type (see @ref AQFS_ITEMHEADER_TYPE_GENERIC_DATA et al).
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param i new value
 */
void AQFS_ItemHeader_SetDataType(AQFS_ITEM_HEADER *ih, uint32_t i);


/*@}*/



/** @name Runtime data not stored in the file.
 */
/*@{*/

/**
 * Return the offset of the start of this AQFS_ITEM_HEADER object in the file (field not stored on disk).
 * @param ih pointer to AQFS_ITEM_HEADER object
 */
uint64_t AQFS_ItemHeader_GetOffsetHeaderStart(const AQFS_ITEM_HEADER *ih);

/**
 * Set the offset of the start of this AQFS_ITEM_HEADER object in the file (field not stored on disk).
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param i new value
 */
void AQFS_ItemHeader_SetOffsetHeaderStart(AQFS_ITEM_HEADER *ih, uint64_t i);

/**
 * Return the offset of the start of the item data in the file (field not stored on disk).
 * @param ih pointer to AQFS_ITEM_HEADER object
 */
uint64_t AQFS_ItemHeader_GetOffsetDataStart(const AQFS_ITEM_HEADER *ih);

/**
 * Set the offset of the start of the item data in the file (field not stored on disk).
 * @param ih pointer to AQFS_ITEM_HEADER object
 * @param i new value
 */
void AQFS_ItemHeader_SetOffsetDataStart(AQFS_ITEM_HEADER *ih, uint64_t i);

/*@}*/





#endif
