/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "aqfs_table_p.h"
#include "aqfs_file.h"

#include <gwenhywfar/memory.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>

#include <string.h>


GWEN_LIST_FUNCTIONS(AQFS_TABLE, AQFS_Table)


AQFS_TABLE *AQFS_Table_new()
{
  AQFS_TABLE *ft;

  GWEN_NEW_OBJECT(AQFS_TABLE, ft);
  ft->refCount=1;
  GWEN_LIST_INIT(AQFS_TABLE, ft);

  return ft;
}



void AQFS_Table_Attach(AQFS_TABLE *ft)
{
  assert(ft && ft->refCount);
  if (ft && ft->refCount) {
    ft->refCount++;
  }
}



void AQFS_Table_free(AQFS_TABLE *ft)
{
  if (ft) {
    assert(ft->refCount);
    if (ft->refCount==1) {
      GWEN_LIST_FINI(AQFS_TABLE, ft);
      ft->refCount=0;
      free(ft->ptrEntries);
      GWEN_FREE_OBJECT(ft);
    }
    else {
      ft->refCount--;
    }
  }
}



int AQFS_Table_GetRefCounter(const AQFS_TABLE *ft)
{
  assert(ft);
  return ft->refCount;
}



AQFS_TABLE *AQFS_Table_dup(const AQFS_TABLE *ftOrig)
{
  AQFS_TABLE *ft;

  assert(ftOrig);
  assert(ftOrig->refCount);
  ft=AQFS_Table_new();
  ft->maxEntries=ftOrig->maxEntries;
  ft->freeEntries=ftOrig->freeEntries;
  ft->highestEntry=ftOrig->highestEntry;
  ft->offsetNextTable=ftOrig->offsetNextTable;
  ft->offsetTableHeaderStart=ftOrig->offsetTableHeaderStart;
  ft->runtimeFlags=ftOrig->runtimeFlags;

  /* copy offset entries */
  if (ftOrig->maxEntries && ftOrig->ptrEntries) {
    uint64_t offsetArraySize;

    offsetArraySize=ftOrig->maxEntries*sizeof(uint64_t);
    ft->ptrEntries=(uint64_t *) malloc(offsetArraySize);
    assert(ft->ptrEntries);
    memmove(ft->ptrEntries, ftOrig->ptrEntries, offsetArraySize);
  }

  return ft;
}



AQFS_TABLE *AQFS_Table_Create(uint64_t maxEntries)
{
  AQFS_TABLE *ft;
  uint64_t offsetArraySize;
  uint64_t *ptr;

  ft=AQFS_Table_new();
  ft->maxEntries=maxEntries;
  ft->freeEntries=maxEntries;

  offsetArraySize=ft->maxEntries*sizeof(uint64_t);

  ptr=(uint64_t *) malloc(offsetArraySize);
  assert(ptr);
  memset(ptr, 0, offsetArraySize);
  AQFS_Table_SetPtrEntries(ft, ptr);

  return ft;
}



uint64_t AQFS_Table_GetMaxEntries(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->maxEntries;
}



void AQFS_Table_SetMaxEntries(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  ft->maxEntries=i;
}



uint64_t AQFS_Table_GetFreeEntries(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->freeEntries;
}



void AQFS_Table_SetFreeEntries(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  ft->freeEntries=i;
}



void AQFS_Table_DecFreeEntries(AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  if (ft->freeEntries>0)
    ft->freeEntries--;
}



void AQFS_Table_IncFreeEntries(AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  if (ft->freeEntries<ft->maxEntries)
    ft->freeEntries++;
}



uint64_t AQFS_Table_GetHighestEntry(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->highestEntry;
}



void AQFS_Table_SetHighestEntry(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  ft->highestEntry=i;
}



void AQFS_Table_CheckAndSetHighestEntry(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  if (i>ft->highestEntry)
    ft->highestEntry=i;
}



uint64_t AQFS_Table_GetOffsetNextTable(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->offsetNextTable;
}



void AQFS_Table_SetOffsetNextTable(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  ft->offsetNextTable=i;
}



uint64_t *AQFS_Table_GetPtrEntries(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->ptrEntries;
}



void AQFS_Table_SetPtrEntries(AQFS_TABLE *ft, uint64_t *ptr)
{
  assert(ft);
  assert(ft->refCount);
  if (ft->ptrEntries && ft->ptrEntries!=ptr)
    free(ft->ptrEntries);
  ft->ptrEntries=ptr;
}



uint64_t AQFS_Table_GetOffsetTableHeaderStart(const AQFS_TABLE *ft)
{
  assert(ft);
  assert(ft->refCount);
  return ft->offsetTableHeaderStart;
}



void AQFS_Table_SetOffsetTableHeaderStart(AQFS_TABLE *ft, uint64_t i)
{
  assert(ft);
  assert(ft->refCount);
  ft->offsetTableHeaderStart=i;
}



int AQFS_Table_ReadFromBuffer(AQFS_TABLE *ft, const uint8_t *buffer, uint64_t len)
{
  uint64_t offsetArraySize;
  uint64_t *ptr;
  const uint64_t *sptr;
  uint64_t *dptr;
  uint64_t i;

  assert(ft);
  assert(ft->refCount);
  if (len<AQFS_TABLE_HEADER_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for valid table data (%d < %d)",
              (int) len, (int) AQFS_TABLE_HEADER_SIZE);
    return GWEN_ERROR_BAD_DATA;
  }

  /* read normal variables */
  ft->maxEntries=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_TABLE_OFFS_MAX_ENTRIES)));
  ft->freeEntries=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_TABLE_OFFS_FREE_ENTRIES)));
  ft->highestEntry=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_TABLE_OFFS_HIGHEST_ENTRY)));
  ft->offsetNextTable=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQFS_TABLE_OFFS_NEXT_TABLE)));

  /* calculate size used by offset array */
  offsetArraySize=ft->maxEntries*sizeof(uint64_t);
  if (len<(AQFS_TABLE_OFFS_ENTRIES+offsetArraySize)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for complete table data (%d < %d)",
              (int) len, (int)(AQFS_TABLE_OFFS_ENTRIES+offsetArraySize));
    return GWEN_ERROR_BAD_DATA;
  }

  /* allocate and copy offset array */
  ptr=(uint64_t *) malloc(offsetArraySize);
  assert(ptr);

  dptr=ptr;
  sptr=(const uint64_t *)(buffer+AQFS_TABLE_OFFS_ENTRIES);
  for (i=0; i<ft->maxEntries; i++) {
    *(dptr++)=GWEN_ENDIAN_LE64TOH(*(sptr++));
  }

  /* set offset array, takes over the given pointer */
  AQFS_Table_SetPtrEntries(ft, ptr);

  return 0;
}



AQFS_TABLE *AQFS_Table_fromBuffer(const uint8_t *buffer, uint64_t len)
{
  AQFS_TABLE *ft;
  int rv;

  ft=AQFS_Table_new();
  rv=AQFS_Table_ReadFromBuffer(ft, buffer, len);
  if (rv<0) {
    AQFS_Table_free(ft);
    return NULL;
  }

  return ft;
}



uint64_t AQFS_Table_toBuffer(const AQFS_TABLE *ft, uint8_t *buffer, uint64_t len)
{
  uint64_t offsetArraySize;
  const uint64_t *sptr;
  uint64_t *dptr;
  int i;

  offsetArraySize=ft->maxEntries*sizeof(uint64_t);
  if (len<(AQFS_TABLE_OFFS_ENTRIES+offsetArraySize)) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for complete table data (%d < %d)",
              (int) len, (int)(AQFS_TABLE_OFFS_ENTRIES+offsetArraySize));
    return 0;
  }

  *((uint64_t *)(buffer+AQFS_TABLE_OFFS_MAX_ENTRIES))=GWEN_ENDIAN_HTOLE64(ft->maxEntries);
  *((uint64_t *)(buffer+AQFS_TABLE_OFFS_FREE_ENTRIES))=GWEN_ENDIAN_HTOLE64(ft->freeEntries);
  *((uint64_t *)(buffer+AQFS_TABLE_OFFS_HIGHEST_ENTRY))=GWEN_ENDIAN_HTOLE64(ft->highestEntry);
  *((uint64_t *)(buffer+AQFS_TABLE_OFFS_NEXT_TABLE))=GWEN_ENDIAN_HTOLE64(ft->offsetNextTable);

  sptr=ft->ptrEntries;
  dptr=(uint64_t *)(buffer+AQFS_TABLE_OFFS_ENTRIES);
  for (i=0; i<ft->maxEntries; i++) {
    *(dptr++)=GWEN_ENDIAN_HTOLE64(*(sptr++));
  }


  return (AQFS_TABLE_OFFS_ENTRIES+offsetArraySize);
}



uint64_t AQFS_Table_Calculate_Size(const AQFS_TABLE *ft)
{
  return AQFS_TABLE_OFFS_ENTRIES+(ft->maxEntries*sizeof(uint64_t));
}




uint32_t AQFS_Table_GetRuntimeFlags(const AQFS_TABLE *ft)
{
  assert(ft);
  return ft->runtimeFlags;
}



void AQFS_Table_SetRuntimeFlags(AQFS_TABLE *ft, uint32_t i)
{
  assert(ft);
  ft->runtimeFlags=i;
}



void AQFS_Table_AddRuntimeFlags(AQFS_TABLE *ft, uint32_t i)
{
  assert(ft);
  ft->runtimeFlags|=i;
}



void AQFS_Table_SubRuntimeFlags(AQFS_TABLE *ft, uint32_t i)
{
  assert(ft);
  ft->runtimeFlags&=~i;
}



AQFS_TABLE_LIST *AQFS_Table_List_dup(const AQFS_TABLE_LIST *ftlOrig)
{
  if (ftlOrig) {
    AQFS_TABLE_LIST *ftl;
    const AQFS_TABLE *ftOrig;

    ftl=AQFS_Table_List_new();

    ftOrig=AQFS_Table_List_First(ftlOrig);
    while (ftOrig) {
      AQFS_TABLE *ft;

      ft=AQFS_Table_dup(ftOrig);
      AQFS_Table_List_Add(ft, ftl);
      ftOrig=AQFS_Table_List_Next(ftOrig);
    }

    return ftl;
  }
  else
    return NULL;
}



void AQFS_Table_Dump(const AQFS_TABLE *ft)
{
  if (ft) {
    fprintf(stderr, "Table\n");
    fprintf(stderr, " RefCount    : %d\n", ft->refCount);
    fprintf(stderr, " Table Pos   : %lu\n", (unsigned long)(ft->offsetTableHeaderStart));
    fprintf(stderr, " MaxEntries  : %lu\n", (unsigned long)(ft->maxEntries));
    fprintf(stderr, " FreeEntries : %lu\n", (unsigned long)(ft->freeEntries));
    fprintf(stderr, " HighestEntry: %lu\n", (unsigned long)(ft->highestEntry));
    fprintf(stderr, " Next Table  : %lu\n", (unsigned long)(ft->offsetNextTable));
  }
}



