/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFileStore (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQFS_TABLE_H
#define AQFS_TABLE_H


#include <gwenhywfar/list1.h>


#define AQFS_TABLE_HEADER_SIZE      32
#define AQFS_TABLE_MAX_ENTRIES      1024

/** table content was changed but not written to file (flag not used by functions in this class) */
#define AQFS_TABLE_RUNTIME_FLAGS_DIRTY  0x01
#define AQFS_TABLE_RUNTIME_FLAGS_ISCOPY 0x02



/**
 * A AQFS_TABLE object stores multiple uint64_t entries, which normally contain the offset
 * of other objects within a AQFS file.
 */
typedef struct AQFS_TABLE AQFS_TABLE;
GWEN_LIST_FUNCTION_DEFS(AQFS_TABLE, AQFS_Table)



#include <aqfilestorage/aqfs_file.h>



/**
 * @file aqfs_table.h
 *
 * This file contains definitions for a AQFS_TABLE object.
 *
 * Tables (type AQFS_TABLE) are basically just lists of offsets, used to store the positions of other objects or tables in a AQFS file.
 *
 * A table is stored within the AQFS file as an item (i.e. @ref AQFS_ITEM_HEADER followed by data created with @ref AQFS_Table_toBuffer).
 *
 * Tables can be used to organize objects within a AQFS_FILE. For that purpose they contain a number of "entries" (i.e. just some
 * numbers, which represent offsets within a AQFS file). A table also knows about the number of entries it can contain at most, the
 * number of used entries within a table and a pointer to the next table.
 *
 */


/** @name Construction, destruction, reading, writing.
 */
/*@{*/

/**
 * Constructor.
 */
AQFS_TABLE *AQFS_Table_new();

/**
 * Destructor.
 * @param ft pointer to AQFS_TABLE object
 */
void AQFS_Table_free(AQFS_TABLE *ft);

/**
 * Increase reference counter
 */
void AQFS_Table_Attach(AQFS_TABLE *ft);


/**
 * Create a deep copy of another table.
 * @return newly created deep copy of the original table.
 * @param ftOrig original to copy
 */
AQFS_TABLE *AQFS_Table_dup(const AQFS_TABLE *ftOrig);


/**
 * Creates a table with the given number of possible entries.
 * @return newly created table.
 * @param maxEntries maximum number of entries the new table should be able to hold
 */
AQFS_TABLE *AQFS_Table_Create(uint64_t maxEntries);


/**
 * Read a AQFS_TABLE object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param ft AQFS_TABLE object to read data into
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
int AQFS_Table_ReadFromBuffer(AQFS_TABLE *ft, const uint8_t *buffer, uint64_t len);

/**
 * Create a AQFS_TABLE object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
AQFS_TABLE *AQFS_Table_fromBuffer(const uint8_t *buffer, uint64_t len);

/**
 * Write content of a AQFS_TABLE object to a buffer.
 * @return number of bytes written, <1 on error
 * @param ft pointer to AQFS_TABLE object
 * @param buffer pointer to a buffer to write to
 * @param len size of the buffer pointed to by buffer
 */
uint64_t AQFS_Table_toBuffer(const AQFS_TABLE *ft, uint8_t *buffer, uint64_t len);

/*@}*/



/** @name Information stored in the file.
 */
/*@{*/

/**
 * Return the maximum number of entries this table can hold.
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_GetMaxEntries(const AQFS_TABLE *ft);

/**
 * Set the maximum number of entries this table can hold.
 * @param ft pointer to AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetMaxEntries(AQFS_TABLE *ft, uint64_t i);

/**
 * Return the number of free entries in this table.
 * For newly created tables this equals the value returned by @ref AQFS_Table_GetMaxEntries.
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_GetFreeEntries(const AQFS_TABLE *ft);

/**
 * Set the number of free entries in this table.
 * @param ft pointer to AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetFreeEntries(AQFS_TABLE *ft, uint64_t i);

/**
 * Decrement the number of free entries in this table by 1.
 * @param ft pointer to AQFS_TABLE object
 */
void AQFS_Table_DecFreeEntries(AQFS_TABLE *ft);

/**
 * Increment the number of free entries in this table by 1.
 * @param ft pointer to AQFS_TABLE object
 */
void AQFS_Table_IncFreeEntries(AQFS_TABLE *ft);


/**
 * Get highest entry ever used in this table.
 * This can be used to increase the speed of adding new entries to a table.
 * If this field is used one has not to loop through every entry of the table to
 * check for a free entry.
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_GetHighestEntry(const AQFS_TABLE *ft);

/**
 * Set highest entry ever used in this table.
 * @param ft AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetHighestEntry(AQFS_TABLE *ft, uint64_t i);

/**
 * Check whether the given argument is higher than the current highest entry (@ref AQFS_Table_GetHighestEntry).
 * If so the highest entry will be set to the given argument.
 * @param ft pointer to AQFS_TABLE object
 * @param i number to check the current highest entry against
 */
void AQFS_Table_CheckAndSetHighestEntry(AQFS_TABLE *ft, uint64_t i);


/**
 * Get the offset of the next table.
 * AQFS_TABLE objects are chained by this field. There is a convenience function to read
 * all tables belonging to a chain into a AQFS_TABLE_LIST (see @ref AQFS_File_ReadTableList).
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_GetOffsetNextTable(const AQFS_TABLE *ft);

/**
 * Set the offset of the next table.
 * @param ft pointer to AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetOffsetNextTable(AQFS_TABLE *ft, uint64_t i);


/**
 * Get a pointer to the array of entries.
 * Each member of that array represents an offset into the AQFS file.
 * If you add an entry to that array (i.e. use a previously unused entry) make sure
 * to call @ref AQFS_Table_CheckAndSetHighestEntry to make sure that the highest entry field
 * is updated.
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t *AQFS_Table_GetPtrEntries(const AQFS_TABLE *ft);

/**
 * Set the array of entries. Takes over the given pointer.
 * The previous entry array will be free'd before setting the new array.
 * The array will be free'd when overwritten later or upon destruction of the table object.
 * @param ft pointer to AQFS_TABLE object
 * @param ptr new value
 */
void AQFS_Table_SetPtrEntries(AQFS_TABLE *ft, uint64_t *ptr);

/*@}*/


/** @name Runtime data not stored in the file.
 */
/*@{*/

/**
 * Return the offset of the start of the AQFS_ITEM_HEADER object in the file for this table (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_GetOffsetTableHeaderStart(const AQFS_TABLE *ft);

/**
 * Set the offset of the start of the AQFS_ITEM_HEADER object in the file for this table (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetOffsetTableHeaderStart(AQFS_TABLE *ft, uint64_t i);


/**
 * Get runtime flags (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 */
uint32_t AQFS_Table_GetRuntimeFlags(const AQFS_TABLE *ft);

/**
 * Set runtime flags (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 * @param i new value
 */
void AQFS_Table_SetRuntimeFlags(AQFS_TABLE *ft, uint32_t i);

/**
 * Convenience function: Add runtime flags (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 * @param i flags to add (actually ORed to the current value)
 */
void AQFS_Table_AddRuntimeFlags(AQFS_TABLE *ft, uint32_t i);

/**
 * Convenience function: Subtract runtime flags (field not stored on disk).
 * @param ft pointer to AQFS_TABLE object
 * @param i flags to remove (actually achieved by inversing this value and ANDing with the current value).
 */
void AQFS_Table_SubRuntimeFlags(AQFS_TABLE *ft, uint32_t i);

/*@}*/


/** @name Helper Functions.
 */
/*@{*/

/**
 * Calculate number of bytes this table needs when storing (needed for @ref AQFS_Table_toBuffer).
 * This only contains the size of the table including the entries. It does not take into account
 * the size of a AQFS_ITEM_HEADER which is of course always prepended when writing items to file.
 * @param ft pointer to AQFS_TABLE object
 */
uint64_t AQFS_Table_Calculate_Size(const AQFS_TABLE *ft);

/**
 * Create a deep copy of a list of tables, i.e. for every table in the original list a deep copy
 * will be created and added to a new list, which is then returned.
 * @return newly created deep copy of the original list
 * @param ftlOrig pointer to the original table list
 */
AQFS_TABLE_LIST *AQFS_Table_List_dup(const AQFS_TABLE_LIST *ftlOrig);


void AQFS_Table_Dump(const AQFS_TABLE *ft);
int AQFS_Table_GetRefCounter(const AQFS_TABLE *ft);


/*@}*/



#endif
