/****************************************************************************
 * This file is part of the project AqFinance.
 * AqDataBase (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "aqfs_table_ptrlist.h"

#include <gwenhywfar/memory.h>
#include <gwenhywfar/debug.h>


/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static GWENHYWFAR_CB void _attachToTable(GWEN_SIMPLEPTRLIST *pl, void *p);
static GWENHYWFAR_CB void _detachFromTable(GWEN_SIMPLEPTRLIST *pl, void *p);



/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */


AQFS_TABLE_PTRLIST *AQFS_Table_PtrList_new(int tableMaxEntries)
{
  AQFS_TABLE_PTRLIST *tl;

  tl=GWEN_SimplePtrList_new(128, 128);
  GWEN_SimplePtrList_SetAttachObjectFn(tl, _attachToTable);
  GWEN_SimplePtrList_SetFreeObjectFn(tl, _detachFromTable);
  GWEN_SimplePtrList_AddFlags(tl, GWEN_SIMPLEPTRLIST_FLAGS_ATTACHTOOBJECTS);
  GWEN_SimplePtrList_AddFlags(tl, GWEN_SIMPLEPTRLIST_FLAGS_DETACHFROMOBJECTS);
  GWEN_SimplePtrList_SetUserIntData(tl, tableMaxEntries);

  return tl;
}



void AQFS_Table_PtrList_free(AQFS_TABLE_PTRLIST *tl)
{
  GWEN_SimplePtrList_free(tl);
}



void AQFS_Table_PtrList_Clear(AQFS_TABLE_PTRLIST *tl)
{
  GWEN_SimplePtrList_Clear(tl);
}



int AQFS_Table_PtrList_GetTableMaxEntries(const AQFS_TABLE_PTRLIST *tl)
{
  return GWEN_SimplePtrList_GetUserIntData(tl);
}



AQFS_TABLE_PTRLIST *AQFS_Table_PtrList_LazyCopy(AQFS_TABLE_PTRLIST *oldList)
{
  return GWEN_SimplePtrList_LazyCopy(oldList);
}



AQFS_TABLE *AQFS_Table_PtrList_GetTableAt(const AQFS_TABLE_PTRLIST *tl, uint64_t idx)
{
  return GWEN_SimplePtrList_GetPtrAt(tl, idx);
}



int AQFS_Table_PtrList_SetTableAt(AQFS_TABLE_PTRLIST *tl, uint64_t idx, AQFS_TABLE *t)
{
  return GWEN_SimplePtrList_SetPtrAt(tl, idx, t);
}



int64_t AQFS_Table_PtrList_AddTable(AQFS_TABLE_PTRLIST *tl, AQFS_TABLE *t)
{
  return GWEN_SimplePtrList_AddPtr(tl, t);
}



uint64_t AQFS_Table_PtrList_GetUsedTables(const AQFS_TABLE_PTRLIST *tl)
{
  return GWEN_SimplePtrList_GetUsedEntries(tl);
}



int64_t AQFS_Table_PtrList_GetLastTablePos(const AQFS_TABLE_PTRLIST *tl)
{
  uint64_t idx;

  idx=GWEN_SimplePtrList_GetUsedEntries(tl);
  if (idx)
    return idx-1;
  return GWEN_ERROR_NO_DATA;
}



int64_t AQFS_Table_PtrList_GetEntryAt(const AQFS_TABLE_PTRLIST *tl, uint64_t idx)
{
  int entriesPerTable;

  entriesPerTable=GWEN_SimplePtrList_GetUserIntData(tl);
  if (entriesPerTable) {
    uint64_t tablePos;
    AQFS_TABLE *t;

    tablePos=idx/entriesPerTable;
    t=AQFS_Table_PtrList_GetTableAt(tl, tablePos);
    if (t) {
      uint64_t *entries;

      entries=AQFS_Table_GetPtrEntries(t);
      if (entries) {
        uint64_t entryPos;

        entryPos=idx%entriesPerTable;
        return entries[entryPos];
      }
    }
    else {
      DBG_ERROR(AQFS_LOGDOMAIN, "No table at table pos %lu", (unsigned long) tablePos);
    }
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "No entriesPerTable");
  }

  return GWEN_ERROR_BUFFER_OVERFLOW;
}



int AQFS_Table_PtrList_SetEntryAt(AQFS_TABLE_PTRLIST *tl, uint64_t idx, uint64_t entry)
{
  int rv;
  int entriesPerTable;

  rv=GWEN_SimplePtrList_EnsureWritability(tl);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return (int64_t) rv;
  }

  entriesPerTable=GWEN_SimplePtrList_GetUserIntData(tl);
  if (entriesPerTable) {
    uint64_t tablePos;
    AQFS_TABLE *t;

    tablePos=idx/entriesPerTable;
    t=AQFS_Table_PtrList_GetTableAt(tl, tablePos);
    if (t) {
      uint64_t *entries;

      /* copy table if necessary (copy-on-write) */
      if (!(AQFS_Table_GetRuntimeFlags(t) & AQFS_TABLE_RUNTIME_FLAGS_ISCOPY)) {
        AQFS_TABLE *pTableCopy;

        pTableCopy=AQFS_Table_dup(t);
        AQFS_Table_PtrList_SetTableAt(tl, tablePos, pTableCopy); /* increases refCounter, need to free */
        t=pTableCopy;
        AQFS_Table_AddRuntimeFlags(t, AQFS_TABLE_RUNTIME_FLAGS_ISCOPY);
        AQFS_Table_free(pTableCopy);
      }

      entries=AQFS_Table_GetPtrEntries(t);
      if (entries) {
        uint64_t entryPos;

        entryPos=idx%entriesPerTable;
        entries[entryPos]=entry;
        AQFS_Table_AddRuntimeFlags(t, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
        return 0;
      }
    } /* if (t) */
    else {
      DBG_ERROR(AQFS_LOGDOMAIN, "No table at position %lu", (unsigned long int) tablePos);
      return GWEN_ERROR_INTERNAL;
    }
  } /* if (entriesPerTable) */
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "No entriesPerTable, internal error");
    return GWEN_ERROR_INTERNAL;
  }

  return GWEN_ERROR_BUFFER_OVERFLOW;

}



int64_t AQFS_Table_PtrList_AddEntry(AQFS_TABLE_PTRLIST *tl, uint64_t entry)
{
  AQFS_TABLE *pTableCurrent=NULL;
  int64_t idxTableCurrent=0;
  int entriesPerTable=AQFS_Table_PtrList_GetTableMaxEntries(tl);
  int rv;

  rv=GWEN_SimplePtrList_EnsureWritability(tl);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    return (int64_t) rv;
  }

  /* get last table */
  idxTableCurrent=AQFS_Table_PtrList_GetLastTablePos(tl);
  DBG_VERBOUS(AQFS_LOGDOMAIN, "Last table pos is %d", (int)idxTableCurrent);
  if (idxTableCurrent>=0)
    pTableCurrent=AQFS_Table_PtrList_GetTableAt(tl, idxTableCurrent);

  /* check last table for existence and free entries, possibly create and add new table */
  if (pTableCurrent==NULL || AQFS_Table_GetFreeEntries(pTableCurrent)==0) {
    /* create new table */
    if (pTableCurrent==NULL) {
      DBG_VERBOUS(AQFS_LOGDOMAIN, "No table, need to create one");
    }
    else if (AQFS_Table_GetFreeEntries(pTableCurrent)==0) {
      DBG_VERBOUS(AQFS_LOGDOMAIN, "Current table has no free entries, need to create new one");
    }

    DBG_VERBOUS(AQFS_LOGDOMAIN, "Creating table with %d entries", entriesPerTable);
    pTableCurrent=AQFS_Table_Create(entriesPerTable);
    AQFS_Table_AddRuntimeFlags(pTableCurrent, AQFS_TABLE_RUNTIME_FLAGS_ISCOPY); /* no need to copy later */

    /* add table to list */
    idxTableCurrent=AQFS_Table_PtrList_AddTable(tl, pTableCurrent);
    if (idxTableCurrent<0) {
      DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) idxTableCurrent);
      AQFS_Table_free(pTableCurrent);
      return idxTableCurrent;
    }
    AQFS_Table_free(pTableCurrent);
  } /* if (pTableCurrent || AQFS_Table_GetFreeEntries(pTableCurrent)==0) */

  /* allocate free entry in current table */
  if (pTableCurrent && AQFS_Table_GetFreeEntries(pTableCurrent)) {
    uint64_t *ptr;
    int64_t index=idxTableCurrent*entriesPerTable;
    int64_t entryPos;

    /* copy table if necessary (copy-on-write) */
    if (!(AQFS_Table_GetRuntimeFlags(pTableCurrent) & AQFS_TABLE_RUNTIME_FLAGS_ISCOPY)) {
      AQFS_TABLE *pTableCopy;

      DBG_VERBOUS(AQFS_LOGDOMAIN, "Copying table at idx %lu", (unsigned long) idxTableCurrent);
      pTableCopy=AQFS_Table_dup(pTableCurrent);
      AQFS_Table_PtrList_SetTableAt(tl, idxTableCurrent, pTableCopy);
      AQFS_Table_free(pTableCopy);
      pTableCurrent=pTableCopy;
      AQFS_Table_AddRuntimeFlags(pTableCurrent, AQFS_TABLE_RUNTIME_FLAGS_ISCOPY);
    }

    ptr=AQFS_Table_GetPtrEntries(pTableCurrent);

    /* find entryPos of free entry in pTableCurrent */
    DBG_VERBOUS(AQFS_LOGDOMAIN, "Current table (ptr=%p, %d entriesPerTable):", (void *)ptr, entriesPerTable);
    /*AQFS_Table_Dump(pTableCurrent);*/
    if (AQFS_Table_GetFreeEntries(pTableCurrent)==AQFS_Table_GetMaxEntries(pTableCurrent)) {
      /** all entries are free, this is simple */
      entryPos=0;
    }
    else {
      if (AQFS_Table_GetHighestEntry(pTableCurrent)+1<entriesPerTable) {
        /* fastest way: Just append to the end */
        DBG_VERBOUS(AQFS_LOGDOMAIN, "Finding free empty the fast way");
        entryPos=AQFS_Table_GetHighestEntry(pTableCurrent)+1;
        if (ptr[entryPos]!=0) {
          DBG_ERROR(AQFS_LOGDOMAIN, "Entry[highest+1] should be 0 but isnt, SNH!");
          return GWEN_ERROR_INTERNAL;
        }
      }
      else {
        /* slower way: find free entry somewhere in the table */
        DBG_VERBOUS(AQFS_LOGDOMAIN, "Finding free empty the slow way");
        for (entryPos=0; entryPos<entriesPerTable; entryPos++) {
          if (ptr[entryPos]==0)
            break;
        }
      }
    }

    DBG_VERBOUS(AQFS_LOGDOMAIN, "New entry will be at index %lu in table %lu (index=%lu, resulting index: %lu)",
                (unsigned long) entryPos,
                (unsigned long) idxTableCurrent,
                (unsigned long) index,
                (unsigned long)(index+entryPos));

    if (entryPos<entriesPerTable) {
      /* store new entry, get index */
      ptr[entryPos]=entry;
      index+=entryPos;
      AQFS_Table_DecFreeEntries(pTableCurrent);
      AQFS_Table_CheckAndSetHighestEntry(pTableCurrent, entryPos);
      AQFS_Table_AddRuntimeFlags(pTableCurrent, AQFS_TABLE_RUNTIME_FLAGS_DIRTY);
      return index;
    }
    else {
      DBG_ERROR(AQFS_LOGDOMAIN, "Free entry not found, internal counter is invalid. SNH!");
      return GWEN_ERROR_INTERNAL;
    }
  }
  else {
    DBG_ERROR(AQFS_LOGDOMAIN, "Still no table? SNH!");
    return GWEN_ERROR_INTERNAL;
  }
}







void _attachToTable(GWEN_UNUSED GWEN_SIMPLEPTRLIST *pl, void *p)
{
  AQFS_TABLE *ft;

  ft=(AQFS_TABLE *) p;
  AQFS_Table_Attach(ft);
}



void _detachFromTable(GWEN_UNUSED GWEN_SIMPLEPTRLIST *pl, void *p)
{
  AQFS_TABLE *ft;

  ft=(AQFS_TABLE *) p;
  AQFS_Table_free(ft);
}



#include "aqfs_table_ptrlist-t.c"



