/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFinance (c) by 2015 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "dbfile_p.h"

#include "file/iolowlevel/layer.h"
#include "file/iologged/layer.h"

#include "aqdatabase/backendsupport/ioquery/layer.h"
#include "aqdatabase/backendsupport/iocached/layer.h"
#include "aqdatabase/backendsupport/iotimed/layer.h"

#include <aqdatabase/aqdb_db_be.h>
#include <aqdatabase/aqdb_object_be.h>
#include <aqdatabase/backendsupport/aqdb_sessionlog.h>
#include <aqdatabase/aqdb_sessiondescr.h>
#include <aqdatabase/aqdb_query.h>
#include <aqdatabase/backendsupport/aqdb_objectquery.h>

#include <gwenhywfar/misc.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/endianfns.h>
#include <gwenhywfar/multicache.h>

#include <errno.h>
#include <string.h>



GWEN_INHERIT(AQDB_DB, AQDB_DB_FILE)



static AQDB_IOLAYER *_createIoLayer(AQDB_DB *db);





AQDB_DB *AQDB_DbFile_new(const char *urlString)
{
  AQDB_DB *db;
  AQDB_DB_FILE *xdb;

  db=AQDB_DB_new(urlString);
  if (db==NULL)
    return NULL;

  GWEN_NEW_OBJECT(AQDB_DB_FILE, xdb);
  if (xdb==NULL) {
    AQDB_DB_free(db);
    return NULL;
  }
  GWEN_INHERIT_SETDATA(AQDB_DB, AQDB_DB_FILE, db, xdb,
                       AQDB_DbFile_FreeData);

  /* set virtual functions */
  AQDB_DB_SetCreateFn(db, AQDB_DbFile_Create);
  AQDB_DB_SetOpenFn(db, AQDB_DbFile_Open);
  AQDB_DB_SetCloseFn(db, AQDB_DbFile_Close);
  AQDB_DB_SetCheckFn(db, AQDB_DbFile_Check);

  AQDB_DB_SetCreateTableFn(db, AQDB_DbFile_CreateTable);
  AQDB_DB_SetOpenTableFn(db, AQDB_DbFile_OpenTable);
  AQDB_DB_SetCloseTableFn(db, AQDB_DbFile_CloseTable);
  AQDB_DB_SetGetTableNameListFn(db, AQDB_DbFile_GetTableNameList);

  AQDB_DB_SetAddObjectFn(db, AQDB_DbFile_AddObject);
  AQDB_DB_SetDeleteObjectFn(db, AQDB_DbFile_DeleteObject);
  AQDB_DB_SetWriteObjectFn(db, AQDB_DbFile_WriteObject);
  AQDB_DB_SetReadObjectFn(db, AQDB_DbFile_ReadObject);

  AQDB_DB_SetBeginEditFn(db, AQDB_DbFile_BeginEdit);
  AQDB_DB_SetEndEditFn(db, AQDB_DbFile_EndEdit);

  AQDB_DB_SetGetColumnListFn(db, AQDB_DbFile_GetColumnList);

  AQDB_DB_SetQuerySubmitFn(db, AQDB_DbFile_QuerySubmit);
  AQDB_DB_SetQueryGetFirstFn(db, AQDB_DbFile_QueryGetFirst);
  AQDB_DB_SetQueryGetNextFn(db, AQDB_DbFile_QueryGetNext);
  AQDB_DB_SetQueryCloseFn(db, AQDB_DbFile_QueryClose);

  AQDB_DB_SetGetRollbackSessionsFn(db, AQDB_DbFile_GetRollbackSessions);
  AQDB_DB_SetUndoSessionFn(db, AQDB_DbFile_UndoSession);

  return db;
}



void GWENHYWFAR_CB AQDB_DbFile_FreeData(void *bp, void *p)
{
  AQDB_DB_FILE *xdb;

  xdb=(AQDB_DB_FILE *)p;
  assert(xdb);

  if (xdb->ioLayer)
    AQDB_IoLayer_free(xdb->ioLayer);

  GWEN_FREE_OBJECT(xdb);
}



AQDB_IOLAYER *_createIoLayer(AQDB_DB *db)
{
  const GWEN_URL *url;
  AQDB_IOLAYER *ioLayer;
  const char *s;
  int rv;

  /* Create io layers (lowlevel <- logged <- cached */
  url=AQDB_DB_GetUrl(db);
  ioLayer=AQDB_IoLayerLowlevel_new(db, url);
  if (ioLayer==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return NULL;
  }
  rv=AQDB_IoLayerLogged_Extend(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }
  rv=AQDB_IoLayerCached_Extend(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }
  rv=AQDB_IoLayerQuery_Extend(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  s=getenv("AQDB_TIMING");
  if (s && *s=='1') {
    rv=AQDB_IoLayerTimed_Extend(ioLayer);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return NULL;
    }
  }

  return ioLayer;
}



AQDB_IOLAYER *AQDB_DbFile_GetIoLayer(const AQDB_DB *db)
{
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  return xdb->ioLayer;
}



int GWENHYWFAR_CB AQDB_DbFile_Create(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_FILE *xdb;
  AQDB_IOLAYER *ioLayer;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Creating database");

  if (xdb->openFlags) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database already open");
    return GWEN_ERROR_OPEN;
  }

  ioLayer=_createIoLayer(db);
  if (ioLayer==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->ioLayer=ioLayer;

  rv=AQDB_IoLayer_Create(xdb->ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_IoLayer_free(xdb->ioLayer);
    xdb->ioLayer=NULL;
    return rv;
  }

  AQDB_IoLayer_free(xdb->ioLayer);
  xdb->ioLayer=NULL;
  xdb->openFlags=0;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database created");
  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_Open(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_FILE *xdb;
  AQDB_IOLAYER *ioLayer;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Opening database");

  if ((flags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE))==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No open flags given (%d)", flags);
    return GWEN_ERROR_INVALID;
  }

  if (xdb->openFlags) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database already open");
    return GWEN_ERROR_OPEN;
  }

  ioLayer=_createIoLayer(db);
  if (ioLayer==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xdb->ioLayer=ioLayer;

  rv=AQDB_IoLayer_Open(xdb->ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* database is open */
  xdb->openFlags=flags;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database is now open");

  return rv;
}



int GWENHYWFAR_CB AQDB_DbFile_Close(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing database");

  if (xdb->openFlags==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "DB not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  rv=AQDB_IoLayer_Close(xdb->ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d), ignoring", rv);
  }

  AQDB_IoLayer_free(xdb->ioLayer);
  xdb->ioLayer=NULL;
  xdb->openFlags=0;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database is now closed");

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_Check(AQDB_DB *db, GWEN_BUFFER *resultTextBuf)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Checking database");

  rv=AQDB_IoLayer_Check(xdb->ioLayer, resultTextBuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_CreateTable(AQDB_DB *db,
                                          const char *name,
                                          uint32_t flags,
                                          const AQDB_COLUMN_LIST *columns)
{
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (xdb->openFlags==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "DB not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  if (xdb->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_INVALID;
  }

  if (AQDB_IoLayer_CreateTable(xdb->ioLayer, name, flags, columns)==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_OpenTable(AQDB_DB *db,
                                        uint32_t aflags,
                                        const char *name,
                                        uint32_t flags,
                                        AQDB_ID *pTableId)
{
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (xdb->openFlags==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "DB not open");
    return GWEN_ERROR_NOT_OPEN;
  }

  if (xdb->isInEditMode==0)
    aflags&=~AQDB_ACCESS_FLAGS_ISWRITER;

  if (AQDB_IoLayer_OpenTable(xdb->ioLayer, aflags, name, flags, pTableId)==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_CloseTable(AQDB_DB *db,
                                         AQDB_ID tableId,
                                         uint32_t flags)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  uint32_t oflags;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table with id %d not found.", tableId);
    return GWEN_ERROR_INVALID;
  }

  oflags=AQDB_Table_GetOpenFlags(dbt);
  if ((oflags & AQDB_ACTION_FLAGS_MASK)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not open.", tableId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_IoLayer_CloseTable(xdb->ioLayer, dbt, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE_LIST *dbTypeTables;
  AQDB_TABLE *dbt;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbTypeTables=AQDB_IoLayer_GetDbTypeTables(xdb->ioLayer);
  if (dbTypeTables && AQDB_Table_List_GetCount(dbTypeTables)) {
    dbt=AQDB_Table_List_First(dbTypeTables);
    while (dbt) {
      const char *s;

      s=AQDB_Table_GetName(dbt);
      if (s && *s && *s!='_')
        GWEN_StringList_AppendString(tableNameList, s, 0, 1);
      dbt=AQDB_Table_List_Next(dbt);
    }
  }

  return 0;
}




int GWENHYWFAR_CB AQDB_DbFile_AddObject(AQDB_DB *db, AQDB_ID tableId, AQDB_OBJECT *object)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (xdb->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_INVALID;
  }

  /* get type table */
  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }
  if (!(AQDB_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_AddObject(xdb->ioLayer, dbt, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_DeleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (xdb->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_INVALID;
  }

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id.");
    return GWEN_ERROR_INVALID;
  }

  /* get type table */
  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }
  if (!(AQDB_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_DeleteObject(xdb->ioLayer, dbt, objectId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object)
{
  AQDB_DB_FILE *xdb;
  AQDB_ID tableId;
  AQDB_ID objectId;
  AQDB_TABLE *dbt;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (xdb->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_INVALID;
  }

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id.");
    return GWEN_ERROR_INVALID;
  }

  /* get type table */
  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }
  if (!(AQDB_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_WriteObject(xdb->ioLayer, dbt, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_ReadObject(AQDB_DB *db,
                                         uint32_t aflags,
                                         AQDB_ID tableId,
                                         AQDB_ID objectId,
                                         AQDB_OBJECT **pObject)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  if (xdb->isInEditMode==0)
    aflags&=~AQDB_ACCESS_FLAGS_ISWRITER;

  rv=AQDB_IoLayer_ReadObject(xdb->ioLayer, aflags, dbt, objectId, pObject);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_BeginEdit(AQDB_DB *db, uint32_t editFlags, const char *user)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Begin edit session");

  if ((xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database not open for writing");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->isInEditMode!=0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is already in edit mode");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_IoLayer_BeginEdit(xdb->ioLayer, editFlags, user);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  xdb->isInEditMode=1;

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_EndEdit(AQDB_DB *db, uint32_t flags)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Ending edit session%s", (flags & AQDB_ACTION_FLAGS_ABORT)?" (ABANDON)":"");

  if (xdb->isInEditMode==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode");
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_IoLayer_EndEdit(xdb->ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  xdb->isInEditMode=0;

  return 0;
}



AQDB_COLUMN_LIST *GWENHYWFAR_CB AQDB_DbFile_GetColumnList(AQDB_DB *db, uint32_t aflags, AQDB_ID tableId)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  AQDB_COLUMN_LIST *cl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (xdb->isInEditMode==0)
    aflags&=~AQDB_ACCESS_FLAGS_ISWRITER;

  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return NULL;
  }

  cl=AQDB_Table_GetColumnList(dbt);
  if (cl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d has not column list, should not happen", (int) tableId);
    return NULL;
  }

  cl=AQDB_Column_List_dup(cl);

  return cl;
}



int GWENHYWFAR_CB AQDB_DbFile_QuerySubmit(AQDB_DB *db,
                                          uint32_t aflags,
                                          AQDB_ID tableId,
                                          AQDB_ID refQueryId,
                                          const char *queryExpression,
                                          uint32_t flags,
                                          AQDB_ID *pQueryId)
{
  AQDB_DB_FILE *xdb;
  AQDB_TABLE *dbt;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (xdb->isInEditMode==0)
    aflags&=~AQDB_ACCESS_FLAGS_ISWRITER;

  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return GWEN_ERROR_GENERIC;
  }

  rv=AQDB_IoLayer_QuerySubmit(xdb->ioLayer, aflags, dbt, refQueryId, queryExpression, flags, pQueryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_QueryGetFirst(AQDB_DB *db, AQDB_ID queryId, AQDB_ID *pIdListPtr, int *pIdListLen)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_QueryGetFirst(xdb->ioLayer, queryId, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_QueryGetNext(AQDB_DB *db,
                                           AQDB_ID queryId,
                                           AQDB_ID *pIdListPtr,
                                           int *pIdListLen)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_QueryGetNext(xdb->ioLayer, queryId, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_QueryClose(AQDB_DB *db, AQDB_ID queryId)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_IoLayer_QueryClose(xdb->ioLayer, queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_GetRollbackSessions(AQDB_DB *db, uint32_t aflags, AQDB_SESSION_DESCR_LIST *sl)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (xdb->isInEditMode==0)
    aflags&=~AQDB_ACCESS_FLAGS_ISWRITER;

  rv=AQDB_IoLayer_GetRollbackSessions(xdb->ioLayer, aflags, sl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_UndoSession(AQDB_DB *db, AQDB_ID sessionId)
{
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Undo session");
  rv=AQDB_IoLayer_UndoSession(xdb->ioLayer, sessionId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbFile_GetMatchingObjects(AQDB_DB *db,
                                   AQDB_ID tableId,
                                   uint32_t accessFlags,
                                   const char *expr,
                                   uint32_t queryFlags,
                                   AQDB_OBJECT_LIST *ol)
{
  AQDB_DB_FILE *xdb;
  AQDB_ID queryId;
  AQDB_ID resultList[64];
  int idListLen;
  int i;
  int rv;
  AQDB_TABLE *dbt;
  uint32_t oflags;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  dbt=AQDB_IoLayer_GetTableById(xdb->ioLayer, tableId);
  if (dbt==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table with id %d not found.", tableId);
    return GWEN_ERROR_INVALID;
  }

  oflags=AQDB_Table_GetOpenFlags(dbt);
  if ((oflags & AQDB_ACTION_FLAGS_MASK)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not open.", tableId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_IoLayer_QuerySubmit(xdb->ioLayer, accessFlags, dbt, 0, expr, queryFlags, &queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  idListLen=64;
  rv=AQDB_IoLayer_QueryGetFirst(xdb->ioLayer, queryId, resultList, &idListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_IoLayer_QueryClose(xdb->ioLayer, queryId);
    return rv;
  }

  for (;;) {
    for (i=0; i<idListLen; i++) {
      AQDB_OBJECT *o=NULL;

      rv=AQDB_IoLayer_ReadObject(xdb->ioLayer, accessFlags, dbt, resultList[i], &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_IoLayer_QueryClose(xdb->ioLayer, queryId);
        return rv;
      }
      assert(o);
      AQDB_Object_List_Add(o, ol);
    }

    idListLen=64;
    rv=AQDB_IoLayer_QueryGetNext(xdb->ioLayer, queryId, resultList, &idListLen);
    if (rv<0) {
      if (rv!=GWEN_ERROR_NOT_FOUND) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_IoLayer_QueryClose(xdb->ioLayer, queryId);
        return rv;
      }
      else
        break;
    }
  }

  rv=AQDB_IoLayer_QueryClose(xdb->ioLayer, queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}





uint32_t AQDB_DbFile_GetOpenFlags(const AQDB_DB *db)
{
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  return xdb->openFlags;
}



void AQDB_DbFile_SetOpenFlags(AQDB_DB *db, uint32_t i)
{
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  xdb->openFlags=i;
}



/* end */


















