/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


/* included from "layer.c" */





int AQDB_IoLayerLogged_OpenDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_ID tid;
  AQDB_TABLE *dbt;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Opening database");

  if (xio->openFn==NULL || xio->openTableFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  rv=xio->openFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* open table for session type */
  tid=0;
  dbt=AQDB_IoLayer_OpenTable(ioLayer, 0, AQDB_DBFILE_TYPENAME_SESSIONS, flags & AQDB_ACTION_FLAGS_MASK, &tid);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xio->tableIdSessionDescr=tid;

  /* open table for changelog type */
  tid=0;
  dbt=AQDB_IoLayer_OpenTable(ioLayer, 0, AQDB_DBFILE_TYPENAME_CHANGELOG, flags & AQDB_ACTION_FLAGS_MASK, &tid);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }
  xio->tableIdChangeLog=tid;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database is now open");
  return 0;
}



int AQDB_IoLayerLogged_CreateDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->createFn==NULL || xio->openFn==NULL || xio->beginEditFn==NULL || xio->endEditFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Missing virtual function in base layer");
    return GWEN_ERROR_GENERIC;
  }

  rv=xio->createFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=xio->openFn(ioLayer, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=xio->beginEditFn(ioLayer, 0, NULL);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=_createSessionTypes(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xio->endEditFn(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    xio->closeFn(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  rv=xio->endEditFn(ioLayer, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xio->endEditFn(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    xio->closeFn(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  rv=xio->closeFn(ioLayer, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    xio->closeFn(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  return 0;
}



int _createSessionTypes(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_TABLE *dbt;
  AQDB_COLUMN_LIST *cl;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  /* create session descrption type */
  cl=AQDB_SessionDescr_CreateColumnList();
  dbt=xio->createTableFn(ioLayer, AQDB_DBFILE_TYPENAME_SESSIONS, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  /* create changelog type */
  cl=AQDB_DbFile_Changelog_CreateColumnList();
  dbt=xio->createTableFn(ioLayer, AQDB_DBFILE_TYPENAME_CHANGELOG, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, cl);
  AQDB_Column_List_free(cl);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}



int AQDB_IoLayerLogged_CloseDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing database");

  if (xio->closeFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Base layer has virtual function CLOSE not set");
    return GWEN_ERROR_GENERIC;
  }

  rv=xio->closeFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database is now closed");
  return 0;
}



AQDB_OBJECT_LIST *AQDB_IoLayerLogged_ReadMatchingObjects(AQDB_IOLAYER *ioLayer,
                                                         AQDB_ID tableId,
                                                         uint32_t accessFlags,
                                                         const char *queryExpression,
                                                         uint32_t queryFlags)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_OBJECT_LIST *objectList;
  GWEN_IDLIST64 *sourceIdList;
  AQDB_TABLE *databaseTypeTable;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->readObjectFn==NULL || xio->getActiveIdsForTableFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "iologged: missing virtual function(s)");
    return NULL;
  }

  databaseTypeTable=AQDB_IoLayer_GetTableById(ioLayer, tableId);
  if (databaseTypeTable==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table with id %lu not found", (unsigned long int) tableId);
    return NULL;
  }

  sourceIdList=xio->getActiveIdsForTableFn(ioLayer, accessFlags, databaseTypeTable, queryFlags);
  if (sourceIdList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No objects in table %s", AQDB_Table_GetName(databaseTypeTable));
    return NULL;
  }

  if (queryExpression)
    objectList=_readMatchingObjects(ioLayer, databaseTypeTable, sourceIdList, accessFlags, queryExpression, queryFlags);
  else
    objectList=_readAllObjects(ioLayer, databaseTypeTable, sourceIdList, accessFlags, queryFlags);

  GWEN_IdList64_free(sourceIdList);

  if (AQDB_Object_List_GetCount(objectList)<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "No matching objects");
    AQDB_Object_List_free(objectList);
    return NULL;
  }

  return objectList;
}



AQDB_OBJECT_LIST *_readMatchingObjects(AQDB_IOLAYER *ioLayer,
                                       AQDB_TABLE *databaseTypeTable,
                                       GWEN_IDLIST64 *sourceIdList,
                                       uint32_t accessFlags,
                                       const char *queryExpression,
                                       uint32_t queryFlags)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_OBJECT_LIST *objectList;
  AQDB_COLUMN_LIST *columnList;
  GWEN_IDLIST64_ITERATOR *it;
  uint64_t id;
  AQDB_QUERY *dbq;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  objectList=AQDB_Object_List_new();

  columnList=AQDB_Table_GetColumnList(databaseTypeTable);
  if (columnList==NULL || AQDB_Column_List_GetCount(columnList)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %s has no columns", AQDB_Table_GetName(databaseTypeTable));
    AQDB_Object_List_free(objectList);
    return NULL;
  }

  /* parse expression */
  dbq=AQDB_ObjectQuery_new(queryExpression, columnList);
  if (dbq==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No query generated (%s)", AQDB_Table_GetName(databaseTypeTable));
    AQDB_Object_List_free(objectList);
    return NULL;
  }

  it=GWEN_IdList64_Iterator_new(sourceIdList);
  id=GWEN_IdList64_Iterator_GetFirstId(it);
  while (id) {
    AQDB_OBJECT *o=NULL;
    int rv;

    rv=xio->readObjectFn(ioLayer, accessFlags, databaseTypeTable, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "Error reading object (%d)", rv);
    }
    else {
      /* an object has been read, check whether it matches */
      rv=AQDB_ObjectQuery_Check(dbq, o);
      if (rv) {
        /* it actually matches, add it to the idlist */
        AQDB_Object_List_Add(o, objectList);
        if (queryFlags & AQDB_QUERY_FLAGS_STOPATFIRST)
          break;
      }
      else
        AQDB_Object_free(o);
    }
    id=GWEN_IdList64_Iterator_GetNextId(it);
  } /* while */
  AQDB_Query_free(dbq);
  GWEN_IdList64_Iterator_free(it);

  if (AQDB_Object_List_GetCount(objectList)<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "No matching objects");
    AQDB_Object_List_free(objectList);
    return NULL;
  }

  return objectList;
}



AQDB_OBJECT_LIST *_readAllObjects(AQDB_IOLAYER *ioLayer,
                                  AQDB_TABLE *databaseTypeTable,
                                  GWEN_IDLIST64 *sourceIdList,
                                  uint32_t accessFlags,
                                  uint32_t queryFlags)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_OBJECT_LIST *objectList;
  GWEN_IDLIST64_ITERATOR *it;
  uint64_t id;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  objectList=AQDB_Object_List_new();

  it=GWEN_IdList64_Iterator_new(sourceIdList);
  id=GWEN_IdList64_Iterator_GetFirstId(it);
  while (id) {
    AQDB_OBJECT *o=NULL;
    int rv;

    rv=xio->readObjectFn(ioLayer, accessFlags, databaseTypeTable, id, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "Error reading object (%d)", rv);
    }
    else {
      AQDB_Object_List_Add(o, objectList);
      if (queryFlags & AQDB_QUERY_FLAGS_STOPATFIRST)
        break;
    }

    /* try next id */
    id=GWEN_IdList64_Iterator_GetNextId(it);
  } /* while */
  GWEN_IdList64_Iterator_free(it);

  if (AQDB_Object_List_GetCount(objectList)<1) {
    DBG_INFO(AQDB_LOGDOMAIN, "No matching objects");
    AQDB_Object_List_free(objectList);
    return NULL;
  }

  return objectList;
}


