/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


/* included from "layer.c" */



int AQDB_IoLayerLogged_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;
  int64_t rd;
  uint64_t newOffset;
  AQDB_ID objectId;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->addObjectFn==NULL || xio->getObjectPosFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }


  /* really add object to database */
  rv=xio->addObjectFn(ioLayer, dbt, object);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* object id is assigned by lower level AddObject() function */
  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id after adding, SNH!");
    return GWEN_ERROR_INTERNAL;
  }

  /* find new object position within the table */
  rd=xio->getObjectPosFn(ioLayer, dbt, objectId-1, 1);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  newOffset=rd;

  if (!(AQDB_Table_GetFlags(dbt) & AQDB_DBFILE_TABLE_FLAGS_NOCHGLOG)) {
    AQDB_SESSION_DESCR *currentSession;

    currentSession=AQDB_IoLayerLogged_GetCurrentSession(ioLayer);
    rv=_writeChangeLog(ioLayer, AQDB_Table_GetTableId(dbt), objectId, 0, newOffset,
                       AQDB_DbFile_ChangelogType_Addition);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return rv;
    }

    /* increase number of additions */
    if (currentSession)
      AQDB_SessionDescr_IncAdditions(currentSession);
  }
  else {
    DBG_DEBUG(AQDB_LOGDOMAIN,
              "Not writing changelog entry for type \"%s\" (NOCHGLOG)",
              AQDB_Table_GetName(dbt));
  }

  return 0;
}



int AQDB_IoLayerLogged_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;
  int64_t rd;
  uint64_t oldOffset;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->deleteObjectFn==NULL || xio->getObjectPosFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  /* find current object position within the table */
  rd=xio->getObjectPosFn(ioLayer, dbt, objectId-1, 1);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  oldOffset=rd;

  rv=xio->deleteObjectFn(ioLayer, dbt, objectId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  if (!(AQDB_Table_GetFlags(dbt) & AQDB_DBFILE_TABLE_FLAGS_NOCHGLOG)) {
    AQDB_SESSION_DESCR *currentSession;

    currentSession=AQDB_IoLayerLogged_GetCurrentSession(ioLayer);
    /* @todo this should be AQDB_DbFile_ChangelogType_Deletion */
    rv=_writeChangeLog(ioLayer, AQDB_Table_GetTableId(dbt), objectId, oldOffset, 0,
                       AQDB_DbFile_ChangelogType_Modification);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return rv;
    }
    /* increase number of deletions */
    if (currentSession)
      AQDB_SessionDescr_IncDeletions(currentSession);
  }
  else {
    DBG_DEBUG(AQDB_LOGDOMAIN,
              "Not writing changelog entry for type \"%s\" (NOCHGLOG)",
              AQDB_Table_GetName(dbt));
  }

  return 0;
}



int AQDB_IoLayerLogged_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                          const AQDB_OBJECT *object)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_ID tableId;
  AQDB_ID objectId;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->writeObjectFn==NULL || xio->getObjectPosFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id.");
    return GWEN_ERROR_INVALID;
  }

  if (1) {
    int rv;
    int64_t rd;
    uint64_t oldOffset;
    uint64_t newOffset;

    /* find current object position within the table */
    rd=xio->getObjectPosFn(ioLayer, dbt, objectId-1, 1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    oldOffset=rd;

    /* write object to database */
    rv=xio->writeObjectFn(ioLayer, dbt, object);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    /* find new object position within the table */
    rd=xio->getObjectPosFn(ioLayer, dbt, objectId-1, 1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    newOffset=rd;

    if (!(AQDB_Table_GetFlags(dbt) & AQDB_DBFILE_TABLE_FLAGS_NOCHGLOG)) {
      AQDB_SESSION_DESCR *currentSession;

      currentSession=AQDB_IoLayerLogged_GetCurrentSession(ioLayer);
      rv=_writeChangeLog(ioLayer, tableId, objectId, oldOffset, newOffset, AQDB_DbFile_ChangelogType_Modification);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
        return rv;
      }

      if (currentSession)
        AQDB_SessionDescr_IncModifications(currentSession);
    }
    else {
      DBG_DEBUG(AQDB_LOGDOMAIN,
                "Not writing changelog entry for type \"%s\" (NOCHGLOG)",
                AQDB_Table_GetName(dbt));
    }
  }

  return 0;
}



int _writeChangeLog(AQDB_IOLAYER *ioLayer, AQDB_ID tableId, AQDB_ID objectId,
                    uint64_t oldOffset, uint64_t newOffset,
                    AQDB_DBFILE_CHANGELOG_TYPE chgType)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_DB *database;
  AQDB_DBFILE_CHANGELOG *chg;
  AQDB_OBJECT *o;
  AQDB_ID tableIdChangeLog;
  int rv;
  AQDB_TABLE *dbtChangeLog;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  database=AQDB_IoLayer_GetDatabase(ioLayer);

  tableIdChangeLog=xio->tableIdChangeLog;
  if (tableIdChangeLog==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "No id for ChangeLog table");
    return GWEN_ERROR_INTERNAL;
  }
  dbtChangeLog=AQDB_IoLayer_GetTableById(ioLayer, tableIdChangeLog);
  if (dbtChangeLog==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No ChangeLog table found");
    return GWEN_ERROR_INTERNAL;
  }

  /* prepare entry for changelog event */
  chg=AQDB_DbFile_Changelog_new();
  AQDB_DbFile_Changelog_SetSessionLogId(chg, AQDB_IoLayer_GetCurrentSessionId(ioLayer));
  AQDB_DbFile_Changelog_SetTableId(chg, tableId);
  AQDB_DbFile_Changelog_SetObjectId(chg, objectId);
  AQDB_DbFile_Changelog_SetOldOffset(chg, oldOffset);
  AQDB_DbFile_Changelog_SetNewOffset(chg, newOffset);
  AQDB_DbFile_Changelog_SetEntryType(chg, chgType);

  /* transform changelog entry to database object */
  o=NULL;
  rv=AQDB_DbFile_Changelog_CreateObject(database, tableIdChangeLog, 0, &o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_Changelog_free(chg);
    return rv;
  }

  rv=AQDB_DbFile_Changelog_WriteObject(chg, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    AQDB_DbFile_Changelog_free(chg);
    return rv;
  }

  /* append change log entry to database */
  rv=xio->addObjectFn(ioLayer, dbtChangeLog, o);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    AQDB_DbFile_Changelog_free(chg);
    return rv;
  }
  AQDB_Object_free(o);
  AQDB_DbFile_Changelog_free(chg);
  return 0;
}




