/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


/* included from "layer.c" */



int AQDB_IoLayerLogged_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;
  uint64_t currentSessionId;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Begin edit session");

  if (xio->beginEditFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  rv=xio->beginEditFn(ioLayer, editFlags, user);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  currentSessionId=AQDB_IoLayer_GetCurrentSessionId(ioLayer);

  /* set username for session logs */
  AQDB_IoLayer_SetCurrentUserName(ioLayer, user);
  xio->editFlags=editFlags;

  /* get session id */
  if (editFlags & AQDB_DBFILE_SESSION_FLAGS_ROLLBACK) {
    /* dont increment session id when in rollback mode (reverse incrementing done in base layer) */
    DBG_DEBUG(AQDB_LOGDOMAIN, "BeginEdit in rollback mode");
    if (currentSessionId>0) {
      currentSessionId--;
      AQDB_IoLayer_SetCurrentSessionId(ioLayer, currentSessionId);
    }
  }

  /* create session */
  if (1) {
    AQDB_SESSION_DESCR *sess;
    GWEN_TIME *ti;

    ti=GWEN_CurrentTime();
    sess=AQDB_SessionDescr_new();
    AQDB_SessionDescr_SetSessionId(sess, currentSessionId);
    AQDB_SessionDescr_SetUserName(sess, user);
    AQDB_SessionDescr_SetDateTime(sess, ti);
    GWEN_Time_free(ti);

    AQDB_IoLayerLogged_SetCurrentSession(ioLayer, sess);
  }

  return 0;
}



int AQDB_IoLayerLogged_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Ending edit session%s", (flags & AQDB_ACTION_FLAGS_ABORT)?" (ABANDON)":"");

  if (xio->endEditFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  if (!(flags & AQDB_ACTION_FLAGS_ABORT)) {
    /* close session */
    rv=_closeSession(ioLayer);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  rv=xio->endEditFn(ioLayer, flags);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerLogged_UndoSession(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId)
{
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Undo session");

  rv=AQDB_IoLayerLogged_BeginEdit(ioLayer, AQDB_DBFILE_SESSION_FLAGS_ROLLBACK, NULL);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  DBG_DEBUG(AQDB_LOGDOMAIN, "Really undoing session");
  rv=_undoSessionsBackTo(ioLayer, sessionId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_IoLayerLogged_EndEdit(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  /* end edit */
  rv=AQDB_IoLayerLogged_EndEdit(ioLayer, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_IoLayerLogged_EndEdit(ioLayer, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerLogged_GetRollbackSessions(AQDB_IOLAYER *ioLayer, uint32_t aflags,
                                           AQDB_SESSION_DESCR_LIST *sl)
{
  AQDB_SESSION_DESCR_LIST *localSessionDescrList;

  localSessionDescrList=_getSessionDescriptionsBackTo(ioLayer, aflags, 0);
  if (localSessionDescrList) {
    AQDB_SessionDescr_List_AddList(sl, localSessionDescrList);
    AQDB_SessionDescr_List_free(localSessionDescrList);
    return 0;
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No session description found");
    return 0;
  }
}



int _closeSession(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;
  int rv;
  AQDB_SESSION_DESCR *currentSession;
  uint32_t editFlags;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing session");

  if (xio->addObjectFn==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Some virtual functions are not set in base layer");
    return GWEN_ERROR_GENERIC;
  }

  /* save session */
  editFlags=xio->editFlags;
  currentSession=AQDB_IoLayerLogged_GetCurrentSession(ioLayer);
  if (currentSession) {
    if (editFlags & AQDB_DBFILE_SESSION_FLAGS_ROLLBACK) {
      DBG_INFO(AQDB_LOGDOMAIN, "Rollback session, not saving session");
    }
    else {
      if (AQDB_SessionDescr_GetAdditions(currentSession) ||
          AQDB_SessionDescr_GetDeletions(currentSession) ||
          AQDB_SessionDescr_GetModifications(currentSession)) {
        AQDB_DB *database;
        AQDB_TABLE *dbt;

        database=AQDB_IoLayer_GetDatabase(ioLayer);
        dbt=AQDB_IoLayer_GetTableByName(ioLayer, AQDB_DBFILE_TYPENAME_SESSIONS);
        if (dbt) {
          AQDB_OBJECT *o=NULL;

          rv=AQDB_SessionDescr_CreateObject(database, AQDB_Table_GetTableId(dbt), 0, &o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            return rv;
          }

          rv=AQDB_SessionDescr_toObject(currentSession, o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            AQDB_Object_free(o);
            return rv;
          }

          rv=xio->addObjectFn(ioLayer, dbt, o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            AQDB_Object_free(o);
            return rv;
          }
          AQDB_Object_free(o);
        }
      }
      else {
        DBG_DEBUG(AQDB_LOGDOMAIN, "No additions, deletions or modifications, not saving session");
      }
    } /* if not in rollback mode */
  } /* if session */
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "No current session");
    return GWEN_ERROR_INTERNAL;
  }

  AQDB_IoLayerLogged_SetCurrentSession(ioLayer, NULL);
  return 0;
}



int _undoSessionsBackTo(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_SESSION_DESCR_LIST *sl;
  int rv;
  AQDB_TABLE *dbt;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Undoing all sessions back to %llu", (unsigned long long) sessionId);

  /* get table info of the session descriptor table */
  dbt=AQDB_IoLayer_GetTableById(ioLayer, xio->tableIdSessionDescr);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Session description table not found, SNH!");
    return GWEN_ERROR_INTERNAL;
  }

  /* read all rollback sessions */
  sl=_getSessionDescriptionsBackTo(ioLayer, AQDB_ACCESS_FLAGS_ISWRITER, sessionId);
  if (sl) {
    AQDB_SESSION_DESCR *sd;

    /* sort by session id descending, so the latest session becomes the first one */
    AQDB_SessionDescr_List_SortBySessionId(sl, 0);

    DBG_DEBUG(AQDB_LOGDOMAIN, "Undoing %d sessions", AQDB_SessionDescr_List_GetCount(sl));
    sd=AQDB_SessionDescr_List_First(sl);
    while (sd) {
      assert(AQDB_SessionDescr_GetDbId(sd));

      DBG_DEBUG(AQDB_LOGDOMAIN, "Undoing session %llu", (unsigned long long) AQDB_SessionDescr_GetSessionId(sd));

      /* undo this session */
      rv=_undoSingleSession(ioLayer, sd);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_SessionDescr_List_free(sl);
        return rv;
      }

      /* delete session entry from session descriptor list */
      AQDB_IoLayer_SetObjectPos(ioLayer, dbt, AQDB_SessionDescr_GetDbId(sd)-1, 0xffffffffffffffffL);

      sd=AQDB_SessionDescr_List_Next(sd);
    } /* while */
    AQDB_SessionDescr_List_free(sl);
  }
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "No session found >= %llu", (unsigned long long) sessionId);
  }

  return 0;
}



AQDB_SESSION_DESCR_LIST *_getSessionDescriptionsBackTo(AQDB_IOLAYER *ioLayer, uint32_t aflags,
                                                       AQDB_ID minimumSessionId)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_OBJECT_LIST *ol;
  char exprbuf[64];
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  snprintf(exprbuf, sizeof(exprbuf)-1, "$sessionId>=\"%llu\"",
           (long long unsigned int) minimumSessionId);
  exprbuf[sizeof(exprbuf)-1]=0;

  ol=AQDB_IoLayerLogged_ReadMatchingObjects(ioLayer,
                                            xio->tableIdSessionDescr,
                                            aflags, /* aflags */
                                            minimumSessionId?exprbuf:NULL, /* expression */
                                            0 /* query flags */);
  if (ol==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No session found >= %llu",
             (unsigned long long) minimumSessionId);
    return NULL;
  }
  else {
    AQDB_SESSION_DESCR_LIST *sessionDescrList;
    AQDB_OBJECT *o;

    sessionDescrList=AQDB_SessionDescr_List_new();
    o=AQDB_Object_List_First(ol);
    while (o) {
      AQDB_SESSION_DESCR *sd=NULL;

      /* create change log entry */
      rv=AQDB_SessionDescr_fromObject(o, &sd);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_Object_List_free(ol);
        AQDB_SessionDescr_List_free(sessionDescrList);
        return NULL;
      }
      AQDB_SessionDescr_SetDbId(sd, AQDB_Object_GetObjectId(o));

      AQDB_SessionDescr_List_Add(sd, sessionDescrList);
      o=AQDB_Object_List_Next(o);
    }

    if (AQDB_SessionDescr_List_GetCount(sessionDescrList)>0)
      return sessionDescrList;
    else {
      AQDB_SessionDescr_List_free(sessionDescrList);
      return NULL;
    }
    AQDB_Object_List_free(ol);
  }

  return NULL;
}



AQDB_DBFILE_CHANGELOG_LIST *_getChangelogsForSession(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_OBJECT_LIST *ol;
  char exprbuf[64];
  int rv;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  snprintf(exprbuf, sizeof(exprbuf)-1, "$sessionLogId==\"%llu\"",
           (long long unsigned int) sessionId);
  exprbuf[sizeof(exprbuf)-1]=0;

  ol=AQDB_IoLayerLogged_ReadMatchingObjects(ioLayer,
                                            xio->tableIdChangeLog,
                                            AQDB_ACCESS_FLAGS_ISWRITER, /* aflags */
                                            exprbuf, /* expression */
                                            0 /* query flags */);
  if (ol==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No matching objects");
    return NULL;
  }

  if (AQDB_Object_List_GetCount(ol)) {
    AQDB_DBFILE_CHANGELOG_LIST *changeLogList;
    AQDB_OBJECT *o;

    changeLogList=AQDB_DbFile_Changelog_List_new();
    o=AQDB_Object_List_First(ol);
    while (o) {
      AQDB_DBFILE_CHANGELOG *c=NULL;

      /* create change log entry */
      rv=AQDB_DbFile_Changelog_fromObject(o, &c);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_Object_List_free(ol);
        AQDB_DbFile_Changelog_List_free(changeLogList);
        return NULL;
      }
      AQDB_DbFile_Changelog_SetDbId(c, AQDB_Object_GetObjectId(o));

      AQDB_DbFile_Changelog_List_Add(c, changeLogList);
      o=AQDB_Object_List_Next(o);
    }
    AQDB_Object_List_free(ol);
    if (AQDB_DbFile_Changelog_List_GetCount(changeLogList)>0)
      return changeLogList;
    else {
      AQDB_DbFile_Changelog_List_free(changeLogList);
      return NULL;
    }
  }

  AQDB_Object_List_free(ol);
  return NULL;
}



int _undoSingleSession(AQDB_IOLAYER *ioLayer, const AQDB_SESSION_DESCR *sd)
{
  AQDB_IOLAYER_LOGGED *xio;
  AQDB_ID sessionId;
  AQDB_DBFILE_CHANGELOG_LIST *changeLogList;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  sessionId=AQDB_SessionDescr_GetSessionId(sd);
  DBG_INFO(AQDB_LOGDOMAIN,
           "Undoing session %llu with %llu additions, %llu deletions and %llu modifications",
           (unsigned long long) sessionId,
           (unsigned long long) AQDB_SessionDescr_GetAdditions(sd),
           (unsigned long long) AQDB_SessionDescr_GetDeletions(sd),
           (unsigned long long) AQDB_SessionDescr_GetModifications(sd));

  changeLogList=_getChangelogsForSession(ioLayer, sessionId);
  if (changeLogList) {
    AQDB_TABLE *dbtChangelog;
    AQDB_DBFILE_CHANGELOG *c;

    /* get table info of the session descriptor table */
    dbtChangelog=AQDB_IoLayer_GetTableById(ioLayer, xio->tableIdChangeLog);
    if (dbtChangelog==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "Changelog table not found, SNH!");
      return GWEN_ERROR_INTERNAL;
    }

    /* Sort backwards by database id */
    AQDB_DbFile_Changelog_List_SortByDbId(changeLogList, 0);
    c=AQDB_DbFile_Changelog_List_First(changeLogList);
    while (c) {
      AQDB_DBFILE_CHANGELOG *cNext;

      cNext=AQDB_DbFile_Changelog_List_Next(c);
      if (AQDB_DbFile_Changelog_GetSessionLogId(c)==sessionId) {
        int rv;

        rv=_reverseChangelogEntry(ioLayer, c);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          return rv;
        }

        /* delete changelog entry from table */
        if (AQDB_DbFile_Changelog_GetDbId(c)) {
          /* setting this value to the table entry has two effects:
           * - the entry will not be reused when new entries are added
           * - this entry will not be returned by query functions.
           */
          AQDB_IoLayer_SetObjectPos(ioLayer, dbtChangelog, AQDB_DbFile_Changelog_GetDbId(c)-1, 0xffffffffffffffffL);
        }

        /* delete changeLog from List */
        AQDB_DbFile_Changelog_List_Del(c);
        AQDB_DbFile_Changelog_free(c);
      } /* if (AQDB_DbFile_Changelog_GetSessionLogId(c)==sessionId) */
      c=cNext;
    } /* while */
    AQDB_DbFile_Changelog_List_free(changeLogList);
  } /* if changeLogList */

  return 0;
}



int _reverseChangelogEntry(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_CHANGELOG *c)
{
  AQDB_ID tableId;
  AQDB_ID objectId;
  AQDB_ID oldOffset;

  tableId=AQDB_DbFile_Changelog_GetTableId(c);
  objectId=AQDB_DbFile_Changelog_GetObjectId(c);
  oldOffset=AQDB_DbFile_Changelog_GetOldOffset(c);

  if (tableId && objectId) {
    AQDB_TABLE *dbt;

    dbt=AQDB_IoLayer_GetTableById(ioLayer, tableId);
    if (dbt==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "Table %llu not found", (unsigned long long int) tableId);
      return GWEN_ERROR_INTERNAL;
    }
    if (AQDB_Table_GetOpenFlags(dbt)==0) {
      AQDB_ID tmpTableId=0;

      DBG_DEBUG(AQDB_LOGDOMAIN, "Opening table %llu on the fly", (unsigned long long int) tableId);
      if (NULL==AQDB_IoLayer_OpenTable(ioLayer,
                                              AQDB_ACCESS_FLAGS_ISWRITER,
                                              AQDB_Table_GetName(dbt),
                                              AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                                              &tmpTableId)) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Error opening table %llu", (unsigned long long int) tableId);
      }
    }
    else {
      DBG_DEBUG(AQDB_LOGDOMAIN, "Table %llu is already open", (unsigned long long int) tableId);
    }

    switch (AQDB_DbFile_Changelog_GetEntryType(c)) {
    case AQDB_DbFile_ChangelogType_Unknown:
    case AQDB_DbFile_ChangelogType_None:
      DBG_ERROR(AQDB_LOGDOMAIN, "Unhandled changeLog type (%d)", AQDB_DbFile_Changelog_GetEntryType(c));
      break;
    case AQDB_DbFile_ChangelogType_Addition:
      /* object was added, delete the entry */
      AQDB_IoLayer_SetObjectPos(ioLayer, dbt, objectId-1, 0xffffffffffffffffL);
      break;
    case AQDB_DbFile_ChangelogType_Deletion:
      /* object was deleted, add it back again */
      AQDB_IoLayer_SetObjectPos(ioLayer, dbt, objectId-1, oldOffset);
      break;
    case AQDB_DbFile_ChangelogType_Modification:
      /* object was modified, set old entry */
      AQDB_IoLayer_SetObjectPos(ioLayer, dbt, objectId-1, oldOffset);
      break;
    }
  } /* if (tableId && objectId) */

  return 0;
}




