/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2021 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "file/iologged/layer_p.h"

#include "file/dbfile_l.h"
#include "file/changelog.h"

#include <aqdatabase/aqdb_query.h>
#include <aqdatabase/backendsupport/aqdb_objectquery.h>
#include <aqdatabase/backendsupport/aqdb_query_result.h>

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED)



#define AQDB_DBFILE_TYPENAME_SESSIONS  "+SessionDescr"
#define AQDB_DBFILE_TYPENAME_CHANGELOG "+ChangeLog"



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB AQDB_IoLayerLogged_FreeData(void *bp, void *p);

static int AQDB_IoLayerLogged_OpenDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags);
static int AQDB_IoLayerLogged_CreateDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags);
static int AQDB_IoLayerLogged_CloseDatabase(AQDB_IOLAYER *ioLayer, uint32_t flags);

int AQDB_IoLayerLogged_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user);
int AQDB_IoLayerLogged_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags);
int AQDB_IoLayerLogged_GetRollbackSessions(AQDB_IOLAYER *ioLayer, uint32_t aflags,
                                           AQDB_SESSION_DESCR_LIST *sl);
int AQDB_IoLayerLogged_UndoSession(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId);

int AQDB_IoLayerLogged_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object);
int AQDB_IoLayerLogged_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId);
int AQDB_IoLayerLogged_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, const AQDB_OBJECT *object);
AQDB_OBJECT_LIST *AQDB_IoLayerLogged_ReadMatchingObjects(AQDB_IOLAYER *ioLayer,
                                                         AQDB_ID tableId,
                                                         uint32_t accessFlags,
                                                         const char *queryExpression,
                                                         uint32_t queryFlags);

static AQDB_OBJECT_LIST *_readMatchingObjects(AQDB_IOLAYER *ioLayer,
                                              AQDB_TABLE *databaseTypeTable,
                                              GWEN_IDLIST64 *sourceIdList,
                                              uint32_t accessFlags,
                                              const char *queryExpression,
                                              uint32_t queryFlags);
static AQDB_OBJECT_LIST *_readAllObjects(AQDB_IOLAYER *ioLayer,
                                         AQDB_TABLE *databaseTypeTable,
                                         GWEN_IDLIST64 *sourceIdList,
                                         uint32_t accessFlags,
                                         uint32_t queryFlags);


static int _createSessionTypes(AQDB_IOLAYER *ioLayer);

static int _writeChangeLog(AQDB_IOLAYER *ioLayer,
                           AQDB_ID tableId, AQDB_ID objectId,
                           uint64_t oldOffset, uint64_t newOffset,
                           AQDB_DBFILE_CHANGELOG_TYPE chgType);

static int _undoSessionsBackTo(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId);
static int _undoSingleSession(AQDB_IOLAYER *ioLayer, const AQDB_SESSION_DESCR *sd);
static int _closeSession(AQDB_IOLAYER *ioLayer);
static AQDB_SESSION_DESCR_LIST *_getSessionDescriptionsBackTo(AQDB_IOLAYER *ioLayer, uint32_t aflags,
                                                              AQDB_ID minimumSessionId);
static AQDB_DBFILE_CHANGELOG_LIST *_getChangelogsForSession(AQDB_IOLAYER *ioLayer, AQDB_ID sessionId);
static int _reverseChangelogEntry(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_CHANGELOG *c);






/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerLogged_Extend(AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;

  AQDB_IoLayer_SetName(ioLayer, "logged");

  GWEN_NEW_OBJECT(AQDB_IOLAYER_LOGGED, xio);
  if (xio==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer, xio, AQDB_IoLayerLogged_FreeData);

  xio->createFn=AQDB_IoLayer_SetCreateFn(ioLayer, AQDB_IoLayerLogged_CreateDatabase);
  xio->openFn=AQDB_IoLayer_SetOpenFn(ioLayer, AQDB_IoLayerLogged_OpenDatabase);
  xio->closeFn=AQDB_IoLayer_SetCloseFn(ioLayer, AQDB_IoLayerLogged_CloseDatabase);

  xio->beginEditFn=AQDB_IoLayer_SetBeginEditFn(ioLayer, AQDB_IoLayerLogged_BeginEdit);
  xio->endEditFn=AQDB_IoLayer_SetEndEditFn(ioLayer, AQDB_IoLayerLogged_EndEdit);

  xio->createTableFn=AQDB_IoLayer_GetCreateTableFn(ioLayer);
  xio->openTableFn=AQDB_IoLayer_GetOpenTableFn(ioLayer);
  xio->closeTableFn=AQDB_IoLayer_GetCloseTableFn(ioLayer);

  xio->getObjectPosFn=AQDB_IoLayer_GetGetObjectPosFn(ioLayer);
  xio->setObjectPosFn=AQDB_IoLayer_GetSetObjectPosFn(ioLayer);

  xio->addObjectFn=AQDB_IoLayer_SetAddObjectFn(ioLayer, AQDB_IoLayerLogged_AddObject);
  xio->deleteObjectFn=AQDB_IoLayer_SetDeleteObjectFn(ioLayer, AQDB_IoLayerLogged_DeleteObject);
  xio->writeObjectFn=AQDB_IoLayer_SetWriteObjectFn(ioLayer, AQDB_IoLayerLogged_WriteObject);
  xio->readObjectFn=AQDB_IoLayer_GetReadObjectFn(ioLayer);

  xio->querySubmitFn=AQDB_IoLayer_GetQuerySubmitFn(ioLayer);
  xio->queryGetFirstFn=AQDB_IoLayer_GetQueryGetFirstFn(ioLayer);
  xio->queryGetNextFn=AQDB_IoLayer_GetQueryGetNextFn(ioLayer);
  xio->queryCloseFn=AQDB_IoLayer_GetQueryCloseFn(ioLayer);

  xio->getActiveIdsForTableFn=AQDB_IoLayer_GetGetActiveIdsForTableFn(ioLayer);

  xio->getRollbackSessionsFn=AQDB_IoLayer_SetGetRollbackSessionsFn(ioLayer,
                                                                          AQDB_IoLayerLogged_GetRollbackSessions);
  xio->undoSessionFn=AQDB_IoLayer_SetUndoSessionFn(ioLayer, AQDB_IoLayerLogged_UndoSession);

  return 0;
}



void AQDB_IoLayerLogged_FreeData(void *bp, void *p)
{
  AQDB_IOLAYER_LOGGED *xio;

  xio=(AQDB_IOLAYER_LOGGED *) p;
  assert(xio);


  GWEN_FREE_OBJECT(xio);
}



AQDB_ID AQDB_IoLayerLogged_GetTableIdSessionDescr(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  return xio->tableIdSessionDescr;
}



void AQDB_IoLayerLogged_SetTableIdSessionDescr(AQDB_IOLAYER *ioLayer, AQDB_ID id)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  xio->tableIdSessionDescr=id;
}



AQDB_ID AQDB_IoLayerLogged_GetTableIdChangeLog(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  return xio->tableIdChangeLog;
}



void AQDB_IoLayerLogged_SetTableIdChangeLog(AQDB_IOLAYER *ioLayer, AQDB_ID id)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  xio->tableIdChangeLog=id;
}



AQDB_SESSION_DESCR *AQDB_IoLayerLogged_GetCurrentSession(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  return xio->currentSession;
}



void AQDB_IoLayerLogged_SetCurrentSession(AQDB_IOLAYER *ioLayer, AQDB_SESSION_DESCR *sdescr)
{
  AQDB_IOLAYER_LOGGED *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOGGED, ioLayer);
  assert(xio);

  if (xio->currentSession)
    AQDB_SessionDescr_free(xio->currentSession);
  xio->currentSession=sdescr;
}





#include "io_db.c"
#include "io_object.c"
#include "io_session.c"








