/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "file/iolowlevel/io_db.h"

#include "file/iolowlevel/io_object.h"
#include "file/iolowlevel/io_table.h"
#include "file/iolowlevel/layer.h"
#include "file/header_l.h"
#include "file/dbfile_l.h"

#include <aqdatabase/aqdb_object_be.h>

#include <aqfilestorage/aqfs_check.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

int _readDatabaseStart(AQDB_IOLAYER *ioLayer);
int _readHeader(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_HEADER *dbh);
int _readTypes(AQDB_IOLAYER *ioLayer);
int _readTypesFromTableList(AQDB_IOLAYER *ioLayer, const AQFS_TABLE_LIST *typeTableList,
                            AQDB_TABLE_LIST *dbtl);

int _createDatabaseStart(AQDB_IOLAYER *ioLayer);

int _createAndWriteStartTable(AQDB_IOLAYER *ioLayer);
int _createAndWriteDatabaseHeader(AQDB_IOLAYER *ioLayer);
int _createAndWriteTypeTable(AQDB_IOLAYER *ioLayer);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerLowlevel_Open(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  const char *dataFileName;
  AQFS_FILE *fsf;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Opening database");

  dataFileName=AQDB_IoLayerLowlevel_GetDataFileName(ioLayer);

  if ((flags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE))==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No open flags given (%d)", flags);
    return GWEN_ERROR_INVALID;
  }

  /* open file */
  fsf=AQFS_File_new(dataFileName);
  rv=AQFS_File_Open(fsf, (flags & AQDB_ACTION_FLAGS_WRITE)?1:0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_File_free(fsf);
    return rv;
  }
  AQDB_IoLayerLowlevel_SetFile(ioLayer, fsf);

  /* read database headers and stuff */
  rv=_readDatabaseStart(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    //AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  /* read types */
  rv=_readTypes(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    //AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  /* database is open */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Database is now open");

  return rv;
}



int AQDB_IoLayerLowlevel_Create(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  const char *dataFileName;
  AQFS_FILE *fsf;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Creating database");

  dataFileName=AQDB_IoLayerLowlevel_GetDataFileName(ioLayer);

  /* create file */
  fsf=AQFS_File_new(dataFileName);
  rv=AQFS_File_Create(fsf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_File_free(fsf);
    return rv;
  }
  AQDB_IoLayerLowlevel_SetFile(ioLayer, fsf);

  rv=_createDatabaseStart(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    //AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing file");
  rv=AQFS_File_Close(fsf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d), ignoring", rv);
    //AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }
  AQDB_IoLayerLowlevel_SetFile(ioLayer, NULL);

  //AQDB_DbFile_ReleaseInternalStructures(db);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Database created");
  return 0;
}



int AQDB_IoLayerLowlevel_Close(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  int rv;
  AQFS_FILE *fsf;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing database");

  fsf=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* close file */
  if (fsf) {
    DBG_DEBUG(AQDB_LOGDOMAIN, "Closing file");
    rv=AQFS_File_Close(fsf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d), ignoring", rv);
    }
  }
  AQDB_IoLayerLowlevel_SetFile(ioLayer, NULL);

  AQDB_IoLayer_SetDbTypeTables(ioLayer, NULL);
  AQDB_IoLayerLowlevel_SetTableListDatabase(ioLayer, NULL);
  AQDB_IoLayerLowlevel_SetTableListTypes(ioLayer, NULL);
  AQDB_IoLayerLowlevel_SetHeader(ioLayer, NULL);

  //AQDB_DbFile_ReleaseInternalStructures(db);

  return 0;
}


int _createDatabaseStart(AQDB_IOLAYER *ioLayer)
{
  int rv;

  rv=_createAndWriteStartTable(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=_createAndWriteDatabaseHeader(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=_createAndWriteTypeTable(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  rv=AQDB_IoLayerLowlevel_FlushTypeData(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}





int _readDatabaseStart(AQDB_IOLAYER *ioLayer)
{
  AQFS_FILE *fsFile;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* create startTable (table containing database info and list of types) */
  if (1) {
    AQFS_TABLE_LIST *ftl;
    uint64_t offset;
    int rv;

    /* read first table */
    ftl=AQFS_Table_List_new();
    offset=AQFS_File_GetFirstItemOffset(fsFile);
    rv=AQFS_File_ReadTableList(fsFile, offset, ftl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQFS_Table_List_free(ftl);
      return rv;
    }

    AQDB_IoLayerLowlevel_SetTableListDatabase(ioLayer, ftl);
  }

  /* read database header */
  if (1) {
    AQDB_DBFILE_HEADER *dbh;
    int rv;

    dbh=AQDB_DbFile_Header_new();
    rv=_readHeader(ioLayer, dbh);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Header_free(dbh);
      return rv;
    }
    AQDB_IoLayer_SetCurrentSessionId(ioLayer, AQDB_DbFile_Header_GetLastSessionId(dbh));
    AQDB_IoLayerLowlevel_SetHeader(ioLayer, dbh);
  }

  /* read type table (table containing list of types) */
  if (1) {
    AQFS_TABLE_LIST *tableListDatabase;
    AQFS_TABLE_LIST *ftl;
    uint64_t offset;
    int rv;

    tableListDatabase=AQDB_IoLayerLowlevel_GetTableListDatabase(ioLayer);

    /* read type list table back */
    ftl=AQFS_Table_List_new();
    offset=AQFS_File_GetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_TYPES);
    rv=AQFS_File_ReadTableList(fsFile, offset, ftl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQFS_Table_List_free(ftl);
      return rv;
    }

    AQDB_IoLayerLowlevel_SetTableListTypes(ioLayer, ftl);
    AQDB_IoLayer_SetDbTypeTables(ioLayer, AQDB_Table_List_new());
  }

  return 0;
}




int _readHeader(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_HEADER *dbh)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE_LIST *tableListDatabase;
  uint8_t *ptr=NULL;
  uint64_t offset;
  int64_t rd;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Reading header");

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  tableListDatabase=AQDB_IoLayerLowlevel_GetTableListDatabase(ioLayer);

  offset=AQFS_File_GetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_HEADER);
  if (offset==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No header in file");
    return GWEN_ERROR_BAD_DATA;
  }

  rd=AQFS_File_ReadDataAt(fsFile, offset, &ptr, AQDB_DBFILE_HEADER_SIZE);
  if (rd<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  else if (rd==0) {
    DBG_INFO(AQFS_LOGDOMAIN, "Empty data");
    return GWEN_ERROR_BAD_DATA;
  }

  /* now ptr contains a pointer to the header data */
  rv=AQDB_DbFile_Header_ReadFromBuffer(dbh, ptr, rd);
  if (rv<0) {
    DBG_INFO(AQFS_LOGDOMAIN, "here (%d)", rv);
    free(ptr);
    return rv;
  }
  free(ptr);

  /* done */
  return 0;
}





int _readTypes(AQDB_IOLAYER *ioLayer)
{
  AQFS_TABLE_LIST *tableListTypes;
  AQDB_TABLE_LIST *dbtl;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Reading database types");

  tableListTypes=AQDB_IoLayerLowlevel_GetTableListTypes(ioLayer);

  dbtl=AQDB_Table_List_new();

  rv=_readTypesFromTableList(ioLayer, tableListTypes, dbtl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Table_List_free(dbtl);
    return rv;
  }

  AQDB_IoLayer_SetDbTypeTables(ioLayer, dbtl);
  return 0;
}



int _readTypesFromTableList(AQDB_IOLAYER *ioLayer, const AQFS_TABLE_LIST *typeTableList,
                            AQDB_TABLE_LIST *dbtl)
{
  AQFS_TABLE *ft;
  uint64_t currentTableId=0;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Reading types");

  /* read every entry from the type table */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Table list has %d entries", AQFS_Table_List_GetCount(typeTableList));
  ft=AQFS_Table_List_First(typeTableList);
  while (ft) {
    uint64_t maxEntries;
    uint64_t idx;
    uint64_t *ptr;

    maxEntries=AQFS_Table_GetMaxEntries(ft);
    DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading entries from table (maxEntries=%d, free entries=%d)",
                (int) maxEntries,
                (int) AQFS_Table_GetFreeEntries(ft));
    ptr=AQFS_Table_GetPtrEntries(ft);
    for (idx=0; idx<maxEntries; idx++) {
      uint64_t offset;

      offset=*(ptr++);
      if (offset>0) {
        AQDB_TABLE *dbt;
        int rv;
        const char *sName;

        DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading type entry at %llu",
                    (unsigned long long int) offset);
        dbt=AQDB_IoLayerLowlevel_Table_new();
        rv=AQDB_IoLayerLowlevel_ReadTableAt(ioLayer, dbt, offset);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          return rv;
        }
        sName=AQDB_Table_GetName(dbt);
        if (sName && *sName) {
          DBG_VERBOUS(AQDB_LOGDOMAIN, "Type at %llu is [%s]",
                      (unsigned long long int) offset,
                      sName);
          if (*sName=='-') {
            /* name starts with a hyphen, don't save changelogs when
             * adding, removing or modifying objects of this type */
            AQDB_Table_AddFlags(dbt, AQDB_DBFILE_TABLE_FLAGS_NOCHGLOG);
          }
        }
        AQDB_Table_SetTableId(dbt, currentTableId+idx+1);
        AQDB_Table_SubOpenFlags(dbt, AQDB_ACTION_FLAGS_MASK);
        DBG_DEBUG(AQDB_LOGDOMAIN, "Adding type \"%s\" with id %llu",
                  AQDB_Table_GetName(dbt),
                  (long long unsigned)(currentTableId+idx+1));
        AQDB_Table_List_Add(dbt, dbtl);
      }
    }

    currentTableId+=maxEntries;
    ft=AQFS_Table_List_Next(ft);
  }

  return 0;
}



int AQDB_IoLayerLowlevel_WriteHeader(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_HEADER *dbh)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE_LIST *tableListDatabase;
  uint8_t *ptr;
  uint64_t len;
  uint64_t offset;
  int64_t rd;
  int rv;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  tableListDatabase=AQDB_IoLayerLowlevel_GetTableListDatabase(ioLayer);

  DBG_DEBUG(AQDB_LOGDOMAIN, "Writing header");
  ptr=(uint8_t *) malloc(AQDB_DBFILE_HEADER_SIZE);
  assert(ptr);
  rd=AQDB_DbFile_Header_toBuffer(dbh, ptr, AQDB_DBFILE_HEADER_SIZE);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    free(ptr);
    return (int) rd;
  }
  len=rd;

  offset=AQFS_File_GetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_HEADER);
  if (offset==0) {
    DBG_DEBUG(AQDB_LOGDOMAIN, " Creating new header in file");
    /* not yet set, append data */
    rd=AQFS_File_AppendData(fsFile, AQFS_ITEMHEADER_TYPE_GENERIC_DATA, ptr, len);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      free(ptr);
      return (int) rd;
    }
    offset=rd;

    /* set entry, write table to file */
    AQFS_File_SetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_HEADER, offset);
    rv=AQFS_File_WriteDirtyTables(fsFile, tableListDatabase);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    DBG_DEBUG(AQDB_LOGDOMAIN, " Overwriting existing header in file");
    /* already set, write data */
    rd=AQFS_File_WriteDataAt(fsFile, offset, AQFS_ITEMHEADER_TYPE_GENERIC_DATA, ptr, len);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      free(ptr);
      return (int) rd;
    }
  }
  free(ptr);

  return 0;
}



int AQDB_IoLayerLowlevel_FlushTypeData(AQDB_IOLAYER *ioLayer)
{
  AQFS_FILE *fsFile;
  AQDB_TABLE_LIST *dbTypeTables;
  AQFS_TABLE_LIST *tableListTypes;
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Flushing type data");

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  tableListTypes=AQDB_IoLayerLowlevel_GetTableListTypes(ioLayer);
  dbTypeTables=AQDB_IoLayer_GetDbTypeTables(ioLayer);

  /* let all types write their column lists */
  if (dbTypeTables) {
    AQDB_TABLE *dbt;

    dbt=AQDB_Table_List_First(dbTypeTables);
    while (dbt) {
      AQFS_TABLE_LIST *ftl;

      DBG_DEBUG(AQDB_LOGDOMAIN, "- Flushing type \"%s\"", AQDB_Table_GetName(dbt));

      ftl=AQDB_IoLayerLowlevel_Table_GetTableListColumns(dbt);
      if (ftl) {
        DBG_DEBUG(AQDB_LOGDOMAIN, "  - Writing dirty column tables");
        rv=AQFS_File_WriteDirtyTables(fsFile, ftl);
        if (rv<0) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
          return rv;
        }
      }

      dbt=AQDB_Table_List_Next(dbt);
    }
  }

  if (tableListTypes) {
    /* write type list */
    DBG_DEBUG(AQDB_LOGDOMAIN, "Writing type list");
    rv=AQFS_File_WriteDirtyTables(fsFile, tableListTypes);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}




int AQDB_IoLayerLowlevel_Check(AQDB_IOLAYER *ioLayer, GWEN_BUFFER *resultTextBuf)
{
  const char *dataFileName;
  int rv;
  uint64_t numOfTables=0;
  uint64_t numOfItems=0;
  int hasErrors=0;
  char txtBuf[2048];

  DBG_DEBUG(AQDB_LOGDOMAIN, "Checking database");

  txtBuf[0]=0;

  dataFileName=AQDB_IoLayerLowlevel_GetDataFileName(ioLayer);

  /* first check */
  rv=AQFS_CheckFileSize(dataFileName);
  if (rv) {
    snprintf(txtBuf, sizeof(txtBuf)-1, "Error in file on size check (%d)\n", rv);
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
    hasErrors++;
  }
  else {
    snprintf(txtBuf, sizeof(txtBuf)-1, "File size check succeeded\n");
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
  }

  /* next check */
  numOfTables=0;
  numOfItems=0;
  rv=AQFS_CheckFileSequentially(dataFileName, &numOfTables, &numOfItems);
  if (rv) {
    snprintf(txtBuf, sizeof(txtBuf)-1, "Error in file on sequential check (%d)\n", rv);
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
    hasErrors++;
  }
  else {
    snprintf(txtBuf, sizeof(txtBuf)-1,
             "Sequential check succeeded (read %llu tables, %llu items)\n",
             (unsigned long long int) numOfTables,
             (unsigned long long int) numOfItems);
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
  }

  /* next check */
  numOfTables=0;
  numOfItems=0;
  rv=AQFS_CheckFileStructure(dataFileName, &numOfTables, &numOfItems);
  if (rv) {
    snprintf(txtBuf, sizeof(txtBuf)-1, "Error in file on structure check (%d)\n", rv);
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
    hasErrors++;
  }
  else {
    snprintf(txtBuf, sizeof(txtBuf)-1,
             "Structure check succeeded (read %llu tables, %llu items)\n",
             (unsigned long long int) numOfTables,
             (unsigned long long int) numOfItems);
    txtBuf[sizeof(txtBuf)-1]=0;
    GWEN_Buffer_AppendString(resultTextBuf, txtBuf);
  }

  if (hasErrors)
    return GWEN_ERROR_BAD_DATA;
  return 0;
}




/* create startTable (table containing database info and list of types) */
int _createAndWriteStartTable(AQDB_IOLAYER *ioLayer)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE *ft;
  AQFS_TABLE_LIST *ftl;
  uint64_t offset;
  int rv;
  int64_t rd;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  ft=AQFS_Table_Create(AQDB_DBFILE_TYPETABLE_MAXENTRIES);

  /* append startTable as first table to database */
  rd=AQFS_File_AppendTable(fsFile, ft);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    AQFS_Table_free(ft);
    return (int) rd;
  }
  AQFS_Table_free(ft);

  /* read first table back */
  ftl=AQFS_Table_List_new();
  offset=AQFS_File_GetFirstItemOffset(fsFile);
  rv=AQFS_File_ReadTableList(fsFile, offset, ftl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_Table_List_free(ftl);
    return rv;
  }
  AQDB_IoLayerLowlevel_SetTableListDatabase(ioLayer, ftl);

  /* add entry for database header to the startTable */
  rd=AQFS_File_AddTableEntry(fsFile, ftl, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }

  /* add entry for type table to the startTable */
  rd=AQFS_File_AddTableEntry(fsFile, ftl, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }

  return 0;
}



int _createAndWriteDatabaseHeader(AQDB_IOLAYER *ioLayer)
{
  AQDB_DBFILE_HEADER *dbh;
  int rv;

  dbh=AQDB_DbFile_Header_new();
  AQDB_DbFile_Header_SetLastSessionId(dbh, AQDB_IoLayer_GetCurrentSessionId(ioLayer));
  rv=AQDB_IoLayerLowlevel_WriteHeader(ioLayer, dbh);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }
  AQDB_IoLayerLowlevel_SetHeader(ioLayer, dbh);

  return 0;
}



/* create type table (table containing list of types) */
int _createAndWriteTypeTable(AQDB_IOLAYER *ioLayer)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE_LIST *tableListDatabase;
  AQFS_TABLE *ft;
  AQFS_TABLE_LIST *ftl;
  uint64_t offset;
  int rv;
  int64_t rd;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  tableListDatabase=AQDB_IoLayerLowlevel_GetTableListDatabase(ioLayer);
  ft=AQFS_Table_Create(AQDB_DBFILE_TYPETABLE_MAXENTRIES);
  rd=AQFS_File_AppendTable(fsFile, ft);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    AQFS_Table_free(ft);
    return (int) rd;
  }
  offset=rd;
  AQFS_Table_free(ft);

  /* store offset of type table in startTable */
  AQFS_File_SetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_TYPES, offset);
  rv=AQFS_File_WriteDirtyTables(fsFile, tableListDatabase);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* read type list table back */
  ftl=AQFS_Table_List_new();
  offset=AQFS_File_GetTableEntry(fsFile, tableListDatabase, AQDB_DBFILE_DBDATA_INDEX_TYPES);
  rv=AQFS_File_ReadTableList(fsFile, offset, ftl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_Table_List_free(ftl);
    return rv;
  }

  AQDB_IoLayerLowlevel_SetTableListTypes(ioLayer, ftl);
  AQDB_IoLayer_SetDbTypeTables(ioLayer, AQDB_Table_List_new());

  return 0;
}



