/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "file/iolowlevel/io_object.h"

#include <aqdatabase/aqdb_object_be.h>


#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>




/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static int64_t _appendTypeObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *o);

static int _readObjectFromBuffer(AQDB_OBJECT *object, const uint8_t *pStart, uint32_t bsize);
static AQDB_OBJECT *_object_fromBuffer(AQDB_ID tableId, AQDB_ID objectId, uint32_t nfields, const uint8_t *pStart,
                                       uint32_t bsize);
static int _object_toBuffer(const AQDB_OBJECT *o, GWEN_BUFFER *buf);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



int AQDB_IoLayerLowlevel_AddObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *object)
{
  int64_t rd;

  rd=_appendTypeObject(ioLayer, dbt, object);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (error %d, table [%d])", (int) rd, (int) AQDB_Table_GetTableId(dbt));
    return (int) rd;
  }

  AQDB_Object_SetObjectId(object, rd+1);
  AQDB_Object_SetTableId(object, AQDB_Table_GetTableId(dbt));

  DBG_VERBOUS(AQDB_LOGDOMAIN, "Added object %llu to table [%s]",
              (unsigned long long)(rd+1), AQDB_Table_GetName(dbt));

  return 0;
}



int AQDB_IoLayerLowlevel_DeleteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_ID objectId)
{
  /* modify the data object entry to in the given table list */
  AQDB_IoLayerLowlevel_Table_SetObjectPos(dbt, objectId-1, 0xffffffffffffffffL);
  DBG_DEBUG(AQDB_LOGDOMAIN, "Deleted object %u", (unsigned int) objectId-1);

  return 0;
}



int AQDB_IoLayerLowlevel_WriteObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, const AQDB_OBJECT *object)
{
  AQFS_FILE *fsFile;
  uint64_t objectId;
  GWEN_BUFFER *tbuf;
  int rv;
  int64_t rd;
  uint64_t offset;
  uint64_t len;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }

  /* write object to a writable buffer */
  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=_object_toBuffer(object, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }
  len=(uint64_t) GWEN_Buffer_GetUsedBytes(tbuf);

  /* append data to file.
   * Objects are never overwritten. The new version of an object is just appended
   * to the data file and its entry in the object table is modified to reflect the
   * new position. As long as the object table list isn't written to the file the
   * database remains consistent.
   */
  rd=AQFS_File_AppendData(fsFile, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                          (const uint8_t *) GWEN_Buffer_GetStart(tbuf),
                          len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rd;
  }
  GWEN_Buffer_free(tbuf);
  offset=rd;

  /* modify the data object entry to in the given table list */
  AQDB_IoLayerLowlevel_Table_SetObjectPos(dbt, objectId-1, offset);
  DBG_VERBOUS(AQDB_LOGDOMAIN, "Written object %u at %llu (%llu bytes) [%s]",
              (unsigned int) objectId-1,
              (unsigned long long int) offset,
              (unsigned long long int) len,
              AQDB_Table_GetName(dbt));

  return 0;
}



int AQDB_IoLayerLowlevel_ReadObject(AQDB_IOLAYER *ioLayer,
                                    uint32_t aflags,
                                    AQDB_TABLE *dbt,
                                    AQDB_ID objectId,
                                    AQDB_OBJECT **pObject)
{
  AQFS_FILE *fsFile;
  int64_t rd;
  uint64_t offset;
  uint64_t len;
  uint8_t *ptr;
  AQDB_OBJECT *o;
  int isWriter;
  AQDB_ID tableId;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  isWriter=(aflags & AQDB_ACCESS_FLAGS_ISWRITER);
  tableId=AQDB_Table_GetTableId(dbt);

  DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading object %d from table %s (%d) as %s",
              (int) objectId, AQDB_Table_GetName(dbt), (int) tableId, isWriter?"WRITER":"READER");

  /* get offset of the given object inside the file */
  rd=AQDB_IoLayerLowlevel_Table_GetObjectPos(dbt, objectId-1, isWriter);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  offset=rd;

  /* read object data */
  rd=AQFS_File_ReadDataAt(fsFile, offset, &ptr, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  len=rd;

  /* create object from buffer */
  o=_object_fromBuffer(tableId, objectId, AQDB_Table_GetColumnCount(dbt), ptr, len);
  if (o==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    free(ptr);
    return GWEN_ERROR_BAD_DATA;
  }
  free(ptr);

  *pObject=o;
  return 0;
}



int AQDB_IoLayerLowlevel_ReadObjectAt(AQDB_IOLAYER *ioLayer,
                                        AQDB_OBJECT *object,
                                        uint64_t offset)
{
  AQFS_FILE *fsFile;
  int64_t rd;
  uint64_t len;
  uint8_t *ptr;
  int rv;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* read object data */
  rd=AQFS_File_ReadDataAt(fsFile, offset, &ptr, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  len=rd;

  rv=_readObjectFromBuffer(object, ptr, len);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    free(ptr);
    return rv;
  }
  free(ptr);

  return 0;
}




int64_t AQDB_IoLayerLowlevel_AppendObject(AQDB_IOLAYER *ioLayer, AQFS_TABLE_LIST *ftl, AQDB_OBJECT *o)
{
  AQFS_FILE *fsFile;
  GWEN_BUFFER *tbuf;
  int rv;
  int64_t rd;
  uint64_t offset;
  uint64_t len;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* write object to a writable buffer */
  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=_object_toBuffer(o, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }
  len=(uint64_t) GWEN_Buffer_GetUsedBytes(tbuf);

  /* append data to file */
  rd=AQFS_File_AppendData(fsFile, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                          (const uint8_t *) GWEN_Buffer_GetStart(tbuf),
                          len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rd;
  }
  GWEN_Buffer_free(tbuf);
  offset=rd;

  /* add an entry for the data object to the given table list */
  rd=AQFS_File_AddTableEntry(fsFile, ftl, offset);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rd;
  }
  DBG_VERBOUS(AQDB_LOGDOMAIN, "Added object %u at %llu (%llu bytes)",
              (unsigned int) rd,
              (unsigned long long int) offset,
              (unsigned long long int) len);

  /* return index of new object into table list */
  return rd;
}



int64_t _appendTypeObject(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQDB_OBJECT *o)
{
  AQFS_FILE *fsFile;
  GWEN_BUFFER *tbuf;
  int rv;
  int64_t rd;
  uint64_t offset;
  uint64_t len;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* write object to a writable buffer */
  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=_object_toBuffer(o, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }
  len=(uint64_t) GWEN_Buffer_GetUsedBytes(tbuf);

  /* append data to file */
  rd=AQFS_File_AppendData(fsFile, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                          (const uint8_t *) GWEN_Buffer_GetStart(tbuf),
                          len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rd;
  }
  GWEN_Buffer_free(tbuf);
  offset=rd;

  /* add an entry for the data object to the given table list */
  rd=AQDB_IoLayerLowlevel_Table_AddObjectPos(dbt, offset);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rd;
  }
  DBG_VERBOUS(AQDB_LOGDOMAIN, "Added type object %u at %llu (%llu bytes) [%s]",
              (unsigned int)(rd+1),
              (unsigned long long int) offset,
              (unsigned long long int) len,
              AQDB_Table_GetName(dbt));

  /* return index of new object into table list */
  return rd;
}



AQDB_OBJECT *_object_fromBuffer(AQDB_ID tableId, AQDB_ID objectId, uint32_t nfields, const uint8_t *pStart,
                                uint32_t bsize)
{
  AQDB_OBJECT *o;
  int rv;

  assert(nfields);
  o=AQDB_Object_new(tableId, objectId, nfields);
  assert(o);
  rv=_readObjectFromBuffer(o, pStart, bsize);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return NULL;
  }

  return o;
}



int _readObjectFromBuffer(AQDB_OBJECT *object, const uint8_t *pStart, uint32_t bsize)
{
  uint32_t j;
  uint32_t nfields;
  const uint8_t *p;
  const uint8_t *pEnd;

  assert(object);

  nfields=AQDB_Object_GetFieldCount(object);
  assert(nfields);

#if 0
  DBG_ERROR(AQDB_LOGDOMAIN, "Reading object from this buffer (%d bytes)", bsize);
  GWEN_Text_LogString((const char *)pStart, bsize, AQDB_LOGDOMAIN, GWEN_LoggerLevel_Error);
#endif

  /* scan buffer and set fields accordingly */
  p=pStart;
  pEnd=p+bsize;
  if (pEnd<p) {
    DBG_INFO(AQDB_LOGDOMAIN, "Empty buffer");
    return GWEN_ERROR_BAD_DATA;
  }

  j=0;
  while ((p<pEnd) && (j<nfields)) {
    uint32_t len;
    uint8_t c;

    c=*(p++);
    switch (c) {
    case 0xf1:
      len=*(p++);
      break;

    case 0xf2:
      len=*(p++);
      len+=(*(p++))<<8;
      break;

    case 0xf3:
      len=*(p++);
      len+=(*(p++))<<8;
      len+=(*(p++))<<16;
      break;

    case 0xf4:
      len=*(p++);
      len+=(*(p++))<<8;
      len+=(*(p++))<<16;
      len+=(*(p++))<<24;
      break;

    case 0xf5:
    case 0xf6:
    case 0xf7:
    case 0xf8:
    case 0xf9:
    case 0xfa:
    case 0xfb:
    case 0xfc:
    case 0xfd:
    case 0xfe:
    case 0xff:
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size specifier for field (%d)", j);
      return GWEN_ERROR_BAD_DATA;

    default:
      len=c;
      len+=(*(p++))<<8;
      break;
    }

#if 0
    DBG_ERROR(AQDB_LOGDOMAIN, "Setting field %d (out of %d):", j, nfields);
    GWEN_Text_LogString((const char *)p, len, AQDB_LOGDOMAIN, GWEN_LoggerLevel_Error);
#endif

    if ((p+len)>(pEnd+1)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size (buffer overflow)");
      return GWEN_ERROR_BAD_DATA;
    }
    AQDB_Object_SetField(object, j, p, len);
    p+=len;
    j++;
  }
  if (p!=pEnd) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Incomplete data in buffer (%d, %d)",
              (int)(p-pStart), (int) bsize);
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}



int _object_toBuffer(const AQDB_OBJECT *o, GWEN_BUFFER *buf)
{
  uint32_t fieldCount;
  uint32_t j;

  fieldCount=AQDB_Object_GetFieldCount(o);
  for (j=0; j<fieldCount; j++) {
    uint8_t *pData;
    uint32_t len;
    int rv;
    uint8_t lh[5];

    rv=AQDB_Object_GetField(o, j, &pData, &len);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    if (len>0xffffff) {
      lh[0]=0xf4;
      *((uint32_t *)(lh+1))=GWEN_ENDIAN_HTOLE32(len);
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 5);
    }
    else if (len>0xffff) {
      lh[0]=0xf3;
      *((uint32_t *)(lh+1))=GWEN_ENDIAN_HTOLE32(len);
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 4);
    }
    else if (len>0xf000) {
      lh[0]=0xf2;
      *((uint16_t *)(lh+1))=GWEN_ENDIAN_HTOLE16(len);
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 3);
    }
    else {
      *((uint16_t *)(lh))=GWEN_ENDIAN_HTOLE16(len);
      GWEN_Buffer_AppendBytes(buf, (const char *) lh, 2);
    }

    if (len>0)
      GWEN_Buffer_AppendBytes(buf, (const char *) pData, len);
  }

  return 0;
}




