/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "file/iolowlevel/io_session.h"

#include "file/iolowlevel/layer.h"
#include "file/iolowlevel/io_db.h"
#include "file/iolowlevel/io_table.h"

#include <aqdatabase/aqdb_object_be.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>


#define AQDB_DBFILE_LOCK_TIMEOUT 30000



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */


static int _lockDbFile(AQDB_IOLAYER *ioLayer);
static int _unlockDbFile(AQDB_IOLAYER *ioLayer);
static int _beginEditAllTables(AQDB_IOLAYER *ioLayer);
static int _endEditAllTables(AQDB_IOLAYER *ioLayer, int abandon);
static int _incrementSessionIdAndWriteHeader(AQDB_IOLAYER *ioLayer);
static int _flushTypeTablesAndSync(AQDB_IOLAYER *ioLayer);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */


int AQDB_IoLayerLowlevel_BeginEdit(AQDB_IOLAYER *ioLayer, uint32_t editFlags, const char *user)
{
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Begin edit session");

  if (AQDB_IoLayerLowlevel_GetFileLock(ioLayer)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Already locked");
    return GWEN_ERROR_INVALID;
  }

  rv=_lockDbFile(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    _unlockDbFile(ioLayer);
    return rv;
  }

  rv=_beginEditAllTables(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    _unlockDbFile(ioLayer);
    return rv;
  }

  rv=_incrementSessionIdAndWriteHeader(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    _unlockDbFile(ioLayer);
    return rv;
  }

  return 0;
}



int AQDB_IoLayerLowlevel_EndEdit(AQDB_IOLAYER *ioLayer, uint32_t flags)
{
  int rv;

  DBG_DEBUG(AQDB_LOGDOMAIN, "End edit session");

  if (AQDB_IoLayerLowlevel_GetFileLock(ioLayer)==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked");
    return GWEN_ERROR_INVALID;
  }

  rv=_endEditAllTables(ioLayer, (flags & AQDB_ACTION_FLAGS_ABORT));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    _unlockDbFile(ioLayer);
    return rv;
  }

  rv=_flushTypeTablesAndSync(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    _unlockDbFile(ioLayer);
    return rv;
  }

  rv=_unlockDbFile(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _lockDbFile(AQDB_IOLAYER *ioLayer)
{
  const char *dataFileName;
  GWEN_FSLOCK *lck;
  GWEN_FSLOCK_RESULT res;

  dataFileName=AQDB_IoLayerLowlevel_GetDataFileName(ioLayer);
  lck=GWEN_FSLock_new(dataFileName, GWEN_FSLock_TypeFile);
  res=GWEN_FSLock_Lock(lck, AQDB_DBFILE_LOCK_TIMEOUT, 0);
  switch (res) {
  case GWEN_FSLock_ResultOk:
    AQDB_IoLayerLowlevel_SetFileLock(ioLayer, lck);
    break;
  case GWEN_FSLock_ResultError:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock error");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_IO;
  case GWEN_FSLock_ResultTimeout:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_TIMEOUT;
  case GWEN_FSLock_ResultBusy:
    DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout (busy)");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_TIMEOUT;
  case GWEN_FSLock_ResultNoLock:
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid lock");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_IO;
  case GWEN_FSLock_ResultUserAbort:
    DBG_INFO(AQDB_LOGDOMAIN, "User aborted");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_USER_ABORTED;
  default:
    DBG_INFO(AQDB_LOGDOMAIN, "Unknown error");
    GWEN_FSLock_free(lck);
    return GWEN_ERROR_GENERIC;
  }

  return 0;
}



int _unlockDbFile(AQDB_IOLAYER *ioLayer)
{
  GWEN_FSLOCK *lck;

  DBG_DEBUG(AQDB_LOGDOMAIN, "Unlock DB file");

  lck=AQDB_IoLayerLowlevel_GetFileLock(ioLayer);
  GWEN_FSLock_Unlock(lck);
  AQDB_IoLayerLowlevel_SetFileLock(ioLayer, NULL);

  return 0;
}



int _beginEditAllTables(AQDB_IOLAYER *ioLayer)
{
  AQDB_TABLE_LIST *dbTypeTables;

  dbTypeTables=AQDB_IoLayer_GetDbTypeTables(ioLayer);
  if (dbTypeTables) {
    AQDB_TABLE *dbt;

    dbt=AQDB_Table_List_First(dbTypeTables);
    while (dbt) {
      int rv;

      rv=AQDB_IoLayerLowlevel_BeginEditTable(ioLayer, dbt);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
      dbt=AQDB_Table_List_Next(dbt);
    }
  }

  return 0;
}



int _endEditAllTables(AQDB_IOLAYER *ioLayer, int abandon)
{
  AQDB_TABLE_LIST *dbTypeTables;

  dbTypeTables=AQDB_IoLayer_GetDbTypeTables(ioLayer);
  if (dbTypeTables) {
    AQDB_TABLE *dbt;

    dbt=AQDB_Table_List_First(dbTypeTables);
    while (dbt) {
      int rv;

      rv=AQDB_IoLayerLowlevel_EndEditTable(ioLayer, dbt, abandon);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }

      dbt=AQDB_Table_List_Next(dbt);
    }
  }

  return 0;
}



int _incrementSessionIdAndWriteHeader(AQDB_IOLAYER *ioLayer)
{
  AQDB_ID currentSessionId;
  AQDB_DBFILE_HEADER *dbh;
  int rv;

  currentSessionId=AQDB_IoLayer_GetCurrentSessionId(ioLayer)+1;
  AQDB_IoLayer_SetCurrentSessionId(ioLayer, currentSessionId);

  dbh=AQDB_IoLayerLowlevel_GetHeader(ioLayer);
  AQDB_DbFile_Header_SetLastSessionId(dbh, currentSessionId);

  rv=AQDB_IoLayerLowlevel_WriteHeader(ioLayer, dbh);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int _flushTypeTablesAndSync(AQDB_IOLAYER *ioLayer)
{
  AQFS_FILE *fsFile;
  int rv;

  /* let all types write their object lists and column lists */
  rv=AQDB_IoLayerLowlevel_FlushTypeData(ioLayer);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  rv=AQFS_File_Sync(fsFile);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}

