/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "file/iolowlevel/io_table.h"

#include "file/iolowlevel/io_object.h"
#include "file/iolowlevel/layer.h"


#include <aqdatabase/aqdb_object_be.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/endianfns.h>



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */


static int _readTableFields(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt);
static int _createTableForColumns(AQDB_TABLE *dbt, AQFS_FILE *fsFile);
static int _writeColumns(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, const AQDB_COLUMN_LIST *columns);
static int _createTableForObjects(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt);
static int64_t _writeTypeTable(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt);
static AQFS_TABLE_PTRLIST *_readObjectTablePtrList(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt,
                                                   AQFS_FILE *fsFile);
static AQFS_TABLE_LIST *_readTableListColumns(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt,
                                              AQFS_FILE *fsFile);
static AQDB_COLUMN_LIST *_readColumnList(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt,
                                         AQFS_TABLE_LIST *ftl,
                                         AQFS_FILE *fsFile);
static AQDB_COLUMN *_readColumn(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, uint64_t offset);
static uint32_t _sampleMaxObjectSizeForColumnsList(const AQDB_COLUMN_LIST *cl);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */


int AQDB_IoLayerLowlevel_ReadTableAt(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, uint64_t offset)
{
  AQFS_FILE *fsFile;
  int rv;
  int64_t rd;
  uint8_t *p=NULL;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  /* read data */
  rd=AQFS_File_ReadDataAt(fsFile, offset, &p, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }

  /* parse it into table */
  rv=AQDB_IoLayerLowlevel_Table_ReadFromBuffer(dbt, p, rd);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    free(p);
    return rv;
  }

  /* done with data */
  free(p);

  /* done entirely */
  return 0;
}




AQDB_TABLE *AQDB_IoLayerLowlevel_CreateTable(AQDB_IOLAYER *ioLayer,
                                             const char *name,
                                             uint32_t flags,
                                             const AQDB_COLUMN_LIST *columns)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE_LIST *tableListTypes;
  AQDB_TABLE_LIST *dbTypeTables;
  AQDB_TABLE *dbt;
  int64_t rd;
  uint64_t offset;
  int rv;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  dbTypeTables=AQDB_IoLayer_GetDbTypeTables(ioLayer);
  tableListTypes=AQDB_IoLayerLowlevel_GetTableListTypes(ioLayer);

  /* create table */
  dbt=AQDB_IoLayerLowlevel_Table_new();
  AQDB_Table_SetName(dbt, name);
  AQDB_Table_SetFlags(dbt, flags);
  AQDB_Table_SetColumnList(dbt, AQDB_Column_List_dup(columns));

  rv=_createTableForColumns(dbt, fsFile);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Table_free(dbt);
    return NULL;
  }

  rv=_writeColumns(ioLayer, dbt, columns);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Table_free(dbt);
    return NULL;
  }

  rv=_createTableForObjects(ioLayer, dbt);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Table_free(dbt);
    return NULL;
  }

  rd=_writeTypeTable(ioLayer, dbt);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int)rd);
    AQDB_Table_free(dbt);
    return NULL;
  }
  offset=rd;

  /* append to list of types in file */
  rd=AQFS_File_AddTableEntry(fsFile, tableListTypes, offset);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int)rd);
    AQDB_Table_free(dbt);
    return NULL;
  }
  AQDB_Table_SetTableId(dbt, rd+1); /* id=index into type table */
  AQDB_Table_SetOpenFlags(dbt, flags);

  /* done */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Created type \"%s\" with id %llu", name, (long long unsigned)(rd+1));
  AQDB_Table_List_Add(dbt, dbTypeTables);

  /* table has been created, so we're in editMode, let table reflect that */
  rv=AQDB_IoLayerLowlevel_BeginEditTable(ioLayer, dbt);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    /* dont free table here, it has already been added to the list */
    return NULL;
  }

  return dbt;
}




AQDB_TABLE *AQDB_IoLayerLowlevel_OpenTable(AQDB_IOLAYER *ioLayer,
                                           uint32_t aflags,
                                           const char *tname,
                                           uint32_t flags,
                                           AQDB_ID *pTableId)
{
  AQDB_TABLE *dbt;
  int rv;
  uint32_t previousOpenFlags;

  /* get table */
  dbt=AQDB_IoLayer_GetTableByName(ioLayer, tname);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%s] not found", tname);
    return NULL;
  }

  previousOpenFlags=AQDB_Table_GetOpenFlags(dbt);
  AQDB_Table_AddOpenFlags(dbt, flags);

  if (previousOpenFlags==0) {
    AQDB_COLUMN_LIST *cl;

    /* read table fields if not already done so */
    cl=AQDB_Table_GetColumnList(dbt);
    if (AQDB_Column_List_GetCount(cl)==0) {
      rv=_readTableFields(ioLayer, dbt);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return NULL;
      }
    }

    if (AQDB_IoLayerLowlevel_GetFileLock(ioLayer)) {
      /* we're in editMode, let table reflect that */
      rv=AQDB_IoLayerLowlevel_BeginEditTable(ioLayer, dbt);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_Table_SetOpenFlags(dbt, 0); /* was 0, set back to 0 */
        /* dont free table here, it has already been added to the list */
        return NULL;
      }
    }
  }

  *pTableId=AQDB_Table_GetTableId(dbt);
  return dbt;
}



int AQDB_IoLayerLowlevel_CloseTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, uint32_t flags)
{
  DBG_DEBUG(AQDB_LOGDOMAIN, "Closing table [%s]", AQDB_Table_GetName(dbt));

#if 0
  AQDB_IoLayerLowlevel_Table_SetSavedTableListObjects(dbt, NULL);
  AQDB_IoLayerLowlevel_Table_SetTableListObjects(dbt, NULL);
  AQDB_IoLayerLowlevel_Table_SetTableListColumns(dbt, NULL);
  AQDB_Table_SetColumnList(dbt, NULL);

#endif

  if (AQDB_IoLayerLowlevel_GetFileLock(ioLayer)) {
    int rv;

    /* we're in editMode, leave it for this table */
    rv=AQDB_IoLayerLowlevel_EndEditTable(ioLayer, dbt, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    }
  }

  AQDB_Table_SetOpenFlags(dbt, 0);

  return 0;
}



int64_t AQDB_IoLayerLowlevel_GetObjectPos(GWEN_UNUSED AQDB_IOLAYER *ioLayer,
                                            const AQDB_TABLE *dbt,
                                            uint64_t index,
                                            int asWriter)
{
  int64_t pos;

  /* begin critical section */
  pos=AQDB_IoLayerLowlevel_Table_GetObjectPos(dbt, index, asWriter);
  /* end critical section */

  return pos;
}



int AQDB_IoLayerLowlevel_SetObjectPos(GWEN_UNUSED AQDB_IOLAYER *ioLayer,
                                        AQDB_TABLE *dbt,
                                        uint64_t index,
                                        uint64_t objectPos)
{
  int rv;

  /* begin critical section */
  rv=AQDB_IoLayerLowlevel_Table_SetObjectPos(dbt, index, objectPos);
  /* end critical section */
  return rv;
}




int AQDB_IoLayerLowlevel_BeginEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt)
{
  AQFS_TABLE_PTRLIST *readersTablePtrListObjects;
  AQFS_TABLE_PTRLIST *writersTablePtrListObjects;
  const char *tableName;

  tableName=AQDB_Table_GetName(dbt);
  writersTablePtrListObjects=AQDB_IoLayerLowlevel_Table_GetWritersTablePtrListObjects(dbt);
  readersTablePtrListObjects=AQDB_IoLayerLowlevel_Table_GetReadersTablePtrListObjects(dbt);

  if (writersTablePtrListObjects!=NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "There already is a writersTablePtrList, already in edit mode?");
    return GWEN_ERROR_INVALID;
  }

  DBG_DEBUG(AQDB_LOGDOMAIN, "BeginEdit(%s)", (tableName)?(tableName):"<noname>");
  writersTablePtrListObjects=AQFS_Table_PtrList_LazyCopy(readersTablePtrListObjects);
  if (writersTablePtrListObjects==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error copying readersTablePtrList");
    return GWEN_ERROR_GENERIC;
  }
  AQDB_IoLayerLowlevel_Table_SetWritersTablePtrListObjects(dbt, writersTablePtrListObjects);
  return 0;
}



int AQDB_IoLayerLowlevel_EndEditTable(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, int abandon)
{
  AQFS_FILE *fsf;
  AQFS_TABLE_PTRLIST *writersTablePtrListObjects;
  const char *tableName;

  fsf=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  tableName=AQDB_Table_GetName(dbt);
  writersTablePtrListObjects=AQDB_IoLayerLowlevel_Table_GetWritersTablePtrListObjects(dbt);

  if (writersTablePtrListObjects==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No writersTablePtrList, not in edit mode?");
    return GWEN_ERROR_INVALID;
  }

  DBG_DEBUG(AQDB_LOGDOMAIN, "EndEdit(%s%s) started", (tableName)?(tableName):"<noname>", abandon?", abandon":"");

  if (abandon==0) {
    int rv;

    DBG_DEBUG(AQDB_LOGDOMAIN, "EndEdit(%s): Append new tables to file", (tableName)?(tableName):"<noname>");
    rv=AQFS_File_AppendNewTablesFromPtrListToFile(fsf, writersTablePtrListObjects);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "Error in AQFS_File_AppendNewTablesFromPtrListToFile: %d", rv);
    }
    else {
      DBG_DEBUG(AQDB_LOGDOMAIN, "EndEdit(%s): Write modified tables to file", (tableName)?(tableName):"<noname>");
      rv=AQFS_File_WritDirtyTablesFromPtrList(fsf, writersTablePtrListObjects);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error in AQFS_File_WritDirtyTablesFromPtrList: %d", rv);
      }
      else {
        AQDB_IoLayerLowlevel_Table_TakeWritersTablePtrListObjects(dbt);
        AQDB_IoLayerLowlevel_Table_SetReadersTablePtrListObjects(dbt, writersTablePtrListObjects);
      }
    }
  }

  AQDB_IoLayerLowlevel_Table_SetWritersTablePtrListObjects(dbt, NULL);

  DBG_DEBUG(AQDB_LOGDOMAIN, "EndEdit(%s) finished", (tableName)?(tableName):"<noname>");
  return 0;
}






int _readTableFields(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt)
{
  AQFS_FILE *fsFile;
  AQFS_TABLE_LIST *ftlColumns;
  AQFS_TABLE_PTRLIST *ftlReadersTablePtrList;
  AQDB_COLUMN_LIST *columnList;
  uint32_t maxObjectSize;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);

  ftlReadersTablePtrList=_readObjectTablePtrList(ioLayer, dbt, fsFile);
  if (ftlReadersTablePtrList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    return GWEN_ERROR_GENERIC;
  }

  ftlColumns=_readTableListColumns(ioLayer, dbt, fsFile);
  if (ftlColumns==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    AQFS_Table_PtrList_free(ftlReadersTablePtrList);
    return GWEN_ERROR_GENERIC;
  }

  columnList=_readColumnList(ioLayer, dbt, ftlColumns, fsFile);
  if (columnList==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    AQFS_Table_List_free(ftlColumns);
    AQFS_Table_PtrList_free(ftlReadersTablePtrList);
    return GWEN_ERROR_GENERIC;
  }

  maxObjectSize=_sampleMaxObjectSizeForColumnsList(columnList);

  /* begin critical section */
  AQDB_IoLayerLowlevel_Table_SetReadersTablePtrListObjects(dbt, ftlReadersTablePtrList);
  AQDB_IoLayerLowlevel_Table_SetTableListColumns(dbt, ftlColumns);
  AQDB_Table_SetColumnList(dbt, columnList);
  AQDB_Table_SetMaxObjectSize(dbt, maxObjectSize);
  /* end critical section */

  return 0;
}



AQFS_TABLE_PTRLIST *_readObjectTablePtrList(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt, AQFS_FILE *fsFile)
{
  uint64_t offset;

  /* read object table ptrlist */
  offset=AQDB_IoLayerLowlevel_Table_GetOffsetObjectTable(dbt);
  if (offset>0) {
    AQFS_TABLE_PTRLIST *ftl;
    int rv;

    ftl=AQFS_Table_PtrList_new(AQDB_DBFILE_OBJECTTABLE_MAXENTRIES);
    rv=AQFS_File_ReadTablePtrList(fsFile, offset, ftl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQFS_Table_PtrList_free(ftl);
      return NULL;
    }
    return ftl;
  }

  return NULL;
}



AQFS_TABLE_LIST *_readTableListColumns(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, AQFS_FILE *fsFile)
{
  uint64_t offset;

  offset=AQDB_IoLayerLowlevel_Table_GetOffsetColumnTable(dbt);
  if (offset>0) {
    AQFS_TABLE_LIST *ftl;
    int rv;

    ftl=AQFS_Table_List_new();
    rv=AQFS_File_ReadTableList(fsFile, offset, ftl);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQFS_Table_List_free(ftl);
      return NULL;
    }
    return ftl;
  }

  return NULL;
}



AQDB_COLUMN_LIST *_readColumnList(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, AQFS_TABLE_LIST *ftl,
                                  AQFS_FILE *fsFile)
{
  if (ftl) {
    AQDB_COLUMN_LIST *cl;
    AQFS_TABLE *ft;
    uint32_t maxObjectSize=0;

    cl=AQDB_Column_List_new();

    /* read columns from all tables */
    ft=AQFS_Table_List_First(ftl);
    while (ft) {
      uint64_t maxEntries;
      uint64_t idx;
      uint64_t *ptr;

      maxEntries=AQFS_Table_GetMaxEntries(ft);
      DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading entries from table (maxEntries=%d, free entries=%d)",
                  (int) maxEntries,
                  (int) AQFS_Table_GetFreeEntries(ft));
      ptr=AQFS_Table_GetPtrEntries(ft);
      for (idx=0; idx<maxEntries; idx++) {
        uint64_t offset;

        offset=*(ptr++);
        if (offset>0) {
          AQDB_COLUMN *col=NULL;

          col=_readColumn(ioLayer, dbt, offset);
          if (col==NULL) {
            DBG_INFO(AQDB_LOGDOMAIN, "here");
            AQDB_Column_List_free(cl);
            return NULL;
          }

          DBG_VERBOUS(AQDB_LOGDOMAIN, "Column at %llu is [%s]",
                      (unsigned long long int) offset,
                      AQDB_Column_GetName(col));
          maxObjectSize+=AQDB_Column_GetMaxLen(col);
          /* add to column list */
          AQDB_Column_List_Add(col, cl);
        }
      }

      ft=AQFS_Table_List_Next(ft);
    } /* while ftl */
    return cl;
  } /* if ftl */

  return NULL;
}



uint32_t _sampleMaxObjectSizeForColumnsList(const AQDB_COLUMN_LIST *cl)
{
  AQDB_COLUMN *col;
  uint32_t maxObjectSize=0;

  col=AQDB_Column_List_First(cl);
  while (col) {
    maxObjectSize+=AQDB_Column_GetMaxLen(col);
    col=AQDB_Column_List_Next(col);
  }

  return maxObjectSize;
}


AQDB_COLUMN *_readColumn(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, uint64_t offset)
{
  AQDB_OBJECT *o;
  AQDB_COLUMN *col=NULL;
  int rv;

  DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading column entry at %llu (%16llx)",
              (unsigned long long int) offset,
              (unsigned long long int) offset);
  /* read object from file */
  o=AQDB_Object_new(AQDB_Table_GetTableId(dbt), 1, AQDB_COLUMN_FIELD_COUNT);
  rv=AQDB_IoLayerLowlevel_ReadObjectAt(ioLayer, o, offset);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return NULL;
  }

  /* create column from object */
  rv=AQDB_Column_fromObject(o, &col);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_free(o);
    return NULL;
  }
  AQDB_Object_free(o);

  return col;
}



int _createTableForColumns(AQDB_TABLE *dbt, AQFS_FILE *fsFile)
{
  AQFS_TABLE *ft;
  int64_t rd;

  /* create table for columns */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Create Table [%s]: Creating table for columns", AQDB_Table_GetName(dbt));
  ft=AQFS_Table_Create(AQDB_DBFILE_COLUMNTABLE_MAXENTRIES);
  rd=AQFS_File_AppendTable(fsFile, ft);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    AQFS_Table_free(ft);
    return (int)rd;
  }
  AQDB_IoLayerLowlevel_Table_SetOffsetColumnTable(dbt, rd);
  AQFS_Table_List_Add(ft, AQDB_IoLayerLowlevel_Table_GetTableListColumns(dbt));

  return 0;
}



int _writeColumns(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt, const AQDB_COLUMN_LIST *columns)
{
  AQDB_COLUMN *col;
  const char *name;

  name=AQDB_Table_GetName(dbt);

  /* write columns */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Create Table [%s]: Writing columns", name);
  col=AQDB_Column_List_First(columns);
  while (col) {
    AQDB_OBJECT *o;
    int rv;
    int64_t rd;

    /* create object with data from column */
    DBG_DEBUG(AQDB_LOGDOMAIN, "  Create Table [%s]: Column to object [%s]", name, AQDB_Column_GetName(col));
    o=AQDB_Object_new(1, 1, AQDB_COLUMN_FIELD_COUNT);
    rv=AQDB_Column_toObject(col, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      return rv;
    }

    /* append object to file */
    DBG_DEBUG(AQDB_LOGDOMAIN, "  Create Table [%s]: Writing column [%s]", name, AQDB_Column_GetName(col));
    rd=AQDB_IoLayerLowlevel_AppendObject(ioLayer, AQDB_IoLayerLowlevel_Table_GetTableListColumns(dbt), o);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      AQDB_Object_free(o);
      return GWEN_ERROR_GENERIC;
    }
    DBG_DEBUG(AQDB_LOGDOMAIN, "Added table object with id %llu", (unsigned long long int) rd);

    /* object no longer needed */
    AQDB_Object_free(o);

    col=AQDB_Column_List_Next(col);
  }

  return 0;
}



int _createTableForObjects(AQDB_IOLAYER *ioLayer, AQDB_TABLE *dbt)
{
  AQFS_FILE *fsFile;
  const char *name;
  AQFS_TABLE *ft;
  AQFS_TABLE_PTRLIST *ftl;
  int64_t rd;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  name=AQDB_Table_GetName(dbt);

  /* create table for objects */
  DBG_DEBUG(AQDB_LOGDOMAIN, "Create Table [%s]: Creating table for objects", name);
  ft=AQFS_Table_Create(AQDB_DBFILE_OBJECTTABLE_MAXENTRIES);
  rd=AQFS_File_AppendTable(fsFile, ft);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    AQFS_Table_free(ft);
    return GWEN_ERROR_GENERIC;
  }
  AQDB_IoLayerLowlevel_Table_SetOffsetObjectTable(dbt, rd);

  ftl=AQFS_Table_PtrList_new(AQDB_DBFILE_OBJECTTABLE_MAXENTRIES);
  AQFS_Table_PtrList_AddTable(ftl, ft);
  AQDB_IoLayerLowlevel_Table_SetReadersTablePtrListObjects(dbt, ftl);
  return 0;
}



int64_t _writeTypeTable(AQDB_IOLAYER *ioLayer, const AQDB_TABLE *dbt)
{
  AQFS_FILE *fsFile;
  const char *name;
  uint8_t *ptr;
  uint64_t len;
  int64_t rd;

  fsFile=AQDB_IoLayerLowlevel_GetFile(ioLayer);
  name=AQDB_Table_GetName(dbt);

  DBG_DEBUG(AQDB_LOGDOMAIN, "  Create Table [%s]: Writing type table", name);
  ptr=(uint8_t *) malloc(AQDB_DBFILE_TABLE_SIZE);
  assert(ptr);
  rd=AQDB_IoLayerLowlevel_Table_toBuffer(dbt, ptr, AQDB_DBFILE_TABLE_SIZE);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    free(ptr);
    return (int)rd;
  }
  len=rd;

  /* write table data to file */
  rd=AQFS_File_AppendData(fsFile, AQFS_ITEMHEADER_TYPE_GENERIC_DATA, ptr, len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    free(ptr);
    return (int)rd;
  }
  free(ptr);
  return rd; /* offset */
}




