/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "file/iolowlevel/layer_p.h"

#include "file/iolowlevel/io_object.h"
#include "file/iolowlevel/io_table.h"
#include "file/iolowlevel/io_query.h"
#include "file/iolowlevel/io_db.h"
#include "file/iolowlevel/io_session.h"

#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL)




/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void GWENHYWFAR_CB _freeData(void *bp, void *p);
static void _unattachFromTables(AQDB_IOLAYER *ioLayer);
static void _createFileNameFromUrl(AQDB_IOLAYER *ioLayer, const GWEN_URL *url);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */



AQDB_IOLAYER *AQDB_IoLayerLowlevel_new(AQDB_DB *database, const GWEN_URL *url)
{
  AQDB_IOLAYER *ioLayer;
  AQDB_IOLAYER_LOWLEVEL *xio;

  ioLayer=AQDB_IoLayer_new();
  AQDB_IoLayer_SetName(ioLayer, "lowlevel");
  AQDB_IoLayer_SetDatabase(ioLayer, database);

  GWEN_NEW_OBJECT(AQDB_IOLAYER_LOWLEVEL, xio);
  if (xio==NULL) {
    AQDB_IoLayer_free(ioLayer);
    return NULL;
  }
  GWEN_INHERIT_SETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer, xio, _freeData);

  _createFileNameFromUrl(ioLayer, url);


  xio->createFn=AQDB_IoLayer_SetCreateFn(ioLayer, AQDB_IoLayerLowlevel_Create);
  xio->openFn=AQDB_IoLayer_SetOpenFn(ioLayer, AQDB_IoLayerLowlevel_Open);
  xio->closeFn=AQDB_IoLayer_SetCloseFn(ioLayer, AQDB_IoLayerLowlevel_Close);
  xio->checkFn=AQDB_IoLayer_SetCheckFn(ioLayer, AQDB_IoLayerLowlevel_Check);

  xio->beginEditFn=AQDB_IoLayer_SetBeginEditFn(ioLayer, AQDB_IoLayerLowlevel_BeginEdit);
  xio->endEditFn=AQDB_IoLayer_SetEndEditFn(ioLayer, AQDB_IoLayerLowlevel_EndEdit);

  xio->createTableFn=AQDB_IoLayer_SetCreateTableFn(ioLayer, AQDB_IoLayerLowlevel_CreateTable);
  xio->openTableFn=AQDB_IoLayer_SetOpenTableFn(ioLayer, AQDB_IoLayerLowlevel_OpenTable);
  xio->closeTableFn=AQDB_IoLayer_SetCloseTableFn(ioLayer, AQDB_IoLayerLowlevel_CloseTable);

  xio->getObjectPosFn=AQDB_IoLayer_SetGetObjectPosFn(ioLayer, AQDB_IoLayerLowlevel_GetObjectPos);
  xio->setObjectPosFn=AQDB_IoLayer_SetSetObjectPosFn(ioLayer, AQDB_IoLayerLowlevel_SetObjectPos);

  xio->addObjectFn=AQDB_IoLayer_SetAddObjectFn(ioLayer, AQDB_IoLayerLowlevel_AddObject);
  xio->deleteObjectFn=AQDB_IoLayer_SetDeleteObjectFn(ioLayer, AQDB_IoLayerLowlevel_DeleteObject);
  xio->writeObjectFn=AQDB_IoLayer_SetWriteObjectFn(ioLayer, AQDB_IoLayerLowlevel_WriteObject);
  xio->readObjectFn=AQDB_IoLayer_SetReadObjectFn(ioLayer, AQDB_IoLayerLowlevel_ReadObject);

  xio->getActiveIdsForTableFn=AQDB_IoLayer_SetGetActiveIdsForTableFn(ioLayer, AQDB_IoLayerLowlevel_GetActiveIds);

  return ioLayer;
}



void _freeData(void *bp, void *p)
{
  AQDB_IOLAYER *ioLayer;
  AQDB_IOLAYER_LOWLEVEL *xio;

  ioLayer=(AQDB_IOLAYER*) bp;
  xio=(AQDB_IOLAYER_LOWLEVEL *) p;

  _unattachFromTables(ioLayer);

  AQFS_File_free(xio->file);
  free(xio->dataFileName);

  GWEN_FREE_OBJECT(xio);
}



void _unattachFromTables(AQDB_IOLAYER *ioLayer)
{
  AQDB_TABLE_LIST *tl;

  tl=AQDB_IoLayer_GetDbTypeTables(ioLayer);
  if (tl) {
    AQDB_TABLE *dbt;

    dbt=AQDB_Table_List_First(tl);
    while(dbt) {
      AQDB_IoLayerLowlevel_Table_Unextend(dbt);
      dbt=AQDB_Table_List_Next(dbt);
    }
  }
}



const char *AQDB_IoLayerLowlevel_GetDataFileName(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->dataFileName;
}



AQFS_FILE *AQDB_IoLayerLowlevel_GetFile(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->file;
}



void AQDB_IoLayerLowlevel_SetFile(AQDB_IOLAYER *ioLayer, AQFS_FILE *f)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (xio->file)
    AQFS_File_free(xio->file);
  xio->file=f;
}



GWEN_FSLOCK *AQDB_IoLayerLowlevel_GetFileLock(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->fileLock;
}



void AQDB_IoLayerLowlevel_SetFileLock(AQDB_IOLAYER *ioLayer, GWEN_FSLOCK *lck)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (xio->fileLock)
    GWEN_FSLock_free(xio->fileLock);
  xio->fileLock=lck;
}



AQFS_TABLE_LIST *AQDB_IoLayerLowlevel_GetTableListDatabase(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->tableListDatabase;
}



void AQDB_IoLayerLowlevel_SetTableListDatabase(AQDB_IOLAYER *ioLayer, AQFS_TABLE_LIST *ftl)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (xio->tableListDatabase)
    AQFS_Table_List_free(xio->tableListDatabase);
  xio->tableListDatabase=ftl;
}



AQFS_TABLE_LIST *AQDB_IoLayerLowlevel_GetTableListTypes(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->tableListTypes;
}



void AQDB_IoLayerLowlevel_SetTableListTypes(AQDB_IOLAYER *ioLayer, AQFS_TABLE_LIST *ftl)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (xio->tableListTypes)
    AQFS_Table_List_free(xio->tableListTypes);
  xio->tableListTypes=ftl;
}



AQDB_DBFILE_HEADER *AQDB_IoLayerLowlevel_GetHeader(const AQDB_IOLAYER *ioLayer)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  return xio->header;
}



void AQDB_IoLayerLowlevel_SetHeader(AQDB_IOLAYER *ioLayer, AQDB_DBFILE_HEADER *dbh)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (xio->header)
    AQDB_DbFile_Header_free(xio->header);
  xio->header=dbh;
}



void _createFileNameFromUrl(AQDB_IOLAYER *ioLayer, const GWEN_URL *url)
{
  AQDB_IOLAYER_LOWLEVEL *xio;

  assert(ioLayer);
  xio=GWEN_INHERIT_GETDATA(AQDB_IOLAYER, AQDB_IOLAYER_LOWLEVEL, ioLayer);
  assert(xio);

  if (url) {
    GWEN_BUFFER *pbuf;
    const char *s;

    pbuf=GWEN_Buffer_new(0, 256, 0, 1);
    s=GWEN_Url_GetServer(url);
    if (s)
      GWEN_Buffer_AppendString(pbuf, s);
    s=GWEN_Url_GetPath(url);
    if (s) {
      if (GWEN_Buffer_GetUsedBytes(pbuf))
        GWEN_Buffer_AppendString(pbuf, GWEN_DIR_SEPARATOR_S);
      GWEN_Buffer_AppendString(pbuf, s);
    }
    xio->dataFileName=strdup(GWEN_Buffer_GetStart(pbuf));
    GWEN_Buffer_free(pbuf);
  }
}



