/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2021 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "file/iolowlevel/table_p.h"

#include <gwenhywfar/endianfns.h>
#include <gwenhywfar/debug.h>



GWEN_INHERIT(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE)



/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static void AQDB_IoLayerLowlevel_Table_FreeData(void *bp, void *p);
static char *my_strndup(const char *src, size_t n);



/* ------------------------------------------------------------------------------------------------
 * code
 * ------------------------------------------------------------------------------------------------
 */


int AQDB_IoLayerLowlevel_Table_Extend(AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  GWEN_NEW_OBJECT(AQDB_IOLAYER_LOWLEVEL_TABLE, xt);
  if (xt==NULL) {
    return GWEN_ERROR_GENERIC;
  }
  GWEN_INHERIT_SETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt, xt,
                       AQDB_IoLayerLowlevel_Table_FreeData);

  xt->tableListColumns=AQFS_Table_List_new();
  xt->readersTablePtrListObjects=AQFS_Table_PtrList_new(AQDB_DBFILE_OBJECTTABLE_MAXENTRIES);

  return 0;
}



void AQDB_IoLayerLowlevel_Table_Unextend(AQDB_TABLE *dbt)
{
  if (AQDB_IoLayerLowlevel_Table_IsOfType(dbt)) {
    GWEN_INHERIT_UNLINK(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  }
}



AQDB_TABLE *AQDB_IoLayerLowlevel_Table_new()
{
  AQDB_TABLE *dbt;
  int rv;

  dbt=AQDB_Table_new();
  assert(dbt);
  rv=AQDB_IoLayerLowlevel_Table_Extend(dbt);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  return dbt;
}



void AQDB_IoLayerLowlevel_Table_FreeData(void *bp, void *p)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  xt=(AQDB_IOLAYER_LOWLEVEL_TABLE *) p;

  AQFS_Table_List_free(xt->tableListColumns);
  AQFS_Table_PtrList_free(xt->writersTablePtrListObjects);
  AQFS_Table_PtrList_free(xt->readersTablePtrListObjects);

  GWEN_FREE_OBJECT(xt);
}



int AQDB_IoLayerLowlevel_Table_IsOfType(const AQDB_TABLE *dbt)
{
  return GWEN_INHERIT_ISOFTYPE(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
}




int AQDB_IoLayerLowlevel_Table_ReadFromBuffer(AQDB_TABLE *dbt, const uint8_t *buffer, uint64_t len)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  char *sName;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (len<AQDB_DBFILE_TABLE_SIZE) {
    DBG_ERROR(AQFS_LOGDOMAIN, "Buffer too small for valid table data (%d < %d)",
              (int) len, (int) AQDB_DBFILE_TABLE_SIZE);
    return GWEN_ERROR_BAD_DATA;
  }

  xt->offsetColumnTable=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQDB_DBFILE_TABLE_OFFS_COLUMN_TABLE)));
  xt->offsetObjectTable=GWEN_ENDIAN_LE64TOH(*((uint64_t *)(buffer+AQDB_DBFILE_TABLE_OFFS_OBJECT_TABLE)));
  sName=my_strndup((const char *)(buffer+AQDB_DBFILE_TABLE_OFFS_NAME), AQDB_DBFILE_TABLE_SIZE_NAME-1);
  AQDB_Table_SetName(dbt, sName);
  free(sName);

  AQFS_Table_PtrList_Clear(xt->readersTablePtrListObjects);
  AQFS_Table_List_Clear(xt->tableListColumns);

  return 0;
}



int64_t AQDB_IoLayerLowlevel_Table_toBuffer(const AQDB_TABLE *dbt, uint8_t *buffer, uint64_t len)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  const char *sName;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (len<AQDB_DBFILE_TABLE_SIZE) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Buffer too small for valid table data (%d < %d)",
              (int) len, (int) AQDB_DBFILE_TABLE_SIZE);
    return GWEN_ERROR_BAD_DATA;
  }
  memset(buffer, 0, AQDB_DBFILE_TABLE_SIZE);

  *((uint64_t *)(buffer+AQDB_DBFILE_TABLE_OFFS_COLUMN_TABLE))=GWEN_ENDIAN_HTOLE64(xt->offsetColumnTable);
  *((uint64_t *)(buffer+AQDB_DBFILE_TABLE_OFFS_OBJECT_TABLE))=GWEN_ENDIAN_HTOLE64(xt->offsetObjectTable);
  sName=AQDB_Table_GetName(dbt);
  if (sName)
    strncpy((char *)(buffer+AQDB_DBFILE_TABLE_OFFS_NAME), sName, AQDB_DBFILE_TABLE_SIZE_NAME-1);
  buffer[AQDB_DBFILE_TABLE_OFFS_NAME+AQDB_DBFILE_TABLE_SIZE_NAME-1]=0;

  return AQDB_DBFILE_TABLE_SIZE;
}



uint64_t AQDB_IoLayerLowlevel_Table_GetOffsetColumnTable(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  return xt->offsetColumnTable;
}



void AQDB_IoLayerLowlevel_Table_SetOffsetColumnTable(AQDB_TABLE *dbt, uint64_t i)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  xt->offsetColumnTable=i;
}



uint64_t AQDB_IoLayerLowlevel_Table_GetOffsetObjectTable(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  return xt->offsetObjectTable;
}



void AQDB_IoLayerLowlevel_Table_SetOffsetObjectTable(AQDB_TABLE *dbt, uint64_t i)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  xt->offsetObjectTable=i;
}



AQFS_TABLE_LIST *AQDB_IoLayerLowlevel_Table_GetTableListColumns(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  return xt->tableListColumns;
}



void AQDB_IoLayerLowlevel_Table_SetTableListColumns(AQDB_TABLE *dbt, AQFS_TABLE_LIST *ftl)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (ftl!=xt->tableListColumns)
    AQFS_Table_List_free(xt->tableListColumns);
  xt->tableListColumns=ftl;
}



AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_GetReadersTablePtrListObjects(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  return xt->readersTablePtrListObjects;
}



void AQDB_IoLayerLowlevel_Table_SetReadersTablePtrListObjects(AQDB_TABLE *dbt, AQFS_TABLE_PTRLIST *ftl)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (ftl!=xt->readersTablePtrListObjects)
    AQFS_Table_PtrList_free(xt->readersTablePtrListObjects);
  xt->readersTablePtrListObjects=ftl;
}



AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_GetWritersTablePtrListObjects(const AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  return xt->writersTablePtrListObjects;
}



AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_TakeWritersTablePtrListObjects(AQDB_TABLE *dbt)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  AQFS_TABLE_PTRLIST *ptrList;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  ptrList=xt->writersTablePtrListObjects;
  xt->writersTablePtrListObjects=NULL;
  return ptrList;
}



void AQDB_IoLayerLowlevel_Table_SetWritersTablePtrListObjects(AQDB_TABLE *dbt, AQFS_TABLE_PTRLIST *ftl)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (ftl!=xt->writersTablePtrListObjects)
    AQFS_Table_PtrList_free(xt->writersTablePtrListObjects);
  xt->writersTablePtrListObjects=ftl;
}



int64_t AQDB_IoLayerLowlevel_Table_GetObjectPos(const AQDB_TABLE *dbt, uint64_t index, int asWriter)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  int64_t offset;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (asWriter) {
    if (xt->writersTablePtrListObjects==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "No writersTablePtrList, not in edit mode?");
      return GWEN_ERROR_INVALID;
    }
    offset=AQFS_Table_PtrList_GetEntryAt(xt->writersTablePtrListObjects, index);
  }
  else
    offset=AQFS_Table_PtrList_GetEntryAt(xt->readersTablePtrListObjects, index);
  if (offset<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d) (as %s)", (int) offset, asWriter?"writer":"reader");
    return offset;
  }

  return offset;
}



int AQDB_IoLayerLowlevel_Table_SetObjectPos(AQDB_TABLE *dbt, uint64_t index, uint64_t objectPos)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  int rv;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  rv=AQFS_Table_PtrList_SetEntryAt(xt->writersTablePtrListObjects, index, objectPos);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int64_t AQDB_IoLayerLowlevel_Table_AddObjectPos(AQDB_TABLE *dbt, uint64_t objectPos)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  int64_t index;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (xt->writersTablePtrListObjects==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No writersTablePtrList, not in edit mode?");
    return (int64_t) GWEN_ERROR_INVALID;
  }

  index=AQFS_Table_PtrList_AddEntry(xt->writersTablePtrListObjects, objectPos);
  if (index<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) index);
    return index;
  }

  return index;
}



int AQDB_IoLayerLowlevel_Table_GetObjectIds(AQDB_TABLE *dbt, GWEN_IDLIST64 *idl, int asWriter)
{
  AQDB_IOLAYER_LOWLEVEL_TABLE *xt;
  AQFS_TABLE_PTRLIST *ptrList;
  uint64_t tableCount;
  uint64_t idxCurrentTable;
  uint64_t currentObjectId=0;
  uint64_t maxEntriesPerTable=0;

  assert(dbt);
  xt=GWEN_INHERIT_GETDATA(AQDB_TABLE, AQDB_IOLAYER_LOWLEVEL_TABLE, dbt);
  assert(xt);

  if (asWriter) {
    ptrList=xt->writersTablePtrListObjects;
    if (ptrList==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "No writersTablePtrList, not in edit mode?");
      return GWEN_ERROR_INVALID;
    }
  }
  else
    ptrList=xt->readersTablePtrListObjects;
  assert(ptrList);

  tableCount=AQFS_Table_PtrList_GetUsedTables(ptrList);
  maxEntriesPerTable=AQFS_Table_PtrList_GetTableMaxEntries(ptrList);

  for (idxCurrentTable=0; idxCurrentTable<tableCount; idxCurrentTable++) {
    AQFS_TABLE *ft;

    ft=AQFS_Table_PtrList_GetTableAt(ptrList, idxCurrentTable);
    if (ft) {
      uint64_t maxEntries;
      uint64_t idx;
      uint64_t *ptr;

      maxEntries=AQFS_Table_GetMaxEntries(ft);
      DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading object entries from table \"%s\" (maxEntries=%d, free entries=%d)",
                  AQDB_Table_GetName(dbt),
                  (int) maxEntries,
                  (int) AQFS_Table_GetFreeEntries(ft));
      ptr=AQFS_Table_GetPtrEntries(ft);
      for (idx=0; idx<maxEntries; idx++) {
        uint64_t offset;

        offset=*(ptr++);
        if (offset>0 && offset !=0xffffffffffffffffL) {
          /* add id */
          GWEN_IdList64_AddId(idl, currentObjectId+idx+1);
        }
      }

      currentObjectId+=maxEntries;
    }
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "No table at position %lu", (unsigned long) idxCurrentTable);
      currentObjectId+=maxEntriesPerTable;
    }
  }

  return 0;
}



char *my_strndup(const char *src, size_t n)
{
  int len;

  len=strlen(src);
  if (len<n)
    return strdup(src);
  else {
    char *cpy;

    cpy=(char *) malloc(n+1);
    assert(cpy);
    memmove(cpy, src, n);
    cpy[n]=0;
    return cpy;
  }
}




