/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifndef AQDB_IOLAYER_LOWLEVEL_TABLE_H
#define AQDB_IOLAYER_LOWLEVEL_TABLE_H


#include "aqdatabase/backendsupport/iolayer.h"
#include "aqdatabase/backendsupport/table.h"

#include <aqfilestorage/aqfs_table.h>



#define AQDB_DBFILE_TABLE_SIZE 80

#define AQDB_DBFILE_OBJECTTABLE_MAXENTRIES 1024
#define AQDB_DBFILE_COLUMNTABLE_MAXENTRIES 64




/** @name Construction, destruction, reading, writing
 *
 */
/*@{*/

AQDB_TABLE *AQDB_IoLayerLowlevel_Table_new();

int AQDB_IoLayerLowlevel_Table_Extend(AQDB_TABLE *dbt);
void AQDB_IoLayerLowlevel_Table_Unextend(AQDB_TABLE *dbt);

int AQDB_IoLayerLowlevel_Table_IsOfType(const AQDB_TABLE *dbt);


/**
 * Read a AQDB_TABLE object from a buffer.
 * @return 0 if ok, error code otherwise
 * @param dbt AQDB_TABLE object to read data into
 * @param buffer pointer to a buffer containing data to read
 * @param len size of the buffer pointed to by buffer
 */
int AQDB_IoLayerLowlevel_Table_ReadFromBuffer(AQDB_TABLE *dbt, const uint8_t *buffer, uint64_t len);


/**
 * Write content of a AQDB_TABLE object to a buffer.
 * @return number of bytes written, <1 on error
 * @param dbt pointer to this AQDB_TABLE object
 * @param buffer pointer to a buffer to write to
 * @param len size of the buffer pointed to by buffer
 */
int64_t AQDB_IoLayerLowlevel_Table_toBuffer(const AQDB_TABLE *dbt, uint8_t *buffer, uint64_t len);

/*@}*/



/** @name Information stored on disk.
 *
 */
/*@{*/

/**
 * Return the offset of the first AQFS_TABLE for column descriptions.
 * Each entry of the AQFS_TABLE referenced by this offset points to the offset of a column description
 * for the type stored in this object.
 * @param dbt pointer to this AQDB_TABLE object
 */
uint64_t AQDB_IoLayerLowlevel_Table_GetOffsetColumnTable(const AQDB_TABLE *dbt);

/**
 * Set the offset of the first AQFS_TABLE for column descriptions (see @ref AQDB_IoLayerLowlevel_Table_GetOffsetColumnTable).
 * @param dbt pointer to this AQDB_TABLE object
 * @param i new value
 */
void AQDB_IoLayerLowlevel_Table_SetOffsetColumnTable(AQDB_TABLE *dbt, uint64_t i);


/**
 * Return the offset of the first AQFS_TABLE for objects.
 * Each entry of the AQFS_TABLE referenced by this offset points to the offset of a data object
 * of the type stored in this object.
 * @param dbt pointer to this AQDB_TABLE object
 */
uint64_t AQDB_IoLayerLowlevel_Table_GetOffsetObjectTable(const AQDB_TABLE *dbt);

/**
 * Set the offset of the first AQFS_TABLE for objects.
 * @param dbt pointer to this AQDB_TABLE object
 * @param i new value
 */
void AQDB_IoLayerLowlevel_Table_SetOffsetObjectTable(AQDB_TABLE *dbt, uint64_t i);


/*@}*/




/** @name Runtime data, not stored on disk.
 *
 */
/*@{*/

/**
 * Get AQFS_TABLE_LIST for the column descriptions.
 *
 * This is a chained list of AQFS_TABLE objects referenced by @ref AQDB_IoLayerLowlevel_Table_GetOffsetColumnTable.
 * Each entry of each table points to a column description.
 * @param dbt pointer to this AQDB_TABLE object
 */
AQFS_TABLE_LIST *AQDB_IoLayerLowlevel_Table_GetTableListColumns(const AQDB_TABLE *dbt);

/**
 * Set AQFS_TABLE_LIST for the column descriptions.
 * @param dbt pointer to this AQDB_TABLE object
 * @param ftl new value
 */
void AQDB_IoLayerLowlevel_Table_SetTableListColumns(AQDB_TABLE *dbt, AQFS_TABLE_LIST *ftl);

/**
 * Get AQFS_TABLE_PTRLIST for the objects.
 *
 * This is a chained list of AQFS_TABLE objects referenced by @ref AQDB_IoLayerLowlevel_Table_GetOffsetObjectTable.
 * Each entry of each table points to an object.
 * @param dbt pointer to this AQDB_TABLE object
 */
AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_GetReadersTablePtrListObjects(const AQDB_TABLE *dbt);


/**
 * Set AQFS_TABLE_PTRLIST for the objects.
 * @param dbt pointer to this AQDB_TABLE object
 * @param ftl new value
 */
void AQDB_IoLayerLowlevel_Table_SetReadersTablePtrListObjects(AQDB_TABLE *dbt, AQFS_TABLE_PTRLIST *ftl);


AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_GetWritersTablePtrListObjects(const AQDB_TABLE *dbt);
AQFS_TABLE_PTRLIST *AQDB_IoLayerLowlevel_Table_TakeWritersTablePtrListObjects(AQDB_TABLE *dbt);

void AQDB_IoLayerLowlevel_Table_SetWritersTablePtrListObjects(AQDB_TABLE *dbt, AQFS_TABLE_PTRLIST *ftl);


/**
 * Get position of the given object within the database file.
 *
 * If asWriter!=0 then the object position is looked up from the internal writersTablePtrListObjects.
 * @return position of the object within the database file
 * @param dbt pointer to this AQDB_TABLE object
 * @param index index of the object within the object table (e.g. object id -1)
 * @param asWriter if !=0 lookup from the temporary writers view of the object list
 */
int64_t AQDB_IoLayerLowlevel_Table_GetObjectPos(const AQDB_TABLE *dbt, uint64_t index, int asWriter);


/**
 * Set position of the given object within the database file.
 *
 * Always uses the temporary writers view of the object list.
 * @return 0 if okay, error code otherwise
 * @param dbt pointer to this AQDB_TABLE object
 * @param index index of the object within the object table (e.g. object id -1)
 * @param objectPos position of the object within the database file
 */
int AQDB_IoLayerLowlevel_Table_SetObjectPos(AQDB_TABLE *dbt, uint64_t index, uint64_t objectPos);


/**
 * Add position of the given object within the database file to the object list.
 *
 * Always uses the temporary writers view of the object list.
 * @return 0 if okay, error code otherwise
 * @param dbt pointer to this AQDB_TABLE object
 * @param index index of the object within the object table (e.g. object id -1)
 * @param objectPos position of the object within the database file
 */
int64_t AQDB_IoLayerLowlevel_Table_AddObjectPos(AQDB_TABLE *dbt, uint64_t objectPos);


int AQDB_IoLayerLowlevel_Table_GetObjectIds(AQDB_TABLE *dbt, GWEN_IDLIST64 *idl, int asWriter);




/*@}*/

#endif


