/***************************************************************************
 begin       : Thu Oct 08 2020
 copyright   : (C) 2020 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "./addobject.h"
#include "./dbopen.h"
#include "./utils.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/args.h>



#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg


/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static int _doIt(AQDB_DB *database, const char *sTableName, const char *sValues);
static AQDB_OBJECT *_createAndReadObject(AQDB_DB *database, AQDB_ID tableId, const char *sValues);
static int _addObject(AQDB_DB *database, AQDB_ID tableId, AQDB_OBJECT *object);



/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */



int DbTool_AddObject(GWEN_DB_NODE *dbArgs, int argc, char **argv)
{
  GWEN_DB_NODE *db;
  int rv;
  const char *dbUrl;
  AQDB_DB *database;
  const char *sTableName;
  const char *sValues;
  const GWEN_ARGS args[]= {
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "dburl",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "D",                          /* short option */
      "dburl",                      /* long option */
      I18S("Specify the database URL"),
      I18S("Specify the database URL")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "table",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "t",                          /* short option */
      "table",                      /* long option */
      I18S("Specify the name of the database table"),
      I18S("Specify the name of the database table")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "values",                     /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "V",                          /* short option */
      "values",                      /* long option */
      I18S("Specify the name of the values to set"),
      I18S("Specify the name of the values to set")
    },
    {
      GWEN_ARGS_FLAGS_HELP | GWEN_ARGS_FLAGS_LAST, /* flags */
      GWEN_ArgsType_Int,             /* type */
      "help",                       /* name */
      0,                            /* minnum */
      0,                            /* maxnum */
      "h",                          /* short option */
      "help",                       /* long option */
      I18S("Show this help screen"),    /* short description */
      I18S("Show this help screen")     /* long description */
    }
  };


  db=GWEN_DB_GetGroup(dbArgs, GWEN_DB_FLAGS_DEFAULT, "local");
  rv=GWEN_Args_Check(argc, argv, 1,
                     0 /*GWEN_ARGS_MODE_ALLOW_FREEPARAM*/,
                     args,
                     db);
  if (rv==GWEN_ARGS_RESULT_ERROR) {
    fprintf(stderr, "ERROR: Could not parse arguments\n");
    return 1;
  }
  else if (rv==GWEN_ARGS_RESULT_HELP) {
    GWEN_BUFFER *ubuf;

    ubuf=GWEN_Buffer_new(0, 1024, 0, 1);
    if (GWEN_Args_Usage(args, ubuf, GWEN_ArgsOutType_Txt)) {
      fprintf(stderr, "ERROR: Could not create help string\n");
      return 1;
    }
    fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(ubuf));
    GWEN_Buffer_free(ubuf);
    return 0;
  }

  dbUrl=GWEN_DB_GetCharValue(db, "dbUrl", 0, NULL);
  sTableName=GWEN_DB_GetCharValue(db, "table", 0, NULL);
  sValues=GWEN_DB_GetCharValue(db, "values", 0, NULL);

  database=DbTool_OpenDatabase(dbUrl, AQDB_ACTION_FLAGS_READ|AQDB_ACTION_FLAGS_WRITE);
  if (database==NULL) {
    DBG_ERROR(0, "Error creating database (%d)", rv);
    return 2;
  }

  rv=_doIt(database, sTableName, sValues);
  if (rv!=0) {
    DBG_ERROR(0, "Error opening table [%s] (%d)", sTableName, rv);
    AQDB_DB_Close(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_free(database);
    return rv;
  }

  rv=AQDB_DB_Close(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error closing database (%d)", rv);
    AQDB_DB_free(database);
    return 3;
  }

  AQDB_DB_free(database);
  return 0;
}



int _doIt(AQDB_DB *database, const char *sTableName, const char *sValues)
{
  int rv;
  AQDB_ID tableId=0;
  AQDB_OBJECT *object;

  rv=AQDB_DB_OpenTable(database, sTableName, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, &tableId);
  if (rv<0) {
    DBG_ERROR(0, "Error opening table [%s] (%d)", sTableName, rv);
    return 2;
  }

  object=_createAndReadObject(database, tableId, sValues);
  if (rv<0) {
    DBG_ERROR(0, "Error creating and reading object values for table \"%s\"", sTableName);
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 2;
  }

  rv=_addObject(database, tableId, object);
  if (rv<0) {
    DBG_ERROR(0, "Error adding object to database (%d)", rv);
    AQDB_Object_free(object);
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 2;
  }

  AQDB_Object_free(object);

  rv=AQDB_DB_CloseTable(database, tableId, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error closing table (%d)", rv);
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 3;
  }

  return 0;
}



AQDB_OBJECT *_createAndReadObject(AQDB_DB *database, AQDB_ID tableId, const char *sValues)
{
  AQDB_COLUMN_LIST *columnList;
  AQDB_OBJECT *object;
  int rv;

  columnList=AQDB_DB_GetColumnList(database, tableId);
  if (columnList==NULL) {
    DBG_ERROR(0, "Error getting column list for table");
    return NULL;
  }

  object=AQDB_Object_new(tableId, 0, AQDB_Column_List_GetCount(columnList));
  if (object==NULL) {
    DBG_ERROR(0, "Error creating object");
    AQDB_Column_List_free(columnList);
    return NULL;
  }

  rv=DbTool_ReadObjectFromString(sValues, columnList, object);
  if (rv<0) {
    DBG_ERROR(0, "Error reading object values");
    AQDB_Object_free(object);
    AQDB_Column_List_free(columnList);
    return NULL;
  }

  AQDB_Column_List_free(columnList);
  return object;
}



int _addObject(AQDB_DB *database, AQDB_ID tableId, AQDB_OBJECT *object)
{
  int rv;

  rv=AQDB_DB_BeginEdit(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error beginning editing database (%d)", rv);
    return rv;
  }

  rv=AQDB_DB_AddObject(database, tableId, object);
  if (rv<0) {
    DBG_ERROR(0, "Error adding object");
    AQDB_DB_EndEdit(database, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  rv=AQDB_DB_EndEdit(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error ending editing database (%d)", rv);
    AQDB_DB_EndEdit(database, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  return 0;
}



