/***************************************************************************
 begin       : Thu Oct 08 2020
 copyright   : (C) 2020 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "./createdb.h"
#include "./dbopen.h"
#include "./utils.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/args.h>



#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg




int DbTool_CreateDb(GWEN_DB_NODE *dbArgs, int argc, char **argv)
{
  GWEN_DB_NODE *db;
  int rv;
  const char *dbUrl;
  AQDB_DB *database;
  const char *sTableName;
  const char *sColumns;
  AQDB_COLUMN_LIST *columnList;
  const GWEN_ARGS args[]= {
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "dburl",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "D",                          /* short option */
      "dburl",                      /* long option */
      I18S("Specify the database URL"),
      I18S("Specify the database URL")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "table",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "t",                          /* short option */
      "table",                      /* long option */
      I18S("Specify the name of the database table"),
      I18S("Specify the name of the database table")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "columns",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "c",                          /* short option */
      "columns",                      /* long option */
      I18S("Specify the columns of the database field"),
      I18S("Specify the columns of the database field")
    },
    {
      GWEN_ARGS_FLAGS_HELP | GWEN_ARGS_FLAGS_LAST, /* flags */
      GWEN_ArgsType_Int,             /* type */
      "help",                       /* name */
      0,                            /* minnum */
      0,                            /* maxnum */
      "h",                          /* short option */
      "help",                       /* long option */
      I18S("Show this help screen"),    /* short description */
      I18S("Show this help screen")     /* long description */
    }
  };


  db=GWEN_DB_GetGroup(dbArgs, GWEN_DB_FLAGS_DEFAULT, "local");
  rv=GWEN_Args_Check(argc, argv, 1,
                     0 /*GWEN_ARGS_MODE_ALLOW_FREEPARAM*/,
                     args,
                     db);
  if (rv==GWEN_ARGS_RESULT_ERROR) {
    fprintf(stderr, "ERROR: Could not parse arguments\n");
    return 1;
  }
  else if (rv==GWEN_ARGS_RESULT_HELP) {
    GWEN_BUFFER *ubuf;

    ubuf=GWEN_Buffer_new(0, 1024, 0, 1);
    if (GWEN_Args_Usage(args, ubuf, GWEN_ArgsOutType_Txt)) {
      fprintf(stderr, "ERROR: Could not create help string\n");
      return 1;
    }
    fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(ubuf));
    GWEN_Buffer_free(ubuf);
    return 0;
  }

  dbUrl=GWEN_DB_GetCharValue(db, "dbUrl", 0, NULL);
  sTableName=GWEN_DB_GetCharValue(db, "table", 0, NULL);
  sColumns=GWEN_DB_GetCharValue(db, "columns", 0, NULL);

  columnList=DbTool_ColumnList_fromString(sColumns);
  if (columnList==NULL) {
    DBG_ERROR(0, "Invalid column specs");
    return 2;
  }

  database=DbTool_CreateDatabase(dbUrl);
  if (database==NULL) {
    DBG_ERROR(0, "Error creating database (%d)", rv);
    return 2;
  }

  rv=AQDB_DB_BeginEdit(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error beginning editing database (%d)", rv);
    AQDB_DB_Close(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_free(database);
    return 2;
  }

  rv=AQDB_DB_CreateTable(database, sTableName, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE, columnList);
  if (rv<0) {
    DBG_ERROR(0, "Error creating table (%d)", rv);
    AQDB_DB_EndEdit(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_Close(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_free(database);
    AQDB_Column_List_free(columnList);
    return 2;
  }

  rv=AQDB_DB_EndEdit(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error ending editing database (%d)", rv);
    AQDB_DB_Close(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_free(database);
    AQDB_Column_List_free(columnList);
    return 2;
  }

  rv=AQDB_DB_Close(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error closing database (%d), ignoring", rv);
  }
  AQDB_DB_free(database);
  AQDB_Column_List_free(columnList);

  return 0;
}





