/***************************************************************************
 begin       : Thu Oct 08 2020
 copyright   : (C) 2020 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "./printobjects.h"
#include "./dbopen.h"
#include "./utils.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/args.h>



#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg


/* ------------------------------------------------------------------------------------------------
 * forward declarations
 * ------------------------------------------------------------------------------------------------
 */

static int _doIt(AQDB_DB *database, const char *sTableName, const char *sQuery, const char *sTemplate);
static int _forEveryObject(AQDB_DB *database, AQDB_ID tableId, const char *sQuery, uint32_t flags,
                           AQDB_COLUMN_LIST *columnList, GWEN_STRINGLIST *slTemplate);
static int _dumpObject(AQDB_OBJECT *o, AQDB_COLUMN_LIST *columnList, GWEN_STRINGLIST *slTemplate);
static int _dumpField(AQDB_OBJECT *o, const char *sFieldName, const AQDB_COLUMN_LIST *columnList, GWEN_BUFFER *destBuf);
static GWEN_STRINGLIST *_getColumnNamesIntoStringList(const AQDB_COLUMN_LIST *columnList);



/* ------------------------------------------------------------------------------------------------
 * implementations
 * ------------------------------------------------------------------------------------------------
 */



int DbTool_PrintObjects(GWEN_DB_NODE *dbArgs, int argc, char **argv)
{
  GWEN_DB_NODE *db;
  int rv;
  const char *dbUrl;
  AQDB_DB *database;
  const char *sTableName;
  const char *sQuery;
  const char *sTemplate;
  const GWEN_ARGS args[]= {
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "dburl",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "D",                          /* short option */
      "dburl",                      /* long option */
      I18S("Specify the database URL"),
      I18S("Specify the database URL")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "table",                      /* name */
      1,                            /* minnum */
      1,                            /* maxnum */
      "t",                          /* short option */
      "table",                      /* long option */
      I18S("Specify the name of the database table"),
      I18S("Specify the name of the database table")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "query",                      /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "q",                          /* short option */
      "query",                      /* long option */
      I18S("Specify a query string to select objects"),
      I18S("Specify query string to select objects")
    },
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "template",                   /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "tt",                         /* short option */
      "template",                   /* long option */
      I18S("Specify list of fields to print"),
      I18S("Specify list of fields to print")
    },
    {
      GWEN_ARGS_FLAGS_HELP | GWEN_ARGS_FLAGS_LAST, /* flags */
      GWEN_ArgsType_Int,             /* type */
      "help",                       /* name */
      0,                            /* minnum */
      0,                            /* maxnum */
      "h",                          /* short option */
      "help",                       /* long option */
      I18S("Show this help screen"),    /* short description */
      I18S("Show this help screen")     /* long description */
    }
  };


  db=GWEN_DB_GetGroup(dbArgs, GWEN_DB_FLAGS_DEFAULT, "local");
  rv=GWEN_Args_Check(argc, argv, 1,
                     0 /*GWEN_ARGS_MODE_ALLOW_FREEPARAM*/,
                     args,
                     db);
  if (rv==GWEN_ARGS_RESULT_ERROR) {
    fprintf(stderr, "ERROR: Could not parse arguments\n");
    return 1;
  }
  else if (rv==GWEN_ARGS_RESULT_HELP) {
    GWEN_BUFFER *ubuf;

    ubuf=GWEN_Buffer_new(0, 1024, 0, 1);
    if (GWEN_Args_Usage(args, ubuf, GWEN_ArgsOutType_Txt)) {
      fprintf(stderr, "ERROR: Could not create help string\n");
      return 1;
    }
    fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(ubuf));
    GWEN_Buffer_free(ubuf);
    return 0;
  }

  dbUrl=GWEN_DB_GetCharValue(db, "dbUrl", 0, NULL);
  sTableName=GWEN_DB_GetCharValue(db, "table", 0, NULL);
  sQuery=GWEN_DB_GetCharValue(db, "query", 0, NULL);
  sTemplate=GWEN_DB_GetCharValue(db, "template", 0, NULL);

  database=DbTool_OpenDatabase(dbUrl, AQDB_ACTION_FLAGS_READ);
  if (database==NULL) {
    DBG_ERROR(0, "Error creating database (%d)", rv);
    return 2;
  }

  rv=_doIt(database, sTableName, sQuery, sTemplate);
  if (rv!=0) {
    DBG_ERROR(0, "Error opening table [%s] (%d)", sTableName, rv);
    AQDB_DB_Close(database, AQDB_ACTION_FLAGS_ABORT);
    AQDB_DB_free(database);
    return rv;
  }

  rv=AQDB_DB_Close(database, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error closing database (%d)", rv);
    AQDB_DB_free(database);
    return 3;
  }

  AQDB_DB_free(database);
  return 0;
}



int _doIt(AQDB_DB *database, const char *sTableName, const char *sQuery, const char *sTemplate)
{
  int rv;
  AQDB_ID tableId=0;
  AQDB_COLUMN_LIST *columnList;
  GWEN_STRINGLIST *slTemplate;

  rv=AQDB_DB_OpenTable(database, sTableName, AQDB_ACTION_FLAGS_READ, &tableId);
  if (rv<0) {
    DBG_ERROR(0, "Error opening table [%s] (%d)", sTableName, rv);
    return 2;
  }

  columnList=AQDB_DB_GetColumnList(database, tableId);
  if (columnList==NULL) {
    DBG_ERROR(0, "Error getting column list for table");
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 2;
  }

  if (sTemplate) {
    slTemplate=GWEN_StringList_fromString(sTemplate, ",", 1);
    if (slTemplate==NULL) {
      DBG_ERROR(0, "Invalid template string \"%s\"", sTemplate);
      AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
      return 1;
    }
  }
  else {
    slTemplate=_getColumnNamesIntoStringList(columnList);
    if (slTemplate==NULL) {
      DBG_ERROR(0, "No column names");
      AQDB_Column_List_free(columnList);
      AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
      return 2;
    }
  }

  rv=_forEveryObject(database, tableId, sQuery, 0, columnList, slTemplate);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_StringList_free(slTemplate);
    AQDB_Column_List_free(columnList);
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 2;
  }

  GWEN_StringList_free(slTemplate);
  AQDB_Column_List_free(columnList);

  rv=AQDB_DB_CloseTable(database, tableId, 0);
  if (rv<0) {
    DBG_ERROR(0, "Error closing table (%d)", rv);
    AQDB_DB_CloseTable(database, tableId, AQDB_ACTION_FLAGS_ABORT);
    return 3;
  }

  return 0;
}



int _forEveryObject(AQDB_DB *database, AQDB_ID tableId, const char *sQuery, uint32_t flags,
                    AQDB_COLUMN_LIST *columnList, GWEN_STRINGLIST *slTemplate)
{
  AQDB_ID qid=0;
  int rv;
  AQDB_ID oids[16];
  int len;

  /* query */
  rv=AQDB_DB_QuerySubmit(database, tableId, 0, sQuery, 0, &qid);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(0, "here (%d)", rv);
      return rv;
    }
    return 0;
  }

  /* browse all objects */
  len=16;
  rv=AQDB_DB_QueryGetFirst(database, qid, oids, &len);
  while (rv==0) {
    int i;

    for (i=0; i<len; i++) {
      AQDB_OBJECT *o=NULL;

      rv=AQDB_DB_ReadObject(database, tableId, oids[i], &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error reading object %d (%d)", oids[i], rv);
        AQDB_DB_QueryClose(database, qid);
        return rv;
      }

      rv=_dumpObject(o, columnList, slTemplate);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "Error printing object %d (%d)", oids[i], rv);
        AQDB_DB_QueryClose(database, qid);
        return rv;
      }
      AQDB_Object_free(o);
    }

    len=16;
    rv=AQDB_DB_QueryGetNext(database, qid, oids, &len);
  }
  AQDB_DB_QueryClose(database, qid);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int _dumpObject(AQDB_OBJECT *o, AQDB_COLUMN_LIST *columnList, GWEN_STRINGLIST *slTemplate)
{
  GWEN_STRINGLISTENTRY *se;
  GWEN_BUFFER *destBuf;
  int first=1;

  destBuf=GWEN_Buffer_new(0, 256, 0, 1);
  se=GWEN_StringList_FirstEntry(slTemplate);
  while (se) {
    const char *s;

    s=GWEN_StringListEntry_Data(se);
    if (s && *s) {
      int rv;

      rv=_dumpField(o, s, columnList, destBuf);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        return rv;
      }
      // print
      if (first)
        fprintf(stdout, "%s", GWEN_Buffer_GetStart(destBuf));
      else
        fprintf(stdout, "\t%s", GWEN_Buffer_GetStart(destBuf));
      GWEN_Buffer_Reset(destBuf);
    }
    first=0;
    se=GWEN_StringListEntry_Next(se);
  } /* while */

  GWEN_Buffer_free(destBuf);
  fprintf(stdout, "\n");
  return 0;
}



int _dumpField(AQDB_OBJECT *object, const char *sFieldName, const AQDB_COLUMN_LIST *columnList, GWEN_BUFFER *destBuf)
{
  AQDB_COLUMN *column;
  int fieldIdx=0;
  AQDB_DATA_TYPE t;
  const char *modifier;
  int rv;

  column=AQDB_Column_List_First(columnList);
  while (column) {
    const char *cname;

    cname=AQDB_Column_GetName(column);
    if (cname && strcasecmp(cname, sFieldName)==0)
      break;
    fieldIdx++;
    column=AQDB_Column_List_Next(column);
  }

  if (column==NULL) {
    DBG_ERROR(0, "Field \"%s\" not found", sFieldName);
    return GWEN_ERROR_INVALID;
  }

  t=AQDB_Column_GetType(column);
  switch (t) {
  case AQDB_DataType_Timestamp:
    modifier="YYYYMMDD-hh:mm:ss";
    break;
  case AQDB_DataType_Bin:
    modifier="hex";
    break;
  default:
    modifier=NULL;
    break;
  }

  rv=AQDB_Object_GetFieldAsStringByType(object, fieldIdx, t, modifier, destBuf);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    return rv;
  }

  return 0;
}



GWEN_STRINGLIST *_getColumnNamesIntoStringList(const AQDB_COLUMN_LIST *columnList)
{
  const AQDB_COLUMN *column;
  GWEN_STRINGLIST *sl;

  sl=GWEN_StringList_new();
  column=AQDB_Column_List_First(columnList);
  while (column) {
    const char *s;

    s=AQDB_Column_GetName(column);
    if (s && *s)
      GWEN_StringList_AppendString(sl, s, 0, 1);
    column=AQDB_Column_List_Next(column);
  }
  if (GWEN_StringList_Count(sl)<1) {
    GWEN_StringList_free(sl);
    DBG_ERROR(0, "No column names");
    return NULL;
  }

  return sl;
}






