/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2019 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "./listtables.h"
#include "./dbopen.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/args.h>



#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg



static int showTable(AQDB_DB *database, const char *tableName, uint32_t flags);
static int showColumns(AQDB_DB *database, const char *tableName, AQDB_ID tableId);
static int countObjects(AQDB_DB *database, const char *tableName, AQDB_ID tableId);
static int scanObjects(AQDB_DB *database, const char *tableName, AQDB_ID tableId);


#define SHOWTABLES_FLAGS_SHOWCOLUMNS  0x0001
#define SHOWTABLES_FLAGS_COUNTOBJECTS 0x0002
#define SHOWTABLES_FLAGS_SCANOBJECTS  0x0004




int DbTool_ShowTables(GWEN_DB_NODE *dbArgs, int argc, char **argv)
{
  GWEN_DB_NODE *db;
  int rv;
  const char *dbUrl;
  AQDB_DB *database=NULL;
  uint32_t flags=0;
  GWEN_STRINGLIST *tableNameList;
  const GWEN_ARGS args[]= {
    {
      GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
      GWEN_ArgsType_Char,           /* type */
      "dburl",                      /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "D",                          /* short option */
      "dburl",                      /* long option */
      I18S("Specify the database URL"),
      I18S("Specify the database URL")
    },
    {
      0,                            /* flags */
      GWEN_ArgsType_Int,            /* type */
      "showColumns",                /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "c",                          /* short option */
      "showcolumns",             /* long option */
      "Show columns for every table",/* short description */
      "Show columns for every table"        /* long description */
    },
    {
      0,                            /* flags */
      GWEN_ArgsType_Int,            /* type */
      "countObjects",               /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "o",                          /* short option */
      "countobjects",               /* long option */
      "Show number of objects for every table",/* short description */
      "Show number of objects for every table"        /* long description */
    },
    {
      0,                            /* flags */
      GWEN_ArgsType_Int,            /* type */
      "scanObjects",                /* name */
      0,                            /* minnum */
      1,                            /* maxnum */
      "s",                          /* short option */
      "scanobjects",               /* long option */
      "Read every object of every table",/* short description */
      "Read every object of every table"        /* long description */
    },
    {
      GWEN_ARGS_FLAGS_HELP | GWEN_ARGS_FLAGS_LAST, /* flags */
      GWEN_ArgsType_Int,             /* type */
      "help",                       /* name */
      0,                            /* minnum */
      0,                            /* maxnum */
      "h",                          /* short option */
      "help",                       /* long option */
      I18S("Show this help screen"),    /* short description */
      I18S("Show this help screen")     /* long description */
    }
  };

  db=GWEN_DB_GetGroup(dbArgs, GWEN_DB_FLAGS_DEFAULT, "local");
  rv=GWEN_Args_Check(argc, argv, 1,
                     0 /*GWEN_ARGS_MODE_ALLOW_FREEPARAM*/,
                     args,
                     db);
  if (rv==GWEN_ARGS_RESULT_ERROR) {
    fprintf(stderr, "ERROR: Could not parse arguments\n");
    return 1;
  }
  else if (rv==GWEN_ARGS_RESULT_HELP) {
    GWEN_BUFFER *ubuf;

    ubuf=GWEN_Buffer_new(0, 1024, 0, 1);
    if (GWEN_Args_Usage(args, ubuf, GWEN_ArgsOutType_Txt)) {
      fprintf(stderr, "ERROR: Could not create help string\n");
      return 1;
    }
    fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(ubuf));
    GWEN_Buffer_free(ubuf);
    return 0;
  }

  dbUrl=GWEN_DB_GetCharValue(db, "dbUrl", 0, NULL);
  if (GWEN_DB_GetIntValue(db, "showColumns", 0, 0))
    flags|=SHOWTABLES_FLAGS_SHOWCOLUMNS;
  if (GWEN_DB_GetIntValue(db, "countObjects", 0, 0))
    flags|=SHOWTABLES_FLAGS_COUNTOBJECTS;
  if (GWEN_DB_GetIntValue(db, "scanObjects", 0, 0))
    flags|=SHOWTABLES_FLAGS_SCANOBJECTS;


  rv=AQDB_Init();
  if (rv) {
    DBG_ERROR_ERR(0, rv);
    return rv;
  }

  rv=AQDB_DbFactory(dbUrl, &database);
  if (rv<0) {
    DBG_ERROR(NULL, "ERROR: Unable to instantiate db (%d)", rv);
    return 2;
  }

  rv=AQDB_DB_Open(database, AQDB_ACTION_FLAGS_READ);
  if (rv<0) {
    DBG_ERROR(NULL, "ERROR: Unable to open db (%d)", rv);
    AQDB_DB_free(database);
    return 2;
  }

  tableNameList=GWEN_StringList_new();

  rv=AQDB_DB_GetTableNameList(database, tableNameList);
  if (rv<0) {
    DBG_ERROR(NULL, "ERROR: Unable to list table names (%d)", rv);
    AQDB_DB_Close(database, 1);
    AQDB_DB_free(database);
    return 2;
  }

  if (GWEN_StringList_Count(tableNameList)) {
    GWEN_STRINGLISTENTRY *se;

    GWEN_StringList_Sort(tableNameList, 1, GWEN_StringList_SortModeNoCase);
    se=GWEN_StringList_FirstEntry(tableNameList);
    while (se) {
      const char *s;

      s=GWEN_StringListEntry_Data(se);
      if (s && *s) {
        showTable(database, s, flags);
      }
      se=GWEN_StringListEntry_Next(se);
    }
  }

  rv=AQDB_DB_Close(database, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to close db (%d)\n", rv);
    AQDB_DB_free(database);
    return 2;
  }


  return 0;
}



int showTable(AQDB_DB *database, const char *tableName, uint32_t flags)
{
  AQDB_ID tableId=0;
  int rv;

  fprintf(stdout, "%s\n", tableName);
  rv=AQDB_DB_OpenTable(database, tableName, AQDB_ACTION_FLAGS_READ, &tableId);
  if (rv<0) {
    DBG_ERROR(0, "Error on open table [%s]: %d", tableName, rv);
    return rv;
  }

  if (flags & SHOWTABLES_FLAGS_SHOWCOLUMNS) {
    rv=showColumns(database, tableName, tableId);
    if (rv<0) {
      DBG_ERROR(0, "Error on showColumns [%s]: %d", tableName, rv);
      AQDB_DB_CloseTable(database, tableId, 0);
      return rv;
    }
  }

  if (flags & SHOWTABLES_FLAGS_COUNTOBJECTS) {
    rv=countObjects(database, tableName, tableId);
    if (rv<0) {
      DBG_ERROR(0, "Error on countObjects [%s]: %d", tableName, rv);
      AQDB_DB_CloseTable(database, tableId, 0);
      return rv;
    }
  }

  if (flags & SHOWTABLES_FLAGS_SCANOBJECTS) {
    rv=scanObjects(database, tableName, tableId);
    if (rv<0) {
      DBG_ERROR(0, "Error on scanObjects [%s]: %d", tableName, rv);
      AQDB_DB_CloseTable(database, tableId, 0);
      return rv;
    }
  }


  AQDB_DB_CloseTable(database, tableId, 0);

  return 0;
}



int showColumns(AQDB_DB *database, const char *tableName, AQDB_ID tableId)
{
  AQDB_COLUMN_LIST *columnList;

  columnList=AQDB_DB_GetColumnList(database, tableId);
  if (columnList==NULL) {
    DBG_ERROR(0, "Error on GetColumnList: No Columns [%s]", tableName);
    return GWEN_ERROR_BAD_DATA;
  }

  if (AQDB_Column_List_GetCount(columnList)) {
    const AQDB_COLUMN *col;

    fprintf(stdout, "  Columns:\n");
    col=AQDB_Column_List_First(columnList);
    while (col) {
      const char *columnName;

      columnName=AQDB_Column_GetName(col);
      fprintf(stdout,
              "    Column: Name=%s, type=%s, maxlen=%d, isMutable=%d\n",
              columnName?columnName:"<unnamed>",
              AQDB_DataType_toString(AQDB_Column_GetType(col)),
              AQDB_Column_GetMaxLen(col),
              AQDB_Column_GetIsMutable(col));
      col=AQDB_Column_List_Next(col);
    }
  }
  AQDB_Column_List_free(columnList);

  return 0;
}


int countObjects(AQDB_DB *database, const char *tableName, AQDB_ID tableId)
{
  int rv;
  AQDB_ID queryId=0;
  size_t objectCount=0;

  fprintf(stdout, "  Objects:\n");

  rv=AQDB_DB_QuerySubmit(database, tableId, 0, NULL, 0, &queryId);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_ERROR(0, "Error on AQDB_DB_QuerySubmit [%s]: %d", tableName, rv);
      return rv;
    }
  }
  else {
    AQDB_ID queryResultsData[512];
    int queryResultsLen;

    queryResultsLen=sizeof(queryResultsData)/sizeof(queryResultsData[0]);
    rv=AQDB_DB_QueryGetFirst(database, queryId, queryResultsData, &queryResultsLen);
    while (rv==0) {
      objectCount+=queryResultsLen;
      queryResultsLen=sizeof(queryResultsData)/sizeof(queryResultsData[0]);
      rv=AQDB_DB_QueryGetNext(database, queryId, queryResultsData, &queryResultsLen);
    }
  }

  fprintf(stdout, "    %llu Objects\n", (long long unsigned int) objectCount);

  AQDB_DB_QueryClose(database, queryId);
  return 0;
}



int scanObjects(AQDB_DB *database, const char *tableName, AQDB_ID tableId)
{
  int rv;
  AQDB_ID queryId=0;
  size_t objectCount=0;

  fprintf(stdout, "  Scanned Objects:\n");

  rv=AQDB_DB_QuerySubmit(database, tableId, 0, NULL, 0, &queryId);
  if (rv<0) {
    if (rv!=GWEN_ERROR_NOT_FOUND) {
      DBG_ERROR(0, "Error on AQDB_DB_QuerySubmit [%s]: %d", tableName, rv);
      return rv;
    }
  }
  else {
    AQDB_ID queryResultsData[512];
    int queryResultsLen;

    queryResultsLen=sizeof(queryResultsData)/sizeof(queryResultsData[0]);
    rv=AQDB_DB_QueryGetFirst(database, queryId, queryResultsData, &queryResultsLen);
    while (rv==0) {
      int i;

      for (i=0; i<queryResultsLen; i++) {
        AQDB_OBJECT *object=NULL;

        rv=AQDB_DB_ReadObject(database, tableId, queryResultsData[i], &object);
        if (rv<0) {
          DBG_ERROR(0, "Error on AQDB_DB_ReadObject %llu [%s]: %d",
                    (long long unsigned int)(queryResultsData[i]),
                    tableName, rv);
        }
        else {
          objectCount++;
          AQDB_Object_free(object);
        }
      } /* for i */

      queryResultsLen=sizeof(queryResultsData)/sizeof(queryResultsData[0]);
      rv=AQDB_DB_QueryGetNext(database, queryId, queryResultsData, &queryResultsLen);
    }
  }

  fprintf(stdout, "    %llu Objects\n", (long long unsigned int) objectCount);

  AQDB_DB_QueryClose(database, queryId);
  return 0;
}





