/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2019 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "./utils.h"
#include "./dbopen.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>
#include <aqdatabase/aqdb_object.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/args.h>
#include <gwenhywfar/stringlist.h>



#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)
#define I18S(msg) msg



static AQDB_COLUMN *_columnFromString(const char *sourceString);
static int _handleAssignment(const char *sourceString, const AQDB_COLUMN_LIST *columnList, AQDB_OBJECT *object);
static int _assignValue(const char *sFieldName, const char *sValue, const AQDB_COLUMN_LIST *columnList,
                        AQDB_OBJECT *object);




AQDB_COLUMN_LIST *DbTool_ColumnList_fromString(const char *sourceString)
{
  GWEN_STRINGLIST *slColumns;

  slColumns=GWEN_StringList_fromString(sourceString, ",", 1);
  if (slColumns==NULL) {
    DBG_ERROR(0, "Invalid type string \"%s\"", sourceString);
    return NULL;
  }

  if (GWEN_StringList_Count(slColumns)) {
    AQDB_COLUMN_LIST *columnList;
    GWEN_STRINGLISTENTRY *se;

    columnList=AQDB_Column_List_new();

    se=GWEN_StringList_FirstEntry(slColumns);
    while (se) {
      const char *s;

      s=GWEN_StringListEntry_Data(se);
      if (s && *s) {
        AQDB_COLUMN *column;

        column=_columnFromString(s);
        if (column)
          AQDB_Column_List_Add(column, columnList);
      }
      se=GWEN_StringListEntry_Next(se);
    } /* while */

    GWEN_StringList_free(slColumns);
    return columnList;
  }
  else {
    DBG_ERROR(0, "No string list created from string \"%s\"", sourceString);
    return NULL;
  }
}



AQDB_COLUMN *_columnFromString(const char *sourceString)
{
  GWEN_STRINGLIST *slFields;
  int fieldCount;

  slFields=GWEN_StringList_fromString(sourceString, ":", 0);
  if (slFields==NULL) {
    DBG_ERROR(0, "Invalid type string \"%s\"", sourceString);
    return NULL;
  }

  fieldCount=GWEN_StringList_Count(slFields);
  if (fieldCount>0) {
    const char *columnName=NULL;
    AQDB_DATA_TYPE columnType=AQDB_DataType_Int;
    const char *s;
    int columnSize=0;
    AQDB_COLUMN *column=NULL;

    columnName=GWEN_StringList_StringAt(slFields, 0);
    if (fieldCount>1) {
      s=GWEN_StringList_StringAt(slFields, 1);
      columnType=AQDB_DataType_fromString(s);

      if (columnType==AQDB_DataType_Unknown) {
        DBG_ERROR(0, "\"%s\" is not a data type", s);
        GWEN_StringList_free(slFields);
        return NULL;
      }

      /* preset default according to type if possible */
      switch (columnType) {
      case AQDB_DataType_Unknown:
        columnSize=0;
        break;
      case AQDB_DataType_None:
        columnSize=0;
        break;
      case AQDB_DataType_Int:
        columnSize=8;
        break;
      case AQDB_DataType_String:
        columnSize=256;
        break;
      case AQDB_DataType_Bin:
        columnSize=256;
        break;
      case AQDB_DataType_Date:
        columnSize=16;
        break;
      case AQDB_DataType_Timestamp:
        columnSize=16;
        break;
      case AQDB_DataType_Amount:
        columnSize=256;
        break;
      }

      if (fieldCount>2) {
        s=GWEN_StringList_StringAt(slFields, 2);
        if (1!=sscanf(s, "%d", &columnSize)) {
          DBG_ERROR(0, "\"%s\" is not a number for column size", s);
          GWEN_StringList_free(slFields);
          return NULL;
        }
      }
    }

    column=AQDB_Column_new(columnType, columnName, columnSize);
    if (column==NULL) {
      DBG_ERROR(0, "Could not create column from \"%s\"", sourceString);
      GWEN_StringList_free(slFields);
      return NULL;
    }

    /* column created */
    GWEN_StringList_free(slFields);
    return column;
  }
  else {
    DBG_ERROR(0, "No fields created from string \"%s\"", sourceString?sourceString:"<empty>");
    GWEN_StringList_free(slFields);
    return NULL;
  }
}



void DbTool_DumpColumnList(const AQDB_COLUMN_LIST *columnList)
{
  if (columnList) {
    AQDB_COLUMN *column;

    column=AQDB_Column_List_First(columnList);
    while (column) {
      DbTool_DumpColumn(column);
      column=AQDB_Column_List_Next(column);
    }
  }
}



void DbTool_DumpColumn(const AQDB_COLUMN *column)
{
  if (column) {
    const char *colName;
    AQDB_DATA_TYPE colType;
    int colSize;

    colName=AQDB_Column_GetName(column);
    colType=AQDB_Column_GetType(column);
    colSize=AQDB_Column_GetMaxLen(column);
    fprintf(stdout, "  %s (%s, %d bytes)\n",
            colName?colName:"<no name>",
            AQDB_DataType_toString(colType),
            colSize);
  }
}





int DbTool_ReadObjectFromString(const char *sourceString,
                                const AQDB_COLUMN_LIST *columnList,
                                AQDB_OBJECT *object)
{
  GWEN_STRINGLIST *slAssignments;

  slAssignments=GWEN_StringList_fromString(sourceString, ",", 1);
  if (slAssignments==NULL) {
    DBG_ERROR(0, "Invalid object string \"%s\"", sourceString);
    return GWEN_ERROR_BAD_DATA;
  }

  if (GWEN_StringList_Count(slAssignments)) {
    GWEN_STRINGLISTENTRY *se;

    se=GWEN_StringList_FirstEntry(slAssignments);
    while (se) {
      const char *s;
      int rv;

      s=GWEN_StringListEntry_Data(se);
      if (s && *s) {
        rv=_handleAssignment(s, columnList, object);
        if (rv<0) {
          DBG_ERROR(0, "Bad assignemt \"%s\" (%d)", s, rv);
          GWEN_StringList_free(slAssignments);
          return rv;
        }
      }
      se=GWEN_StringListEntry_Next(se);
    } /* while */
  }
  else {
    DBG_ERROR(0, "Invalid object string \"%s\" (no assignments in string)", sourceString);
    GWEN_StringList_free(slAssignments);
    return GWEN_ERROR_BAD_DATA;
  }

  GWEN_StringList_free(slAssignments);
  return 0;
}



int _handleAssignment(const char *sourceString, const AQDB_COLUMN_LIST *columnList, AQDB_OBJECT *object)
{
  GWEN_STRINGLIST *slAssignment;

  slAssignment=GWEN_StringList_fromString(sourceString, "=", 1);
  if (slAssignment==NULL) {
    DBG_ERROR(0, "Invalid assignment \"%s\"", sourceString);
    return GWEN_ERROR_BAD_DATA;
  }

  if (GWEN_StringList_Count(slAssignment)==2) {
    const char *sFieldName;
    const char *sValue;
    int rv;

    sFieldName=GWEN_StringList_StringAt(slAssignment, 0);
    if (!(sFieldName && *sFieldName)) {
      DBG_ERROR(0, "Empty or missing field name in assignemnt string \"%s\"", sourceString);
      GWEN_StringList_free(slAssignment);
      return GWEN_ERROR_BAD_DATA;
    }
    sValue=GWEN_StringList_StringAt(slAssignment, 1);

    rv=_assignValue(sFieldName, sValue, columnList, object);
    if (rv<0) {
      DBG_INFO(0, "here (%d)", rv);
      GWEN_StringList_free(slAssignment);
      return rv;
    }
  }
  else {
    DBG_ERROR(0, "Invalid object string \"%s\" (invalid assignment in string)", sourceString);
    return GWEN_ERROR_BAD_DATA;
  }

  GWEN_StringList_free(slAssignment);
  return 0;
}



int _assignValue(const char *sFieldName, const char *sValue, const AQDB_COLUMN_LIST *columnList, AQDB_OBJECT *object)
{
  AQDB_COLUMN *column;
  int fieldIdx=0;
  AQDB_DATA_TYPE t;
  const char *modifier;
  int rv;

  column=AQDB_Column_List_First(columnList);
  while (column) {
    const char *cname;

    cname=AQDB_Column_GetName(column);
    if (cname && strcasecmp(cname, sFieldName)==0)
      break;
    fieldIdx++;
    column=AQDB_Column_List_Next(column);
  }

  if (column==NULL) {
    DBG_ERROR(0, "Field \"%s\" not found", sFieldName);
    return GWEN_ERROR_INVALID;
  }

  t=AQDB_Column_GetType(column);
  switch (t) {
  case AQDB_DataType_Timestamp:
    modifier="YYYYMMDD-hh:mm:ss";
    break;
  case AQDB_DataType_Bin:
    modifier="hex";
    break;
  default:
    modifier=NULL;
    break;
  }

  rv=AQDB_Object_SetFieldFromStringByType(object, fieldIdx, t, sValue, modifier);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    return rv;
  }

  return 0;
}



